//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//

using System;
using System.Collections.Generic;
using System.Text;
using System.IO ;
using System.Runtime.InteropServices;
using System.Diagnostics;
using Bwg.Utils ;

namespace Bwg.Audio.Acm
{
    /// <summary>
    /// This class is a class that uses the codecs installed into the audio compression
    /// manager to convert audio data from one form to another.
    /// </summary>
    public unsafe class AcmConversion
    {
        // [DllImport("BwgAudioScanner.dll")]
        // internal static extern int BwgScanBlock(IntPtr buf, int count, ref short lmin, ref short lmax);

        // [DllImport("BwgAudioScanner.dll")]
        // internal static extern int BwgBeginScan();

        // [DllImport("BwgAudioScanner.dll")]
        // internal static extern int BwgEndScan();
        // private long m_conv_calls;

        #region private data structures

        /// <summary>
        /// This object is the source of data to convert
        /// </summary>
        private AcmDataSource m_source;

        /// <summary>
        /// Handle to the driver to use in the conversion
        /// </summary>
        private AcmDriver m_driver;

        /// <summary>
        /// The size of the input buffer.
        /// </summary>
        private uint m_convert_input_buffer_size;

        /// <summary>
        /// The size of the output buffer
        /// </summary>
        private uint m_convert_output_buffer_size;

        /// <summary>
        /// The format of the source data
        /// </summary>
        private byte[] m_source_format;

        /// <summary>
        /// The format of the destination data
        /// </summary>
        private byte[] m_dest_format;

        /// <summary>
        /// The handle for the conversion stream
        /// </summary>
        private uint m_stream_handle;

        /// <summary>
        /// This is the stream header used to convert the data
        /// </summary>
        private AcmInterface.ACMSTREAMHEADER m_header;

        /// <summary>
        /// If true, the header from above is valid and prepared
        /// </summary>
        private bool m_header_valid;

        /// <summary>
        /// This is a list of data buffers holding converted data
        /// </summary>
        private IList<AcmDataBuffer> m_buffers;

        /// <summary>
        /// The flags for the conversion call
        /// </summary>
        private uint m_convert_flags;

        #endregion

        #region constructors
        /// <summary>
        /// A converter that converts data from an input file to an output stream
        /// </summary>
        public AcmConversion(AcmDataSource src)
        {
            m_source = src;
            m_buffers = new List<AcmDataBuffer>();
            m_header = new AcmInterface.ACMSTREAMHEADER();
            m_header_valid = false;
            // m_conv_calls = 0;
        }

        #endregion

        #region public methods
        /// <summary>
        /// Set the input and output buffer sizes for the conversion process.  If either
        /// of these values is zero, the buffer size is set by the system.
        /// </summary>
        /// <param name="inbufsize">size of the input buffer</param>
        /// <param name="outbufsize">size of the output buffer</param>
        public void SetBufferSizes(uint inbufsize, uint outbufsize)
        {    
            m_convert_input_buffer_size = inbufsize;
            m_convert_output_buffer_size = outbufsize;
        }

        /// <summary>
        /// Set the format for the input and output data.  The byte arrays
        /// are the contents of the WAVEFORMATEX or derived structure that defines
        /// the format of the data.
        /// </summary>
        /// <param name="srcfmt">the format of the source data</param>
        /// <param name="destfmt">the format of the destination data</param>
        public void SetFormats(byte[] srcfmt, byte[] destfmt)
        {
            m_source_format = srcfmt;
            m_dest_format = destfmt;
        }

        /// <summary>
        /// Close the ACM converter and free all resource associated with the converter
        /// </summary>
        /// <returns>true if the close was sucessful, false otherwise</returns>
        public bool CloseAcmConverter()
        {
            uint result;

            // If a header has been prepared, unprepare the header
            // and free the buffers
            if (m_header_valid)
            {
                m_header.cbSrcLength = m_convert_input_buffer_size;
                m_header.cbDstLength = m_convert_output_buffer_size;
                fixed (AcmInterface.ACMSTREAMHEADER* p = &m_header)
                {
                    result = AcmInterface.acmStreamUnprepareHeader(m_stream_handle, p, 0);
                }
                Marshal.FreeHGlobal(m_header.pbDst);
                Marshal.FreeHGlobal(m_header.pbSrc);
                m_header_valid = false;
            }

            // Close the stream
            result = AcmInterface.acmStreamClose(m_stream_handle, 0);
            if (result != 0)
                throw new AcmConversonException("acmStreamClose", result);

            return true;
        }
        
        /// <summary>
        /// This method opens the conversion stream. The AcmDriver object, if not null
        /// gives the driver to use to do the conversion.
        /// </summary>
        /// <param name="driver">the driver to use to perform the conversion</param>
        /// <returns>true if the converter was opened sucessfully, false otherwise</returns>
        public bool OpenAcmConverter(AcmDriver driver)
        {
            uint result;

            // Remember the driver handle
            m_driver = driver;

            if (m_source_format == null)
                return false;

            if (m_dest_format == null)
                return false;

            // Get the driver handle, if a specific driver was specified
            uint handle = 0;
            if (driver != null)
            {
                if (!driver.IsOpen)
                {
                    if (!driver.Open())
                        throw new AcmConversonException("AcmDriver.Open()", 0);
                }
                handle = driver.Handle;
            }

            fixed (byte* srcptr = &m_source_format[0])
            {
                fixed (byte* destptr = &m_dest_format[0])
                {
                    fixed (uint* hptr = &m_stream_handle)
                    {
                        // Open a conversion stream
                        result = AcmInterface.acmStreamOpen(hptr, handle, srcptr, destptr, IntPtr.Zero, 0, 0, 0x04);
                        if (result != 0)
                            return false;
                    }
                }
            }

            m_convert_flags = AcmInterface.ACM_STREAMCONVERTF_BLOCKALIGN | AcmInterface.ACM_STREAMCONVERTF_START;

            return true;
        }

        /// <summary>
        /// This function converts data until the requested amount of data is available, or until
        /// there is no more data to convert.  This function will call the data source (AcmDataSource)
        /// as much as necessary to convert the amount of data requested.
        /// </summary>
        /// <param name="data">the buffer for the converted data</param>
        /// <param name="count">the amount of data to convert</param>
        /// <returns>the amount of data converted</returns>
        public int ConvertData(IntPtr data, int count)
        {
            uint result;

            if (!m_header_valid)
                CreateStreamHeader();

            m_header.cbSrcLength = 0;
            bool eof = false;
            // BwgBeginScan();
            while (DataBuffered < count)
            {
                int dread;
                // short lmin = 0, lmax = 0;

                // Now, stream data into input buffer
                if (!eof)
                {
                    int len = (int)(m_convert_input_buffer_size - m_header.cbSrcLength);
                    dread = ReadDataIntoPtr(m_header.pbSrc, (int)m_header.cbSrcLength, len);
                }
                else
                    dread = 0;

                if (dread != 0)
                    m_header.cbSrcLength += (uint)dread;
                else
                {
                    // No additional data was read from the data file, so this is the last conversion
                    // operation we will do
                    m_convert_flags = AcmInterface.ACM_STREAMCONVERTF_END | AcmInterface.ACM_STREAMCONVERTF_BLOCKALIGN;

                    // This will mean we read no more data, but continue to convert the existing data until it is
                    // all converted
                    eof = true;
                }

                // Do a conversion of the buffer
                uint remaining = m_header.cbSrcLength;
                if (eof && remaining == 0)
                    break;

                fixed (AcmInterface.ACMSTREAMHEADER* p = &m_header)
                {
                    result = AcmInterface.acmStreamConvert(m_stream_handle, p, m_convert_flags);
                }
                if (result != 0)
                    throw new AcmConversonException("acmStreamConvert", result);

                if (m_header.cbDstLengthUsed > m_header.cbDstLength)
                    throw new AcmConversonException("acmStreamConvert(toomuch)", 0);

                m_convert_flags = AcmInterface.ACM_STREAMCONVERTF_BLOCKALIGN;

                // Copy converted data into buffer
                // BwgScanBlock(m_header.pbDst, (int)m_header.cbDstLengthUsed, ref lmin, ref lmax);
                // m_conv_calls++;
                // if (m_conv_calls == 11641)
                    // Debug.WriteLine("Here");
                // Debug.WriteLine("ConvCalls " + m_conv_calls.ToString() + ", min " + lmin.ToString() + ", max " + lmax.ToString());
                if (m_header.cbDstLengthUsed > 0)
                    MoveConvertedDataIntoQueue();

                // Copy non-converted data to head of queue
                IntPtr srcptr = new IntPtr(m_header.pbSrc.ToInt32() + m_header.cbSrcLengthUsed);
                AcmInterface.RtlMoveMemory(m_header.pbSrc, srcptr, (int)(m_header.cbSrcLength - m_header.cbSrcLengthUsed));
                m_header.cbSrcLength -= m_header.cbSrcLengthUsed;

                // If no more data was converted, or if there is no more data to convert, we are done
                if (eof && (m_header.cbSrcLength == remaining || m_header.cbSrcLength == 0))
                    break;
            }
            // BwgEndScan();

            return CopyOut(data, count);
        }

        /// <summary>
        /// Rewind the conversion source
        /// </summary>
        /// <returns>true if the rewind was successful, false otherwise</returns>
        public bool Rewind()
        {
            // m_conv_calls = 0;

            if (!m_source.Rewind())
                return false;

            if (!CloseAcmConverter())
                return false;

            if (!OpenAcmConverter(m_driver))
                return false;

            return true;
        }

        #endregion

        #region public properties
        /// <summary>
        /// Handle to the converter
        /// </summary>
        public uint Handle
        {
            get
            {
                return m_stream_handle;
            }
        }

        /// <summary>
        /// The amount of data currently buffered by the converter object.
        /// </summary>
        public int DataBuffered
        {
            get
            {
                int total = 0;
                foreach (AcmDataBuffer buf in m_buffers)
                    total += buf.DataRemaining;

                return total;
            }
        }
        #endregion

        #region private functions
        /// <summary>
        /// This function copies data that is buffered out into the given
        /// data pointer.  The amount of data to copy is given by the count
        /// parameter.
        /// </summary>
        /// <param name="data"></param>
        /// <param name="count"></param>
        /// <returns></returns>
        private int CopyOut(IntPtr data, int count)
        {
            int off = 0 ;

            while (m_buffers.Count > 0 && off < count)
            {
                AcmDataBuffer buf = m_buffers[0];
                int remain = buf.DataRemaining;

                if (remain < count - off)
                {
                    // Copy the entire buffer
                    buf.CopyTo(data, off, remain);
                    m_buffers.RemoveAt(0);
                    buf.Dispose();
                    off += remain;
                }
                else
                {
                    // We only need a partial buffer
                    int length = count - off;
                    buf.CopyTo(data, off, length);
                    off += length;
                }
            }

            return off;
        }

        /// <summary>
        /// Create the stream header for the conversion
        /// </summary>
        private void CreateStreamHeader()
        {
            uint result;
            uint insize;

            if (m_convert_input_buffer_size == 0 && m_convert_output_buffer_size != 0)
            {
                result = AcmInterface.acmStreamSize(m_stream_handle, m_convert_output_buffer_size, &insize, AcmInterface.ACM_STREAMSIZEF_DESTINATION);
                if (result != 0)
                    m_convert_input_buffer_size = m_convert_output_buffer_size;
                else
                    m_convert_input_buffer_size = insize;
            }
            else if (m_convert_input_buffer_size != 0 && m_convert_output_buffer_size == 0)
            {
                result = AcmInterface.acmStreamSize(m_stream_handle, m_convert_input_buffer_size, &insize, AcmInterface.ACM_STREAMSIZEF_SOURCE);
                if (result != 0)
                    m_convert_output_buffer_size = m_convert_input_buffer_size;
                else
                    m_convert_output_buffer_size = insize;
            }
            else if (m_convert_input_buffer_size == 0 && m_convert_output_buffer_size == 0)
            {
                m_convert_input_buffer_size = 2352 * 16;
                m_convert_output_buffer_size = 2352 * 16;
            }


            m_header.cbStruct = (uint)Marshal.SizeOf(m_header);
            m_header.cbDstLength = m_convert_output_buffer_size;
            m_header.pbDst = Marshal.AllocHGlobal((int)m_header.cbDstLength);

            m_header.cbSrcLength = m_convert_input_buffer_size;
            m_header.pbSrc = Marshal.AllocHGlobal((int)m_header.cbSrcLength);

            fixed (AcmInterface.ACMSTREAMHEADER* p = &m_header)
            {
                result = AcmInterface.acmStreamPrepareHeader(m_stream_handle, p, 0);
            }

            if (result != 0)
            {
                m_header_valid = false;
                Marshal.FreeHGlobal(m_header.pbDst);
                Marshal.FreeHGlobal(m_header.pbSrc);
                throw new AcmConversonException("acmStreamPrepareHeader", result);
            }
            else
                m_header_valid = true;
        }

        /// <summary>
        /// Move the data converted into the buffer
        /// </summary>
        private void MoveConvertedDataIntoQueue()
        {
            AcmDataBuffer buffer = new AcmDataBuffer((int)m_header.cbDstLengthUsed);
            AcmInterface.RtlMoveMemory(buffer.DataBuffer, m_header.pbDst, (int)m_header.cbDstLengthUsed);
            m_buffers.Add(buffer);
        }

        /// <summary>
        /// Read data from the data source into the buffer for conversion
        /// </summary>
        /// <param name="dest">the destination buffer for the data</param>
        /// <param name="offset">the offset in the buffer to hold the data</param>
        /// <param name="length">the length of the data to read</param>
        /// <returns>the amount of data read</returns>
        private int ReadDataIntoPtr(IntPtr dest, int offset, int length)
        {
            int amount;
            IntPtr ptr = new IntPtr(dest.ToInt32() + offset);

            if (!m_source.Read(ptr, length, out amount))
            {
                int err = Marshal.GetLastWin32Error();
                throw new Exception(err.ToString() + " - error reading data from OS file");
            }

            return amount;
        }

        #endregion
    }
}
