//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//

using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using System.Runtime.InteropServices;
using Bwg.Logging;

namespace Bwg.Audio.Acm
{
    public unsafe class AcmManager
    {
        #region private member variables

        /// <summary>
        /// The list of drivers found that support the desired format
        /// </summary>
        private AcmDriver [] m_drivers;

        /// <summary>
        /// Logger for logging messages
        /// </summary>
        private Logger m_logger;
        #endregion

        #region public constants
        public const uint CD_Sample_Rate = 44100;
        public const ushort CD_Channel_Count = 2;
        public const ushort CD_BitsPerSample = 16;
        #endregion

        #region constructor
        public AcmManager(Logger l)
        {
            string str ;

            m_logger = l;

            m_drivers = new AcmDriver[0];

            m_logger.LogMessage(new UserMessage(UserMessage.Category.Debug, 0, "AcmManager - searching for codecs"));
            FindDrivers();
            str = "AcmManager - found " + m_drivers.GetLength(0).ToString() + " codecs";
            m_logger.LogMessage(new UserMessage(UserMessage.Category.Debug, 0, str)) ;
        }
        #endregion

        #region public properties
        /// <summary>
        /// This property returns the array of drivers that match the desired format
        /// </summary>
        public AcmDriver[] Drivers
        {
            get
            {
                return m_drivers;
            }
        }
        #endregion

        #region public methods

        public unsafe byte[] GetCDQualityWaveFormat()
        {
            AcmInterface.WAVEFORMATEX dest = new AcmInterface.WAVEFORMATEX();
            dest.cbSize = 0;
            dest.nAvgBytesPerSec = CD_Sample_Rate * CD_Channel_Count * CD_BitsPerSample / 8;
            dest.nBlockAlign = (ushort)(CD_Channel_Count * CD_BitsPerSample / 8);
            dest.nChannels = CD_Channel_Count;
            dest.nSamplesPerSec = CD_Sample_Rate;
            dest.wBitsPerSample = CD_BitsPerSample;
            dest.wFormatTag = (ushort)AcmInterface.WaveFormat.WAVE_FORMAT_PCM;

            byte[] data = new byte[Marshal.SizeOf(dest)];
            IntPtr srcptr = new IntPtr(&dest);
            Marshal.Copy(srcptr, data, 0, data.GetLength(0));

            return data;
        }

        public bool CanConvert(AcmDriver drv, AcmFormatDetails src)
        {
            bool can = false ;

            foreach (AcmFormatDetails dest in drv.FormatDescriptors)
            {
                if (CanConvert(drv, src, dest))
                {
                    can = true;
                    break;
                }
            }
            return can;
        }

        public bool CanConvert(AcmDriver drv, AcmFormatDetails src, AcmFormatDetails dest)
        {
            // If we don't support it with a header writer, it does not matter that ACM can
            // convert the binary data
            if (!IsSupported(dest.Tag))
                return false;

            byte [] srcwave ;
            byte [] destwave ;

            if (src == null)
                srcwave = GetCDQualityWaveFormat();
            else
                srcwave = ConvertToByteArray(src.WaveFormatExPtr);

            destwave = ConvertToByteArray(dest.WaveFormatExPtr);

            return CanConvert(drv, srcwave, destwave);
        }

        public bool CanConvert(AcmDriver drv, byte[] src, byte[] dest)
        {
            AcmConversion conv = new AcmConversion(null);
            conv.SetBufferSizes(0, 0);
            conv.SetFormats(src, dest);

            bool b = conv.OpenAcmConverter(drv);
            if (b)
                conv.CloseAcmConverter();

            return b;
        }

        public byte[] ConvertToByteArray(IntPtr waveptr)
        {
            AcmInterface.WAVEFORMATEX fmt = new AcmInterface.WAVEFORMATEX();
            AcmInterface.WAVEFORMATEX* p = (AcmInterface.WAVEFORMATEX*)waveptr.ToPointer();
            int size = Marshal.SizeOf(fmt) + p->cbSize;
            byte[] result = new byte[size];
            Marshal.Copy(waveptr, result, 0, result.GetLength(0));

            return result;
        }

        public byte[] GetFormat(AcmFormatDesc desc)
        {
            foreach (AcmDriver drv in m_drivers)
            {
                if (drv.SupportsFormatTag(desc.Format))
                {
                    byte[] fmtout = drv.GetFormat(desc);
                    if (fmtout != null)
                        return fmtout;
                }
            }

            return null;
        }

        public bool SupportsType(AcmInterface.WaveFormat fmt)
        {
            foreach (AcmDriver drv in m_drivers)
            {
                if (drv.SupportsFormatTag(fmt))
                    return true;
            }
            return false;
        }

        public AcmDriver FindDriverById(uint id)
        {
            foreach (AcmDriver drv in m_drivers)
            {
                if (drv.DriverID == id)
                    return drv;
            }
            return null;
        }

        public void LogConverteCreate(AcmConversion c)
        {
            string str ;
            uint res;
            uint id ;

            res = AcmInterface.acmDriverID(c.Handle, &id, 0) ;
            if (res != 0)
                str = "error calling acmDriverID() to find driver associated with conversion";
            else
            {
                AcmDriver drv = FindDriverById(id);
                if (drv != null)
                    str = "created converter based on driver '" + drv.ToString() + "'";
                else
                    str = "created converter based on driver id " + id.ToString() + ", lookup failed";
            }

            m_logger.LogMessage(new UserMessage(UserMessage.Category.Debug, 0, str));
        }
        #endregion

        #region private methods
        private bool IsSupported(AcmInterface.WaveFormat fmt)
        {
            return (fmt == AcmInterface.WaveFormat.WAVE_FORMAT_MPEGLAYER3 ||
                    fmt == AcmInterface.WaveFormat.WAVE_FORMAT_PCM);
        }

        private void FindDrivers()
        {
            uint result = 0;

            try
            {
                result = AcmInterface.acmDriverEnum(new AcmInterface.AcmDriverEnumCallback(EnumDriverCallback), 0, 0);
            }
            catch (Exception)
            {
            }
        }

        private bool EnumDriverCallback(uint id, uint dwInstance, uint dwSupport)
        {
            string str;
            AcmDriver drv = new AcmDriver(m_logger, id);

            if (drv.Initialize())
            {
                Array.Resize(ref m_drivers, m_drivers.GetLength(0) + 1);
                m_drivers[m_drivers.GetLength(0) - 1] = drv;

                str = "AcmManager - found codecs '" + drv.LongName + "'";
                m_logger.LogMessage(new UserMessage(UserMessage.Category.Debug, 1, str));
            }
            else
            {
                str = "AcmManager - error initializing codecs with id '" + id.ToString() + "'";
                m_logger.LogMessage(new UserMessage(UserMessage.Category.Error, 0, str));
            }
            return true;
        }
        #endregion
    }
}
