//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//

using System;
using System.Collections.Generic;
using System.Text;
using System.Runtime.InteropServices;
using System.Diagnostics;
using System.IO;
using System.ComponentModel;
using Bwg.Audio.Acm ;
using Bwg.Logging ;

namespace Bwg.Audio
{
    public unsafe class AcmReader : AudioReader
    {
        [DllImport("BwgAudioScanner.dll")]
        internal static extern int BwgScanBlock(IntPtr buf, int count, ref short lmin, ref short lmax);

        #region private member variables

        /// <summary>
        /// This member is the data source for the data conversion process
        /// </summary>
        private AcmFileSource m_source;

        /// <summary>
        /// This is the last progress number reported
        /// </summary>
        private int m_progress;

        /// <summary>
        /// This is the total size of the file in bytes
        /// </summary>
        private long m_file_size;

        /// <summary>
        /// This is the header reader that will parse the file header and
        /// determine the format of the data and where it starts.  This header
        /// reader may also parse a set of properties from the file that includes
        /// things like the title of the song, etc.
        /// </summary>
        private AudioHeaderReader m_header_reader;

        /// <summary>
        /// This is the audio compressions manager (ACM) class that converts the audio
        /// data from out format to another.  This is basically a wrapper for the stream
        /// object created by the acmStreamOpen() function.
        /// </summary>
        private AcmConversion m_converter;

        /// <summary>
        /// This is the number of bytes of audio data in the audio file being read
        /// </summary>
        private long m_byte_count;

        #endregion

        #region protected member variables
        /// <summary>
        /// This is the ACM manager
        /// </summary>
        protected AcmManager m_acm_manager;

        /// <summary>
        /// The name of the file we are reading
        /// </summary>
        protected string m_filename;

        #endregion

        #region public properties

        public override string Name
        {
            get
            {
                return m_filename;
            }
        }

        public override int PreferredReadSize
        {
            get
            {
                // TODO - is there something better here?
                return 2352;
            }
        }

        /// <summary>
        /// This property returns the size of the audio file in bytes.  This
        /// size is the size of the raw audio data in terms of the output format.
        /// </summary>
        public override long ByteCount
        {
            get
            {
                return m_byte_count;
            }
        }

        /// <summary>
        /// This is the header reader that will parse the file header and
        /// determine the format of the data and where it starts.  This header
        /// reader may also parse a set of properties from the file that includes
        /// things like the title of the song, etc.
        /// </summary>
        public override string FileTypeString
        {
            get
            {
                return m_header_reader.FileTypeString;
            }
        }
        #endregion

        #region constructors
        internal AcmReader(AcmManager mgr, string filename, AudioHeaderReader header)
        {
            m_acm_manager = mgr;
            m_filename = filename;
            m_header_reader = header;

            m_converter = null;

            m_header_reader.ReadHeader(filename);
        }
        #endregion

        #region public member functions
        public override int Read(IntPtr buffer, int count)
        {
            Debug.Assert(m_converter != null);
            return m_converter.ConvertData(buffer, count);
        }

        public override bool Rewind()
        {
            return m_converter.Rewind();
        }

        public override bool ScanInputFile(out short tlmin, out short tlmax, out short trmin, out short trmax)
        {
            FileInfo info = new FileInfo(m_filename);
            short lmin = 0, lmax = 0;

            tlmin = 0;
            tlmax = 0;
            trmin = 0;
            trmax = 0;

            try
            {
                m_file_size = info.Length;
            }
            catch (Exception)
            {
                m_file_size = 0;
            }

            //
            // Why 2352?, why not.  Does not really matter here, we just want some kind
            // of buffer size to allow us to scan the file
            //
            int bufsize = 2352;
            IntPtr data = IntPtr.Zero;
            long datacnt;

            // The converter should be null
            if (!CreateConverter())
                return false;

            // Now open the converter
            if (!m_converter.OpenAcmConverter(null))
                return false;

            m_acm_manager.LogConverteCreate(m_converter);

            try
            {
                // Create a buffer for the data
                data = Marshal.AllocHGlobal((int)bufsize);

                // Now, read data until it is all read
                datacnt = 0;
                int count;
                do
                {
                    count = Read(data, bufsize);
                    BwgScanBlock(data, count, ref lmin, ref lmax);
                    datacnt += count;
                    CalcProgress();
                } while (count != 0);
            }
            finally
            {
                if (data != IntPtr.Zero)
                    Marshal.FreeHGlobal(data);
            }

            m_byte_count = datacnt;

            m_converter.Rewind();

            return true;
        }

        #endregion

        #region private member functions

        private byte[] GetOutputFormat()
        {
            byte[] ret = null;

            ret = m_acm_manager.GetCDQualityWaveFormat();
            fixed (void* ptr = &ret[0])
            {
                AcmInterface.WAVEFORMATEX* fmt = (AcmInterface.WAVEFORMATEX*)ptr;
                m_bits_per_channel = fmt->wBitsPerSample;
                m_number_of_channels = fmt->nChannels;
                m_sample_rate = (int)fmt->nSamplesPerSec;
            }

            return ret;
        }

        private bool CreateConverter()
        {
            m_source = new AcmFileSource();
            if (!m_source.Open(m_filename, m_header_reader.AudioDataPosition))
                return false;

            m_converter = new AcmConversion(m_source);
            m_converter.SetBufferSizes(m_header_reader.BufferSize, 0);
            m_converter.SetFormats(m_header_reader.DataFormat, GetOutputFormat());

            return true;
        }

        private void CalcProgress()
        {
            int prog;

            if (m_file_size == 0 || ScanProgressChanged == null)
                return ;

            prog = (int)((m_source.FilePosition * 100) / m_file_size);
            if (prog != m_progress)
            {
                m_progress = prog;
                ProgressChangedEventArgs args = new ProgressChangedEventArgs(m_progress, null);
                ScanProgressChanged(this, args);
            }
        }


        #endregion
    }
}
