//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//
using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Runtime.InteropServices ;
using Bwg.Audio.Acm ;
using Bwg.Utils;

namespace Bwg.Audio
{
    public class AcmWriter : AudioWriter
    {

        #region private constants
        private const int AudioConvertBlockSize = 16384 ;
        #endregion

        #region private member variables

        /// <summary>
        /// The ACM manager
        /// </summary>
        private AcmManager m_manager;

        /// <summary>
        /// The source of the data to write.  This data should be CD format
        /// data.
        /// </summary>
        private AudioReader m_reader;

        /// <summary>
        /// The driver for converting the data from CD format output to the
        /// desired file output format.
        /// </summary>
        private AcmDriver m_driver;

        /// <summary>
        /// The output format desired
        /// </summary>
        private AcmFormatDetails m_format;

        /// <summary>
        /// The converter for converting data to the desired format
        /// </summary>
        private AcmConversion m_converter;

        /// <summary>
        /// This class is responsible for writing the file header
        /// </summary>
        private AudioHeaderWriter m_header_writer;

        /// <summary>
        /// Properties associated with the file
        /// </summary>
        private IDictionary<string, object> m_props;

        #endregion

        #region constructor
        public AcmWriter(AcmManager mgr, IDictionary<string, object> props, 
                         AudioReader rdr, AcmDriver drv, AcmFormatDetails fmt)
        {
            m_manager = mgr;
            m_reader = rdr;
            m_driver = drv;
            m_format = fmt;

            m_props = props;
        }
        #endregion

        #region public properties
        public override string FileTypeString
        {
            get
            {
                string result = "Unknown" ;

                switch (m_format.Tag)
                {
                    case AcmInterface.WaveFormat.WAVE_FORMAT_MPEGLAYER3:
                        result = "MP3";
                        break;

                    case AcmInterface.WaveFormat.WAVE_FORMAT_PCM:
                        result = "WAVE";
                        break;

                    case AcmInterface.WaveFormat.WAVE_FORMAT_MSAUDIO1:
                    case AcmInterface.WaveFormat.WAVE_FORMAT_MSAUDIO2:
                        result = "WMA";
                        break;
                }
                return result;
            }
        }
        #endregion

        #region public methods
        public void SetProperty(string key, object v)
        {
            m_props[key] = v;
        }

        public override bool Close()
        {
            if (m_converter != null)
                m_converter.CloseAcmConverter();

            return true;
        }

        public override bool Write(string filename)
        {
            bool ret = true ;

            // Create the header writer.  This tells us early on if we support the
            // given output format.
            if (!CreateHeaderWriter())
                return false;

            // Create the source for the data for the converter
            AcmReaderSource src = new AcmReaderSource(m_reader) ;

            // Create a converter to converting the data from the CD quality data we
            // get to the output format desired
            m_converter = new AcmConversion(src);
            m_converter.SetBufferSizes(0, 0);
            m_converter.SetFormats(m_manager.GetCDQualityWaveFormat(), m_manager.ConvertToByteArray(m_format.WaveFormatExPtr));
            if (!m_converter.OpenAcmConverter(m_driver))
                return false;

            // Add the default extension based on the file type
            filename += "." + m_header_writer.Extension;

            FileStream strm = null;
            try
            {
                // Create the file stream
                strm = new FileStream(filename, FileMode.Create, FileAccess.ReadWrite);

                // Write the file header
                if (!m_header_writer.WriteHeader(m_props, strm))
                    return false ;

                // Close the stream and 
                strm.Close();
                strm = null;

                OsFileWriter oswr = new OsFileWriter();
                if (!oswr.Open(filename, OsFileWriter.OpenType.Append))
                    return false;

                if (!oswr.Seek(0, OsFileWriter.SeekType.End))
                    return false;

                // Convert the data and write it
                int length;
                if (!WriteData(oswr, out length))
                    return false ;

                // Close the OS file
                oswr.Close();

                // Reopen the stream file
                strm = new FileStream(filename, FileMode.Open, FileAccess.ReadWrite);

                // Patch the header if necessary, based on the amount of data written
                if (!m_header_writer.PatchHeader(m_props, strm, length))
                    return false ;
            }
            catch (Exception)
            {
                ret = false;
            }
            finally
            {
                if (strm != null)
                    strm.Close() ;
            }

            return ret;
        }
        #endregion

        #region private methods
        private bool WriteData(OsFileWriter oswr, out int length)
        {
            bool ret = true;
            IntPtr buffer = Marshal.AllocHGlobal(AudioConvertBlockSize) ;
            length = 0;

            try
            {
                while (true)
                {
                    int written ;
                    int cvted = m_converter.ConvertData(buffer, AudioConvertBlockSize);
                    if (cvted == 0)
                        break;

                    if (!oswr.WriteData(buffer, cvted, out written))
                    {
                        ret = false;
                        break;
                    }

                    length += written;
                }
            }
            catch (Exception)
            {
                ret = false;
            }
            finally
            {
                Marshal.FreeHGlobal(buffer);
            }
            return ret;
        }

        private bool CreateHeaderWriter()
        {
            bool ret = true;

            switch (m_format.Tag)
            {
                case AcmInterface.WaveFormat.WAVE_FORMAT_PCM:
                    m_header_writer = new WaveHeaderWriter(m_format);
                    break;

                case AcmInterface.WaveFormat.WAVE_FORMAT_MPEGLAYER3:
                    m_header_writer = new Mp3HeaderWriter(m_format);
                    break;

                default:
                    m_header_writer = null;
                    ret = false;
                    break;
            }

            return ret;
        }
        #endregion
    }
}
