//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//
using System;
using System.Collections.Generic;
using System.Text;
using System.ComponentModel ;
using System.Runtime.InteropServices;
using System.Diagnostics;
using Bwg.Audio.Acm ;

namespace Bwg.Audio
{
    /// <summary>
    /// This is a base class for an audio file reader.  An audio file readers is responsible for
    /// dealing with a specific audio file format and returning information about the audio file.  THe
    /// reader is also responsible for providing a CD compatible (44100 samples per second at 16 bits and two
    /// channels) audio stream when requested.
    /// </summary>
    public abstract class AudioReader
    {
        #region public properties

        /// <summary>
        /// This property gives a human readable for of this data source
        /// </summary>
        public abstract string Name
        {
            get;
        }

        /// <summary>
        /// This property gives a preferred block size in bytes for reads from 
        /// this reader.  This is useful if you are reading large blocks of data from
        /// an audio reader and want to do this is a somewhat optimal manner
        /// </summary>
        public abstract int PreferredReadSize
        {
            get;
        }

        /// <summary>
        /// This property returns a human readable describing the type of the file.  This should
        /// be something like "MP3 File" or "WAV File"
        /// </summary>
        public abstract string FileTypeString
        {
            get;
        }

        /// <summary>
        /// This property returns the number of channels in the audio stream
        /// </summary>
        public ushort NumberOfChannels
        {
            get
            {
                return m_number_of_channels;
            }
        }

        /// <summary>
        /// This property returns the number of samples per second in the audio stream
        /// </summary>
        public int SampleRate
        {
            get
            {
                return m_sample_rate;
            }
        }

        /// <summary>
        /// This property returns the number of bits per sample in the audio stream
        /// </summary>
        public ushort BitsPerChannel
        {
            get
            {
                return m_bits_per_channel;
            }
        }

        /// <summary>
        /// This returns the total number of bytes per sample for the audio stream
        /// </summary>
        public ushort BytesPerSample
        {
            get
            {
                return (ushort)(m_bits_per_channel * NumberOfChannels / 8) ;
            }
        }

        /// <summary>
        /// This property returns the size of the audio file in bytes.  This
        /// size is the size of the raw audio data in terms of the output format.
        /// </summary>
        public abstract long ByteCount
        {
            get;
        }

        #endregion

        #region protected member variables

        /// <summary>
        /// The number of channels supported by this audio file
        /// </summary>
        protected ushort m_number_of_channels;

        /// <summary>
        /// The sample rate in samples per second
        /// </summary>
        protected int m_sample_rate;

        /// <summary>
        /// The number of bits per sample
        /// </summary>
        protected ushort m_bits_per_channel;

        #endregion

        #region public member variables
        public ProgressChangedEventHandler ScanProgressChanged;
        #endregion

        #region constructor
        /// <summary>
        /// This method creates a new file reader object.
        /// </summary>
        public AudioReader()
        {
            m_number_of_channels = 0;
            m_sample_rate = 0;
            m_bits_per_channel = 0;
        }
        #endregion

        #region public member functions
        /// <summary>
        /// This function reads data from an audio file, decodes the date and stores
        /// raw audio samples into the byte array.  The data is stored as a set of
        /// 16 bit words, each of pair of which represents a single sample.
        /// </summary>
        /// <param name="data">the array that will hold the sample data</param>
        /// <returns>the number of bytes read into the array</returns>
        public abstract int Read(IntPtr buffer, int count) ;

        /// <summary>
        /// Rewind the stream to start at the beginning
        /// </summary>
        /// <returns>true if the rewind was sucessful, false otherwise</returns>
        public abstract bool Rewind() ;

        /// <summary>
        /// This function is called to scan the input file and gather any information required
        /// from the file.  It should be called before any reading
        /// </summary>
        /// <returns></returns>
        public abstract bool ScanInputFile(out short lmin, out short lmax, out short rmin, out short rmax);

        /// <summary>
        /// Convert a count of bytes to the largest number of whole samples
        /// that can be stored in the array
        /// </summary>
        /// <param name="bytes">the number of bytes</param>
        /// <returns>the number of samples</returns>
        public int BytesToSamples(int bytes)
        {
            return (bytes << 3) / BitsPerChannel / NumberOfChannels;
        }

        /// <summary>
        /// Convert a number of samples to the size of the memory area in
        /// bytes that is needed to hold the samples
        /// </summary>
        /// <param name="samples">the number of samples</param>
        /// <returns>the number of bytes</returns>
        public int SamplesToBytes(int samples)
        {
            return BitsPerChannel * NumberOfChannels * samples / 8;
        }

        #endregion
    }
}
