//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//

using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Runtime.InteropServices;
using Bwg.Audio.Acm;

namespace Bwg.Audio
{
    unsafe class WaveHeaderReader : AudioHeaderReader
    {
        #region private member variables
        /// <summary>
        /// The format of the data in the wave file
        /// </summary>
        AcmInterface.WaveFormat m_format;

        private int m_number_of_channels;
        private int m_sample_rate;
        private int m_bits_per_channel;

        private static uint m_fmt_tag = 'f' | ('m' << 8) | ('t' << 16) | (' ' << 24);

        #endregion

        #region constructor
        public WaveHeaderReader(AcmManager mgr): base(mgr)
        {
        }
        #endregion

        #region public member functions
        public override void ReadHeader(string filename)
        {
            FileStream strm = new FileStream(filename, FileMode.Open, FileAccess.Read);
            ProbeFile(filename, strm);

            m_audio_data_position = strm.Position;

            m_file_type_string = "WAV File, " + m_sample_rate + "Hz, " + m_number_of_channels + " channels, ";
            m_file_type_string += m_bits_per_channel + " bits";

            CreateFormat();

            //
            // Default buffer size for converting data is 1K samples of audio
            //
            m_buffer_size = (uint)(m_bits_per_channel * m_number_of_channels / 8 * 1024);
        }
        #endregion

        #region private member functions

        private void CreateFormat()
        {
            AcmInterface.WAVEFORMATEX fmt = new AcmInterface.WAVEFORMATEX();
            fmt.cbSize = 0;
            fmt.nChannels = (ushort)m_number_of_channels;
            fmt.nSamplesPerSec = (uint)m_sample_rate;
            fmt.wBitsPerSample = (ushort)m_bits_per_channel;
            fmt.nBlockAlign = (ushort)(fmt.nChannels * fmt.wBitsPerSample / 8);
            fmt.nAvgBytesPerSec = (uint)(fmt.nChannels * fmt.wBitsPerSample * fmt.nSamplesPerSec / 8);
            fmt.wFormatTag = (ushort)m_format;

            m_data_format = new byte[Marshal.SizeOf(fmt)];
            IntPtr ptr = new IntPtr(&fmt);
            Marshal.Copy(ptr, m_data_format, 0, m_data_format.GetLength(0));
        }

        private uint Extract4(byte[] data, int offset)
        {
            return (uint)(data[offset] | (data[offset + 1] << 8) | (data[offset + 2] << 16) | (data[offset + 3] << 24));
        }

        private ushort Extract2(byte[] data, int offset)
        {
            return (ushort)(data[offset] | (data[offset + 1] << 8));
        }

        private string ExtractString(byte[] data, int offset, int len)
        {
            string str = "";

            while (len-- > 0)
                str += (char)data[offset++];

            return str;
        }

        private bool ReadHeader(FileStream instrm, out uint id, out uint size)
        {
            byte[] data = new byte[8];

            id = 0;
            size = 0;
            if (instrm.Read(data, 0, 8) != 8)
                return false;

            id = Extract4(data, 0);
            size = Extract4(data, 4);

            return true;
        }

        private void ProbeFile(string filename, FileStream instrm)
        {
            byte[] data = new byte[20];
            bool found = false;
            uint id;
            uint size;

            // Read the RIFF file header
            if (instrm.Read(data, 0, 12) != 12)
                throw new InvalidFileFormat(filename, "unexpected end of file reading main header");

            if (ExtractString(data, 0, 4) != "RIFF")
                throw new InvalidFileFormat(filename, "this is not a wave file, expected 'RIFF' in header");

            if (ExtractString(data, 8, 4) != "WAVE")
                throw new InvalidFileFormat(filename, "this is not a wave file, expected 'WAVE' in header");

            do
            {
                // Now, read a chunk file header
                if (!ReadHeader(instrm, out id, out size))
                    throw new InvalidFileFormat(filename, "invalid file format");

                if (size < 16)
                    throw new InvalidFileFormat(filename, "first header is of invalid size");

                // Resize the array to match the length we read from the file
                if (size > data.GetLength(0))
                    Array.Resize(ref data, (int)size);

                if (instrm.Read(data, 0, (int)size) != size)
                    throw new InvalidFileFormat(filename, "unexpected end of file reading wave format record");

                if (id == m_fmt_tag)
                    found = true;

            } while (!found);

            m_format = (AcmInterface.WaveFormat)Extract2(data, 0) ;
            if (!m_acm_manager.SupportsType(m_format))
                throw new InvalidFileFormat(filename, "there is not codec to read files of type '" + m_format.ToString() + "'");

            m_number_of_channels = Extract2(data, 2);
            m_sample_rate = (int)Extract4(data, 4);
            m_bits_per_channel = Extract2(data, 14);

            // Now, find the data
            while (true)
            {
                if (!ReadHeader(instrm, out id, out size))
                    throw new InvalidFileFormat(filename, "invalid file format");

                if (id == 0x61746164)           // 'data'
                    break;

                // If this is not the data resorce, seek to the next record
                instrm.Seek(size, SeekOrigin.Current);
            }

            m_audio_data_position = instrm.Position;
        }
        #endregion
    }
}
