using System;
using System.Collections.Generic;
using System.Text;
using System.Runtime.InteropServices;
using System.ComponentModel;
using System.Diagnostics;
using System.IO ;
using Bwg.Utils;
using Bwg.Audio.WindowsMedia;

namespace Bwg.Audio
{
    /// <summary>
    /// This class is and AudioReader
    /// </summary>
    class WindowsMediaReader : AudioReader, IDisposable
    {
        #region private member variables

        /// <summary>
        /// The stream reader for WMA files
        /// </summary>
        private WmaStream m_reader;

        /// <summary>
        /// The name of the file
        /// </summary>
        private string m_filename;

        /// <summary>
        /// The progress made in the scan to date
        /// </summary>
        private int m_progress;

        /// <summary>
        /// The size of the stream in bytes
        /// </summary>
        private long m_file_size;

        /// <summary>
        /// If true, a temp file has been created.
        /// </summary>
        private bool m_temp_file_created;

        /// <summary>
        /// The name of the temp file that contains the data
        /// </summary>
        private string m_temp_file_name;

        /// <summary>
        /// The stream that contains the data ;
        /// </summary>
        private Bwg.Utils.OsFileReader m_data_stream;

        /// <summary>
        /// This is the size that the reader prefers
        /// </summary>
        private int m_preferred_read_size;
        #endregion

        #region constructor
        public WindowsMediaReader(string filename)
        {
            m_filename = filename;
            m_temp_file_created = false;

            try
            {
                m_reader = new WmaStream(filename);
            }
            catch (COMException ex)
            {
                if ((uint)(ex.ErrorCode) == 0xC00D0BBD)
                {
                    //
                    // This is a DRM failure.  We just tell the user that the file is
                    // protected and we cannot use it.
                    //
                    string name = Path.GetFileName(filename);
                    string str = "the file '" + name + "' is copy protected.  This program can only read unprotected WMA files.";
                    throw new Exception(str);
                }
                else
                    throw;
            }

            WaveFormat fmt = m_reader.Format;
            m_bits_per_channel = (ushort)fmt.wBitsPerSample ;
            m_number_of_channels = (ushort)fmt.nChannels;
            m_sample_rate = fmt.nSamplesPerSec;

            m_preferred_read_size = m_reader.SampleSize;
            m_data_stream = null;
            m_temp_file_name = string.Empty;

            if (m_bits_per_channel != 16)
                throw new Exception("Currently the windows media reader only supports output of 16 bits per channel");

            if (m_number_of_channels != 2)
                throw new Exception("Currently the windows media reader only supports stereo output");

            if (m_sample_rate != 44100)
                throw new Exception("Current the windows media reader only supports CD compatible output (44100 Hz)") ;
        }

        ~WindowsMediaReader()
        {
            Dispose(false);
        }
        #endregion

        #region public methods

        private void Dispose(bool disposing)
        {
            if (m_reader != null)
            {
                m_reader.Dispose();
                m_reader = null;
            }

            if (m_temp_file_created)
            {
                m_data_stream.Close();
                File.Delete(m_temp_file_name);
                m_temp_file_created = false;
            }
        }

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        public override long ByteCount
        {
            get 
            {
                if (m_file_size == 0 && m_reader != null)
                    m_file_size = m_reader.Length;

                return m_file_size ;
            }
        }

        public override string FileTypeString
        {
            get 
            {
                return "Windows Media Audio (wma)";
            }
        }

        public override string Name
        {
            get 
            {
                return m_filename;
            }
        }

        public override int PreferredReadSize
        {
            get 
            {
                return m_preferred_read_size ;
            }
        }

        public override int Read(IntPtr buffer, int count)
        {
            Debug.Assert(m_data_stream != null);

            int amountread;
            m_data_stream.ReadData(buffer, count, out amountread);
            return amountread;
        }

        public override bool Rewind()
        {
            m_reader.Seek(0, System.IO.SeekOrigin.Begin);
            return true;
        }

        public override bool ScanInputFile(out short lmin, out short lmax, out short rmin, out short rmax)
        {
            short lvalue;
            short rvalue;
            byte[] buffer = new byte[m_reader.SampleSize];
            long size = ByteCount;
            long sofar = 0;

            lmin = short.MaxValue;
            lmax = short.MinValue;
            rmin = short.MaxValue;
            rmax = short.MinValue;

            m_temp_file_name = Path.GetTempFileName();
            OsFileWriter writer = new OsFileWriter();
            if (!writer.Open(m_temp_file_name, OsFileWriter.OpenType.Create))
                return false;

            m_temp_file_created = true;

            m_progress = 0;


            m_reader.Seek(0, System.IO.SeekOrigin.Begin);

            int offset;
            while (true)
            {
                int count = m_reader.Read(buffer, 0, buffer.GetLength(0));
                if (count == 0)
                    break;

                int amountwritten;
                writer.WriteData(buffer, count, out amountwritten);
                if (amountwritten != count)
                    return false;

                if ((count % BytesPerSample) != 0)
                    throw new Exception("internal error - the stream did not return a proper length of data");

                offset = 0 ;

                while (offset < count)
                {
                    //
                    // TODO: this assumes stereo, 16 bit samples
                    //
                    lvalue = (short)(buffer[offset + 0] | (buffer[offset + 1] << 8));
                    rvalue = (short)(buffer[offset + 2] | (buffer[offset + 3] << 8));

                    if (lvalue < lmin)
                        lmin = lvalue;
                    if (lvalue > lmax)
                        lmax = lvalue;
                    if (rvalue < rmin)
                        rmin = rvalue;
                    if (rvalue > rmax)
                        rmax = rvalue;

                    offset += BytesPerSample;
                }
                sofar += count;
                CalcProgress(sofar, size);
            }

            m_reader.Close();
            m_reader = null;

            writer.Close();
            m_data_stream = new OsFileReader();
            if (!m_data_stream.Open(m_temp_file_name))
                return false;

            return true;
        }

        #endregion

        #region private methods
        private void CalcProgress(long sofar, long size)
        {
            int prog;

            if (ScanProgressChanged == null)
                return;

            prog = (int)((sofar * 100) / size);
            if (prog != m_progress)
            {
                m_progress = prog;
                ProgressChangedEventArgs args = new ProgressChangedEventArgs(m_progress, null);
                ScanProgressChanged(this, args);
            }
        }
        #endregion
    }
}
