//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//
using System;
using System.Collections.Generic;
using System.Text;
using Bwg.Scsi;

namespace Bwg.Burner
{
    /// <summary>
    /// This operations structure provides operations for the
    /// CD/R media type
    /// 
    /// This class support capacity operations, burn operations are
    /// implemented in the common base class for CD-R and CD-RW.  Erase
    /// is not supported for CD-R (for obvious reasons).
    /// 
    /// </summary>
    public class CDRDiskOperations : CDR_RWDiskOperations
    {
        /// <summary>
        /// This constructors just calls the base class with the drive object.
        /// </summary>
        /// <param name="d"></param>
        public CDRDiskOperations(Drive d) : base(d)
        {
        }

        /// <summary>
        /// This property returns the capacity of the disk in sectors
        /// </summary>
        public override long CapacityInSectors
        {
            get
            {
                DiscInformation info;

                if (MyDevice.ReadDiscInformation(out info) != Device.CommandStatus.Success)
                    return Drive.D74minDiskCapacity;

                return info.LastPossibleLeadoutStartAddress.SectorCount;
            }
        }

        /// <summary>
        /// The total writable capacity of the disk
        /// </summary>
        public override long WritableCapacityInSectors
        {
            get
            {
                return CapacityInSectors;
            }
        }

        /// <summary>
        /// The remaining writable capacity of this disk.  If the CD-R has been finalized, then the 
        /// remaining capacity is zero, otherwise, we use the next writable address of the invisible
        /// track to find the capacity remaining.
        /// </summary>
        public override long RemainingCapacityInWritableSectors
        {
            get
            {
                DiscInformation info;

                if (MyDevice.ReadDiscInformation(out info) != Device.CommandStatus.Success)
                    return 0;

                if (info.DiscStatus == DiscInformation.DiscStatusType.FinalizedDisc ||
                    info.DiscStatus == DiscInformation.DiscStatusType.Other)
                    return 0;

                // To find the remaining capacity, find the next writable address of the
                // invisible track
                TrackInformation tinfo;
                if (MyDevice.ReadTrackInformation(Device.ReadTrackType.TrackNumber, Device.InvisibleTrack, out tinfo) != Device.CommandStatus.Success)
                    return 0;

                return WritableCapacityInSectors - tinfo.NextWritableAddress;
            }
        }

        /// <summary>
        /// This method returns true if additional data can be added to the media.
        /// </summary>
        public override bool IsBurnSupported
        {
            get
            {
                DiscInformation info;

                if (MyDevice.ReadDiscInformation(out info) == Device.CommandStatus.Success &&
                    (info.DiscStatus == DiscInformation.DiscStatusType.EmptyDisc || 
                    info.DiscStatus == DiscInformation.DiscStatusType.IncompleteDisc))
                    return true ;

                return false ;
            }
        }

        /// <summary>
        /// Always returns false as CD-R media cannot be erased.
        /// </summary>
        public override bool IsEraseSupported
        {
            get
            {
                return false;
            }
        }

        /// <summary>
        /// Always returns true as multi-session is always supported on CD-R media.
        /// </summary>
        public override bool IsMultiSessionSupported
        {
            get
            {
                return true;
            }
        }

        /// <summary>
        /// This property returns true if the drive supported session at once recording
        /// </summary>
        public override bool IsSessionAtOnceSupported
        {
            get
            {
                return m_drive.GetFeatureState(Feature.FeatureType.CDSessionAtOnce) == Drive.FeatureState.PresentAndCurrent;
            }
        }

        /// <summary>
        /// This property returns true if the drive supports track at once recording
        /// </summary>
        public override bool IsTrackAtOnceSupported
        {
            get
            {
                return m_drive.GetFeatureState(Feature.FeatureType.CDTrackAtOnce) == Drive.FeatureState.PresentAndCurrent; ;
            }
        }

        /// <summary>
        /// Erase operation for the CDR media.  This operation is invalid.
        /// </summary>
        /// <param name="et">erase type</param>
        /// <returns>always returns an error as this is not supported</returns>
        public override DiskOperationError Erase(Drive.EraseType et)
        {
            return NotSupported("erase");
        }
    }
}
