//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//
using System;
using System.Collections.Generic;
using System.Text;
using Bwg.Scsi;
using Bwg.Logging;

namespace Bwg.Burner
{
    class CDRWDiskOperations : CDR_RWDiskOperations
    {
        public CDRWDiskOperations(Drive d) : base(d)
        {
        }

        /// <summary>
        /// This property returns the capacity of the disk in sectors
        /// </summary>
        public override long CapacityInSectors
        {
            get
            {
                AtipInfo info;

                if (MyDevice.ReadAtip(out info) != Device.CommandStatus.Success)
                    return Drive.D74minDiskCapacity;

                return info.LastPossibleStartOfLeadout.SectorCount;
            }
        }

        public override long WritableCapacityInSectors
        {
            get
            {
                return CapacityInSectors;
            }
        }

        public override long RemainingCapacityInWritableSectors
        {
            get
            {
                DiscInformation info;

                if (MyDevice.ReadDiscInformation(out info) != Device.CommandStatus.Success)
                    return 0;

                if (info.DiscStatus == DiscInformation.DiscStatusType.FinalizedDisc ||
                    info.DiscStatus == DiscInformation.DiscStatusType.Other)
                    return 0;

                // To find the remaining capacity, find the next writable address of the
                // invisible track
                TrackInformation tinfo;
                if (MyDevice.ReadTrackInformation(Device.ReadTrackType.TrackNumber, Device.InvisibleTrack, out tinfo) != Device.CommandStatus.Success)
                    return 0;

                return WritableCapacityInSectors - tinfo.NextWritableAddress;
            }
        }

        public override bool IsBurnSupported
        {
            get
            {
                return true;
            }
        }

        public override bool IsEraseSupported
        {
            get
            {
                return true;
            }
        }

        public override bool IsMultiSessionSupported
        {
            get
            {
                return true;
            }
        }

        public override bool IsTrackAtOnceSupported
        {
            get
            {
                return m_drive.GetFeatureState(Feature.FeatureType.CDTrackAtOnce) == Drive.FeatureState.PresentAndCurrent;
            }
        }

        public override bool IsSessionAtOnceSupported
        {
            get
            {
                return m_drive.GetFeatureState(Feature.FeatureType.CDSessionAtOnce) == Drive.FeatureState.PresentAndCurrent;
            }
        }

        public override DiskOperationError Erase(Drive.EraseType et)
        {
            DiskOperationError status;

            if (MyLogger != null)
            {
                string args = et.ToString();
                MyLogger.LogMessage(new UserMessage(UserMessage.Category.Debug, 0, "calling Bwg.Burner.Drive.Erase(" + args + ")"));
            }

            Device.CommandStatus st;

            status = m_drive.MakeDriveReady();
            if (status != null)
                return status;

            DiscInformation info;
            st = MyDevice.ReadDiscInformation(out info);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            // This should really never happen, but just in case
            if (!info.Erasable)
                return new BurnerError("the media in the drive is not erasable");

            //
            // It is OK for OPC to fail on blank media
            //
            MyDevice.LogSenseState = false;
            st = MyDevice.SendOpcInformation(false, null);
            MyDevice.LogSenseState = true;

            int cnt = 0;
            while (true)
            {
                if (et == Drive.EraseType.Full)
                    st = MyDevice.Blank(true, Device.BlankType.FullDisc, 0);
                else
                    st = MyDevice.Blank(true, Device.BlankType.MinimalDisc, 0);

                if (st == Device.CommandStatus.Success)
                    break;

                if (st == Device.CommandStatus.DeviceFailed && MyDevice.GetSenseAsc() == 4 && MyDevice.GetSenseAscq() == 8 && cnt++ < 8)
                {
                    MyDevice.CloseTrackSession(false, Device.CloseTrackSessionType.CloseSession, 0);
                    continue;
                }

                return CreateErrorObject(st);
            }

            st = m_drive.WaitForDeviceReady(true);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            st = MyDevice.PreventAllowMediumRemoval(Device.PreventAllow.Allow);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            st = MyDevice.ReadDiscInformation(out info);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st);

            return null;
        }
    }
}
