using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;

namespace Bwg.Burner
{
    /// <summary>
    /// This class represents ghe genre for a song or disk.  Genre consists of a two digit
    /// genre code followed by an optional string.
    /// </summary>
    [Serializable()]
    public class CDTextDataGenre : CDTextData
    {
        #region private member variables

        /// <summary>
        /// The two byte genre code
        /// </summary>
        byte[] m_code;

        /// <summary>
        /// The text of the CD text data field
        /// </summary>
        string m_data;

        /// <summary>
        /// The encoding to use for the text, null means ASCII
        /// </summary>
        Encoding m_encoding;
        #endregion

        #region constructor
        /// <summary>
        /// Create a data entry value based on a string
        /// </summary>
        /// <param name="code">the two byte code for genre</param>
        /// <param name="str">string value</param>
        /// <param name="enc">encoding to be used, null means ASCII</param>
        public CDTextDataGenre(byte[] code, string str, Encoding enc)
        {
            m_data = str;
            m_encoding = enc;
            m_code = new byte[2];
            Debug.Assert(code.GetLength(0) == 2);
            Array.Copy(code, m_code, 2);
        }

        /// <summary>
        /// Create a new blank genre data object
        /// </summary>
        public CDTextDataGenre()
        {
            m_data = string.Empty;
            m_code = new byte[2];
            m_encoding = null;
        }
        #endregion

        #region public properties
        /// <summary>
        /// The text that is the value of the data associated with a cd text field.
        /// </summary>
        public string Text
        {
            get
            {
                return m_data;
            }
        }

        /// <summary>
        /// This property returns the array of bytes that represents this string field
        /// </summary>
        public override byte[] Data
        {
            get
            {
                Encoding enc = m_encoding;
                if (enc == null)
                    enc = Encoding.ASCII;

                byte [] strbytes = enc.GetBytes(m_data);
                byte[] result = new byte[strbytes.GetLength(0) + 2];
                Array.Copy(m_code, result, 2);
                Array.Copy(strbytes, 0, result, 2, strbytes.GetLength(0));

                return result;
            }
        }

        /// <summary>
        /// This property returns the size of the data in bytes
        /// </summary>
        public override int DataLength
        {
            get
            {
                Encoding enc = m_encoding;
                if (enc == null)
                    enc = Encoding.ASCII;

                return enc.GetByteCount(m_data) + 2;
            }
        }
        #endregion

        #region Public Methods
        /// <summary>
        /// Returns a string representation of the data value
        /// </summary>
        /// <returns>returns the string</returns>
        public override string ToString()
        {
            string str;

            str = "0x" + m_code[0].ToString("X2") + m_code[1].ToString("X2");
            if (m_data != string.Empty)
                str += " , \"" + m_data + "\"";

            return str;
        }

        /// <summary>
        /// Add a set of data bytes to this data element
        /// </summary>
        /// <param name="pos">the position of the data in the stream</param>
        /// <param name="data">the data values</param>
        public override void AddData(byte pos, byte[] data)
        {
            Encoding enc = m_encoding;
            if (enc == null)
                enc = Encoding.ASCII;

            if (pos == 0)
            {
                m_code[0] = data[0];
                m_code[1] = data[1];


                m_data = enc.GetString(data, 2, data.GetLength(0) - 2);
            }
            else
            {
                m_data += enc.GetString(data);
            }
        }
        #endregion
    }
}
