using System;
using System.Collections.Generic;
using System.Text;
using System.Runtime.InteropServices;
using System.Diagnostics;
using Bwg.Scsi;

namespace Bwg.Burner
{
    class CDTextDataSource : DataSource
    {
        #region private variables
        /// <summary>
        /// The image that contains the CD Text to be burned onto the disk
        /// </summary>
        private DiskBurnImage m_image;

        /// <summary>
        /// The start sector for the CD text data
        /// </summary>
        private long m_start_sector;

        /// <summary>
        /// The raw encoded data to be sent to the drive.  It was not been mapped to the
        /// bits required.
        /// </summary>
        private byte[] m_data;
        #endregion

        public CDTextDataSource(DiskBurnImage image, long start)
        {
            m_image = image;
            m_start_sector = start;
        }

        public override long SendData(Bwg.Scsi.WriteBufferPool pool, long start_sector, int sector_size, long lba)
        {
            long sent = 0;
            int offset = 0;

            if (m_data == null)
                EncodeBuffers();

            Debug.Assert(m_data != null);

            long seccount = -150 - m_start_sector;
            Debug.Assert(seccount < int.MaxValue) ;

            while (seccount > 0)
            {
                WriteBuffer buf = pool.GetFreeBuffer() ;

                int todo = (int)seccount ;
                if (buf.BufferSize < todo * sector_size)
                    todo = buf.BufferSize / sector_size ;

                buf.Zero() ;

                //
                // Copy data to the buffer
                //
                offset = CopyDataToPoolBuffer(offset, buf.BufferPtr, todo * sector_size);

                //
                // Set the parameters for this block of data
                //
                buf.DataSize = todo ;
                buf.SectorSize = sector_size ;              // Should be 96 bytes per sector
                buf.LogicalBlockAddress = lba ;             // The target logical block address
                buf.SourceString = "CDText data source";

                // Send the buffer of data to the burner
                pool.SendBufferToDevice(buf);

                seccount-= todo ;
                lba += todo;
                sent += todo;
            }

            return sent;
        }

        #region public properties

        public override long SectorCount
        {
            get
            {
                return -150 - m_start_sector;
            }
        }

        public override string Name
        {
            get
            {
                return "CDText Leadin";
            }
        }
        #endregion

        #region private methods

        private int CopyDataToPoolBuffer(int offset, IntPtr ptr, int size)
        {
            int buffer_offset = 0;

            while (buffer_offset < size)
            {
                int tocopy = m_data.GetLength(0) - offset;
                if (buffer_offset + tocopy > size)
                    tocopy = size - buffer_offset;

                Marshal.Copy(m_data, offset, ptr, tocopy);

                offset += tocopy;
                buffer_offset += tocopy;

                if (offset == m_data.GetLength(0))
                    offset = 0;
            }

            return offset;
        }

        private void EncodeBuffers()
        {
            //
            // Create the encoder.  This converts the CD text stored in the burn image into
            // a set of 18 byte text packs.
            //
            CDTextEncoderDecoder enc = new CDTextEncoderDecoder(m_image.CDTextList, m_image.TrackCount);

            //
            // Set the languages to use for CD text
            //
            for (int i = 0; i < m_image.CDTextLanguageCount; i++)
                enc.AddLanguage(m_image.GetCDTextLanguage(i));

            //
            // Set the CDText for the tracks
            //
            int tno = 1;
            foreach (TrackBurnImage ti in m_image)
                enc.SetTrackCDText(tno++, ti.CDTextList);

            //
            // Encode the CD text into a set of CDTextBlocks
            //
            IList<CDTextBlock> blocks = enc.Encode();
            Debug.Assert(blocks.Count < 256);

            //
            // Now, convert the blocks of 18 bytes into an array of bytes to write to the
            // disk.
            //
            EncodeDataToSubchannels(blocks);
        }

        private void EncodeDataToSubchannels(IList<CDTextBlock> blocks)
        {
            int n = blocks.Count * 18 ;
            int k = blocks.Count % 4;

            if (k == 0)
                n *= 1 ;
            else if (k == 2)
                n *= 2;
            else
                n *= 4;

            n = n * 4 / 3;

            m_data = new byte[n];

            int i = 0;
            int j = 0;

            while (j < n)
            {
                int blkno = (i / 18) % blocks.Count;
                int offset = i % 18 ;
                CDTextBlock blk = blocks[blkno];
                EightToSix(blk.GetPacketByte(offset), blk.GetPacketByte(offset + 1), blk.GetPacketByte(offset + 2), m_data, j);

                i += 3;
                j += 4;
            }
        }

        private void EightToSix(byte b1, byte b2, byte b3, byte [] buf, int offset)
        {
            buf[offset] = (byte)((b1 >> 2) & 0x3f);
            buf[offset + 1] = (byte)(((b1 & 0x03) << 4) | ((b2 >> 4) & 0x0f));
            buf[offset + 2] = (byte)(((b2 & 0x0f) << 2) | ((b3 >> 6) & 0x03));
            buf[offset + 3] = (byte)(b3 & 0x3f) ;
        }
        #endregion
    }
}
