//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//

using System;
using System.Collections.Generic;
using System.Text;
using System.Threading ;
using System.Diagnostics;
using Bwg.Scsi ;
using Bwg.Logging;

namespace Bwg.Burner
{
    class DataGenerator
    {
        #region public properties
        /// <summary>
        /// This gives the status of the data generation process
        /// </summary>
        public bool Status
        {
            get
            {
                return m_status;
            }
        }

        /// <summary>
        /// This is an error message if the status property is false
        /// </summary>
        public Exception Exception
        {
            get
            {
                return m_ex;
            }
        }

        /// <summary>
        /// Return the number of data source contained in the generator
        /// </summary>
        public int Count
        {
            get
            {
                return m_streams.Count;
            }
        }

        #endregion

        #region private memory variables

        /// <summary>
        /// This is true until the first page has been delivered
        /// </summary>
        private bool m_first_page;

        /// <summary>
        /// The status of the data generation
        /// </summary>
        private bool m_status;

        /// <summary>
        /// The message associated with any generation error
        /// </summary>
        private Exception m_ex;

        /// <summary>
        /// The thread generating the data for the burn operation
        /// </summary>
        private Thread m_generator_thread;

        /// <summary>
        /// The pool that this class will feed data into, note the burner will take the
        /// data out of the pool
        /// </summary>
        private WriteBufferPool m_pool;

        /// <summary>
        /// The list of data streams that will send data
        /// </summary>
        private IList<StreamElement> m_streams;

        /// <summary>
        /// The logger for logging message
        /// </summary>
        private Logger m_logger;

        /// <summary>
        /// The threshold required to start
        /// </summary>
        private int m_threshold;

        #endregion

        #region constructor
        public DataGenerator(Logger l, int count, int size, int mult, int threshold, int secsize)
        {
            Debug.Assert(mult == 0 || secsize > 0);

            m_logger = l;
            m_streams = new List<StreamElement>();
            m_pool = new WriteBufferPool(count, size, mult, secsize);
            m_first_page = true;
            m_threshold = threshold;
        }
        #endregion

        #region public methods

        /// <summary>
        /// Add a source to the data generator
        /// </summary>
        /// <param name="n">a human readable name for the source</param>
        /// <param name="src"></param>
        /// <param name="start"></param>
        /// <param name="length"></param>
        public void AddSource(string n, TrackBurnImage src, long start, long length)
        {
            StreamElement elem = new StreamElement(n, src, start, length);
            m_streams.Add(elem);
        }

        /// <summary>
        /// Return the stream element given the index.
        /// </summary>
        /// <param name="index"></param>
        /// <returns></returns>
        public StreamElement GetSource(int index)
        {
            return m_streams[index];
        }

        /// <summary>
        /// This method creates a new thread to start the data generation process and generate
        /// pages of data for the burner.
        /// </summary>
        /// <returns></returns>
        public bool StartGeneratorThread()
        {
            m_generator_thread = new Thread(new ThreadStart(DataGenerationProc));
            m_generator_thread.Priority = ThreadPriority.Normal;
            m_generator_thread.Start() ;

            return true;
        }

        /// <summary>
        /// This method stops (aborts) the data generation thread.
        /// </summary>
        /// <returns></returns>
        public bool StopGeneratorThread()
        {
            if (m_generator_thread != null)
                m_generator_thread.Abort();

            return true;
        }

        /// <summary>
        /// This method returns the next data buffer that contain data.
        /// </summary>
        /// <returns>pointer to a buffer</returns>
        public WriteBuffer GetNextDataBuffer()
        {
            while (m_first_page && !m_pool.EndOfData && m_pool.PercentUsed < m_threshold)
            {
                /* Empty Loop */
            }

            m_first_page = false;
            return m_pool.GetNextDataBuffer();
        }

        /// <summary>
        /// This method frees a previously acquired buffer
        /// </summary>
        /// <param name="buf"></param>
        public void FreeWriteBuffer(WriteBuffer buf)
        {
            m_pool.FreeWriteBuffer(buf);
        }
        #endregion

        #region private methods
        private void SendZeros(string name, long count, int sector_size, long lba)
        {
            while (count > 0)
            {
                long todo = count;

                WriteBuffer buf = m_pool.GetFreeBuffer();
                if (buf.BufferSize < count * sector_size)
                    todo = buf.BufferSize / sector_size;

                //
                // Nothing here should exceed the size of a 32 bit integer.  However, we include
                // checks because any count of sectors is represented as a 64 bit number in anticipation
                // of larger and larger disks.
                //
                Debug.Assert(todo < Int32.MaxValue);

                buf.Zero();
                buf.DataSize = (int)todo;
                buf.SectorSize = sector_size;
                buf.LogicalBlockAddress = lba;
                lba = long.MaxValue;
                buf.SourceString = "DataGenerator:SendZeros():" + name;
                count -= todo;
                m_pool.SendBufferToDevice(buf);
            }
        }

        private void DataGenerationProc()
        {
            string str;
            long currentpos = m_streams[0].StartSector;
            m_status = true;

            if (m_logger != null)
            {
                str = "Beginnging data generation process";
                m_logger.LogMessage(new UserMessage(UserMessage.Category.Debug, 1, str));
            }

            for (int i = 0; i < m_streams.Count; i++)
            {
                long sent = 0;
                long tmp;

                StreamElement elem = m_streams[i];
                long lba = elem.StartSector;

                currentpos = lba;

                // First, send the pregap if there is one
                if (elem.Source.PreTrackGap > 0)
                {
                    if (m_logger != null)
                    {
                        str = "Track '" + elem.Source.Name + "' sending pregap" ;
                        str += ", location = " + currentpos.ToString();
                        str += ", size = " + elem.Source.PreTrackGap.ToString();
                        m_logger.LogMessage(new UserMessage(UserMessage.Category.Debug, 1, str));
                    }
                    Debug.Write("Source '" + elem.Name + "' sending pregap of " + elem.Source.PreTrackGap.ToString());
                    Debug.Write(" at sector " + currentpos.ToString());
                    Debug.WriteLine("");

                    SendZeros("PreTrack Gap", elem.Source.PreTrackGap, elem.Source.SectorSize, lba);
                    lba = long.MaxValue;
                    sent += elem.Source.PreTrackGap;
                    currentpos += elem.Source.PreTrackGap;
                }

                // Now send the data source data
                try
                {
                    if (m_logger != null)
                    {
                        str = "Track '" + elem.Source.Name + "' sending data";
                        str += ", location = " + currentpos.ToString();
                        str += ", size = " + elem.Source.DataToSend.ToString();
                        m_logger.LogMessage(new UserMessage(UserMessage.Category.Debug, 1, str));
                    }

                    Debug.Write("Source '" + elem.Name + "' sending maindata of " + elem.Source.Data.SectorCount.ToString());
                    Debug.Write(" at sector " + currentpos.ToString());
                    Debug.WriteLine("");

                    tmp = elem.Source.Data.SendData(m_pool, elem.StartSector + sent, elem.Source.SectorSize, lba);
                    Debug.WriteLine("Sent " + tmp.ToString() + " sectors of data");
                    sent += tmp;
                    currentpos += tmp;
                }
                catch (ThreadAbortException ex)
                {
                    // If the user asked to abort the process, we handle this at a
                    // higher level
                    throw ex;
                }
                catch (Exception ex)
                {
                    str = "Track '" + elem.Source.Name + "' exception while sending data";
                    str += "(" + ex.Message + ")";
                    str += ", location = " + currentpos.ToString();
                    m_logger.LogMessage(new UserMessage(UserMessage.Category.Debug, 1, str));
                    Debug.WriteLine("Exception: " + str);

                    m_status = false;
                    m_ex = ex;
                }

                Debug.Assert(sent <= elem.Source.TotalLengthSectors);

                if (sent < elem.Length)
                {
                    int tosend = (int)(elem.Length - sent);

                    Debug.Write("Source '" + elem.Name + "' sending maindata of " + tosend.ToString());
                    Debug.Write(" at sector " + currentpos.ToString());
                    Debug.WriteLine("");

                    if (m_logger != null)
                    {
                        str = "Track '" + elem.Source.Name + "' sending post data zeros";
                        str += ", location = " + currentpos.ToString();
                        str += ", size = " + tosend.ToString();
                        m_logger.LogMessage(new UserMessage(UserMessage.Category.Debug, 1, str));
                    }
                    Debug.WriteLine("Sending " + tosend.ToString() + " sectors of zero data");
                    SendZeros("End of tracking padding", tosend, elem.Source.SectorSize, long.MaxValue);
                    sent += tosend;
                    currentpos += tosend;
                }

                if (m_logger != null)
                {
                    str = "Stream element " + elem.Name + " produced " + sent.ToString() + " pages";
                    m_logger.LogMessage(new UserMessage(UserMessage.Category.Debug, 0, str)) ;
                }
            }

            m_pool.EndOfData = true;
        }
        #endregion
    }
}
