//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//

using System;
using System.Collections ;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using Bwg.Scsi;
using Bwg.Logging;

namespace Bwg.Burner
{
    /// <summary>
    /// This class represents a disk to be burned.  A disk is a set of information about the
    /// disk in general as well as a set of tracks to be burned to the disk
    /// </summary>
    public class DiskBurnImage : IEnumerable<TrackBurnImage>, IEnumerable
    {
        #region public types
        /// <summary>
        /// This indicates to the disc, the type of target burner
        /// </summary>
        public enum DiskTargetType
        {
            /// <summary>
            /// The target is a DVD disk, no pregaps necessary, or a CD with no GAPS required
            /// </summary>
            DVD,

            /// <summary>
            /// The target is a CD, written as a complete session at once
            /// </summary>
            CD_SAO,

            /// <summary>
            /// The target is a CD, written a track at once
            /// </summary>
            CD_TAO,
        } ;
        #endregion

        #region private types

        private const byte LeadOutTrackNumber = 0xAA;

        [FlagsAttribute]
        enum TrackControl : byte
        {
            PreEmphasis = 0x10,
            DigitalCopyPermitted = 0x20,
            DataTrack = 0x40,
            FourChannelAudio = 0x80,
        }

        [FlagsAttribute]
        enum TrackAdr : byte
        {
            StartTime = 0x01,
            CatalogCode = 0x02,
            ISRCCode = 0x03
        } ;

        enum TrackDataForm : byte
        {
            AudioHost = 0x00,           // 2352 bytes per sector, data provided by the host
            AudioDrive = 0x01,          // 2352 bytes per sector, zeros provided by the drive
            DataHost = 0x10,            // 2048 bytes per sector, data provided by the host
            DataDrive = 0x14,           // 2048 bytes per sector, zeros provided by the drive
        } ;
        #endregion

        #region private member variables

        /// <summary>
        /// The target media for this disk
        /// </summary>
        private DiskTargetType m_target;

        /// <summary>
        /// The list of tracks on this disk.  This contains only user visible tracks,
        /// and does not contains leadin or leadout tracks.
        /// </summary>
        private IList<TrackBurnImage> m_tracks;

        /// <summary>
        /// This is the starting sector on the disk.  This value is -150 for CD based media
        /// and zero for DVD media
        /// </summary>
        private long m_start_sector;

        /// <summary>
        /// This is the number of the first track in this disk image.  For sessions other than the
        /// first session, the start track will be greater than zero
        /// </summary>
        private ushort m_start_track;

        /// <summary>
        /// The logger for logging messages
        /// </summary>
        private Logger m_logger;

        /// <summary>
        /// If true, we generate a two second gap before any audio tracks
        /// </summary>
        private bool m_auto_generate_audio_pregap;

        /// <summary>
        /// The media control number for this disk
        /// </summary>
        private byte[] m_mcn;

        /// <summary>
        /// The list of CD text fields attached to this disk
        /// </summary>
        private IList<CDTextField> m_cdtext;

        /// <summary>
        /// The languages to be supported by the cd text.
        /// </summary>
        private IList<CDTextEncoderDecoder.LanguageType> m_languages;

        /// <summary>
        /// The data source for the cd text
        /// </summary>
        private CDTextDataSource m_cdtext_source;

        #endregion

        #region public properties

        /// <summary>
        /// This property returns true if the disk, or any of its tracks have CDText
        /// attached.
        /// </summary>
        public bool HasCDText
        {
            get
            {
                if (m_cdtext.Count > 0)
                    return true;

                foreach (TrackBurnImage ti in m_tracks)
                {
                    if (ti.Type == TrackBurnImage.TrackType.Audio && ti.CDTextList.Count > 0)
                        return true;
                }
                return false;
            }
        }

        /// <summary>
        /// Returns the list of CD Text attached to this disc
        /// </summary>
        public IList<CDTextField> CDTextList
        {
            get
            {
                return m_cdtext;
            }
        }

        /// <summary>
        /// This property controls the auto generation of audio pregaps
        /// </summary>
        public bool AutoGenerateAudioPregap
        {
            get
            {
                return m_auto_generate_audio_pregap;
            }
            set
            {
                m_auto_generate_audio_pregap = value;
            }
        }

        /// <summary>
        /// This property gives the track media type for this disk.  Setting this type
        /// adjusts the start sector and track pre-track gaps for the disk.
        /// </summary>
        public DiskTargetType TargetMedia
        {
            get
            {
                return m_target;
            }
            set
            {
                m_target = value;
                UpdateTracks();
            }
        }

        /// <summary>
        /// This method returns true if the target is a CD (and not DVD)
        /// </summary>
        public bool IsTargetCD
        {
            get
            {
                return m_target == DiskTargetType.CD_SAO || m_target == DiskTargetType.CD_TAO;
            }
        }

        /// <summary>
        /// Returns TRUE if this disk contains any audio tracks
        /// </summary>
        public bool ContainsAudio
        {
            get
            {
                foreach (TrackBurnImage t in m_tracks)
                {
                    if (t.Type == TrackBurnImage.TrackType.Audio)
                        return true;
                }
                return false;
            }
        }

        /// <summary>
        /// This property sets or retreives the starting track for this disk image
        /// </summary>
        public ushort StartTrack
        {
            get
            {
                return m_start_track;
            }
            set
            {
                m_start_track = value;
            }
        }

        /// <summary>
        /// The start sector for the disk, either -150 or zero
        /// </summary>
        public long StartSector
        {
            get
            {
                return m_start_sector;
            }
            set
            {
                m_start_sector = value;
            }
        }

        /// <summary>
        /// This property returns the number of tracks in the disk
        /// </summary>
        public int TrackCount
        {
            get
            {
                return m_tracks.Count;
            }
        }

        /// <summary>
        /// This method returns the amount data to send to the writer
        /// </summary>
        public long DataToSend
        {
            get
            {
                MinuteSecondFrame minsize = new MinuteSecondFrame(0, 4, 0);
                long total = 0;

                if (IsDiskPureAudio && HasCDText)
                {
                    Debug.Assert(m_cdtext_source != null);
                    total += m_cdtext_source.SectorCount;
                }

                foreach (TrackBurnImage t in m_tracks)
                {
                    long length = t.DataOnlyLength;

                    if (IsTargetCD && length < CDR_RWDiskOperations.MinimumTrackLengthInSectors)
                        length = CDR_RWDiskOperations.MinimumTrackLengthInSectors ;

                    total += length + t.PreTrackGap;
                }


                return total;
            }
        }

        /// <summary>
        /// This property returns the length of the disk in total sectors.
        /// </summary>
        public long LengthSectors
        {
            get
            {
                long total = 0;

                // Now add each track, this includes any gap that is placed
                // after the track
                foreach (TrackBurnImage t in m_tracks)
                    total += t.TotalLengthSectors;

                return total;
            }
        }

        /// <summary>
        /// This method returns true if the disk is a pure audio disk.
        /// </summary>
        public bool IsDiskPureData
        {
            get
            {
                bool pure = true;

                foreach (TrackBurnImage t in m_tracks)
                {
                    if (t.Type == TrackBurnImage.TrackType.Audio ||
                        t.Type == TrackBurnImage.TrackType.Raw)
                    {
                        pure = false;
                        break;
                    }
                }

                Debug.Assert(pure == false || m_tracks.Count == 1);
                return pure;
            }
        }


        /// <summary>
        /// This property returns TRUE if this disk is a pure audio disk, otherwise it returns
        /// false
        /// </summary>
        public bool IsDiskPureAudio
        {
            get
            {
                bool pure = true;

                foreach (TrackBurnImage t in m_tracks)
                {
                    if (t.Type != TrackBurnImage.TrackType.Audio)
                    {
                        pure = false;
                        break;
                    }
                }

                return pure;
            }
        }

        /// <summary>
        /// If true, this disk has a media control number
        /// </summary>
        public bool HasMCN
        {
            get
            {
                return m_mcn != null;
            }
        }

        /// <summary>
        /// Returns the media control number
        /// </summary>
        public byte[] MCN
        {
            get
            {
                return m_mcn;
            }
            set
            {
                if (value.GetLength(0) != 13)
                    throw new Exception("the mcn should be a byte array containing exactly 13 bytes");

                m_mcn = value;
            }
        }

        /// <summary>
        /// This property returns the number of languages defined for CD Text.
        /// </summary>
        public int CDTextLanguageCount
        {
            get
            {
                return m_languages.Count;
            }
        }

        #endregion

        #region constructors
        /// <summary>
        /// The constructor for the disk
        /// </summary>
        public DiskBurnImage(Logger l)
        {
            m_logger = l;
            m_auto_generate_audio_pregap = true;
            m_tracks = new List<TrackBurnImage>();
            m_mcn = null;
            m_cdtext = new List<CDTextField>();
            m_languages = new List<CDTextEncoderDecoder.LanguageType>();
            m_cdtext_source = null;
        }
        #endregion

        #region public data members
        /// <summary>
        /// This event is fired when a track is added
        /// </summary>
        public EventHandler<EventArgs> TrackAdded;

        /// <summary>
        /// This event is fired when a track is removed
        /// </summary>
        public EventHandler<EventArgs> TrackRemoved;
        #endregion

        #region public methods

        #region CDText related methods

        /// <summary>
        /// Given an index, return the language at the given idnex
        /// </summary>
        /// <param name="index">the index of interest</param>
        /// <returns>the langauge</returns>
        public CDTextEncoderDecoder.LanguageType GetCDTextLanguage(int index)
        {
            return m_languages[index];
        }

        /// <summary>
        /// Add a new text language for CD Text and return the index to the language
        /// </summary>
        /// <param name="l">the langauge to add</param>
        /// <returns>the index of this language, or -1 if 8 languages already exist</returns>
        public byte AddCDTextLanguage(CDTextEncoderDecoder.LanguageType l)
        {
            if (m_languages.Count == 8)
                return 0xff;

            m_languages.Add(l);
            return (byte)(m_languages.Count - 1);
        }

        /// <summary>
        /// Return the index associated with the index given
        /// </summary>
        /// <param name="l"></param>
        /// <returns></returns>
        public byte GetCDTextLanguageIndex(CDTextEncoderDecoder.LanguageType l)
        {
            for (byte i = 0; i < m_languages.Count; i++)
            {
                if (m_languages[i] == l)
                    return i;
            }
            return 0xff;
        }

        /// <summary>
        /// Find a CD text field given its type and language
        /// </summary>
        /// <param name="t">the type of field to find</param>
        /// <param name="blocknr">the language index of the field to find</param>
        /// <returns>the cd text field with the given language index and type, or null if it does not exist</returns>
        public CDTextField FindCDTextField(CDTextField.TextFieldType t, byte blocknr)
        {
            foreach (CDTextField tmp in m_cdtext)
            {
                if (tmp.LanguageIndex == blocknr && tmp.Type == t)
                    return tmp;
            }
            return null;
        }

        /// <summary>
        /// Add a CD text field to this disk image
        /// </summary>
        /// <param name="f">the field to add</param>
        /// <returns>true if the field is unique, false if it already exists</returns>
        public bool AddCDTextField(CDTextField f)
        {
            if (FindCDTextField(f.Type, f.LanguageIndex) != null)
                return false;

            m_cdtext.Add(f);
            return true;
        }

        /// <summary>
        /// Remove a given CD Text entry given its type and language
        /// </summary>
        /// <param name="t">the type of CD text field</param>
        /// <param name="blocknr">the language index of the CD text field</param>
        /// <returns>true if removed, false if it does not exist</returns>
        public bool RemoveCDTextField(CDTextField.TextFieldType t, byte blocknr)
        {
            for (int i = 0; i < m_cdtext.Count; i++)
            {
                if (m_cdtext[i].Type == t && m_cdtext[i].LanguageIndex == blocknr)
                {
                    m_cdtext.RemoveAt(i);
                    return true;
                }
            }

            return false;
        }
        #endregion\

        #region Track related methods

        /// <summary>
        /// Retreive a track from the disk based on track number
        /// </summary>
        /// <param name="index"></param>
        /// <returns></returns>
        public TrackBurnImage this[int index]
        {
            get
            {
                return m_tracks[index];
            }
        }

        /// <summary>
        /// Return an enumerator for iterating over all of the tracks on the
        /// disk.
        /// </summary>
        /// <returns>iterator</returns>
        public IEnumerator<TrackBurnImage> GetEnumerator()
        {
            return m_tracks.GetEnumerator();
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            return m_tracks.GetEnumerator();
        }

        /// <summary>
        /// Add a new track to the disk
        /// </summary>
        /// <param name="t">the track to add</param>
        public void AddTrack(TrackBurnImage t)
        {
            if (m_logger != null)
            {
                string str = "Adding track '" + t.Name + "' to disk image";
                m_logger.LogMessage(new UserMessage(UserMessage.Category.Debug, 1, str));
            }
            m_tracks.Add(t);

            UpdateTracks();

            if (TrackAdded != null)
                TrackAdded(this, EventArgs.Empty);
        }

        /// <summary>
        /// Remove a track from the disk
        /// </summary>
        /// <param name="t">the track to remove</param>
        public void RemoveTrack(TrackBurnImage t)
        {
            if (m_logger != null)
            {
                string str = "Removing track '" + t.Name + "' to disk image";
                m_logger.LogMessage(new UserMessage(UserMessage.Category.Debug, 1, str));
            }

            m_tracks.Remove(t);

            UpdateTracks();

            if (TrackRemoved != null)
                TrackRemoved(this, EventArgs.Empty);
        }

        /// <summary>
        /// Move a track up in the disk image and update the related track parameters
        /// </summary>
        /// <param name="t"></param>
        public void MoveTrackUp(TrackBurnImage t)
        {
            int index = m_tracks.IndexOf(t);
            Debug.Assert(index >= 0 && index < m_tracks.Count);

            m_tracks.RemoveAt(index);
            m_tracks.Insert(index - 1, t);
        }

        /// <summary>
        /// Move a track down in the disk image and updated the related track parameters
        /// </summary>
        /// <param name="t"></param>
        public void MoveTrackDown(TrackBurnImage t)
        {
            int index = m_tracks.IndexOf(t);
            Debug.Assert(index >= 0 && index < m_tracks.Count);

            m_tracks.RemoveAt(index);
            m_tracks.Insert(index + 1, t);
        }
        #endregion

        #endregion

        #region internal members functions

        /// <summary>
        /// Creates the image of the cue sheet based on the tracks in the disk
        /// </summary>
        /// <returns>byte array containing cue sheet</returns>
        internal byte[] CreateCueSheet(byte track_num, ref DataGenerator gen, long cdtext_start)
        {
            byte track_data_type, track_data_form, gap_data_form ;
            byte[] cuesheet = new byte[0];
            long currentTime ;
            long timelen;
            int index = 0 ;

            //
            // This should only be called for SAO CD mastering
            //
            Debug.Assert(m_target == DiskTargetType.CD_SAO);

            //
            // The current time for the current track, the 150 is part of the logical block to
            // MSF translation.
            //
            currentTime = m_start_sector;

            track_data_form = 0x01;
            if (IsDiskPureAudio)
            {
                track_data_type = (byte)TrackAdr.StartTime;                                     // Pure audio track
                if (HasCDText)
                    track_data_form = 0x41;                                                     // sec size 2352, data size 96
                else
                    track_data_form = 0x01;                                                     // sec size 2352, data size 0
            }
            else if (IsDiskPureData)
            {
                track_data_type = (byte)TrackControl.DataTrack | (byte)TrackAdr.StartTime;      // Data track
                track_data_form = 0x14 ;                                                        // sec size 2048, data size 0
            }
            else
            {
                track_data_type = (byte)0;
                //
                // No support yet for disks that mix audio and data tracks
                //
                Debug.Assert(false);
            }

            //
            // The time field in the leadin does not reflect the actual MSF location of the leadin
            // data.  The time field should set to zero for the leadin.
            //
            AddTrack(ref cuesheet, index++, track_data_type, 0, 0, track_data_form, 0, new MinuteSecondFrame(0));

            if (HasCDText && IsDiskPureAudio)
            {
                //
                // Add a source for the CD Text
                //
                if (m_cdtext_source == null)
                    m_cdtext_source = new CDTextDataSource(this, cdtext_start);

                TrackBurnImage ti = new TrackBurnImage(TrackBurnImage.TrackType.CDText, m_cdtext_source) ;
                gen.AddSource("cdtext", ti, cdtext_start, m_cdtext_source.SectorCount);
            }

            //
            // If the disk has an media control number, add it to the cue sheet next
            //
            if (HasMCN)
            {
                AddTrack(ref cuesheet, index++, 2, m_mcn[0], m_mcn[1], m_mcn[2], m_mcn[3], m_mcn[4], m_mcn[5], m_mcn[6]);
                AddTrack(ref cuesheet, index++, 2, m_mcn[7], m_mcn[8], m_mcn[9], m_mcn[10], m_mcn[11], m_mcn[12], 0);
            }

            //
            // Add the data from the track
            //
            foreach (TrackBurnImage t in m_tracks)
            {
                // Get the length of the track
                timelen = t.DataOnlyLength;

                //
                // If the track length does not meet the minimum track length, force the
                // length up to the minimum
                //
                if (timelen < CDR_RWDiskOperations.MinimumTrackLengthInSectors)
                    timelen = CDR_RWDiskOperations.MinimumTrackLengthInSectors;

                //
                // Add the track to the disk
                //
                if (t.Type == TrackBurnImage.TrackType.Audio)
                {
                    track_data_type = (byte)TrackControl.DigitalCopyPermitted | (byte)TrackAdr.StartTime ;
                    if (t.PreEmphasis)
                        track_data_type |= (byte)TrackControl.PreEmphasis;
                    track_data_form = (byte)TrackDataForm.AudioHost;
                    gap_data_form = (byte)TrackDataForm.AudioHost ;
                }
                else
                {
                    track_data_type = (byte)TrackControl.DataTrack | (byte)TrackAdr.StartTime ;
                    track_data_form = (byte)TrackDataForm.DataHost ;
                    gap_data_form = (byte)TrackDataForm.DataHost;
                }

                byte tckidx = 0 ;


                //
                // This is the data source for the sectors associated with this track
                //
                gen.AddSource(t.Name, t, currentTime, timelen + t.PreTrackGap);

                // If the track has a pre-gap pad, add it to the track with the current track no and the next
                // index number
                if (t.PreTrackGap > 0)
                {
                    AddTrack(ref cuesheet, index++, track_data_type, track_num, tckidx++, gap_data_form, 0, new MinuteSecondFrame(currentTime + 150));
                    currentTime += t.PreTrackGap;
                }
                else
                {
                    tckidx++;
                }


                // Now add the track to the cue sheet
                AddTrack(ref cuesheet, index++, track_data_type, track_num, tckidx, track_data_form, 0, new MinuteSecondFrame(currentTime + 150));

                // If the track has an ISRC code, add it to the cue sheet
                if (t.HasIsrc)
                {
                    byte [] isrc = t.ISRC;
                    AddTrack(ref cuesheet, index++, 3, track_num, isrc[0], isrc[1], isrc[2], isrc[3], isrc[4], isrc[5]);
                    AddTrack(ref cuesheet, index++, 3, track_num, isrc[6], isrc[7], isrc[8], isrc[9], isrc[10], isrc[11]);
                }

                // Bump the track number
                track_num++;

                // And increase the time
                currentTime += timelen;
            }

            //
            // Add the leadout
            //
            track_data_form = 0x01;
            if (IsDiskPureAudio)
            {
                track_data_type = (byte)TrackAdr.StartTime;                               // Pure audio track
                track_data_form = 0x01;
            }
            else
            {
                track_data_type = (byte)TrackControl.DataTrack | (byte)TrackAdr.StartTime;      // Data track
                track_data_form = 0x14;
            }
            AddTrack(ref cuesheet, index++, track_data_type, LeadOutTrackNumber, 1, track_data_form, 0, new MinuteSecondFrame(currentTime + 150));
            return cuesheet;
        }
        #endregion

        #region private member functions

        /// <summary>
        /// Something changed, update the start time of each track
        /// </summary>
        private void UpdateTracks()
        {
            string str;
            long sector = m_start_sector;

            if (m_logger != null)
            {
                str = "UpdateTracks: start sector is " + sector.ToString();
                m_logger.LogMessage(new UserMessage(UserMessage.Category.Debug, 2, str));
            }

            for(int i = 0 ; i < m_tracks.Count ; i++)
            {
                TrackBurnImage t = m_tracks[i];

                if ((i == 0 && IsTargetCD) || (t.Type == TrackBurnImage.TrackType.Audio && m_auto_generate_audio_pregap))
                {
                    //
                    // The first track always gets a two second pre-gap.  Also, audio tracks get a two
                    // second pre-gap if the auto pre-gap is turned on.
                    //
                    t.PreTrackGap = CDR_RWDiskOperations.DefaultPregapLengthInSectors;
                }
                else
                    t.PreTrackGap = 0;

                t.TrackNumber = i + m_start_track ;
                t.StartSector = sector;

                if (m_logger != null)
                {
                    str = "Track '" + t.Name + "', number = " + t.TrackNumber.ToString();
                    str += ", start = " +  t.StartSector.ToString() ;
                    m_logger.LogMessage(new UserMessage(UserMessage.Category.Debug, 2, str));
                }

                //
                // The return value from TotalLengthSectors will include the pregap set above
                //
                sector += t.TotalLengthSectors;
            }
        }

        private void AddTrack(ref byte[] sheet, int t, byte b1, byte b2, byte b3, byte b4, byte b5, MinuteSecondFrame time)
        {
            // Resize the array if necessary
            if (sheet.GetLength(0) <= t * 8)
                Array.Resize(ref sheet, (t + 1) *  8) ;

            sheet[t * 8 + 0] = b1;
            sheet[t * 8 + 1] = b2;
            sheet[t * 8 + 2] = b3;
            sheet[t * 8 + 3] = b4;
            sheet[t * 8 + 4] = b5;
            sheet[t * 8 + 5] = (byte)time.Minutes;
            sheet[t * 8 + 6] = time.Seconds;
            sheet[t * 8 + 7] = time.Frames;
        }

        private void AddTrack(ref byte[] sheet, int t, byte b1, byte b2, byte b3, byte b4, byte b5, byte b6, byte b7, byte b8)
        {
            // Resize the array if necessary
            if (sheet.GetLength(0) <= t * 8)
                Array.Resize(ref sheet, (t + 1) * 8);

            sheet[t * 8 + 0] = b1;
            sheet[t * 8 + 1] = b2;
            sheet[t * 8 + 2] = b3;
            sheet[t * 8 + 3] = b4;
            sheet[t * 8 + 4] = b5;
            sheet[t * 8 + 5] = b6;
            sheet[t * 8 + 6] = b7;
            sheet[t * 8 + 7] = b8;
        }
        #endregion
    }
}
