//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//
using System;
using System.Collections.Generic;
using System.Text;
using Bwg.Scsi ;
using Bwg.Logging ;

namespace Bwg.Burner
{
    /// <summary>
    /// This class acts as an abstraction of the media type within the burner
    /// drive.  This class contains methods for all operations that have a
    /// media specific implementation.
    /// </summary>
    public abstract class DiskOperations
    {
        /// <summary>
        /// The drive that we are talking to.
        /// </summary>
        protected Drive m_drive;

        /// <summary>
        /// Create a new operations object
        /// </summary>
        /// <param name="d"></param>
        public DiskOperations(Drive d)
        {
            m_drive = d;
        }

        /// <summary>
        /// This property returns the size of the disk in sectors.  It always returns a
        /// value, so if a disk error occurs, a reasonable default is returned.  This value
        /// is the total capacity of the media in sectors, regardless of whether they are
        /// writable or not.
        /// 
        /// Note, this is the total capacity so it should return somewhere between 650 MB
        /// and 700 Mb for CDs and around 4.4 Gb for DVD media.  The value returned will be
        /// a sector count.
        /// </summary>
        public abstract long CapacityInSectors
        {
            get;
        }

        /// <summary>
        /// This property returns the size of a disk in writable sectors.  It always returns
        /// a value, so if a disk error occurs, a reasonable default is returned.  This value
        /// is the total writable capacity of a disk.  For CD/R and CD/RW this should return
        /// around 650M - 700M for a normal sized disk.  For CD/ROM media, this should return
        /// zero.  For DVD-R, DVD-RW, DVD+R, DVD+RW media this will return around 4.4 GB and
        /// for DVD-ROM it should return 0
        /// </summary>
        public abstract long WritableCapacityInSectors
        {
            get;
        }

        /// <summary>
        /// This property should return the remaining capacity in writable sectors.  This property
        /// assumes that the current data will be appeneded to the current disk.  This value can
        /// never exceed the WritableCapacityInSectors.  Also, for media that does not support
        /// additional sessions (e.g. finalized media) this property should return zero.
        /// </summary>
        public abstract long RemainingCapacityInWritableSectors
        {
            get;
        }

        /// <summary>
        /// This property returns true if the media is capable of supporting write
        /// protect.
        /// </summary>
        public abstract bool IsWriteProtectSupported
        {
            get;
        }

        /// <summary>
        /// This property returns true if the loaded disk is write protected
        /// </summary>
        public abstract bool IsWriteProtected
        {
            get;
        }

        /// <summary>
        /// This property returns true if the loaded disk has a write protect password
        /// </summary>
        public abstract bool HasWriteProtectPassword
        {
            get;
        }

        /// <summary>
        /// This property provides the write protect password for a given media
        /// </summary>
        public abstract string WriteProtectPassword
        {
            set;
        }

        /// <summary>
        /// This property returns true if the media is capable of being erased.
        /// </summary>
        public abstract bool IsEraseSupported
        {
            get;
        }

        /// <summary>
        /// This property returns true if the media is capable of being formatted
        /// </summary>
        public abstract bool IsFormatSupported
        {
            get;
        }

        /// <summary>
        /// This property returns true if the media is capable of being burned.  This property
        /// returns true even if an erase may be required before a burn can be initiated.
        /// </summary>
        public abstract bool IsBurnSupported
        {
            get;
        }

        /// <summary>
        /// This property returns true if the media is capable of containing multiple sessions
        /// </summary>
        public abstract bool IsMultiSessionSupported
        {
            get;
        }

        /// <summary>
        /// This property returns true if track at once burning is supported given the media + drive
        /// + software.
        /// </summary>
        public abstract bool IsTrackAtOnceSupported
        {
            get;
        }

        /// <summary>
        /// This property returns true if the media + drive + software is capable of supporting
        /// sessions at once burning.
        /// </summary>
        public abstract bool IsSessionAtOnceSupported
        {
            get;
        }

        /// <summary>
        /// Erase the media that is in the drive.
        /// </summary>
        /// <param name="et">type of erase</param>
        /// <returns>an error object if an error occurred, null if no error occurred</returns>
        public abstract DiskOperationError Erase(Drive.EraseType et);

        /// <summary>
        /// Burn the contents of the disk structured as defined by the Disk object onto the
        /// media.
        /// </summary>
        /// <param name="bt">the burn type for CD/R or CD/RW, must be Session for DVD</param>
        /// <param name="d">the disk to burn</param>
        /// <param name="multi_session">if true, leave the disk open for future sessions (if media supports this)</param>
        /// <param name="speed">the speed for the burn operation</param>
        /// <param name="bytes">an output, the number of bytes burned</param>
        /// <returns></returns>
        public abstract DiskOperationError BurnDisk(Drive.BurnType bt, DiskBurnImage d, bool multi_session, SpeedInfo speed, out ulong bytes);

        /// <summary>
        /// Create the appropriate error for an operation that is not supported by the given
        /// media type.
        /// </summary>
        /// <param name="oper">the operation that is not supporte</param>
        /// <returns>an error object with an appropriate error string</returns>
        protected DiskOperationError NotSupported(string oper)
        {
            string str = "the operation '" + oper + "' is not supported by media ";
            str += "'" + m_drive.CurrentProfile.ToString() + "'";
            return new BurnerError(str);
        }

        /// <summary>
        /// Create an error object based on an error detected on the low level SCSI
        /// device.  The error is either an IOCTL error or a device SCSI error.  This
        /// function figures this out and creates the appropriate error message.
        /// </summary>
        /// <param name="st"></param>
        /// <returns></returns>
        protected DiskOperationError CreateErrorObject(Device.CommandStatus st)
        {
            return Drive.CreateErrorObject(st, MyDevice);
        }

        /// <summary>
        /// The device the burner is connected to
        /// </summary>
        protected Device MyDevice
        {
            get
            {
                return m_drive.Device;
            }
        }

        /// <summary>
        /// The logger associated with this operations, we get it from the device
        /// </summary>
        protected Logger MyLogger
        {
            get
            {
                return m_drive.Device.GetLogger();
            }
        }
    }
}
