//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//

using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;
using System.Windows.Forms;
using System.ComponentModel;
using System.Runtime.InteropServices;
using System.Diagnostics ;
using System.IO ;
using Bwg.Scsi ;
using Bwg.Logging;
using Bwg.Hardware;

namespace Bwg.Burner
{
    /// <summary>
    /// This class represents a CD/DVD drive.  It provide access to high level functionality of the
    /// drive (e.g. erase or burn) and low level functionality (SCSI cmds) can be access via the associated
    /// device.  This class is created from a device where the device object provides low level access to
    /// the drive.
    /// 
    /// This object provides methods and properties to determine what kind of media is loaded, what the 
    /// state of the media is, and what operations you can perform on the media.  This object also provides
    /// the method to perform the two basic operations, erase and burn.
    /// </summary>
    public class Drive : IDisposable
    {
        #region Public Types

        /// <summary>
        /// Capacity of a 63 minute CD/R or CD/RW disk in sectors
        /// </summary>
        public const long D63minDiskCapacity = 283500;

        /// <summary>
        /// Capacity of a 74 minute CD/R or CD/RW disk in sectors
        /// </summary>
        public const long D74minDiskCapacity = 333000;

        /// <summary>
        /// Capacity of a 80 minute CD/R or CD/RW disk in sectors
        /// </summary>
        public const long D80minDiskCapacity = 360000;

        /// <summary>
        /// Capacity of a 4.7 Gig DVD in sectors
        /// </summary>
        public const long D47GigDVDMinusRCapacity = 2295104;

        /// <summary>
        /// Capacity of a 4.7 Gig DVD in sectors
        /// </summary>
        public const long D47GigDVDPlusRCapacity = 2297888;

        /// <summary>
        /// The amount of information that can be stored on a DVD+R DL disk
        /// </summary>
        public const long D47GigDVDPlusRDLCapacity = 2 * 2086912;

        /// <summary>
        /// The amount of information that can be stored on a single side of a DVD+R DL disk
        /// </summary>
        public const long D47GigDVDPlusRDLOneSideCapacity = 2086912;

        /// <summary>
        /// This number of sectors on a blu-ray RE disk
        /// </summary>
        public const long D25GigBDRECapacity = 11826176;

        /// <summary>
        /// This type represents the type of connection from the computer to the drive
        /// </summary>
        public enum ConnectionType
        {
            /// <summary>
            /// We do not know how the drive is connected
            /// </summary>
            Unknown,

            /// <summary>
            /// And IDE connected device
            /// </summary>
            IDE,

            /// <summary>
            /// A SCSI connected device
            /// </summary>
            SCSI,
            
            /// <summary>
            /// A USB connected device
            /// </summary>
            USB,

            /// <summary>
            /// A firware connected device
            /// </summary>
            Firewire
        } ;

        /// <summary>
        /// This type is the state of a given feature
        /// </summary>
        public enum FeatureState
        {
            /// <summary>
            /// The feature is both present and current
            /// </summary>
            PresentAndCurrent,

            /// <summary>
            /// This feautue is present, but is not current
            /// </summary>
            PresentNotCurrent,

            /// <summary>
            /// This drive does not support this feature
            /// </summary>
            NotPresent,

            /// <summary>
            /// An error occured while trying to retreive feature info
            /// </summary>
            Error
        } ;

        /// <summary>
        /// The drive type
        /// </summary>
        public enum MMCDriveType
        {
            /// <summary>
            /// This is a MMC-3 drive
            /// </summary>
            MMC_3,

            /// <summary>
            /// This is a MMC-4 type drive
            /// </summary>
            MMC_4
        } ;

        /// <summary>
        /// The various profiles that may be supported by the hardware
        /// </summary>
        public enum SupportedProfiles
        {
            /// <summary>
            /// No profile
            /// </summary>
            NoCurrentProfile = 0x0000,

            /// <summary>
            /// A non-removable disk
            /// </summary>
            NonRemovableDisk = 0x0001,

            /// <summary>
            /// A removable disk
            /// </summary>
            RemovableDisk = 0x0002,

            /// <summary>
            /// Magneto-Optical erasable disk
            /// </summary>
            MagnetoOpticalErasable = 0x0003,

            /// <summary>
            /// An optical write once drive
            /// </summary>
            OpticalWriteOnce = 0x0004,

            /// <summary>
            /// AS_MO ???
            /// </summary>
            AS_MO = 0x0005,

            /// <summary>
            /// A basic CDROM drive
            /// </summary>
            CD_ROM = 0x0008,

            /// <summary>
            /// A CD-R drive
            /// </summary>
            CD_R = 0x0009,

            /// <summary>
            /// A CD-RW drive
            /// </summary>
            CD_RW = 0x000A,

            /// <summary>
            /// A DVD-ROM drive
            /// </summary>
            DVD_ROM = 0x0010,

            /// <summary>
            /// A DVD-R Sequential Recording
            /// </summary>
            DVD_R_SequentialRecording = 0x0011,

            /// <summary>
            /// A DVD-RAM drive
            /// </summary>
            DVD_RAM = 0x0012,

            /// <summary>
            /// A DVD-RW with Restricted overwrite capabilities
            /// </summary>
            DVD_RW_RestrictedOverwrite = 0x0013,

            /// <summary>
            /// A DVD-RW with sequential recording only
            /// </summary>
            DVD_RW_SequentialRecording = 0x0014,

            /// <summary>
            /// A DVD-R with two layers, sequential recording only
            /// </summary>
            DVD_R_DualLayer_SequentialRecording = 0x0015,

            /// <summary>
            /// A DVD-R with two layer, layer jump recording
            /// </summary>
            DVD_R_DualLayer_LayerJumpRecording = 0x0016,

            /// <summary>
            /// Dual layer DVD + RW
            /// </summary>
            DVD_RW_DualLayer = 0x0017,

            /// <summary>
            /// A DVD+RW drive
            /// </summary>
            DVD_PLUS_RW = 0x001A,

            /// <summary>
            /// A DVD+R drive
            /// </summary>
            DVD_PLUS_R = 0x001B,

            /// <summary>
            /// A double density CD-ROM
            /// </summary>
            DDCD_ROM = 0x0020,

            /// <summary>
            /// A double density CD-R drive
            /// </summary>
            DDCD_R = 0x0021,

            /// <summary>
            /// A double density CD-RW drive
            /// </summary>
            DDCD_RW = 0x0022, 

            /// <summary>
            /// DVD + RW dual layer
            /// </summary>
            DVD_PLUS_RW_DUAL_LAYER = 0x002A,

            /// <summary>
            /// DVD + R dual layer
            /// </summary>
            DVD_PLUS_R_DUAL_LAYER = 0x002B,

            /// <summary>
            /// Blue Ray ROM disk
            /// </summary>
            BD_ROM = 0x0040,

            /// <summary>
            /// Blue Ray Sequential recording
            /// </summary>
            BD_R_SEQUENTIAL_RECORDING = 0x0041,

            /// <summary>
            /// Blue Ray Random recording
            /// </summary>
            BD_R_RANDOM_RECORDING = 0x0042,

            /// <summary>
            /// Blue Ray read capability ???
            /// </summary>
            BD_RE = 0x0043,

            /// <summary>
            /// High density DVD ROM
            /// </summary>
            HD_DVD_ROM = 0x0050,

            /// <summary>
            /// High density DVD-R
            /// </summary>
            HD_DVD_R = 0x0051,

            /// <summary>
            /// High density DVD-RAM
            /// </summary>
            HD_DVD_RAM = 0x0052,

            /// <summary>
            /// A blank DVD-RW that can be either.  This is not a real profile returned from 
            /// the drive, but a synthesized profile
            /// </summary>
            DVD_RW_RestrictedOverwrite_And_Sequential = 65534,

            /// <summary>
            /// Other, non-conforming drive type
            /// </summary>
            Other = 65535
        } ;

        /// <summary>
        /// The type of erase to perform
        /// </summary>
        public enum EraseType
        {
            /// <summary>
            /// Erase only those sectors required to make the disk appear erased
            /// </summary>
            Minimal,

            /// <summary>
            /// Erase all sectors on the disk
            /// </summary>
            Full
        } ;

        /// <summary>
        /// The type of burn to perform, track at a time, or disk at a time
        /// </summary>
        public enum BurnType
        {
            /// <summary>
            /// The caller does not care.  Choose the method that works best.  Generally
            /// session at once will be used unless it is not supported, then track at once
            /// burning will be done.
            /// </summary>
            DontCare,

            /// <summary>
            /// Burn CD-R or CD-RW a track at a time
            /// </summary>
            Track,
            
            /// <summary>
            /// Burn the disk a session at a time
            /// </summary>
            Session
        } ;
        #endregion

        #region Public Data Members

        /// <summary>
        /// The list of profiles supported by the physical device
        /// </summary>
        public IList<SupportedProfiles> ProfileList;

        /// <summary>
        /// The current profile
        /// </summary>
        public SupportedProfiles CurrentProfile;

        /// <summary>
        /// This event is signaled when the progress has changed
        /// </summary>
        public ProgressChangedEventHandler ProgressChanged;

        /// <summary>
        /// This variable gives the type
        /// </summary>
        public MMCDriveType MMCType;

        /// <summary>
        /// If true, the drive is locked
        /// </summary>
        public bool IsLocked;

        #endregion

        #region Private Data Members

        /// <summary>
        /// The physical device we are talking to
        /// </summary>
        private Device m_device;

        /// <summary>
        /// This is the type of connection to the drive (IDE, USB, SCSI, Firewire, etc.)
        /// </summary>
        private ConnectionType m_connection;

        /// <summary>
        /// This is the total length of buffer availabile with the drive for data.  This
        /// value is given in bytes.
        /// </summary>
        private int m_buffer_length;

        /// <summary>
        /// This is the amount of space available yet in the buffer for the drive.  This value
        /// is given in bytes.
        /// </summary>
        private int m_buffer_avail;

        /// <summary>
        /// If true, the latest call to ReadBufferLevels() was sucessful and the m_buffer_length
        /// and m_buffer_avail values are valid.
        /// </summary>
        private bool m_buffer_valid;

        /// <summary>
        /// This value contains the start time for an operation where progress is measured in
        /// real time.
        /// </summary>
        private DateTime m_start;

        /// <summary>
        /// This value contains the total duration for an operation where progress is measured
        /// in real time.  The duration is given in seconds.
        /// </summary>
        private int m_total_duration ;

        /// <summary>
        /// This variable contains a reference to a class derived from the DiskOperations class
        /// that provides the basic functionality for the drive.  The DiskOperations class is a
        /// function of the specific media type that is loaded in the drive.
        /// </summary>
        private DiskOperations m_opers;

        /// <summary>
        /// This variable holds the name of the drive in a human readable form.
        /// </summary>
        private string m_drive_model;

        /// <summary>
        /// This variable holds the name of the media manufacturer as read from the media  Only valid for DVD+R and DVD+RW.
        /// </summary>
        private string m_media_manufacturer;

        /// <summary>
        /// This variable holds the type of the media as read from the media.  Only valid for DVD+R and DVD+RW.
        /// </summary>
        private string m_media_type;

        /// <summary>
        /// This variable holds the media version as read from the media.  Only valid for DVD+R and DVD+RW.
        /// </summary>
        private byte m_media_version;

        /// <summary>
        /// This is the current operation that is underway
        /// </summary>
        private ProgressText m_current_operation;

        /// <summary>
        /// When the write was started
        /// </summary>
        private DateTime m_operation_start;

        /// <summary>
        /// The timer is active
        /// </summary>
        private bool m_timing_active;

        /// <summary>
        /// The amount of time spent writing data to the drive
        /// </summary>
        private TimeSpan m_operation_time;

        /// <summary>
        /// The number of bytes of data burned in the last burn operation
        /// </summary>
        private ulong m_burn_bytes;

        /// <summary>
        /// If true, all speed stuff is done via mode page 0x2a and SetCDSpeed().  
        /// If false GetPerformance and SetStreaming is used
        /// </summary>
        private bool m_mode_page_speed;

        #endregion

        #region public properties

        /// <summary>
        /// This property returns the drives connection type
        /// </summary>
        public ConnectionType Connection
        {
            get
            {
                return m_connection;
            }
        }

        /// <summary>
        /// The total number of bytes sent to the drive in the burn process.
        /// </summary>
        public ulong BurnBytes
        {
            get
            {
                return m_burn_bytes;
            }
        }

        /// <summary>
        /// The total time spent with the last operation
        /// </summary>
        public TimeSpan OperationTime
        {
            get
            {
                return m_operation_time;
            }
        }

        /// <summary>
        /// The actual speed of the burn as caculated by dividing the total amount of
        /// data burned divided by the time it took to burn and ratio'ing this with the
        /// 1x value for either CD or DVD depending on the media type.
        /// </summary>
        public string BurnSpeed
        {
            get
            {
                double bytes_per_second = m_burn_bytes / m_operation_time.TotalSeconds;
                double speed ;

                if (!IsDvd)
                    speed = bytes_per_second / 176400.0;
                else
                    speed = bytes_per_second / 1385000.0;

                return speed.ToString("G2") + "X";
            }
        }

        /// <summary>
        /// Returns true if the data about the buffer is valid
        /// </summary>
        public bool IsBufferValid
        {
            get
            {
                return m_buffer_valid;
            }
        }

        /// <summary>
        /// This length of the buffer in the drive
        /// </summary>
        public int BufferLength
        {
            get
            {
                return m_buffer_length;
            }
        }

        /// <summary>
        /// The amount of buffer space available
        /// </summary>
        public int BufferAvailable
        {
            get
            {
                return m_buffer_avail;
            }
        }

        /// <summary>
        /// This method returns true if media is loaded
        /// </summary>
        public bool IsMediaLoaded
        {
            get
            {
                return CurrentProfile != SupportedProfiles.NoCurrentProfile;
            }
        }

        /// <summary>
        /// This property returns true if the burner supports erasing the current media
        /// </summary>
        public bool IsEraseSupported
        {
            get
            {
                if (m_opers == null)
                    Initialize();

                if (m_opers == null)
                    return false;

                return m_opers.IsEraseSupported;
            }
        }

        /// <summary>
        /// This method returns true if the media loaded supports a format operation
        /// </summary>
        public bool IsFormatSupported
        {
            get
            {
                if (m_opers == null)
                    Initialize();

                if (m_opers == null)
                    return false;

                return m_opers.IsFormatSupported;
            }
        }


        /// <summary>
        /// This property returns true if burning is supported on the loaded media.
        /// </summary>
        public bool IsBurnSupported
        {
            get
            {
                if (m_opers == null)
                    Initialize();

                if (m_opers == null)
                    return false;

                return m_opers.IsBurnSupported;
            }
        }

        /// <summary>
        /// This property returns true if multi-session disks are supported for the given
        /// media type.  This is a combination of the drive + media + software support for
        /// the given media.
        /// </summary>
        public bool IsMultiSessionSupported
        {
            get
            {
                if (m_opers == null)
                    Initialize();

                if (m_opers == null)
                    return false;

                return m_opers.IsMultiSessionSupported;
            }
        }

        /// <summary>
        /// This propertry returns true if the media supports track at once burning
        /// </summary>
        public bool IsTrackAtOnceSupported
        {
            get
            {
                if (m_opers == null)
                    Initialize();

                if (m_opers == null)
                    return false;

                return m_opers.IsTrackAtOnceSupported;
            }
        }

        /// <summary>
        /// This property return true if the media supports sessions at once burning
        /// </summary>
        public bool IsSessionAtOnceSupported
        {
            get
            {
                if (m_opers == null)
                    Initialize();

                if (m_opers == null)
                    return false;

                return m_opers.IsSessionAtOnceSupported;
            }
        }

        /// <summary>
        /// This property is a string representing the current operation.
        /// </summary>
        internal ProgressText CurrentOperation
        {
            get
            {
                return m_current_operation;
            }
            set
            {
                m_current_operation = value;
            }
        }

        /// <summary>
        /// This returns the drive model name and version number
        /// </summary>
        public string DriveModel
        {
            get
            {
                if (m_drive_model == null || m_drive_model.Length == 0)
                    Initialize();
                return m_drive_model;
            }
        }
        
        /// <summary>
        /// This property returns the capacity of the media in the drive in sectors.
        /// </summary>
        public long CapacityInSectors
        {
            get
            {
                if (m_opers == null)
                    Initialize();

                if (m_opers == null)
                    return 0;

                return m_opers.CapacityInSectors;
            }
        }

        /// <summary>
        /// This property returns the maximum writable capacity of the media in the
        /// drive in sectors.
        /// </summary>
        public long WritableCapacityInSectors
        {
            get
            {
                if (m_opers == null)
                    Initialize();

                if (m_opers == null)
                    return 0;

                return m_opers.WritableCapacityInSectors;
            }
        }

        /// <summary>
        /// This property returns the remaining writable capacity of the media in the
        /// drive in sectors.
        /// </summary>
        public long RemainingCapacityInWritableSectors
        {
            get
            {
                if (m_opers == null)
                    Initialize();

                if (m_opers == null)
                    return 0;

                return m_opers.RemainingCapacityInWritableSectors;
            }
        }

        /// <summary>
        /// Return the low level device assocaited with this drive
        /// </summary>
        public Device Device
        {
            get
            {
                return m_device;
            }
        }

        /// <summary>
        /// Return string describing the media manufacturer
        /// </summary>
        public string MediaManufacturer
        {
            get
            {
                return m_media_manufacturer;
            }
        }

        /// <summary>
        /// This property returns the media type as read from the DVD media
        /// </summary>
        public string MediaType
        {
            get
            {
                return m_media_type;
            }
        }

        /// <summary>
        /// This property returns the media version as read from the DVD media
        /// </summary>
        public byte MediaVersion
        {
            get
            {
                return m_media_version;
            }
        }

        /// <summary>
        /// This property returns TRUE if the disk in the drive is a CDRW.  This method
        /// works around the fact that some drives return a profile of CD_ROM for a CDRW
        /// media that is finalized, and other drives return CD_RW.
        /// </summary>
        public bool IsCDRW
        {
            get
            {
                if (CurrentProfile == SupportedProfiles.CD_RW)
                    return true;

                //
                // My HP DVD Writer DVD300E drive returns a profile of CD_ROM for a finalized
                // CD/RW disk.  Therefore if the profile is CD_ROM and the media is erasable
                // we return true here.  I think this is a "feature" so that a finalized disk
                // looks like CD_ROM and the computer cannot tell the difference.  However, it
                // makes is more difficult to tell what kind of media is really in the drive.
                //
                if (CurrentProfile == SupportedProfiles.CD_ROM)
                {
                    DiscInformation info;
                    Device.CommandStatus st = m_device.ReadDiscInformation(out info);
                    if (st != Device.CommandStatus.Success)
                        return false;

                    if (info.Erasable)
                        return true;
                }
                return false;
            }
        }


        /// <summary>
        /// Return the status of the disk in the drive as a string
        /// </summary>
        public string StatusString
        {
            get
            {
                string str ;

                if (CurrentProfile == SupportedProfiles.NoCurrentProfile)
                    Initialize();

                DiscInformation info;
                m_device.LogSenseState = false;
                m_device.ReadDiscInformation(out info);
                m_device.LogSenseState = true;

                switch (CurrentProfile)
                {
                    case SupportedProfiles.NoCurrentProfile:
                        str = "No media loaded";
                        break;
                    case SupportedProfiles.CD_R:
                        str = "CD R Disk";
                        break;
                    case SupportedProfiles.CD_ROM:
                        str = "CD ROM Disk";
                        break;
                    case SupportedProfiles.CD_RW:
                        str = "CD RW Disk";
                        break;
                    case SupportedProfiles.DVD_PLUS_R:
                        str = "DVD+R Disk";
                        break;
                    case SupportedProfiles.DVD_PLUS_RW:
                        str = "DVD+RW Disk";
                        break;
                    case SupportedProfiles.DVD_PLUS_R_DUAL_LAYER:
                        str = "DVD+RW DL Disk";
                        break;
                    case SupportedProfiles.DVD_RAM:
                        str = "DVD RAM Disk";
                        break;
                    case SupportedProfiles.DVD_ROM:
                        str = "DVD ROM Disk";
                        break;
                    case SupportedProfiles.DVD_R_SequentialRecording:
                        str = "DVD-R Disk";
                        break;
                    case SupportedProfiles.DVD_R_DualLayer_SequentialRecording:
                        str = "DVD-R DL Sequential Disk";
                        break;
                    case SupportedProfiles.DVD_R_DualLayer_LayerJumpRecording:
                        str = "DVD-R DL Layer Jump Recording";
                        break;
                    case SupportedProfiles.DVD_RW_RestrictedOverwrite:
                        str = "DVD-RW Restricted Overwrite Disk";
                        break;
                    case SupportedProfiles.DVD_RW_SequentialRecording:
                        str = "DVD-RW Sequential Disk";
                        break;
                    case SupportedProfiles.DVD_RW_RestrictedOverwrite_And_Sequential:
                        str = "DVD-RW Sequential Disk";
                        break;
                    case SupportedProfiles.BD_R_RANDOM_RECORDING:
                        str = "BD-R Random Recording";
                        break;
                    case SupportedProfiles.BD_R_SEQUENTIAL_RECORDING:
                        str = "BD-R Sequential Recording";
                        break;
                    case SupportedProfiles.BD_RE:
                        str = "BD-RE";
                        break;
                    case SupportedProfiles.BD_ROM:
                        str = "BD-ROM";
                        break;
                    case SupportedProfiles.HD_DVD_R:
                        str = "HD DVD-R";
                        break;
                    case SupportedProfiles.HD_DVD_RAM:
                        str = "HD DVD-RAM";
                        break;
                    case SupportedProfiles.HD_DVD_ROM:
                        str = "HD DVD-ROM";
                        break;
                    default:
                        str = "unknown disk format";
                        break;
                }
                
                if (info != null)
                    str += " - " + info.DiscStatus.ToString();

                if (m_media_manufacturer.Length > 0)
                    str += "   Media: " + m_media_manufacturer + " " + m_media_type + " Version " + m_media_version.ToString() ;
                return str;
            }
        }

        /// <summary>
        /// This property returns true if the drive can read CDText
        /// </summary>
        public bool CanReadCDText
        {
            get
            {
                DiskOperationError err;
                Feature feat;

                err = GetFeature(Feature.FeatureType.CDRead, out feat);
                if (err != null)
                    return false;

                if (feat.Data.GetLength(0) == 0)
                    return false;

                if ((feat.Data[0] & 1) == 0)
                    return false;

                return true;
            }
        }
        
        /// <summary>
        /// This property returns true if the drive is capable of doing DAP when doing a CD read
        /// </summary>
        public bool CanDoDAP
        {
            get
            {
                DiskOperationError err;
                Feature feat;

                err = GetFeature(Feature.FeatureType.CDRead, out feat);
                if (err != null)
                    return false;

                if (feat.Data.GetLength(0) == 0)
                    return false;

                if ((feat.Data[0] & 0x80) == 0)
                    return false;

                return true;
            }
        }
        #endregion

        #region constructor
        /// <summary>
        /// This is the constructor for this drive.  This constructor requires a pointer to a SCSI device
        /// that is used to talk to the specific device.
        /// </summary>
        /// <param name="dev">the SCSI device we are connected to</param>
        public Drive(Device dev)
        {
            m_device = dev;
            m_mode_page_speed = false;
        }
        #endregion

        #region Public Member Functions

        /// <summary>
        /// Format the media that is in the drive with the given format parameter list
        /// </summary>
        /// <param name="plist">the parameter list to use during the format</param>
        /// <returns>the error code, or null if success</returns>
        public DiskOperationError FormatUnit(FormatParameterList plist)
        {
            DiskOperationError status = null ;
            Device.CommandStatus st;

            m_current_operation = new ProgressText("Formatting Disk", "Formatting");

            m_device.PreventAllowMediumRemoval(Device.PreventAllow.Prevent);
            st = m_device.FormatUnit(plist);
            if (st != Device.CommandStatus.Success)
                status = Drive.CreateErrorObject(st, m_device);

            st = WaitForDeviceReady(true);
            if (st != Device.CommandStatus.Success)
                status = Drive.CreateErrorObject(st, m_device);

            m_device.PreventAllowMediumRemoval(Device.PreventAllow.Allow);
            TellProgress(100);

            return status;
        }

        /// <summary>
        /// Return the min, max, and preferred user area size for the bd-re disk
        /// </summary>
        /// <param name="minimum">min number of blocks</param>
        /// <param name="maximum">max number of blocks</param>
        /// <param name="preferred">preferred number of blocks</param>
        /// <param name="mediasize">media size with no spare area</param>
        /// <returns></returns>
        public DiskOperationError GetBDRESpareSizeValues(out long minimum, out long maximum, out long preferred, out long mediasize)
        {
            Device.CommandStatus st;
            bool Seen30 = false;
            bool Seen31 = false;

            minimum = -1 ;
            maximum = -1 ;
            preferred = -1 ;
            mediasize = -1 ;

            if (CurrentProfile != SupportedProfiles.BD_RE)
                return new BurnerError("GetBDRESpareSizeValues is only supported for BD-RE media");

            IList<CapacityDescriptor> caplist;
            st = m_device.ReadFormatCapacities(true, out caplist);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st, m_device);

            int i = 0;
            while (i < caplist.Count)
            {
                if (caplist[i].DescType == CapacityDescriptor.DescriptorType.Reserved)
                {
                    if (caplist[i].FormatType == 0x30)
                    {
                        // The first one is the preferred
                        preferred = caplist[i].NumberOfBlocks;
                        i++;

                        if (caplist[i].FormatType != 0x30)
                            return new BurnerError("the three BD-RE descriptors not found");

                        maximum = caplist[i].NumberOfBlocks;
                        i++;

                        if (caplist[i].FormatType != 0x30)
                            return new BurnerError("the three BD-RE descriptors not found");

                        minimum = caplist[i].NumberOfBlocks;
                        i++;
                    }
                    else if (caplist[i].FormatType == 0x31)
                    {
                        mediasize = caplist[i].NumberOfBlocks;
                        i++;
                        Seen31 = true;
                    }
                    else
                        i++;
                }
                else
                    i++;
            }

            if (!Seen30)
                return new BurnerError("no BD-RE (0x30) descriptors found");

            if (!Seen31)
                return new BurnerError("no BD-RE (0x31) descriptors found");

            return null;
        }

        /// <summary>
        /// Read the CD text for a given track, or read the global CD text if the
        /// track number given is zero.
        /// </summary>
        /// <param name="cdtext">the cdtext object containing the decoded text</param>
        /// <returns></returns>
        public DiskOperationError ReadCDText(out CDTextEncoderDecoder cdtext)
        {
            DiskOperationError error = null;
            Device.CommandStatus st;
            byte[] data;

            cdtext = null;

            // The global information stored in the leadin
            st = m_device.ReadCDText(out data);
            if (st == Device.CommandStatus.DeviceFailed)
            {
                data = new byte[1 * 96];

                // Try reading subchannel data from the leadin
                st = m_device.ReadSubChannel(4, 0xf0000000, 1, ref data);
                if (st != Device.CommandStatus.Success)
                    return CreateErrorObject(st, m_device);
            }
            try
            {
                cdtext = new CDTextEncoderDecoder(data, 4);
            }
            catch (Exception ex)
            {
                cdtext = null;
                error = new BurnerError(ex.Message);
            }
            return error;
        }

        /// <summary>
        /// For this drive and media, return the list of burning speeds that
        /// are supported
        /// </summary>g
        public DiskOperationError GetWriteSpeeds(out SpeedInfo[] speeds)
        {
            Device.CommandStatus st;
            SpeedDescriptorList slist;

            speeds = new SpeedInfo[0];

            if (m_mode_page_speed)
            {
                ModeTable table;

                st = m_device.ModeSense(Device.PageControl.Current, 0x2A, out table);
                if (st != Device.CommandStatus.Success)
                    return CreateErrorObject(st, m_device);

                ModePage pg = table.Pages[0];
                if (pg.PageCode != 0x2A)
                    return new BurnerError("Requested mode page 0x2A, got page " + pg.PageCode.ToString("X2"));


                if (pg.Length > 31)
                {
                    ushort len = pg.ModeGet16(0x30);
                    ushort addr = 0x32;
                    while (len-- > 0)
                    {
                        Device.RotationalControl rot = (Device.RotationalControl)(pg.PageData[addr + 1] & 0x03);
                        ushort speed = pg.ModeGet16(addr + 2);

                        SpeedInfo info = new SpeedInfo(rot, speed, !IsDvd);

                        addr += 4;
                    }
                }
            }
            else
            {
                st = m_device.GetSpeed(out slist);
                if (st == Device.CommandStatus.DeviceFailed && m_device.IsScsiError(0x24, 0))
                {
                    //
                    // Our preferred method of dealing with drive speeds is to use GetPerformance()
                    // and SetStreaming().  However, if this method fails with INVALID value in CDB, we
                    // switch to an alternate strategy using mode page 0x2a and SetCDSpeed
                    //
                    m_mode_page_speed = true;
                    return GetWriteSpeeds(out speeds);
                }
                else if (st != Device.CommandStatus.Success)
                    return CreateErrorObject(st, m_device);

                foreach (SpeedDescriptor d in slist)
                {
                    SpeedInfo info = new SpeedInfo(d, !IsDvd);
                    AddOne(ref speeds, info);
                }
            }

            return null;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <returns></returns>
        public override string ToString()
        {
            return DriveModel;
        }

        /// <summary>
        /// Called to close the underlying SCSI device.
        /// </summary>
        public void Dispose()
        {
            m_device.Dispose();
        }

        /// <summary>
        /// Read the buffer levels from the drive and store them
        /// </summary>
        public void ReadBufferLevels()
        {
            Device.CommandStatus st;

            int blen, bavail;
            st = m_device.ReadBufferCapacity(false, out blen, out bavail);
            if (st == Device.CommandStatus.Success)
            {
                m_buffer_valid = true;
                m_buffer_length = blen;
                m_buffer_avail = bavail;
            }
            else
                m_buffer_valid = false;
        }

        /// <summary>
        /// This method begins the timing of a CD/DVD operation
        /// </summary>
        public void OperationStart()
        {
            m_timing_active = true;
            m_operation_start = DateTime.Now;
        }

        /// <summary>
        /// This method ends the timing of a CD/DVD operation and adds the time elapsed since the last
        /// OperationStart() to the total operation time.
        /// </summary>
        public void OperationStop()
        {
            Debug.Assert(m_timing_active);

            m_operation_time += DateTime.Now - m_operation_start;
            m_timing_active = false;
        }

        /// <summary>
        /// This method initializes the operation time to zero in order to time a new
        /// operation that is about to start.
        /// </summary>
        public void OperationStatInitialize()
        {
            // Reset the time taken to zero
            m_operation_time = new TimeSpan(0);

            // Reset the bytes sent to zero
            m_burn_bytes = 0;
        }

        /// <summary>
        /// This method returns the state of a given feature
        /// </summary>
        /// <param name="ft">the feature of interest</param>
        /// <returns></returns>
        public FeatureState GetFeatureState(Feature.FeatureType ft)
        {
            FeatureState state = FeatureState.NotPresent;
            FeatureList flist;
            DiskOperationError status = GetFeatureList(Device.GetConfigType.AllFeatures, out flist) ;

            if (status != null)
                return FeatureState.Error;

            foreach (Feature f in flist.Features)
            {
                if (f.Code == ft)
                {
                    if (f.Current)
                        state = FeatureState.PresentAndCurrent;
                    else
                        state = FeatureState.PresentNotCurrent;

                    break;
                }
            }

            return state;
        }

        /// <summary>
        /// Retreive a given feature from the drive.
        /// </summary>
        /// <param name="ft">the type of the feature to retreive</param>
        /// <param name="feat">the return value for the feature</param>
        /// <returns>null if the operation completed sucessfully, othersize a disk operation error is returned</returns>
        public DiskOperationError GetFeature(Feature.FeatureType ft, out Feature feat)
        {
            FeatureList flist;
            DiskOperationError status ;

            feat = null;
            status = GetFeatureList(Device.GetConfigType.AllFeatures, out flist);

            if (status != null)
                return status;

            foreach (Feature f in flist.Features)
            {
                if (f.Code == ft)
                {
                    feat = f;
                    break;
                }
            }

            return null;
        }

        /// <summary>
        /// This property returns true if the loaded media is dual layer media
        /// </summary>
        public bool IsDualLayer
        {
            get
            {
                return CurrentProfile == SupportedProfiles.DVD_PLUS_R_DUAL_LAYER ||
                        CurrentProfile == SupportedProfiles.DVD_PLUS_RW_DUAL_LAYER ||
                        CurrentProfile == SupportedProfiles.DVD_R_DualLayer_LayerJumpRecording ||
                        CurrentProfile == SupportedProfiles.DVD_R_DualLayer_SequentialRecording ||
                        CurrentProfile == SupportedProfiles.DVD_RW_DualLayer;
            }
        }

        /// <summary>
        /// This property returns true if the media loaded is DVD media
        /// </summary>
        public bool IsDvd
        {
            get
            {
                return CurrentProfile == SupportedProfiles.DVD_PLUS_R ||
                        CurrentProfile == SupportedProfiles.DVD_PLUS_RW ||
                        CurrentProfile == SupportedProfiles.DVD_PLUS_R_DUAL_LAYER ||
                        CurrentProfile == SupportedProfiles.DVD_R_SequentialRecording ||
                        CurrentProfile == SupportedProfiles.DVD_RAM ||
                        CurrentProfile == SupportedProfiles.DVD_ROM ||
                        CurrentProfile == SupportedProfiles.DVD_RW_RestrictedOverwrite ||
                        CurrentProfile == SupportedProfiles.DVD_RW_RestrictedOverwrite_And_Sequential ||
                        CurrentProfile == SupportedProfiles.DVD_RW_SequentialRecording;
            }
        }

        /// <summary>
        /// This property returns true if the media loaded is CD media
        /// </summary>
        public bool IsCd
        {
            get
            {
                return CurrentProfile == SupportedProfiles.CD_R ||
                        CurrentProfile == SupportedProfiles.CD_RW ||
                        CurrentProfile == SupportedProfiles.CD_ROM;
            }
        }

        /// <summary>
        /// This property returns true if the media loaded is BluRay media
        /// </summary>
        public bool IsBluRay
        {
            get
            {
                return CurrentProfile == SupportedProfiles.BD_R_RANDOM_RECORDING ||
                        CurrentProfile == SupportedProfiles.BD_R_SEQUENTIAL_RECORDING ||
                        CurrentProfile == SupportedProfiles.BD_RE ||
                        CurrentProfile == SupportedProfiles.BD_ROM;
            }
        }

        /// <summary>
        /// This property returns true if the media loaded is HD DVD media
        /// </summary>
        public bool IsHdDvd
        {
            get
            {
                return CurrentProfile == SupportedProfiles.HD_DVD_R ||
                        CurrentProfile == SupportedProfiles.HD_DVD_RAM ||
                        CurrentProfile == SupportedProfiles.HD_DVD_ROM;
            }
        }

        /// <summary>
        /// Initialize the burner device, finding information about the hardware
        /// </summary>
        /// <returns></returns>
        public DiskOperationError Initialize()
        {
            Device.CommandStatus st;
            FeatureList flist;

            InquiryResult res ;
            st = m_device.Inquiry(out res);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st, m_device);

            m_connection = ConnectionType.Unknown;
            HardwareManager mgr = new HardwareManager();
            if (mgr.OpenMachine(string.Empty))
            {
                HardwareDevice dev = mgr.FindDevice("CDROM", res.VendorIdentification + " " + res.ProductIdentification);
                if (dev != null)
                {
                    if (dev.IsFirewire)
                        m_connection = ConnectionType.Firewire;
                    else if (dev.IsIde)
                        m_connection = ConnectionType.IDE;
                    else if (dev.IsScsi)
                        m_connection = ConnectionType.SCSI;
                    else if (dev.IsUsb)
                        m_connection = ConnectionType.USB;
                }
            }

            m_drive_model = m_connection.ToString() + ":" + res.VendorIdentification + " " + res.ProductIdentification + " " + res.ProductRevision;
            m_drive_model = m_drive_model.Replace('\0', ' ');
            m_drive_model = m_drive_model.Trim();


            // This is not really used for anything yet
            if (res.ResponseDataFormat == 2)
                MMCType = MMCDriveType.MMC_3;
            else
                MMCType = MMCDriveType.MMC_4;

            st = m_device.GetConfiguration(Device.GetConfigType.OneFeature, 0, out flist);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st, m_device);

            Feature f = flist.Features[0];
            ParseProfileList(f.Data);

            DiscInformation info;
            m_device.LogSenseState = false;
            st = m_device.ReadDiscInformation(out info);
            m_device.LogSenseState = true;

            m_opers = null ;
            switch(CurrentProfile)
            {
                case SupportedProfiles.CD_R:
                    m_opers = new CDRDiskOperations(this) ;
                    break ;

                case SupportedProfiles.CD_ROM:
                    if (info.Erasable || info.DiscStatus == DiscInformation.DiscStatusType.EmptyDisc)
                        m_opers = new CDRWDiskOperations(this);
                    else
                        m_opers = new CDRomDiskOperations(this);
                    break ;

                case SupportedProfiles.CD_RW:
                    m_opers = new CDRWDiskOperations(this);
                    break ;

                case SupportedProfiles.DVD_PLUS_R:
                case SupportedProfiles.DVD_PLUS_R_DUAL_LAYER:
                    m_opers = new DVDPlusRDiskOperations(this);
                    break ;

                case SupportedProfiles.DVD_PLUS_RW:
                    m_opers = new DVDPlusRWDiskOperations(this);
                    break ;

                case SupportedProfiles.DVD_RW_RestrictedOverwrite:
                case SupportedProfiles.DVD_RW_SequentialRecording:
                case SupportedProfiles.DVD_RW_RestrictedOverwrite_And_Sequential:
                    m_opers = new DVDMinusRWDiskOperations(this);
                    break;

                case SupportedProfiles.DVD_R_SequentialRecording:
                case SupportedProfiles.DVD_R_DualLayer_SequentialRecording:
                case SupportedProfiles.DVD_R_DualLayer_LayerJumpRecording:
                    m_opers = new DVDMinusRDiskOperations(this);
                    break;

                case SupportedProfiles.DVD_ROM:
                    m_opers = new CDRomDiskOperations(this);
                    break;

                case SupportedProfiles.BD_RE:
                    m_opers = new BDREDiskOperations(this);
                    break;

                // Not supported yet.  DVD-RAM shows up as either DVD_RAM or if the media is not
                // formatted, RemoveableDisk (at least in the one DVD-RAM drive I have)
                case SupportedProfiles.DVD_RAM:
                case SupportedProfiles.RemovableDisk:
                    
                // Not supported yet.  BD disk is coming online slowly
                case SupportedProfiles.BD_R_RANDOM_RECORDING:
                case SupportedProfiles.BD_R_SEQUENTIAL_RECORDING:
                case SupportedProfiles.BD_ROM:

                // Not supported yet.  I do not have a HDDVD drive
                case SupportedProfiles.HD_DVD_R:
                case SupportedProfiles.HD_DVD_RAM:
                case SupportedProfiles.HD_DVD_ROM:
                    m_opers = new NotSupportedDiskOperations(this);
                    break;

                case SupportedProfiles.NoCurrentProfile:
                    m_opers = new NotSupportedDiskOperations(this);
                    break;

                default:
                    return new BurnerError(CurrentProfile.ToString() + " unsupported media type in drive") ;
            }

            m_media_manufacturer = string.Empty ;
            m_media_type = string.Empty ;
            m_media_version = 0xff ;
            if (CurrentProfile == SupportedProfiles.DVD_PLUS_R || CurrentProfile == SupportedProfiles.DVD_PLUS_RW)
            {
                byte[] data = new byte[260];
                st = m_device.ReadDvdStructure(0, 0, 0x11, ref data);
                if (st == Device.CommandStatus.Success)
                {
                    m_media_manufacturer = "";
                    for (int i = 23; i < 31; i++)
                        m_media_manufacturer += (char)data[i];
                    m_media_type += (char)data[31];
                    m_media_type += (char)data[32];
                    m_media_type += (char)data[33];
                    m_media_version = data[34];
                }
            }

            return null;
        }

        /// <summary>
        /// This method returns a list of features for the device.
        /// </summary>
        /// <param name="type">the type of features to reterive</param>
        /// <param name="list">the output list of features</param>
        /// <returns></returns>
        public DiskOperationError GetFeatureList(Device.GetConfigType type, out FeatureList list)
        {
            Device.CommandStatus st;

            //
            // TODO -
            //
            // Need to check to see if feature data exceeds 64k and read the features out in
            // multiple chunks with differing start feature numbers.  I have not seen a drive do
            // this yet, but I only have a small number of drives that I am playing with.
            //
            st = m_device.GetConfiguration(type, 0, out list);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st, m_device);

            return null;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="et"></param>
        /// <returns></returns>
        public DiskOperationError Erase(EraseType et)
        {
            DiskOperationError status;
            string str;

            str = "Bwg.Burner.Drive.Erase() called on device '" + DriveModel + "'";
            m_device.GetLogger().LogMessage(new UserMessage(UserMessage.Category.Debug, 0, str));

            if (m_opers == null)
            {
                DiskOperationError err = Initialize();
                if (err != null)
                    return err;
            }

            m_buffer_valid = false;
            m_current_operation = new ProgressText("Erasing Disk", "Erasing");
            TellProgress(0);
            m_device.PreventAllowMediumRemoval(Device.PreventAllow.Prevent);
            status = m_opers.Erase(et);
            WaitDeviceReady(10);
            m_device.PreventAllowMediumRemoval(Device.PreventAllow.Allow);
            TellProgress(100);

            return status;
        }

        /// <summary>
        /// Burn a disk based on the given disk image.
        /// </summary>
        /// <param name="bt">the type of burn, either track at a time or session at a time, only applies to CD/R and CD/RW</param>
        /// <param name="d">the disk image to burn to the disk</param>
        /// <param name="multi_session">if true, leave the disk open for multiple sessions (if media is capable)</param>
        /// <param name="speed">the speed to burn at</param>
        /// <returns>a disk error object if an error occurs, otherwise null</returns>
        public DiskOperationError BurnDisk(Drive.BurnType bt, DiskBurnImage d, bool multi_session, SpeedInfo speed)
        {
            DiskOperationError status;
            string str ;

            str = "Bwg.Burner.Drive.BurnDisk() called on device '" + DriveModel + "'";
            m_device.GetLogger().LogMessage(new UserMessage(UserMessage.Category.Debug, 0, str));
            str = "    Options: burntype=" + bt.ToString() + ", multisession=" + multi_session.ToString();
            str += ", speed=" + speed;
            m_device.GetLogger().LogMessage(new UserMessage(UserMessage.Category.Debug, 0, str));

            if (m_opers == null)
            {
                DiskOperationError err = Initialize();
                if (err != null)
                    return err;
            }

            m_current_operation = new ProgressText("Burning Disc", "Burning");
            TellProgress(0);

            OperationStatInitialize();

            m_device.PreventAllowMediumRemoval(Device.PreventAllow.Prevent);
            status = m_opers.BurnDisk(bt, d, multi_session, speed, out m_burn_bytes);
            m_device.PreventAllowMediumRemoval(Device.PreventAllow.Allow);

            return status;
        }

        /// <summary>
        /// Write the data associated with a data track to a file
        /// </summary>
        /// <param name="t">The track to write</param>
        /// <param name="filename">The file to write the data to</param>
        public void WriteTrackToFile(Track t, string filename)
        {
            Device.CommandStatus st;
            int bufsize = 16 ;
            byte[] buffer = new byte [2048 * bufsize];
            FileStream strm = new FileStream(filename, FileMode.Create, FileAccess.Write);
            int toread = t.Info.TrackSize;
            int addr = t.Info.TrackStartAddress;
            int previous = 0;

            Debug.Assert(t.IsDataMode1 || t.IsDataMode2Form1);

            m_current_operation = new ProgressText("Extracting Track Data To File", "Extracting");
            TellProgress(0);

            while (toread > 0)
            {
                int len = toread;
                if (len > bufsize)
                    len = bufsize;

                int pcnt = (t.Info.TrackSize - toread) * 100 / t.Info.TrackSize;
                if (pcnt != previous)
                {
                    previous = pcnt;
                    TellProgress(previous);
                }

                st = m_device.Read(false, false, (uint)addr, (uint)len, ref buffer);
                if (st != Device.CommandStatus.Success)
                {
                    string str;

                    str = "Read failed at LBA " + addr.ToString();
                    throw new Exception(str);
                }

                strm.Write(buffer, 0, len * 2048);
                toread -= len;
                addr += len;
            }

            strm.Close();
        }

        #endregion

        #region internal functions (internal to this assembly)

        /// <summary>
        /// Close a track or a session
        /// </summary>
        /// <param name="closetype"></param>
        /// <param name="track"></param>
        /// <returns></returns>
        public DiskOperationError CloseTrackSession(Device.CloseTrackSessionType closetype, ushort track)
        {
            bool ready = false ;
            Device.CommandStatus st;
            string str;

            switch(closetype)
            {
                case Device.CloseTrackSessionType.CloseSessionMinimalRadius:
                case Device.CloseTrackSessionType.CloseSession:
                    str = "Close Sessions";
                    break ;

                case Device.CloseTrackSessionType.CloseTrack:
                    str = "Close Track";
                    break ;

                case Device.CloseTrackSessionType.FinalizeDiscWithMinimalRadius:
                case Device.CloseTrackSessionType.FinalizeDiscCompatible:
                    str = "Finalize Disk";
                    break ;

                case Device.CloseTrackSessionType.StopBackGroundFormat:
                    str = "Stop De-Icing";
                    break ;

                default:
                    str = "Unknown operation";
                    break;
            }

            m_current_operation = new ProgressText(str, str);
            TellProgress(0);

            st = m_device.CloseTrackSession(true, closetype, track);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st, m_device);

            m_device.LogSenseState = false;
            while (!ready)
            {
                ReadBufferLevels();

                st = m_device.TestUnitReady(out ready);
                if (st != Device.CommandStatus.Success)
                {
                    m_device.LogSenseState = true;
                    return CreateErrorObject(st, m_device);
                }

                if (m_device.HasSenseProgressInformation)
                {
                    double pcnt = (double)m_device.ProgressInformation / 65535.0 * 100;
                    TellProgress((int)pcnt) ;
                }
            }

            return null;
        }

        internal DiskOperationError SynchronizeCache()
        {
            bool ready = false;
            Device.CommandStatus st;

            OperationStart();
            st = m_device.SynchronizeCache(true);
            if (st != Device.CommandStatus.Success)
            {
                OperationStop();
                return CreateErrorObject(st, m_device);
            }

            m_device.LogSenseState = false;
            while (!ready)
            {
                ReadBufferLevels();

                st = m_device.TestUnitReady(out ready);
                if (st != Device.CommandStatus.Success)
                {
                    m_device.LogSenseState = true;
                    OperationStop();
                    return CreateErrorObject(st, m_device);
                }
            }
            OperationStop();

            return null;
        }

        /// <summary>
        /// Wait for the drive to become ready, possible marking progress while
        /// we wait.
        /// </summary>
        /// <param name="check_progress"></param>
        /// <returns></returns>
        public Device.CommandStatus WaitForDeviceReady(bool check_progress)
        {
            Device.CommandStatus st;
            bool ready;
            int last_sent = 0;

            m_device.LogSenseState = false;
            do
            {
                st = m_device.TestUnitReady(out ready);
                if (st != Device.CommandStatus.Success)
                {
                    m_device.LogSenseState = true;
                    return st;
                }
                if (ready)
                    break;

                //
                // If check_progress is true, we get the sense information from the drive and
                // see if we are receiving progress information
                //
                if (check_progress)
                {
                    double pcnt = 0.0;
                    bool pcntset = false ;

                    if (m_device.HasSenseProgressInformation)
                    {
                        pcnt = (double)m_device.ProgressInformation / 65535.0 * 100;
                        pcntset = true ;
                    }
                    else
                    {
                        st = m_device.RequestSense();
                        if (st == Device.CommandStatus.Success)
                        {
                            // Now, send progress indicators
                            if (m_device.HasSenseProgressInformation)
                            {
                                pcnt = (double)m_device.ProgressInformation / 65535.0 * 100;
                                pcntset = true ;
                            }
                        }
                    }
                    if (pcntset)
                        TellProgress((int)pcnt);
                }
                else if (m_total_duration != 0)
                {
                    // If we are not checking progress, we may be doing total duration
                    TimeSpan elapsed = DateTime.Now - m_start;
                    int pcnt = (int)(elapsed.TotalSeconds * 100 / m_total_duration);
                    if (pcnt > 100)
                        pcnt = 100;
                    if (pcnt != last_sent)
                    {
                        TellProgress(pcnt);
                        last_sent = pcnt;
                    }
                }

                // Sleep for a while and let the drive do its thing
                Thread.Sleep(1000);

            } while (true);

            m_device.LogSenseState = true;
            return Device.CommandStatus.Success;
        }

        internal void SetTimedOperation(int duration)
        {
            m_total_duration = duration;
            m_start = DateTime.Now;
        }

        internal DiskOperationError SendData(DataGenerator gen, TrackBurnImage t)
        {
            long totalcnt = t.DataToSend;
            long count = totalcnt;
            int previous = 0;
            Device.CommandStatus st;
            long lba = long.MaxValue;

            string lname = "Writing Track '" + t.Name + "'";
            m_current_operation = new ProgressText(lname, "Writing");

            string str = "burning track, total of " + count.ToString() + " 2K pages";
            m_device.GetLogger().LogMessage(new UserMessage(UserMessage.Category.Debug, 0, str));

            //
            // Now, create the data generator for the disk and start a
            // thread to begin to fill the data buffers
            //
            try
            {
                gen.StartGeneratorThread();

                OperationStart();
                while (true)
                {
                    WriteBuffer buf = gen.GetNextDataBuffer();
                    if (buf == null)
                        break;

                    if (buf.LogicalBlockAddress != long.MaxValue)
                        lba = buf.LogicalBlockAddress;

                    Debug.Assert(lba != long.MaxValue);

                    while (true)
                    {
                        m_device.DontDisplayIgnoreLongWriteInProgress = true;

                        st = m_device.Write(false, false, (int)lba, buf.SectorSize, buf.DataSize, buf.BufferPtr);
                        if (st == Device.CommandStatus.DeviceFailed && m_device.GetSenseAsc() == 0x04 && m_device.GetSenseAscq() == 0x08)
                        {
                            //
                            // This failure is a long write in progress.  This happens on some drives with some media.  We just
                            // ignore the failure and try again.
                            //
                            continue;
                        }

                        if (st != Device.CommandStatus.Success)
                        {
                            m_device.DontDisplayIgnoreLongWriteInProgress = false;
                            return CreateErrorObject(st, m_device);
                        }

                        m_device.DontDisplayIgnoreLongWriteInProgress = false;
                        break ;
                    }

                    ReadBufferLevels();

                    count -= (uint)buf.DataSize;
                    lba += buf.DataSize;
                    gen.FreeWriteBuffer(buf);

                    double pcnt = ((double)(totalcnt - count) / (double)totalcnt) * 100.0;
                    if ((int)pcnt != previous)
                    {
                        previous = (int)pcnt;
                        TellProgress(previous);
                    }
                }
                OperationStop();
            }
            catch (ThreadAbortException ex)
            {
                gen.StopGeneratorThread();
                throw ex;
            }

            if (gen.Status == false)
                return new DataGeneratorExceptionError(gen.Exception);

            return null;
        }

        internal DiskOperationError SendData(DataGenerator gen, DiskBurnImage disk)
        {
            long totalcnt = disk.DataToSend;
            long count = totalcnt;
            int previous = 0;
            Device.CommandStatus st;
            long lba = long.MaxValue;

            m_current_operation = new ProgressText("Writing Disc", "Writing");

            string str = "burning disk, total of " + count.ToString() + " 2K pages in image";
            m_device.GetLogger().LogMessage(new UserMessage(UserMessage.Category.Debug, 0, str));

            for (int i = 0; i < gen.Count; i++)
            {
                StreamElement elem = gen.GetSource(i);
                Debug.WriteLine("Stream: name = '" + elem.Name + "' length = " + elem.Length.ToString() + ", start = " + elem.StartSector.ToString());
            }

            //
            // Now, create the data generator for the disk and start a
            // thread to begin to fill the data buffers
            //
            bool last = false;
            try
            {
                gen.StartGeneratorThread();

                OperationStart();
                while (count > 0)
                {
                    if (last)
                        throw new Exception("Disk provide data too short");

                    WriteBuffer buf = gen.GetNextDataBuffer();
                    if (buf == null)
                    {
                        // This is an internal error - the disk should provide enough
                        // data to complete the disk
                        //
                        throw new Exception("Disk did not provide enough data");
                    }

                    if (buf.LogicalBlockAddress != long.MaxValue)
                    {
                        Debug.WriteLine("\tLBA changed from " + lba.ToString() + " to " + buf.LogicalBlockAddress.ToString());
                        lba = buf.LogicalBlockAddress;
                    }

                    Debug.Assert(lba != long.MaxValue);

                    while (true)
                    {
                        m_device.DontDisplayIgnoreLongWriteInProgress = true;

                        st = m_device.Write(false, false, (int)lba, buf.SectorSize, buf.DataSize, buf.BufferPtr);
                        if (st == Device.CommandStatus.DeviceFailed && m_device.GetSenseAsc() == 0x04 && m_device.GetSenseAscq() == 0x08)
                        {
                            // This is a long write in progress
                            continue;
                        }

                        if (st != Device.CommandStatus.Success)
                        {
                            m_device.DontDisplayIgnoreLongWriteInProgress = false;
                            return CreateErrorObject(st, m_device);
                        }

                        m_device.DontDisplayIgnoreLongWriteInProgress = false;
                        break;
                    }

                    count -= buf.DataSize;
                    lba += buf.DataSize;
                    gen.FreeWriteBuffer(buf);

                    ReadBufferLevels();

                    double pcnt = ((double)(totalcnt - count) / (double)totalcnt) * 100.0;
                    if ((int)pcnt != previous)
                    {
                        previous = (int)pcnt;
                        TellProgress(previous);
                    }
                }
                OperationStop();
            }
            catch (ThreadAbortException ex)
            {
                OperationStop();
                gen.StopGeneratorThread();
                throw ex;
            }

            if (gen.GetNextDataBuffer() != null)
                throw new Exception("Disk provided too much data");

            return null;
        }

        /// <summary>
        /// Create an error object based on the error that occurred
        /// </summary>
        /// <param name="st">the error status</param>
        /// <param name="dev">the device</param>
        /// <returns>the error object of the appropriate type</returns>
        static public DiskOperationError CreateErrorObject(Device.CommandStatus st, Device dev)
        {
            DiskOperationError err = null;

            if (st == Device.CommandStatus.IoctlFailed)
            {
                err = new DeviceIoctlError("DevControl", dev.LastError);
            }
            else if (st == Device.CommandStatus.DeviceFailed)
            {
                err = new ScsiDeviceError(dev);
            }
            else if (st == Device.CommandStatus.NotSupported)
            {
                err = new BurnerError("internal error - the operation request is not supported by the device");
            }

            return err;
        }

        internal DiskOperationError MakeDriveReady()
        {
            Device.CommandStatus st;
            bool ready;

            st = m_device.TestUnitReady(out ready);
            if (st == Device.CommandStatus.Success && ready)
                return null;

            m_device.StartStopUnit(false, Device.PowerControl.NoChange, Device.StartState.LoadDisc);
            Thread.Sleep(2000);

            for (int i = 0; i < 5; i++)
            {
                st = m_device.StartStopUnit(false, Device.PowerControl.NoChange, Device.StartState.StartDisc);
                if (st != Device.CommandStatus.Success)
                    return CreateErrorObject(st, m_device);

                st = m_device.TestUnitReady(out ready);
                if (st == Device.CommandStatus.Success && ready)
                    return null;
            }

            return CreateErrorObject(st, m_device);
        }

        internal DiskOperationError GetBurnProofSetting(out bool proof)
        {
            Device.CommandStatus st;
            ModeTable table;

            proof = false;

            st = m_device.ModeSense(Device.PageControl.Current, 0x2A, out table);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st, m_device);

            ModePage pg = table.Pages[0];
            if ((pg.PageData[4] & 0x80) != 0)
                proof = true;

            return null;
        }

        #endregion

        #region Private Member Functions

        void TellProgress(int percent)
        {
            if (ProgressChanged != null)
            {
                ProgressChangedEventArgs args = new ProgressChangedEventArgs(percent, m_current_operation);
                ProgressChanged(this, args);
            }
        }


        internal DiskOperationError SetSpeed(SpeedInfo speed)
        {
            Device.CommandStatus st;

            if (speed == null)
                st = m_device.SetCdSpeed(Device.RotationalControl.PureCav, Device.OptimumSpeed, Device.OptimumSpeed);
            else if (speed.IsModePage)
            {
                st = m_device.SetCdSpeed(speed.Rot, Device.OptimumSpeed, speed.Value);
            }
            else
            {
                st = m_device.SetStreaming(speed.Desc);
            }

            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st, m_device);

            ModeTable table ;
            st = m_device.ModeSense(Device.PageControl.Current, 0x2A, out table);
            if (st != Device.CommandStatus.Success)
                return CreateErrorObject(st, m_device);

            ModePage pg = table.Pages[0];
            int speedvalue = pg.ModeGet16(28);

            if (speed != null && speedvalue != speed.Desc.WriteSpeed)
            {
                if (!speed.IsModePage)
                {
                    st = m_device.SetCdSpeed(speed.Desc.WRC, (ushort)speed.Desc.ReadSpeed, (ushort)speed.Desc.WriteSpeed);
                    st = m_device.ModeSense(Device.PageControl.Current, 0x2A, out table);
                    if (st != Device.CommandStatus.Success)
                        return CreateErrorObject(st, m_device);

                    pg = table.Pages[0];
                    speedvalue = pg.ModeGet16(28);
                }

                if (speedvalue != speed.Desc.WriteSpeed)
                {
                    string str = "cannot set requested write speed, requested " + speed.Desc.WriteSpeed.ToString() + ", got " + speedvalue.ToString();
                    m_device.GetLogger().LogMessage(new UserMessage(UserMessage.Category.Warning, 0, str));
                }
            }
            return null;
        }

        private void AddOne(ref SpeedInfo[] A, SpeedInfo info)
        {
            Array.Resize(ref A, A.GetLength(0) + 1);
            A[A.GetLength(0) - 1] = info;
        }

        private bool WaitDeviceReady(int seconds)
        {
            bool ready;

            for(int i = 0 ; i < seconds ; i++)
            {
                m_device.LogSenseState = false;
                Device.CommandStatus st = m_device.TestUnitReady(out ready) ;
                m_device.LogSenseState = true; 
                if (st == Device.CommandStatus.Success && ready)
                    return true;

                Thread.Sleep(1000) ;
            }

            return false;
        }

        private void ParseProfileList(byte[] data)
        {
            int offset = 0;

            CurrentProfile = SupportedProfiles.NoCurrentProfile;
            ProfileList = new List<SupportedProfiles>();

            while (offset < data.GetLength(0))
            {
                ushort p = (ushort)((data[offset] << 8) | data[offset + 1]);
                ProfileList.Add((SupportedProfiles)p);

                //
                // There are times when there is more than one current profile.  This is true
                // for blank DVD-RW media that can be either restricted overwrite or sequential
                //
                if ((data[offset + 2] & 0x01) != 0 && CurrentProfile == SupportedProfiles.NoCurrentProfile)
                    CurrentProfile = (SupportedProfiles)p;
                else if (CurrentProfile == SupportedProfiles.DVD_RW_SequentialRecording &&
                         (SupportedProfiles)p == SupportedProfiles.DVD_RW_RestrictedOverwrite)
                {
                    CurrentProfile = SupportedProfiles.DVD_RW_RestrictedOverwrite_And_Sequential;
                }
                else if (CurrentProfile == SupportedProfiles.DVD_RW_RestrictedOverwrite &&
                       (SupportedProfiles)p == SupportedProfiles.DVD_RW_SequentialRecording)
                {
                    CurrentProfile = SupportedProfiles.DVD_RW_RestrictedOverwrite_And_Sequential;
                }

                offset += 4;
            }
        }

        #endregion 
    }
}
