//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//
using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Drawing ;
using Bwg.Logging;
using Bwg.Utils;

namespace Bwg.Controls
{
    /// <summary>
    /// This is a base form class for all BwgBurn based applications.  It provides some helper
    /// functions for finding application files.
    /// </summary>
    public class BwgBurnBaseForm : FormWithSettings
    {
        #region protected variables
        /// <summary>
        /// The error logger for the application
        /// </summary>
        protected Logger m_logger;

        /// <summary>
        /// The memory sink for the application
        /// </summary>
        protected MemorySink m_sink;

        #endregion

        #region constructor
        /// <summary>
        /// Constructor for the base form class.
        /// </summary>
        public BwgBurnBaseForm()
        {
            bool b;

            KeyPreview = true;

            // Create the logger and the assocaited sink
            m_logger = new Logger();
            m_sink = new MemorySink();

            m_logger.SetSink(UserMessage.Category.Error, m_sink);
            m_logger.SetSink(UserMessage.Category.Info, m_sink);

            //
            // Initialize the logger settings based on the user settings loaded above
            //
            b = GetStartupBool(PropertyNameLoggerDebugMessages, true);
            if (b)
                m_logger.SetSink(UserMessage.Category.Debug, m_sink);
            else
                m_logger.RemoveSink(UserMessage.Category.Debug);

            m_logger.SetLevel(UserMessage.Category.Debug, GetStartupInteger(PropertyNameLoggerDebugLevel, 9));

            b = GetStartupBool(PropertyNameLoggerWarningMessages, true);
            if (b)
                m_logger.SetSink(UserMessage.Category.Warning, m_sink);
            else
                m_logger.RemoveSink(UserMessage.Category.Warning);

            m_settings.WriteSettingsFile();

        }
        #endregion

        #region protected functions

        /// <summary>
        /// Move a control to the right of the target by moving it horizontally
        /// </summary>
        /// <param name="tomove">the control to move</param>
        /// <param name="target">the control to target</param>
        /// <param name="space">the space between the controls</param>
        protected void MoveRightOf(Control tomove, Control target, int space)
        {
            Point loc = tomove.Location;
            loc.X = target.Location.X + target.Size.Width + space;
            tomove.Location = loc;
        }

        /// <summary>
        /// Move a control to the right of the target and resize to control to move to fit the
        /// given total width.
        /// </summary>
        /// <param name="tomove">the control to move</param>
        /// <param name="target">the control to target</param>
        /// <param name="space">the space between the controls</param>
        /// <param name="total">The total width of the two controls with space</param>
        protected void MoveRightOfExpand(Control tomove, Control target, int space, int total)
        {
            Size s = new Size(total - target.Width - space, tomove.Height);

            MoveRightOf(tomove, target, space);
            tomove.Size = s;
        }

        /// <summary>
        /// Save the current logger settings
        /// </summary>
        protected void SaveLoggerSettings()
        {
            SetStartupBool(PropertyNameLoggerDebugMessages, m_logger.HasSink(UserMessage.Category.Debug));
            SetStartupInteger(PropertyNameLoggerDebugLevel, m_logger.GetLevel(UserMessage.Category.Debug));
            SetStartupBool(PropertyNameLoggerWarningMessages, m_logger.HasSink(UserMessage.Category.Warning));
        }

        /// <summary>
        /// This method returns the version number for the BwgBurn package
        /// </summary>
        /// <returns>version number string</returns>
        protected string GetVersion()
        {
            return "0.6.5";
        }

        /// <summary>
        /// Called on help
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="args"></param>
        protected virtual void OnControlHelp(object sender, EventArgs args)
        {
        }

        /// <summary>
        /// This method is called when a key is pressed
        /// </summary>
        /// <param name="e"></param>
        protected override void OnKeyDown(KeyEventArgs e)
        {
            if (e.KeyCode == Keys.F1)
            {
                Point pt = MousePosition ;
                Control c = FindSelectedControl(this, pt);
                if (c != null)
                    OnControlHelp(c, null);
                e.Handled = true;
            }
            base.OnKeyDown(e);
        }

        /// <summary>
        /// Return the full path to the CHM file that contains help information for
        /// the various BwgBurn applications.
        /// </summary>
        /// <returns></returns>
        protected string GetHelpFilePath()
        {
            string dir;

#if DEBUG
            dir = Path.GetDirectoryName(Application.ExecutablePath); ;
            dir = Path.GetDirectoryName(dir);
            dir = Path.GetDirectoryName(dir);
            dir = Path.GetDirectoryName(dir);
            dir = Path.Combine(dir, "help");
#else
            dir = Path.GetDirectoryName(Application.ExecutablePath); ;
            dir = Path.GetDirectoryName(dir) ;
            dir = Path.Combine(dir, "docs") ;
#endif
            dir = Path.Combine(dir, "bwgburn.chm");

            return dir;
        }

        /// <summary>
        /// Return the full path to the Copyright file that contains the GNU
        /// copyright information
        /// </summary>
        /// <returns></returns>
        protected string GetCopyrightFile()
        {
            string dir;
#if DEBUG
            dir = Path.GetDirectoryName(Application.ExecutablePath); ;
            dir = Path.GetDirectoryName(dir);
            dir = Path.GetDirectoryName(dir);
            dir = Path.GetDirectoryName(dir);
#else
            dir = Path.GetDirectoryName(Application.ExecutablePath); ;
            dir = Path.GetDirectoryName(dir) ;
            dir = Path.Combine(dir, "docs") ;
#endif
            dir = Path.Combine(dir, "Copyright.rtf");

            return dir;
        }

        /// <summary>
        /// This method returns the base directory of the music library
        /// </summary>
        /// <returns></returns>
        protected string GetMusicLibraryDirectory()
        {
            return Environment.GetFolderPath(Environment.SpecialFolder.MyMusic);
        }
        #endregion

        #region private functions
        private Control FindSelectedControl(Control c, Point pt)
        {
            foreach (Control ctl in c.Controls)
            {
                if (ctl.Visible)
                {
                    Rectangle b = ctl.Parent.RectangleToScreen(ctl.Bounds);
                    if (b.Contains(pt))
                    {
                        // See if a child contains the point
                        Control child = FindSelectedControl(ctl, pt);
                        if (child == null)
                            return ctl;
                        else
                            return child;
                    }
                }
            }

            return null;
        }

        private void SaveSettings()
        {
            string filename = Path.Combine(Application.StartupPath, "settings");
        }

        private void LoadSettings()
        {
        }
        #endregion
    }
}
