//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.IO;
using System.Windows.Forms;
using Bwg.Isofs ;
using Bwg.Utils;

namespace Bwg.Controls
{
    public partial class DiskImageListView : ListView
    {
        #region private member variables
        private DiskImageElement.NameType m_nametype;
        private DiskImageContents m_image;
        private DiskImageDirectory m_dir;
        private Point m_menu_point;
        private ShellIconMgr m_icon_mgr;
        #endregion

        #region public member variables
        /// <summary>
        /// The list of controls to disable while adding to the image
        /// </summary>
        public IList<Control> DisableControls;
        #endregion

        #region constructor
        /// <summary>
        /// Create the disk image directory view list
        /// </summary>
        public DiskImageListView()
        {
            InitializeComponent();
            SetupColumns();

            SmallImageList = FileImages;

            ContextMenu = new ContextMenu();
            ContextMenu.MenuItems.Add(new MenuItem("Delete", new EventHandler(DeleteMenuItemHandler)));
            ContextMenu.MenuItems.Add(new MenuItem("Rename", new EventHandler(RenameMenuItemHandler)));

            LabelEdit = true ;
            AllowDrop = true;

            m_nametype = DiskImageElement.NameType.Local;

            DisableControls = new List<Control>();
        }
        #endregion

        #region public properties

        /// <summary>
        /// This property gets/sets the type of anme to use in the disk image tree
        /// </summary>
        public DiskImageElement.NameType NameType
        {
            get
            {
                return m_nametype;
            }
            set
            {
                m_nametype = value;
                RefreshAll();
            }
        }

        /// <summary>
        /// This property sets or removes the icon manager assocaited with this list view
        /// </summary>
        public ShellIconMgr IconManager
        {
            get
            {
                return m_icon_mgr;
            }
            set
            {
                m_icon_mgr = value;
                if (m_icon_mgr != null)
                    SmallImageList = m_icon_mgr.ImageList;
            }
        }


        /// <summary>
        /// This property is the disk image creator that is being displayed
        /// </summary>
        public DiskImageContents ImageCreator
        {
            get
            {
                return m_image;
            }
            set
            {
                if (m_image != null)
                    m_image.ImageChanged -= new EventHandler<ImageChangedEventArgs>(Creator_ImageChanged);

                m_image = value;
                if (m_image != null)
                    m_image.ImageChanged += new EventHandler<ImageChangedEventArgs>(Creator_ImageChanged);
            }
        }

        /// <summary>
        /// This property is the disk image Directory being displayed
        /// </summary>
        public DiskImageDirectory Directory
        {
            get
            {
                return m_dir;
            }
            set
            {
                if (m_dir != null)
                    m_dir.Changed -= new EventHandler<DiskImageDirectoryChangedArgs>(DirectoryChanged);

                m_dir = value;

                Items.Clear();

                if (m_dir != null)
                {
                    m_dir.Changed += new EventHandler<DiskImageDirectoryChangedArgs>(DirectoryChanged);
                    UpdateList();
                }
            }
        }
        #endregion

        #region private or protected functions

        private void RefreshAll()
        {
            foreach(ListViewItem item in Items)
            {
                if (item.Tag != null && item.Tag is DiskImageElement)
                {
                    DiskImageElement elem = (DiskImageElement)item.Tag;
                    item.Text = elem.GetName(m_nametype);
                }
            }
        }

        /// <summary>
        /// This is called when a key is pressed
        /// </summary>
        /// <param name="e"></param>
        protected override void OnKeyDown(KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Delete)
            {
                foreach (ListViewItem item in SelectedItems)
                {
                    if (item.Tag != null && item.Tag is DiskImageElement)
                        m_image.Remove((DiskImageElement)item.Tag);
                }
            }
            else if (e.KeyCode == Keys.A && e.Control)
            {
                foreach (ListViewItem item in Items)
                    item.Selected = true;
            }
            else
            {
                base.OnKeyDown(e);
            }
        }

        /// <summary>
        /// This method is called when a disk image list item is double clicked
        /// </summary>
        /// <param name="e"></param>
        protected override void OnDoubleClick(EventArgs e)
        {
            if (SelectedItems[0].Tag != null && SelectedItems[0].Tag is DiskImageDirectory)
            {
                Directory = (DiskImageDirectory)SelectedItems[0].Tag;
            }
            base.OnDoubleClick(e);
        }

        private void DirectoryChanged(object sender, DiskImageDirectoryChangedArgs args)
        {
            Items.Clear() ;
            UpdateList() ;
        }

        private ListViewItem FindItemByElem(DiskImageElement elem)
        {
            foreach(ListViewItem item in Items)
            {
                if (item.Tag != null && item.Tag is DiskImageElement && (DiskImageElement)item.Tag == elem)
                    return item ;
            }
            return null ;
        }

        private void Creator_ImageChanged(object sender, ImageChangedEventArgs args)
        {
            ListViewItem item = FindItemByElem(args.Element);
            if (item != null)
            {
                switch (args.Change)
                {
                    case ImageChangedEventArgs.ChangeType.Removed:
                        Items.Remove(item);
                        break;

                    case ImageChangedEventArgs.ChangeType.Renamed:
                        item.Text = args.Element.GetName(m_nametype);
                        break;
                }
            }
        }

        /// <summary>
        /// This method is called when the down button is pressed with the moust
        /// </summary>
        /// <param name="e"></param>
        protected override void OnMouseDown(MouseEventArgs e)
        {
            m_menu_point = e.Location;
            if (HitTest(m_menu_point).Item != null)
                base.OnMouseDown(e);
        }

        private void DeleteMenuItemHandler(object sender, EventArgs args)
        {
            ListViewHitTestInfo info = HitTest(m_menu_point);
            if (info.Item != null && info.Item.Tag is DiskImageElement)
                m_image.Remove((DiskImageElement)info.Item.Tag);
        }

        private void RenameMenuItemHandler(object sender, EventArgs args)
        {
        }

        private void SetupColumns()
        {
            ColumnHeader hdr;

            FullRowSelect = true;
            View = View.Details;
            Columns.Clear();

            hdr = new ColumnHeader("Name");
            hdr.Width = 300;
            hdr.Text = "Name";
            Columns.Add(hdr);

            hdr = new ColumnHeader("Size");
            hdr.Width = 120;
            hdr.Text = "Size";
            Columns.Add(hdr);
        }

        private void AddItem(DiskImageDirectory d)
        {
            ListViewItem item = new ListViewItem(d.GetName(m_nametype));
            item.Tag = d;
            item.ImageIndex = m_icon_mgr.GetFolderIconIndex(string.Empty, false);
            item.ImageIndex = 0;
            Items.Add(item);
        }

        private void AddItem(DiskImageFile f)
        {
            ListViewItem item = new ListViewItem(f.GetName(m_nametype));
            item.ImageIndex = m_icon_mgr.GetFileIconIndex(f.FileSystemLocation);
            item.Tag = f;
            item.SubItems.Add(f.FileSizeInBytes.ToString());
            Items.Add(item);
        }

        private void RemoveItem(DiskImageFile f)
        {
            ListViewItem todel = null;

            foreach (ListViewItem item in Items)
            {
                if (item.Tag == f)
                {
                    todel = item;
                    break;
                }
            }
            if (todel != null)
                Items.Remove(todel);
        }

        private void UpdateList()
        {
            Items.Clear();
            if (m_dir != null)
            {
                foreach (DiskImageElement elem in m_dir)
                {
                    if (elem is DiskImageDirectory)
                        AddItem((DiskImageDirectory)elem);
                }
                foreach (DiskImageElement elem in m_dir)
                {
                    if (elem is DiskImageFile)
                        AddItem((DiskImageFile)elem);
                }
            }
        }

        /// <summary>
        /// Called when the size of the control changes.  This is used to resize the columns
        /// in the window
        /// </summary>
        /// <param name="e">arguments to the event - not used</param>
        protected override void OnSizeChanged(EventArgs e)
        {
            if (Columns.Count > 0)
                Columns[1].Width = Width - Columns[0].Width - 4;

            base.OnSizeChanged(e);
        }

        /// <summary>
        /// Called when a file or directory is dropped on the image view
        /// </summary>
        /// <param name="drgevent">the arguments giving the directory or file dropped</param>
        protected override void OnDragDrop(DragEventArgs drgevent)
        {
            base.OnDragDrop(drgevent);

            object obj = drgevent.Data.GetData("FileNameW", true);
            string[] namearray = (string[])obj;

            Application.UseWaitCursor = true;
            foreach (Control c in DisableControls)
                c.Enabled = false;

            // Now we must add the file or directory to the ISO image based
            // on the drop location
            Point pt = PointToClient(new Point(drgevent.X, drgevent.Y));
            ListViewHitTestInfo info = HitTest(pt);

            // The base ISO path
            DiskImageDirectory dir ;
            if (info.Item != null && info.Item.Tag is DiskImageDirectory)
                dir = (DiskImageDirectory)info.Item.Tag;
            else
                dir = m_dir;

            foreach (string fname in namearray)
            {
                if (File.Exists(fname))
                {
                    string filename = Path.GetFileName(fname);
                    string isopath = dir.GetPath(m_nametype) + "/" + filename;
                    m_image.Add(isopath, fname);
                }
                else if (System.IO.Directory.Exists(fname))
                {
                    string filename = Path.GetFileName(fname);
                    string isopath = dir.GetPath(m_nametype) + "/" + filename;
                    DiskImageElement elem = m_image.Add(isopath, fname);
                }
                else
                {
                    string msg = "The dropped element '" + fname + "' does not exist";
                    MessageBox.Show(msg, "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }

            Directory = m_dir;

            Application.UseWaitCursor = false;
            foreach (Control c in DisableControls)
                c.Enabled = true;
        }

        /// <summary>
        /// Called when a drag and drop operation is drug over the image window
        /// </summary>
        /// <param name="drgevent"></param>
        protected override void OnDragOver(DragEventArgs drgevent)
        {
            base.OnDragOver(drgevent);
            drgevent.Effect = DragDropEffects.Copy;
        }

        #endregion
    }
}
