//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using Bwg.Utils ;

namespace Bwg.Controls
{
    /// <summary>
    /// This class provides a view of a directory in the local file system as a ListView
    /// </summary>
    public partial class FileSystemListView : ListView
    {
        private string m_dir_path;
        private IList<string> m_ext_list;
        private bool m_include_dirs;
        private Rectangle m_drag_rect;
        private ShellIconMgr m_icon_mgr;

        /// <summary>
        /// Construct the file system view control
        /// </summary>
        public FileSystemListView()
        {
            InitializeComponent();
            SetupColumns();

            m_ext_list = null;
            m_include_dirs = true;

            m_icon_mgr = new ShellIconMgr();
        }

        /// <summary>
        /// This property sets or removes the icon manager assocaited with this list view
        /// </summary>
        public ShellIconMgr IconManager
        {
            get
            {
                return m_icon_mgr;
            }
            set
            {
                m_icon_mgr = value;
                SmallImageList = m_icon_mgr.ImageList;
            }
        }

        /// <summary>
        /// This property indicates whether or not directories are included in the
        /// list.
        /// </summary>
        public bool IncludeDirs
        {
            get
            {
                return m_include_dirs;
            }
            set
            {
                m_include_dirs = value;
            }
        }

        /// <summary>
        /// The path of the directory to display
        /// </summary>
        public string Path
        {
            get
            {
                return m_dir_path;
            }
            set
            {
                m_dir_path = value;
                UpdateList();
            }
        }

        /// <summary>
        /// Add an extension to the list of extensions that are 
        /// displayed in this window
        /// </summary>
        /// <param name="ext"></param>
        public void AddExt(string ext)
        {
            if (m_ext_list == null)
                m_ext_list = new List<string>();

            m_ext_list.Add(ext);
        }

        /// <summary>
        /// This method is called when a key is pressed.  This event is used to look for the
        /// Ctrl-A key sequence to select everything in the current window
        /// </summary>
        /// <param name="e"></param>
        protected override void OnKeyDown(KeyEventArgs e)
        {
            if (e.KeyCode == Keys.A && e.Control)
            {
                foreach(ListViewItem item in Items)
                    item.Selected = true;
            }

            base.OnKeyDown(e);
        }

        private void SetupColumns()
        {
            ColumnHeader hdr;

            FullRowSelect = true;
            View = View.Details;
            Columns.Clear();

            hdr = new ColumnHeader("Name");
            hdr.Width = 300;
            hdr.Text = "Name";
            Columns.Add(hdr);

            hdr = new ColumnHeader("Size");
            hdr.Width = 120;
            hdr.Text = "Size";
            Columns.Add(hdr);
        }

        private void UpdateList()
        {
            Items.Clear();
            if (!Directory.Exists(m_dir_path))
                return;

            if (m_include_dirs)
            {
                try
                {
                    string[] dirs = Directory.GetDirectories(m_dir_path);

                    foreach (string d in dirs)
                    {
                        ListViewItem item = new ListViewItem(System.IO.Path.GetFileName(d));
                        item.Tag = d;
                        item.ImageIndex = m_icon_mgr.GetFolderIconIndex(d, false);
                        Items.Add(item);
                    }
                }
                catch (Exception ex)
                {
                    ListViewItem item = new ListViewItem("cannot list directories - " + ex.Message);
                    item.Tag = null;
                    Items.Add(item);
                }
            }

            try
            {
                string[] files = Directory.GetFiles(m_dir_path);
                foreach (string f in files)
                {
                    if (m_ext_list != null)
                    {
                        string ext = System.IO.Path.GetExtension(f);
                        if (ext.Length > 0 && ext[0] == '.')
                            ext = ext.Substring(1);
                        if (!m_ext_list.Contains(ext))
                            continue;
                    }
                    FileInfo info = new FileInfo(f);
                    string filename = System.IO.Path.GetFileName(f);
                    ListViewItem item = new ListViewItem(filename);
                    item.ImageIndex = m_icon_mgr.GetFileIconIndex(filename);
                    item.Tag = f;
                    item.SubItems.Add(info.Length.ToString());
                    Items.Add(item);
                }
            }
            catch (Exception ex)
            {
                ListViewItem item = new ListViewItem("cannot list directories - " + ex.Message);
                item.Tag = null;
                Items.Add(item);
            }
        }

        /// <summary>
        /// Called when the size of the control changes
        /// </summary>
        /// <param name="e">the args for the event handler</param>
        protected override void OnSizeChanged(EventArgs e)
        {
            if (Columns.Count > 0)
                Columns[1].Width = Width - Columns[0].Width - 4;
            base.OnSizeChanged(e);
        }

        /// <summary>
        /// Called when the mouse button is pressed over the control
        /// </summary>
        /// <param name="e">arguments giving the location of the press</param>
        protected override void OnMouseDown(MouseEventArgs e)
        {
            ListViewHitTestInfo info = HitTest(e.Location);
            if (info.Item != null)
            {
                m_drag_rect = new Rectangle(e.Location, SystemInformation.DragSize);
            }
            else
            {
                m_drag_rect = Rectangle.Empty;
            }
            base.OnMouseDown(e);
        }

        /// <summary>
        /// Called when the mouse button is released
        /// </summary>
        /// <param name="e">the location of the mouse up event</param>
        protected override void OnMouseUp(MouseEventArgs e)
        {
            m_drag_rect = Rectangle.Empty;
            base.OnMouseUp(e);
        }

        /// <summary>
        /// Called when the mouse if moved in the control
        /// </summary>
        /// <param name="e">the location of the mouse move event</param>
        protected override void OnMouseMove(MouseEventArgs e)
        {
            DragDropEffects effects;

            if ((e.Button & MouseButtons.Left) == MouseButtons.Left && m_drag_rect != Rectangle.Empty
                && !m_drag_rect.Contains(e.Location))
            {
                ListViewHitTestInfo info = HitTest(e.Location);

                if (info.Item != null && info.Item.Tag is string)
                {
                    int i = 0 ;
                    string[] arr = new string[SelectedItems.Count];
                    foreach(ListViewItem item in SelectedItems)
                        arr[i++] = (string)item.Tag;

                    DataObject dobj = new DataObject(DataFormats.FileDrop, arr);
                    effects = DoDragDrop(dobj, DragDropEffects.All);
                }
            }
            base.OnMouseMove(e);
        }

        /// <summary>
        /// This is called when a double click happens
        /// </summary>
        /// <param name="e"></param>
        protected override void OnMouseDoubleClick(MouseEventArgs e)
        {
            ListViewHitTestInfo info = HitTest(e.Location);
            if (info.Item != null)
            {
                if (info.Item.Tag != null && info.Item.Tag is string)
                {
                    if (Directory.Exists((string)(info.Item.Tag)))
                        Path = (string)info.Item.Tag;
                }
            }

            base.OnMouseDoubleClick(e);
        }
    }
}
