//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using Bwg.Utils;

namespace Bwg.Controls
{
    public partial class FileSystemTreeView : TreeView
    {
        /// <summary>
        /// The type of folder for special folders
        /// </summary>
        public enum FSLocation
        {
            /// <summary>
            /// The computer desktop
            /// </summary>
            Desktop,

            /// <summary>
            /// The My Documents folder
            /// </summary>
            MyDocuments,

            /// <summary>
            /// The MyComputer folder
            /// </summary>
            MyComputer,

            /// <summary>
            /// 
            /// </summary>
            Personel,
        } ;

        private Rectangle m_drag_rect;
        private TreeNode m_drag_node;
        private const string DummyNodeText = "DUMMY";
        private ShellIconMgr m_icon_mgr;

        /// <summary>
        /// Create the file system tree view control
        /// </summary>
        public FileSystemTreeView()
        {
            InitializeComponent();
            m_drag_rect = Rectangle.Empty;
        }

        /// <summary>
        /// This property is the ICON manager for the icons
        /// </summary>
        public ShellIconMgr IconManager
        {
            get
            {
                return m_icon_mgr;
            }
            set
            {
                m_icon_mgr = value;
                if (m_icon_mgr != null)
                    ImageList =  m_icon_mgr.ImageList;
            }
        }

        /// <summary>
        /// Set the selected path for an element.
        /// </summary>
        /// <param name="path">the path to the node to be the selected node</param>
        public void SetSelectedPath(string path)
        {
            char[] seps = new char[1] { '\\' };
            string[] elems = path.Split(seps);

            TreeNodeCollection coll = Nodes;
            foreach (string str in elems)
            {
                TreeNode found = null;

                // Find str in the collectoin
                foreach (TreeNode n in coll)
                {
                    if (n.Text == str)
                    {
                        found = n;
                        break;
                    }
                }
                if (found == null)
                    break;

                // Expand the node we just found so its contents are valid
                found.Expand();
                TreeViewCancelEventArgs args = new TreeViewCancelEventArgs(found, false, TreeViewAction.Expand);
                OnBeforeExpand(args);

                SelectedNode = found;
                coll = found.Nodes;
            }
        }

        /// <summary>
        /// Get a path to the selected node
        /// </summary>
        /// <returns>a string representing a path</returns>
        public string GetSelectedPath()
        {
            string result = string.Empty;
            TreeNode node = SelectedNode;

            while (node != null)
            {
                result = Path.Combine(node.Text, result);
                node = node.Parent;
            }

            return result;
        }

        /// <summary>
        /// Populate the contents of the tree view control from the local disk
        /// </summary>
        public void Populate()
        {
            TreeNode desktop = new TreeNode("Desktop");
            desktop.ImageIndex = m_icon_mgr.GetFolderIconIndex(string.Empty, false);
            desktop.Tag = FSLocation.Desktop;
            desktop.Nodes.Add(DummyNodeText);
            Nodes.Add(desktop);

        }

        private bool IsDummy(TreeNode n)
        {
            return n.Nodes.Count == 1 && n.Nodes[0].Text == DummyNodeText;
        }

        private void ExpandDesktop(TreeNode n)
        {
            TreeNode node;

            // Remove the dummy node
            n.Nodes.Clear();

            node = new TreeNode("My Documents");
            node.Tag = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments);
            node.ImageIndex = m_icon_mgr.GetFolderIconIndex((string)node.Tag, false);
            node.Nodes.Add(DummyNodeText);
            n.Nodes.Add(node);

            node = new TreeNode("My Computer");
            node.Tag = FSLocation.MyComputer;
            node.ImageIndex = m_icon_mgr.GetFolderIconIndex(string.Empty, false);
            node.Nodes.Add(DummyNodeText);
            n.Nodes.Add(node);

            node = new TreeNode("My Music");
            node.Tag = Environment.GetFolderPath(Environment.SpecialFolder.MyMusic);
            node.ImageIndex = m_icon_mgr.GetFolderIconIndex(string.Empty, false);
            node.Nodes.Add(DummyNodeText);
            n.Nodes.Add(node);

            node = new TreeNode("My Pictures");
            node.Tag = Environment.GetFolderPath(Environment.SpecialFolder.MyPictures);
            node.ImageIndex = m_icon_mgr.GetFolderIconIndex(string.Empty, false);
            node.Nodes.Add(DummyNodeText);
            n.Nodes.Add(node);

            String path = Environment.GetFolderPath(Environment.SpecialFolder.Desktop);
            ExpandDirInNode(n, path);
        }

        private void ExpandMyComputer(TreeNode n)
        {
            // Remove the dummy node
            n.Nodes.Clear();

            foreach (DriveInfo di in DriveInfo.GetDrives())
            {
                string text = "";

                switch (di.DriveType)
                {
                    case DriveType.CDRom:
                        text = "CDROM (" + di.Name[0] + ":)";
                        break;
                    case DriveType.Fixed:
                        text = "Local Disk (" + di.Name[0] + ":)";
                        break;
                    default:
                        text = "Disk (" + di.Name[0] + ":)";
                        break;
                }

                TreeNode dsknode = new TreeNode(text);
                string path = di.Name;
                dsknode.Tag = path;
                dsknode.ImageIndex = m_icon_mgr.GetFolderIconIndex(path, false);
                dsknode.Nodes.Add(DummyNodeText);
                n.Nodes.Add(dsknode);
            }
        }

        /// <summary>
        /// Populate the contents of the node that is about to be expanded.
        /// </summary>
        /// <param name="e">the arguments giving the node to be expanded</param>
        protected override void OnBeforeExpand(TreeViewCancelEventArgs e)
        {
            if (!IsDummy(e.Node))
                return;

            if (e.Node.Tag is FSLocation)
            {
                switch ((FSLocation)e.Node.Tag)
                {
                    case FSLocation.Desktop:
                        ExpandDesktop(e.Node);
                        break;

                    case FSLocation.MyComputer:
                        ExpandMyComputer(e.Node);
                        break;
                }

                e.Node.ImageIndex = m_icon_mgr.GetFolderIconIndex(string.Empty, true);
            }
            else if (e.Node.Tag is string)
            {
                e.Node.Nodes.Clear();
                try
                {
                    ExpandDirInNode(e.Node, (string)e.Node.Tag);
                    e.Node.ImageIndex = m_icon_mgr.GetFolderIconIndex((string)e.Node.Tag, true);
                }
                catch (Exception ex)
                {
                    string str = "cannot expand this directory - " + ex.Message ;
                    MessageBox.Show(str, "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    e.Cancel = true;
                }
            }
            else
            {
                string str = "internal error in file system tree, node tag is not of correct type" ;
                MessageBox.Show(str, "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
                e.Cancel = true;
            }
        }

        /// <summary>
        /// Called after a node is collapsed, used to remove the contents of the node
        /// as node contents are generated when opened.
        /// </summary>
        /// <param name="e"></param>
        protected override void OnAfterCollapse(TreeViewEventArgs e)
        {
            e.Node.Nodes.Clear();

            // Add the dummy node
            e.Node.Nodes.Add(DummyNodeText);
            e.Node.ImageIndex = m_icon_mgr.GetFolderIconIndex(string.Empty, false);
        }

        private void ExpandDirInNode(TreeNode node, string path)
        {
            string[] dirs = Directory.GetDirectories(path);
            foreach (string dir in dirs)
            {
                TreeNode dirnode = new TreeNode(Path.GetFileName(dir));
                dirnode.Tag = dir;
                dirnode.Nodes.Add(DummyNodeText);
                node.Nodes.Add(dirnode);
            }
        }

        /// <summary>
        /// Called when the mouse button is pressed in this control
        /// </summary>
        /// <param name="e">the location of the event</param>
        protected override void OnMouseDown(MouseEventArgs e)
        {
            TreeViewHitTestInfo info = HitTest(e.Location);
            if (info.Node != null)
            {
                m_drag_rect = new Rectangle(e.Location, SystemInformation.DragSize);
                m_drag_node = info.Node;
            }
            else
            {
                m_drag_rect = Rectangle.Empty;
            }

            base.OnMouseDown(e);
        }

        /// <summary>
        /// called when the mouse button is released
        /// </summary>
        /// <param name="e">the location of the event</param>
        protected override void OnMouseUp(MouseEventArgs e)
        {
            m_drag_rect = Rectangle.Empty;
            base.OnMouseUp(e);
        }

        /// <summary>
        /// Called when the mouse is moved over the control
        /// </summary>
        /// <param name="e">the location of the event</param>
        protected override void OnMouseMove(MouseEventArgs e)
        {
            DragDropEffects effects;

            if ((e.Button & MouseButtons.Left) == MouseButtons.Left && m_drag_rect != Rectangle.Empty && !m_drag_rect.Contains(e.Location))
            {
                if (m_drag_node != null && m_drag_node.Tag is string)
                {
                    string[] arr = new string[1];
                    arr[0] = (string)m_drag_node.Tag;
                    DataObject dobj = new DataObject(DataFormats.FileDrop, arr);
                    effects = DoDragDrop(dobj, DragDropEffects.All);
                }
            }

            base.OnMouseMove(e);
        }
    }
}
