using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.IO;
using Bwg.Utils;

namespace Bwg.Controls
{
    /// <summary>
    /// This class is a form that has the ability to save and load program settings.
    /// </summary>
    public class FormWithSettings : Form
    {
        #region protected variables
        /// <summary>
        /// This is a map from names to variables as read from the settings
        /// file.
        /// </summary>
        protected StartupSettings m_settings;
        #endregion

        #region protected constants startup var names

        #region Logger Constants
        /// <summary>
        /// Name of the bool startup setting that indicates the display of debug messages
        /// </summary>
        public const string PropertyNameLoggerDebugMessages = "Logger:Debug:Messages";

        /// <summary>
        /// Name of the int startup setting that gives the debug message level
        /// </summary>
        public const string PropertyNameLoggerDebugLevel = "Logger:Debug:Level";

        /// <summary>
        /// Name of the bool startup setting that indicates the display of warning messages
        /// </summary>
        public const string PropertyNameLoggerWarningMessages = "Logger:Warning:Messages";
        #endregion

        #region FreeDB Constants
        /// <summary>
        /// The name of the string startup setting that contains the name of the free db host.
        /// </summary>
        public const string PropertyNameFreeDBHostName = "FreeDB:Hostname";

        /// <summary>
        /// The name of the int startup setting that contains the port number for the free db host.
        /// </summary>
        public const string PropertyNameFreeDBPort = "FreeDB:Port";

        /// <summary>
        /// The name of the string startup setting that contains the name of the freedb protocol
        /// to use.
        /// </summary>
        public const string PropertyNameFreeDBProtocol = "FreeDB:Protocol";

        /// <summary>
        /// The name of the string startup settting that contains the CGI path to use with HTTP
        /// based CDDB V1 protocol.
        /// </summary>
        public const string PropertyNameFreeDBCgiPath = "FreeDB:CgiPath";
        #endregion

        #region CreateDataDisk Constants
        /// <summary>
        /// Name of the int startup setting that gives the default burn type for the CreateDataDisk
        /// application
        /// </summary>
        public const string PropertyNameCreateDataDiskBurnType = "CreateDataDisk:BurnType";

        /// <summary>
        /// Name of the bool startup setting that gives the default multi-session behavior for the
        /// CreateDataDisk application
        /// </summary>
        public const string PropertyNameCreateDataDiskMultiSession = "CreateDataDisk:MultiSession";
        #endregion

        #region BurnIsoImage Constants
        /// <summary>
        /// Name of the int startup setting that gives the default burn type for the BurnIsoImage
        /// application
        /// </summary>
        protected const string PropertyNameBurnIsoImageBurnType = "BurnIsoImage:BurnType";

        /// <summary>
        /// Name of the bool startup setting that gives the default multi-session behavior for the
        /// BurnIsoImage application
        /// </summary>
        protected const string PropertyNameBurnIsoImageMultiSession = "BurnIsoImage:MultiSession";
        #endregion

        #region CreateAudioDisk Constants
        /// <summary>
        /// Name of the int startup setting that gives the default burn type for the CreateDataDisk
        /// application
        /// </summary>
        public const string PropertyNameCreateAudioDiskBurnType = "CreateAudioDisk:BurnType";

        /// <summary>
        /// Name of the bool startup setting that gives the default multi-session behavior for the
        /// CreateDataDisk application
        /// </summary>
        public const string PropertyNameCreateAudioDiskMultiSession = "CreateAudioDisk:MultiSession";

        /// <summary>
        /// Name of the string startup setting that gives the path in the file system for the computer
        /// view tree.  This is used to preserve the user's current location across different runs of
        /// the tool
        /// </summary>
        public const string PropertyNameStartupPath = "CreateAudioDisk:StartupPath";
        #endregion

        #region RipAudio Constants
        /// <summary>
        /// Name of the string property that gives the default audio codec for ripping audio
        /// </summary>
        public const string PropertyNameRipAudioCodec = "RipAudio:Codec";

        /// <summary>
        /// Name of the string property that gives the default audio format for ripped audio
        /// </summary>
        public const string PropertyNameRipAudioFormat = "RipAudio:Format";

        /// <summary>
        /// Name of the string property that gives the default music library path
        /// </summary>
        public const string PropertyNameRipAudioMusicLibraryPath = "RipAudio:LibraryPath";

        /// <summary>
        /// Name of the boolean property that determines if we are to eject when a rip is done
        /// </summary>
        public const string PropertyNameRipAudioEjectWhenDone = "RipAudio:EjectWhenDone";
        #endregion

        #region Copy Disk Constants

        /// <summary>
        /// The burn type for copy disk
        /// </summary>
        public const string PropertyNameCopyDiskBurnType = "CopyDisk:BurnType";
        #endregion

        #region General Constants

        /// <summary>
        /// Location of a temporary directory with plenty of disk space
        /// </summary>
        public const string PropertyNameTemporaryDirectory = "General:TemporaryDirectory";
        #endregion

        #endregion

        /// <summary>
        /// Create a form that has the ability to load and store settings
        /// </summary>
        public FormWithSettings()
        {
            // Load the user settings
            string username = Environment.UserName + ".set";
            string dbfile = Path.Combine(Application.StartupPath, username);
            m_settings = new StartupSettings(dbfile);
            m_settings.ReadSettingsFile();
        }

        /// <summary>
        /// Saves a value given the key across program invocations
        /// </summary>
        /// <param name="key"></param>
        /// <param name="value"></param>
        protected void SetStartupString(string key, string value)
        {
            m_settings.SetValue(key, value);
            m_settings.WriteSettingsFile();
        }

        /// <summary>
        /// Saves a value given the key
        /// </summary>
        /// <param name="key"></param>
        /// <param name="value"></param>
        protected void SetStartupInteger(string key, int value)
        {
            m_settings.SetValue(key, value);
            m_settings.WriteSettingsFile();
        }

        /// <summary>
        /// Saves a boolean value given the key
        /// </summary>
        /// <param name="key">the key of the value to save</param>
        /// <param name="value">the value to save</param>
        protected void SetStartupBool(string key, bool value)
        {
            m_settings.SetValue(key, value);
            m_settings.WriteSettingsFile();
        }

        /// <summary>
        /// Return a value given the key from the startup settings database.
        /// </summary>
        /// <param name="key">the key for the value</param>
        /// <param name="def">the default value if the key does not exist</param>
        /// <returns>string value</returns>
        protected string GetStartupString(string key, string def)
        {
            object v = m_settings.GetValue(key);
            if (v != null && v is string)
                return (string)v;

            return def;
        }

        /// <summary>
        /// Return a boolean given a key from the startup settings database.
        /// </summary>
        /// <param name="key">the key for the value</param>
        /// <param name="def">the default value if the key does not exist</param>
        /// <returns>boolean value</returns>
        protected bool GetStartupBool(string key, bool def)
        {
            object v = m_settings.GetValue(key);
            if (v != null && v is bool)
                return (bool)v;

            return def;
        }

        /// <summary>
        /// Returns an integer value given the key
        /// </summary>
        /// <param name="key"></param>
        /// <param name="def"></param>
        /// <returns></returns>
        protected int GetStartupInteger(string key, int def)
        {
            object v = m_settings.GetValue(key);
            if (v != null && v is Int32)
                return (int)v;

            return def;
        }
    }
}
