//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics;

namespace Bwg.Controls
{
    /// <summary>
    /// This clas is a progress bar that can also display text information
    /// about what is happening.
    /// </summary>
    public partial class TextProgressBar : Control
    {
        #region public types

        /// <summary>
        /// The style to draw the control. 
        /// </summary>
        public enum DrawingMode
        {
            /// <summary>
            /// Draw as a series of stripes
            /// </summary>
            Striped,

            /// <summary>
            /// Draw as a solid bar
            /// </summary>
            Solid,
        } ;

        /// <summary>
        /// The mode to paint the control bar
        /// </summary>
        public enum PaintingMode
        {
            /// <summary>
            /// Draw the colors darkest in the middle and tapering to lighter
            /// colors on either side
            /// </summary>
            HalfGradient,

            /// <summary>
            /// Draw the colors darkest on one edge and tapering to a lighter color
            /// on the other side
            /// </summary>
            FullGradient,

            /// <summary>
            /// Draw the colors as a single solid color
            /// </summary>
            Solid
        } ;

        #endregion

        #region private member variables
        Color m_text;
        Color m_center_light;
        Color m_center_dark;
        Color m_outline;
        long m_minimum ;
        long m_maximum;
        long m_value;
        int m_stripe_width;
        int m_stripe_spacing;
        DrawingMode m_drawing_mode;
        PaintingMode m_paint_mode;
        bool m_add_percent;
        #endregion

        #region constructors

        /// <summary>
        /// Create the new progress bar
        /// </summary>
        public TextProgressBar()
        {
            BackColor = Color.White;

            m_text = Color.Yellow;
            m_center_light = Color.FromArgb(0, 192, 0);
            m_center_dark = Color.FromArgb(192, 255, 192);
            m_outline = Color.DarkGreen;
            m_minimum = 0;
            m_maximum = 100;
            m_value = 0;
            m_stripe_width = 6;
            m_stripe_spacing = 3;
            m_drawing_mode = DrawingMode.Striped;
            m_paint_mode = PaintingMode.HalfGradient;
            m_add_percent = false;

            DoubleBuffered = true;

            InitializeComponent();
        }
        #endregion

        #region public properties

        #region Behavior Properties
        /// <summary>
        /// The minimum value for the progress value
        /// </summary>
        [Category("Behavior")]
        [Description("The minimum value for this bar")]
        public long Minimum
        {
            get
            {
                return m_minimum;
            }
            set
            {
                m_minimum = value;
                Invalidate();
            }
        }

        /// <summary>
        /// The maximum value for the progress value
        /// </summary>
        [Category("Behavior")]
        [Description("The maximum value for this bar")]
        public long Maximum
        {
            get
            {
                return m_maximum;
            }
            set
            {
                m_maximum = value;
                Invalidate();
            }
        }

        /// <summary>
        /// The actual value for the progress
        /// </summary>
        [Category("Behavior")]
        [Description("The value for this bar")]
        public long Value
        {
            get
            {
                return m_value;
            }
            set
            {
                m_value = value;
                if (m_value < m_minimum)
                    m_value = m_minimum;
                if (m_value > m_maximum)
                    m_value = m_maximum;

                Invalidate();
            }
        }
        #endregion

        #region Appearance Properties

        /// <summary>
        /// The text string to display in the progress control
        /// </summary>
        [Category("Appearance")]
        [Description("The text to display in the control")]
        public override string Text
        {
            get
            {
                return base.Text;
            }
            set
            {
                base.Text = value;
                Invalidate();
            }
        }

        /// <summary>
        /// If this property is true, the percentage done is added to the text
        /// </summary>
        [Category("Appearance")]
        [Description("The font to use for drawing the text")]
        public bool AddPercent
        {
            get
            {
                return m_add_percent;
            }
            set
            {
                m_add_percent = value;
                Invalidate();
            }
        }

        /// <summary>
        /// The drawmode for the control.
        /// </summary>
        [Category("Appearance")]
        [Description("The method for drawing the bar")]
        public DrawingMode DrawMode
        {
            get
            {
                return m_drawing_mode;
            }
            set
            {
                m_drawing_mode = value;
                Invalidate();
            }
        }

        /// <summary>
        /// The painting mode for the control
        /// </summary>
        [Category("Appearance")]
        [Description("The style for painting the bar")]
        public PaintingMode PaintMode
        {
            get
            {
                return m_paint_mode;
            }
            set
            {
                m_paint_mode = value;
                Invalidate();
            }
        }

        /// <summary>
        /// For drawing in striped mode, the width of the stripes
        /// </summary>
        [Category("Appearance")]
        [Description("The width of the stripes in the progress bar")]
        public int StripeWidth
        {
            get
            {
                return m_stripe_width;
            }
            set
            {
                m_stripe_width = value;
                Invalidate();
            }
        }

        /// <summary>
        /// For drawing in the striped mode, the spacing of the stripes
        /// </summary>
        [Category("Appearance")]
        [Description("The spacing of the stripes in the progress bar")]
        public int StripeSpacing
        {
            get
            {
                return m_stripe_spacing;
            }
            set
            {
                m_stripe_spacing = value;
                Invalidate();
            }
        }
        
        /// <summary>
        /// The color of the inner part of the bar if the paint mode is
        /// gradient.
        /// </summary>
        [Category("Appearance")]
        [Description("The color of the bar")]
        public Color BarInnerColorLight
        {
            get
            {
                return m_center_light;
            }
            set
            {
                m_center_light = value;
                Invalidate();
            }
        }

        /// <summary>
        /// The color of the bar if the paint mode is solid, or for the
        /// outer part of the bar if the paint mode is gradient
        /// </summary>
        [Category("Appearance")]
        [Description("The color of the bar")]
        public Color BarInnerColorDark
        {
            get
            {
                return m_center_dark;
            }
            set
            {
                m_center_dark = value;
                Invalidate();
            }
        }

        /// <summary>
        /// The color of the outline of the bar
        /// </summary>
        [Category("Appearance")]
        [Description("The color of the bar")]
        public Color BarOutlineColor
        {
            get
            {
                return m_outline;
            }
            set
            {
                m_outline = value;
                Invalidate();
            }
        }

        /// <summary>
        /// The color of the text in the bar
        /// </summary>
        [Category("Appearance")]
        [Description("The color of the text")]
        public Color ColorText
        {
            get
            {
                return m_text;
            }
            set
            {
                m_text = value;
                Invalidate();
            }
        }
        #endregion

        #endregion

        #region paint functions

        /// <summary>
        /// Called when the control is resized.  This method just validates the
        /// control area to cause a redraw when the control is resized.
        /// </summary>
        /// <param name="e"></param>
        protected override void OnSizeChanged(EventArgs e)
        {
            base.OnSizeChanged(e);
            Invalidate();
        }

        /// <summary>
        /// Called to paint the progress bar
        /// </summary>
        /// <param name="pe"></param>
        protected override void OnPaint(PaintEventArgs pe)
        {
            Rectangle outline = new Rectangle(0, 0, Width - 1, Height - 1);
            using (Brush background = new SolidBrush(BackColor))
            {
                pe.Graphics.FillRectangle(background, outline);
            }

            using (Pen p = new Pen(m_outline))
            {
                pe.Graphics.DrawRectangle(p, outline);
            }

            // If min and max are he same, we are done
            if (Maximum == Minimum)
            {
                if (Width > Height)
                {
                    DrawTextHorizontal(pe.Graphics, outline);
                }
                else
                {
                    DrawTextVertical(pe.Graphics, outline);
                }
            }
            else
            {
                if (Width > Height)
                {
                    if (m_drawing_mode == DrawingMode.Striped)
                        DrawStripedHorizontal(pe.Graphics, outline);
                    else
                        DrawSolidHorizontal(pe.Graphics, outline);

                    DrawTextHorizontal(pe.Graphics, outline);
                }
                else if (Height > Width)
                {
                    if (m_drawing_mode == DrawingMode.Striped)
                        DrawStripedVertical(pe.Graphics, outline);
                    else
                        DrawSolidVertical(pe.Graphics, outline);

                    DrawTextVertical(pe.Graphics, outline);
                }
            }
        }

        private string GetString()
        {
            string mystr = Text;
            if (m_add_percent)
            {
                double pcnt = (double)(Value - Minimum) / (double)(Maximum - Minimum) * 100.0;
                mystr += " (" + pcnt.ToString("F0") + "%)";
            }

            return mystr;
        }

        private void DrawStripedHorizontal(Graphics g, Rectangle outline)
        {
            int pitch = m_stripe_width + m_stripe_spacing;
            int total = (Width - 2)/ pitch;
            int todraw ;
            Brush b1 = null, b2 = null;

            long ltodraw = total * (Value - Minimum) / (Maximum - Minimum);

            //
            // The above operation scales the value from a 64 bit user supplied space to the
            // drawing space.  Therefore after the scale everything should fit into a 32 bit 
            // integer with no problem.
            //
            Debug.Assert(ltodraw < Int32.MaxValue);
            todraw = (int)ltodraw;

            if (m_paint_mode == PaintingMode.Solid)
                b1 = new SolidBrush(m_center_dark);
            else if (m_paint_mode == PaintingMode.FullGradient)
                b1 = new LinearGradientBrush(outline, m_center_light, m_center_dark, 90.0f);
            else
            {
                b1 = new LinearGradientBrush(outline, m_center_light, m_center_dark, 90.0f);
                b2 = new LinearGradientBrush(outline, m_center_dark, m_center_light, 90.0f);
            }

            for (int i = 0; i < todraw; i++)
            {
                if (m_paint_mode == PaintingMode.Solid || m_paint_mode == PaintingMode.FullGradient)
                {
                    Rectangle r = new Rectangle(i * pitch + 1 + m_stripe_spacing, 1, m_stripe_width, Height - 2);
                    g.FillRectangle(b1, r);
                }
                else
                {
                    int h = Height - 2;
                    Rectangle r ;

                    r = new Rectangle(i * pitch + 1 + m_stripe_spacing, 1, m_stripe_width, h / 2);
                    g.FillRectangle(b1, r);

                    r = new Rectangle(i * pitch + 1 + m_stripe_spacing, h / 2 + 1, m_stripe_width, h / 2);
                    g.FillRectangle(b2, r);
                }
            }

            b1.Dispose();
            if (m_paint_mode == PaintingMode.HalfGradient)
                b2.Dispose();
        }

        private void DrawSolidHorizontal(Graphics g, Rectangle outline)
        {
            Brush b1 = null, b2 = null;

            if (m_paint_mode == PaintingMode.Solid)
                b1 = new SolidBrush(m_center_dark);
            else if (m_paint_mode == PaintingMode.FullGradient)
                b1 = new LinearGradientBrush(outline, m_center_light, m_center_dark, 90.0f);
            else
            {
                b1 = new LinearGradientBrush(outline, m_center_light, m_center_dark, 90.0f);
                b2 = new LinearGradientBrush(outline, m_center_dark, m_center_light, 90.0f);
            }

            int draw_width = (int)((long)(Width - 2) * (long)(Value - Minimum) / (Maximum - Minimum));

            if (m_paint_mode == PaintingMode.Solid || m_paint_mode == PaintingMode.FullGradient)
            {
                Rectangle r = new Rectangle(1, 1, draw_width, Height - 2);
                g.FillRectangle(b1, r);
            }
            else
            {
                int h = Height - 2;
                Rectangle r ;
                
                r = new Rectangle(1, 1, draw_width, h/2);
                g.FillRectangle(b1, r);

                r = new Rectangle(1, h / 2, draw_width, h / 2 + 1);
                g.FillRectangle(b2, r);
            }

            b1.Dispose();
            if (m_paint_mode == PaintingMode.HalfGradient)
                b2.Dispose();
        }

        private void DrawTextHorizontal(Graphics g, Rectangle outline)
        {
            using (Brush b = new SolidBrush(m_text))
            {
                string mystr = GetString();
                SizeF sf = g.MeasureString(mystr, Font);
                RectangleF rf = new RectangleF((outline.Width - sf.Width) / 2, (outline.Height - sf.Height) / 2, sf.Width, sf.Height);
                g.DrawString(mystr, Font, b, rf);
            }
        }

        private void DrawStripedVertical(Graphics g, Rectangle outline)
        {
            int pitch = m_stripe_width + m_stripe_spacing;
            int total = (Height - 2) / pitch;
            int todraw;
            Brush b1 = null, b2 = null;

            long ltodraw = total * (Value - Minimum) / (Maximum - Minimum);

            //
            // The above operation scales the value from a 64 bit user supplied space to the
            // drawing space.  Therefore after the scale everything should fit into a 32 bit 
            // integer with no problem.
            //
            Debug.Assert(ltodraw < Int32.MaxValue);
            todraw = (int)ltodraw;

            if (m_paint_mode == PaintingMode.Solid)
                b1 = new SolidBrush(m_center_dark);
            else if (m_paint_mode == PaintingMode.FullGradient)
                b1 = new LinearGradientBrush(outline, m_center_light, m_center_dark, 0.0f);
            else
            {
                b1 = new LinearGradientBrush(outline, m_center_light, m_center_dark, 0.0f);
                b2 = new LinearGradientBrush(outline, m_center_dark, m_center_light, 0.0f);
            }

            for (int i = 0; i < todraw; i++)
            {
                if (m_paint_mode == PaintingMode.Solid || m_paint_mode == PaintingMode.FullGradient)
                {
                    Rectangle r = new Rectangle(1, i * pitch + 1 + m_stripe_spacing, Width - 2, m_stripe_width);
                    g.FillRectangle(b1, r);
                }
                else
                {
                    int w = Width - 2;
                    Rectangle r;

                    r = new Rectangle(1, i * pitch + 1 + m_stripe_spacing, w / 2, m_stripe_width);
                    g.FillRectangle(b2, r);

                    r = new Rectangle(w / 2 + 1, i * pitch + 1 + m_stripe_spacing, w / 2, m_stripe_width);
                    g.FillRectangle(b1, r);
                }
            }

            b1.Dispose();
            if (m_paint_mode == PaintingMode.HalfGradient)
                b2.Dispose();
        }

        private void DrawSolidVertical(Graphics g, Rectangle outline)
        {
            Brush b1 = null, b2 = null;

            if (m_paint_mode == PaintingMode.Solid)
                b1 = new SolidBrush(m_center_dark);
            else if (m_paint_mode == PaintingMode.FullGradient)
                b1 = new LinearGradientBrush(outline, m_center_light, m_center_dark, 0.0f);
            else
            {
                b1 = new LinearGradientBrush(outline, m_center_light, m_center_dark, 0.0f);
                b2 = new LinearGradientBrush(outline, m_center_dark, m_center_light, 0.0f);
            }

            long ldraw_height = (Height - 2)* (Value - Minimum) / (Maximum - Minimum);
            Debug.Assert(ldraw_height < Int32.MaxValue);
            int draw_height = (int)ldraw_height;

            if (m_paint_mode == PaintingMode.Solid || m_paint_mode == PaintingMode.FullGradient)
            {
                Rectangle r = new Rectangle(1, 1, Width - 2, draw_height);
                g.FillRectangle(b1, r);
            }
            else
            {
                int w = Width - 2;
                Rectangle r;

                r = new Rectangle(1, 1, w/2, draw_height) ;
                g.FillRectangle(b2, r);

                r = new Rectangle(w / 2, 1, w, draw_height);
                g.FillRectangle(b1, r);
            }

            b1.Dispose();
            if (m_paint_mode == PaintingMode.HalfGradient)
                b2.Dispose();
        }

        private void DrawTextVertical(Graphics g, Rectangle outline)
        {
            using (Brush b = new SolidBrush(m_text))
            {
                string mystr = GetString();
                StringFormat fmt = new StringFormat(StringFormatFlags.DirectionVertical);
                SizeF sf = g.MeasureString(mystr, Font, outline.Width, fmt);
                RectangleF rf = new RectangleF((outline.Width - sf.Width) / 2, (outline.Height - sf.Height) / 2, sf.Width, sf.Height);
                g.DrawString(mystr, Font, b, rf, fmt);
            }
        }

        /// <summary>
        /// Called to paint the background for the progress bar
        /// </summary>
        /// <param name="pevent"></param>
        protected override void OnPaintBackground(PaintEventArgs pevent)
        {
            base.OnPaintBackground(pevent);
        }

        #region size methods
        /// <summary>
        /// Returns the default size for this control, this is used for the
        /// designer interface within Visual Studio (I think).
        /// </summary>
        protected override Size DefaultSize
        {
            get
            {
                return new Size(240, 23);
            }
        }
        #endregion

        #endregion
    }
}
