//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Threading;
using System.Diagnostics;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;
using Bwg.Controls;
using Bwg.Logging;
using Bwg.Audio;
using Bwg.Scsi ;
using Bwg.Burner ;
using Bwg.Utils ;

namespace Bwg.CreateAudioDisk
{
    public partial class CreateAudioDiskForm : BwgBurnBaseForm
    {
        /// <summary>
        /// The file name for the project
        /// </summary>
        private string m_filename;

        /// <summary>
        /// If true, the project has been changed since saved
        /// </summary>
        private bool m_changed;

        /// <summary>
        /// The disk image to burn to the disk
        /// </summary>
        private DiskBurnImage m_disk;

        /// <summary>
        /// The audio manager that manages conversion from one form of audio to another
        /// </summary>
        private AudioManager m_manager;

        /// <summary>
        /// The drive based on the currently selected device
        /// </summary>
        private Drive m_drive;

        /// <summary>
        /// The currently selected device
        /// </summary>
        private Device m_device;

        /// <summary>
        /// The timer used to time elapsed burn time
        /// </summary>
        private System.Windows.Forms.Timer m_timer;

        /// <summary>
        /// Start time for the burn
        /// </summary>
        private DateTime m_start;

        /// <summary>
        /// The thread burning the disk
        /// </summary>
        private Thread m_burner_thread;

        /// <summary>
        /// The burn type (disk at once or track at once)
        /// </summary>
        private Drive.BurnType m_burn_type;

        /// <summary>
        /// The burn speed
        /// </summary>
        private SpeedInfo m_burn_speed;

        /// <summary>
        /// If true, do a multi-session burn
        /// </summary>
        private bool m_multi_session;

        /// <summary>
        /// The ICON manager for displaying icons in the window
        /// </summary>
        private ShellIconMgr m_icon_mgr;

        private ProcessingSongForm m_processing_form;


        private delegate void TimerTickDelegate();
        private delegate void DoneDelegate(DiskOperationError status);
        private delegate void ProgressHandlerDelegate(object sender, ProgressChangedEventArgs args);

        public CreateAudioDiskForm()
        {
            m_filename = string.Empty;
            m_changed = false;

            m_burn_type = (Drive.BurnType)GetStartupInteger(PropertyNameCreateAudioDiskBurnType, (int)Drive.BurnType.DontCare);
            m_multi_session = GetStartupBool(PropertyNameCreateAudioDiskMultiSession, false);

            InitializeComponent();

            LoggerWindow.Sink = m_sink;
            LoggerWindow.SetupColumns();

            m_logger.LogMessage(new UserMessage(UserMessage.Category.Info, 0, "Audio Disk Creator Version " + GetVersion()));

            m_icon_mgr = new ShellIconMgr() ;

            m_manager = new AudioManager(m_logger);
            SongList.Manager = m_manager;

            m_disk = new DiskBurnImage(m_logger);
            m_disk.TargetMedia = DiskBurnImage.DiskTargetType.CD_SAO;
            SongList.Disk = m_disk;
            SongList.DoubleClick +=new EventHandler(SongList_DoubleClick);

            m_disk.TrackAdded += new EventHandler<EventArgs>(DiskChangedAddedRemoved);
            m_disk.TrackRemoved += new EventHandler<EventArgs>(DiskChangedAddedRemoved);

            string[] extlist = m_manager.Extensions;
            foreach (string str in extlist)
                ComputerList.AddExt(str);

            DeviceSelector.Logger = m_logger;
            DeviceSelector.SelectedIndexChanged += new EventHandler(DeviceSelector_SelectedIndexChanged);
            DeviceSelector.InitializeDeviceList(false);

            ComputerTree.AfterSelect += new TreeViewEventHandler(ComputerTree_AfterSelect);
            ComputerTree.IconManager = m_icon_mgr;
            ComputerTree.Populate();

            ComputerList.IconManager = m_icon_mgr;
            ComputerList.DoubleClick += new EventHandler(ComputerList_DoubleClick);

            DiskChanged(null, null);

            m_timer = new System.Windows.Forms.Timer();

            SetStartupPath();
        }

        protected void ComputerList_DoubleClick(object sender, EventArgs args)
        {
            string[] songs = new string[0];

            foreach (ListViewItem item in ComputerList.SelectedItems)
            {
                if (item.Tag is string)
                {
                    int size = songs.GetLength(0);
                    Array.Resize(ref songs, size + 1);
                    songs[size] = (string)item.Tag;
                }
            }

            if (songs.GetLength(0) > 0)
                SongList.AddSongs(songs);
        }

        protected void SongList_DoubleClick(object sender, EventArgs args)
        {
            if (SongList.SelectedItems.Count == 0)
                return ;

            if (SongList.SelectedItems.Count > 1)
            {
                MessageBox.Show("Cannot edit the properties for multiple tracks at once") ;
                return ;
            }

            ListViewItem item = SongList.SelectedItems[0];
            if (!(item.Tag is TrackBurnImage))
            {
                MessageBox.Show("Internal error #1 - ListViewItem.tag is not TrackBurnImage") ;
                return ;
            }

            byte [] isrc = null ;
            TrackBurnImage ti = (TrackBurnImage)item.Tag;
            if (ti.HasIsrc)
                isrc = ti.ISRC;
            Debug.Assert(ti.Data is AudioDataSource);
            AudioDataSource src = (AudioDataSource)ti.Data;
            AudioReader rdr = src.Reader;

            //
            // TODO: expand this to support multiple languages and languages other than 
            //       English
            //
            CDTextTrackProps cdprops = new CDTextTrackProps();
            CDTextField field;

            byte index = m_disk.GetCDTextLanguageIndex(CDTextEncoderDecoder.LanguageType.ENGLISH);
            if (index == 0xff)
            {
                index = m_disk.AddCDTextLanguage(CDTextEncoderDecoder.LanguageType.ENGLISH);
                if (index == 0xff)
                    return;
            }

            field = ti.FindCDTextField(CDTextField.TextFieldType.Arranger, index) ;
            if (field != null)
            {
                Debug.Assert(field.Value is CDTextDataString);
                cdprops.Arranger = (CDTextDataString)field.Value;
            }

            field = ti.FindCDTextField(CDTextField.TextFieldType.Composer, index);
            if (field != null)
            {
                Debug.Assert(field.Value is CDTextDataString);
                cdprops.Composer = (CDTextDataString)field.Value;
            }

            field = ti.FindCDTextField(CDTextField.TextFieldType.Message, index);
            if (field != null)
            {
                Debug.Assert(field.Value is CDTextDataString);
                cdprops.Message = (CDTextDataString)field.Value;
            }

            field = ti.FindCDTextField(CDTextField.TextFieldType.Performer, index);
            if (field != null)
            {
                Debug.Assert(field.Value is CDTextDataString);
                cdprops.Performer = (CDTextDataString)field.Value;
            }

            field = ti.FindCDTextField(CDTextField.TextFieldType.Songwriter, index) ;
            if (field != null)
            {
                Debug.Assert(field.Value is CDTextDataString);
                cdprops.Songwriter = (CDTextDataString)field.Value;
            }

            field = ti.FindCDTextField(CDTextField.TextFieldType.Title, index) ;
            if (field != null)
            {
                Debug.Assert(field.Value is CDTextDataString);
                cdprops.Title = (CDTextDataString)field.Value;
            }

            //
            // Now, see what we can extract from the song file
            //

            TrackProps props = new TrackProps(isrc, cdprops);
            if (props.ShowDialog() == DialogResult.OK)
            {
                m_changed = true;

                ti.ISRC = props.ISRC;

                ti.RemoveCDTextField(CDTextField.TextFieldType.Arranger, index) ;
                if (cdprops.Arranger.Length > 0)
                {
                    field = new CDTextField(index, CDTextField.TextFieldType.Arranger, cdprops.Arranger) ;
                    ti.AddCDTextField(field) ;
                }

                ti.RemoveCDTextField(CDTextField.TextFieldType.Composer, index) ;
                if (cdprops.Composer.Length > 0)
                {
                    field = new CDTextField(index, CDTextField.TextFieldType.Composer, cdprops.Composer) ;
                    ti.AddCDTextField(field) ;
                }

                ti.RemoveCDTextField(CDTextField.TextFieldType.Message, index) ;
                if (cdprops.Message.Length > 0)
                {
                    field = new CDTextField(index, CDTextField.TextFieldType.Message, cdprops.Message) ;
                    ti.AddCDTextField(field) ;
                }

                ti.RemoveCDTextField(CDTextField.TextFieldType.Performer, index) ;
                if (cdprops.Performer.Length > 0)
                {
                    field = new CDTextField(index, CDTextField.TextFieldType.Performer, cdprops.Performer) ;
                    ti.AddCDTextField(field) ;
                }

                ti.RemoveCDTextField(CDTextField.TextFieldType.Songwriter, index) ;
                if (cdprops.Songwriter.Length > 0)
                {
                    field = new CDTextField(index, CDTextField.TextFieldType.Songwriter, cdprops.Songwriter) ;
                    ti.AddCDTextField(field) ;
                }

                ti.RemoveCDTextField(CDTextField.TextFieldType.Title, index) ;
                if (cdprops.Title.Length > 0)
                {
                    field = new CDTextField(index, CDTextField.TextFieldType.Title, cdprops.Title) ;
                    ti.AddCDTextField(field) ;
                }            
            }
        }

        protected override void OnClosing(CancelEventArgs e)
        {
            string path = ComputerTree.GetSelectedPath();
            SetStartupString(PropertyNameStartupPath, path);
            base.OnClosing(e);
        }

        private void SetStartupPath()
        {
            string path = GetStartupString(PropertyNameStartupPath, string.Empty);
            if (path == string.Empty)
                return;

            ComputerTree.SetSelectedPath(path) ;
        }

        private void DiskChangedAddedRemoved(object sender, EventArgs args)
        {
            m_changed = true ;
            DiskChanged(sender, args) ;
        }

        private void DiskChanged(object sender, EventArgs args)
        {
            string str ;
            MinuteSecondFrame total ;

            DiskSpace.Minimum = 0;
            if (m_drive == null || !m_drive.IsMediaLoaded)
            {
                DiskSpace.Minimum = 0;
                DiskSpace.Maximum = Drive.D74minDiskCapacity;
                total = new MinuteSecondFrame(DiskSpace.Maximum);
                str = "[No media loaded, assuming 74 min cd] " ;
                MinuteSecondFrame dlen = new MinuteSecondFrame(m_disk.LengthSectors);
                str += "Time Used " + dlen.ToString("M:S") + " of " + total.ToString("M:S");
            }
            else
            {
                if (!m_drive.IsCd)
                {
                    DiskSpace.Minimum = 0;
                    DiskSpace.Maximum = 0;
                    str = "Media not valid for audio CD";
                }
                else
                {
                    DiskSpace.Minimum = 0;
                    DiskSpace.Maximum = m_drive.RemainingCapacityInWritableSectors;
                    DiskSpace.Value = m_disk.LengthSectors;
                    total = new MinuteSecondFrame(DiskSpace.Maximum);
                    MinuteSecondFrame dlen = new MinuteSecondFrame(m_disk.LengthSectors);
                    str = "Time Used " + dlen.ToString("M:S") + " of " + total.ToString("M:S");
                }
            }

            DiskSpace.Text = str;
        }

        private void ComputerTree_AfterSelect(object sender, TreeViewEventArgs args)
        {
            if (args.Node != null)
            {
                if (args.Node.Tag is string)
                    ComputerList.Path = (string)args.Node.Tag;
                else if (args.Node.Tag is FileSystemTreeView.FSLocation)
                {
                    FileSystemTreeView.FSLocation loc = (FileSystemTreeView.FSLocation)args.Node.Tag;
                    if (loc == FileSystemTreeView.FSLocation.Desktop)
                        ComputerList.Path = Environment.GetFolderPath(Environment.SpecialFolder.Desktop);
                }
            }
        }

        private void DeviceSelector_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (DeviceSelector.SelectedItem == null)
            {
                if (m_drive != null)
                    m_drive.Dispose();

                if (m_device != null)
                    m_device.Dispose();

                DiskStatus.Text = "No disk selected";
            }
            else
            {
                string dname = (string)DeviceSelector.SelectedItem;
                if (m_drive != null)
                {
                    m_drive.Dispose();
                    m_drive = null;
                }

                if (m_device != null)
                {
                    m_device.Dispose();
                    m_device = null;
                }

                m_device = new Device(m_logger);
                if (!m_device.Open(dname[0]))
                {
                    DiskStatus.Text = "Error opening device, " + Device.Win32ErrorToString(m_device.LastError);
                    m_device.Dispose();
                    m_device = null;
                }
                else
                {
                    m_drive = new Drive(m_device);
                    DiskOperationError status = m_drive.Initialize();
                    if (status != null)
                    {
                        m_drive.Dispose();
                        m_device.Dispose();
                        m_drive = null;
                        m_device = null;
                        DiskStatus.Text = "Error initializing drive - " + status.Message;
                    }
                    else
                        DiskStatus.Text = m_drive.StatusString;
                }
            }

            SpeedInfo[] speeds;
            if (m_drive == null || m_drive.GetWriteSpeeds(out speeds) != null || speeds.GetLength(0) == 0)
                m_burn_speed = null;
            else
                m_burn_speed = speeds[0];

            DiskChanged(null, null);
        }

        private void DiskSettings_Click(object sender, EventArgs e)
        {
            bool warning, debug;

            if (m_logger.HasSink(UserMessage.Category.Debug))
                debug = true;
            else
                debug = false;

            if (m_logger.HasSink(UserMessage.Category.Warning))
                warning = true;
            else
                warning = false;

            int level = m_logger.GetLevel(UserMessage.Category.Debug);

            SpeedInfo[] speeds;
            DiskOperationError status = m_drive.GetWriteSpeeds(out speeds);
            if (status != null)
            {
                string str = "Cannot retreive speed data from drive - " + status.Message;
                MessageBox.Show(str, "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            DriveSettings form = new DriveSettings(debug, level, warning, m_multi_session, m_burn_type, speeds,
                                                   m_drive.IsTrackAtOnceSupported, m_drive.IsSessionAtOnceSupported, 
                                                   m_drive.IsMultiSessionSupported);
            if (form.ShowDialog(this) == DialogResult.OK)
            {
                m_multi_session = form.LeaveSessionOpen;
                if (form.DebugMessagesFlag)
                {
                    m_logger.SetSink(UserMessage.Category.Debug, m_sink);
                    m_logger.SetLevel(UserMessage.Category.Debug, form.DebugLevel);
                }
                else
                    m_logger.RemoveSink(UserMessage.Category.Debug);

                if (form.WarningMessagesFlag)
                    m_logger.SetSink(UserMessage.Category.Warning, m_sink);
                else
                    m_logger.RemoveSink(UserMessage.Category.Warning);

                m_burn_type = form.BurnType;
                m_burn_speed = form.BurnSpeedValue;

                SaveLoggerSettings();
                SetStartupInteger(PropertyNameCreateAudioDiskBurnType, (int)m_burn_type);
                SetStartupBool(PropertyNameCreateAudioDiskMultiSession, m_multi_session);
            }
        }

        private void Burn_Click(object sender, EventArgs e)
        {
            bool erase = false ;

            if (m_drive == null)
            {
                MessageBox.Show("CD/DVD drive not selected, please select a drive");
                return;
            }

            if (m_disk.TrackCount == 0)
            {
                MessageBox.Show("Cannot burn disk, no songs selected", "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            DiskOperationError status  = m_drive.Initialize();
            if (status != null)
            {
                string str = "Error intializing drive - " + status.Message;
                MessageBox.Show(str, "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            DiscInformation info;
            Device.CommandStatus st = m_device.ReadDiscInformation(out info);
            if (st != Device.CommandStatus.Success)
            {
                MessageBox.Show("Error reading disc information - see the log for more information", "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            //
            // We include CD/ROM because at least my HP DVD/CD burner sees a CD/R or CD/RW as an empty
            // or erasable CD_ROM drive.
            //
            if (m_drive.CurrentProfile != Drive.SupportedProfiles.CD_RW &&
                m_drive.CurrentProfile != Drive.SupportedProfiles.CD_R &&
                m_drive.CurrentProfile != Drive.SupportedProfiles.CD_ROM)
            {
                MessageBox.Show("An audio disk can only be created with CD/R or CD/RW media.  Please replace the media and select burn again!",
                                "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            if (m_drive.CurrentProfile == Drive.SupportedProfiles.CD_RW &&
                info.DiscStatus != DiscInformation.DiscStatusType.EmptyDisc)
            {
                // Propmt the user to format
                DialogResult res = MessageBox.Show("This disk is not erased.  Do you wish to erase this disk?",
                                                   "Erase Disk?", MessageBoxButtons.YesNo, MessageBoxIcon.Question);
                if (res == DialogResult.No)
                {
                    MessageBox.Show("Please insert a blank disk and press burn again", "Not A Blank Disk", MessageBoxButtons.OK, MessageBoxIcon.Information);
                    return;
                }
                erase = true;
            }

            if (m_drive.CurrentProfile == Drive.SupportedProfiles.CD_RW ||
                m_drive.CurrentProfile == Drive.SupportedProfiles.CD_R ||
                m_drive.CurrentProfile == Drive.SupportedProfiles.CD_ROM)
            {
                EnableControls(false);

                // Start a timer to display elapsed time
                m_start = DateTime.Now;
                m_timer.Interval = 1000;
                m_timer.Tick += new EventHandler(TimerTickHandler);
                m_timer.Tag = this;
                m_timer.Start();

                // Now, create a thread to burn the disk
                m_burner_thread = new Thread(new ParameterizedThreadStart(this.BurnProc));
                m_burner_thread.Start(erase);
            }
        }

        private static void TimerTickHandler(object sender, EventArgs args)
        {
            System.Windows.Forms.Timer t = (System.Windows.Forms.Timer)sender;
            CreateAudioDiskForm form = (CreateAudioDiskForm)t.Tag;
            if (form.InvokeRequired)
            {
                form.Invoke(new TimerTickDelegate(form.TimerTick));
            }
            else
                form.TimerTick();
        }

        private void BurnProc(object param)
        {
            bool erase = false;
            DiskOperationError status;

            if (param is bool)
                erase = (bool)param;

            object[] args = new object[1];

            m_drive.ProgressChanged += new ProgressChangedEventHandler(ProgressHandler);

            if (erase)
                status = m_drive.Erase(Drive.EraseType.Minimal);

            status = m_drive.BurnDisk(m_burn_type, m_disk, m_multi_session, m_burn_speed);

            m_drive.Device.StartStopUnit(false, Device.PowerControl.NoChange, Device.StartState.EjectDisc);
            args[0] = status;
            Invoke(new DoneDelegate(Done), args);
        }

        private void TimerTick()
        {
            TimeSpan elapsed = DateTime.Now - m_start;
            Elapsed.Text = "Elapsed Time: " + elapsed.Minutes.ToString() + ":" + elapsed.Seconds.ToString("D2");

            if (m_drive.BufferLength != 0)
            {
                BufferLevel.Maximum = (int)m_drive.BufferLength;
                BufferLevel.Value = (int)(m_drive.BufferLength - m_drive.BufferAvailable);
            }
            Application.DoEvents();
        }
        
        private void Done(DiskOperationError status)
        {
            string msg;
            m_timer.Stop();
            Operation.Text = "Idle";
            BufferLevel.Value = 0;

            EnableControls(true);

            string caption = "Error!";
            MessageBoxIcon icon;

            if (status == null)
            {
                msg = "The burn operation completed sucessfully\r\n";
                msg += "Target Speed: " + m_burn_speed + "\r\n";
                msg += "Actual Speed: " + m_drive.BurnSpeed;
                caption = "Complete";
                icon = MessageBoxIcon.Information;
            }
            else
            {
                msg = status.Message;
                caption = "Error!";
                icon = MessageBoxIcon.Error;
            }

            MessageBox.Show(msg, caption, MessageBoxButtons.OK, icon);
        }

        private void ProgressHandler(object sender, ProgressChangedEventArgs e)
        {
            if (InvokeRequired)
            {
                object[] param = new object[2];
                param[0] = sender;
                param[1] = e;
                Invoke(new ProgressHandlerDelegate(ProgressHandler), param);
            }
            else
            {
                BurnProgress.Value = e.ProgressPercentage;

                if (e.UserState is ProgressText)
                {
                    ProgressText pt = (ProgressText)e.UserState ;
                    BurnProgress.Text = pt.LongName;
                    TopLevelControl.Text = pt.ShortName + " ( " + e.ProgressPercentage.ToString() + "% )";
                    Operation.Text = pt.ShortName;
                }
                else if (e.UserState is string)
                {
                    BurnProgress.Text = (string)e.UserState;
                    TopLevelControl.Text = "Burning";
                    Operation.Text = "Burning";
                }
            }
        }

        private void EnableControls(bool st)
        {
            DiskSettings.Enabled = st;
            SongList.Enabled = st;
            ComputerTree.Enabled = st;
            ComputerList.Enabled = st;
            Burn.Enabled = st;
            DeviceSelector.Enabled = st;
            UpArrow.Enabled = st;
            DownArrow.Enabled = st;
            Delete.Enabled = st;
        }

        private void UpArrow_Click(object sender, EventArgs e)
        {
            foreach(ListViewItem item in SongList.SelectedItems)
                SongList.MoveUp(item);
        }

        private void DownArrow_Click(object sender, EventArgs e)
        {
            foreach (ListViewItem item in SongList.SelectedItems)
                SongList.MoveUp(item);
        }

        private void Delete_Click(object sender, EventArgs e)
        {
            foreach (ListViewItem item in SongList.SelectedItems)
                SongList.Delete(item);
        }

        private void newToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (m_changed)
            {
                DialogResult res;

                res = MessageBox.Show("You have changed the current project.  " +
                                      "Creating a new project will cause you to lose the changes to the current project." +
                                      "Do you want to save the current project before creating a new project?",
                                      "Lose Changes", MessageBoxButtons.YesNoCancel, MessageBoxIcon.Question) ;
                if (res == DialogResult.Cancel)
                    return ;

                if (res == DialogResult.Yes)
                    saveToolStripMenuItem_Click(null, null);
            }

            m_disk = new DiskBurnImage(m_logger) ;
            SongList.Disk = m_disk ;
        }

        private void openToolStripMenuItem_Click(object sender, EventArgs e)
        {
            OpenFileDialog dialog = new OpenFileDialog() ;
            dialog.Multiselect = false;
            dialog.Filter = "BwgBurn Audio Project Files (*.bap)|*.bap|All files (*.*)|*.*";
            dialog.RestoreDirectory = true;
            dialog.AddExtension = true;
            dialog.CheckFileExists = true;
            dialog.DefaultExt = "bwg";
            dialog.FilterIndex = 0;

            if (dialog.ShowDialog() == DialogResult.OK)
                LoadProjectFile(dialog.FileName) ;
        }

        private void saveToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (m_filename == string.Empty || m_changed)
            {
                if (m_filename == string.Empty)
                    saveAsToolStripMenuItem_Click(null, null) ;
                else
                    SaveProjectFile(m_filename) ;
            }
        }

        private void saveAsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            SaveFileDialog dialog = new SaveFileDialog() ;
            dialog.AddExtension = true;
            dialog.CheckPathExists = true;
            dialog.InitialDirectory = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments);
            dialog.RestoreDirectory = true;
            dialog.Filter = "BwgBurn Audio Project Files (*.bap)|*.bap|All files (*.*)|*.*";
            dialog.FilterIndex = 0 ;
            dialog.OverwritePrompt = true ;
            dialog.ValidateNames = true ;
            DialogResult ret = dialog.ShowDialog();
            if (ret == DialogResult.Cancel)
                return ;

            SaveProjectFile(dialog.FileName) ;
        }

        private void SaveProjectFile(string filename)
        {
            AudioProject p = new AudioProject(m_disk) ;
            m_filename = filename;

            try
            {
                FileStream strm = new FileStream(filename, FileMode.Create, FileAccess.Write);
                BinaryFormatter formatter = new BinaryFormatter();
                formatter.Serialize(strm, p);
                strm.Close();

                m_changed = false ;
            }
            catch (Exception ex)
            {
                string str = "Error writing project file '" + filename + "' - " + ex.Message;
                MessageBox.Show(str, "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void ScanChanged(object sender, ProgressChangedEventArgs args)
        {
            if (m_processing_form != null)
            {
                m_processing_form.ScanProgressBar.Value = args.ProgressPercentage;
            }
        }

        private void LoadProjectFile(string filename)
        {
            AudioProject p ;
            m_filename = filename ;

            try
            {
                FileStream strm = new FileStream(filename, FileMode.Open, FileAccess.Read);
                BinaryFormatter formatter = new BinaryFormatter();
                p = (AudioProject)formatter.Deserialize(strm);
                strm.Close();

                m_disk = new DiskBurnImage(m_logger);
                if (p.HasMCN)
                    m_disk.MCN = p.MCN;

                m_disk.TargetMedia = DiskBurnImage.DiskTargetType.CD_SAO;

                foreach (CDTextEncoderDecoder.LanguageType l in p.CDTextLanguageList)
                    m_disk.AddCDTextLanguage(l);

                foreach (CDTextField tf in p.CDTextList)
                    m_disk.AddCDTextField(tf);

                if (p.Count > 0)
                {
                    m_processing_form = new ProcessingSongForm();
                    m_processing_form.Show();

                    foreach (AudioProjectTrack pt in p)
                    {
                        string fname = pt.FileName;
                        if (!File.Exists(fname))
                        {
                            string str = "The file '" + fname + "' does not exist and is being omitted from the disk";
                            MessageBox.Show(str, "Missing File!", MessageBoxButtons.OK, MessageBoxIcon.Error);
                        }
                        else
                        {
                            AudioReader rdr = null;
                            try
                            {
                                short lmin, lmax, rmin, rmax;
                                rdr = m_manager.GetReader(fname);

                                m_processing_form.SongTitle.Text = Path.GetFileName(fname);
                                Application.DoEvents();

                                rdr.ScanProgressChanged += new ProgressChangedEventHandler(ScanChanged);
                                rdr.ScanInputFile(out lmin, out lmax, out rmin, out rmax);
                                rdr.ScanProgressChanged -= new ProgressChangedEventHandler(ScanChanged);
                                AudioDataSource src = new AudioDataSource(rdr);

                                TrackBurnImage ti = new TrackBurnImage(TrackBurnImage.TrackType.Audio, src);

                                ti.PreTrackGap = pt.PreTrackGap;
                                ti.ISRC = pt.ISRC;

                                foreach (CDTextField tf in pt.CDTextList)
                                    ti.AddCDTextField(tf);

                                m_disk.AddTrack(ti);
                            }
                            catch (Exception ex)
                            {
                                if (rdr != null)
                                    rdr.ScanProgressChanged -= new ProgressChangedEventHandler(ScanChanged);

                                string str = "Cannot create audio track '" + fname + "' - " + ex.Message + ", removing from audio burn image";
                                MessageBox.Show(str, "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
                            }
                        }
                    }
                    m_processing_form.Dispose();
                    m_processing_form = null;
                }

                SongList.Disk = m_disk;
                m_changed = false ;
                DiskChanged(null, null) ;
            }
            catch (Exception ex)
            {
                string str = "Error reading project file '" + filename + "' - " + ex.Message;
                MessageBox.Show(str, "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void DiskProps_Click(object sender, EventArgs e)
        {
            CDTextDiskProps dprops = new CDTextDiskProps();

            byte index = m_disk.GetCDTextLanguageIndex(CDTextEncoderDecoder.LanguageType.ENGLISH) ;
            if (index == 0xff)
            {
                index = m_disk.AddCDTextLanguage(CDTextEncoderDecoder.LanguageType.ENGLISH) ;
                if (index == 0xff)
                    return ;
            }

            CDTextField field;

            field = m_disk.FindCDTextField(CDTextField.TextFieldType.Arranger, index);
            if (field != null)
                dprops.Arranger = (CDTextDataString)field.Value;

            field = m_disk.FindCDTextField(CDTextField.TextFieldType.Composer, index);
            if (field != null)
                dprops.Composer = (CDTextDataString)field.Value;

            field = m_disk.FindCDTextField(CDTextField.TextFieldType.Genre, index);
            if (field != null)
                dprops.Genre = (CDTextDataGenre)field.Value;

            field = m_disk.FindCDTextField(CDTextField.TextFieldType.Message, index);
            if (field != null)
                dprops.Message = (CDTextDataString)field.Value;

            field = m_disk.FindCDTextField(CDTextField.TextFieldType.Performer, index);
            if (field != null)
                dprops.Performer = (CDTextDataString)field.Value;

            field = m_disk.FindCDTextField(CDTextField.TextFieldType.Songwriter, index);
            if (field != null)
                dprops.Performer = (CDTextDataString)field.Value;

            field = m_disk.FindCDTextField(CDTextField.TextFieldType.Title, index);
            if (field != null)
                dprops.Title = (CDTextDataString)field.Value;

            DiskProps props = new DiskProps(null, dprops);

            if (props.ShowDialog() == DialogResult.OK)
            {
                m_disk.RemoveCDTextField(CDTextField.TextFieldType.Arranger, index);
                if (dprops.Arranger.Length > 0)
                {
                    field = new CDTextField(index, CDTextField.TextFieldType.Arranger, dprops.Arranger);
                    m_disk.AddCDTextField(field);
                }

                m_disk.RemoveCDTextField(CDTextField.TextFieldType.Composer, index);
                if (dprops.Composer.Length > 0)
                {
                    field = new CDTextField(index, CDTextField.TextFieldType.Composer, dprops.Composer);
                    m_disk.AddCDTextField(field);
                }

                m_disk.RemoveCDTextField(CDTextField.TextFieldType.Message, index);
                if (dprops.Message.Length > 0)
                {
                    field = new CDTextField(index, CDTextField.TextFieldType.Message, dprops.Message);
                    m_disk.AddCDTextField(field);
                }

                m_disk.RemoveCDTextField(CDTextField.TextFieldType.Performer, index);
                if (dprops.Performer.Length > 0)
                {
                    field = new CDTextField(index, CDTextField.TextFieldType.Performer, dprops.Performer);
                    m_disk.AddCDTextField(field);
                }

                m_disk.RemoveCDTextField(CDTextField.TextFieldType.Songwriter, index);
                if (dprops.Songwriter.Length > 0)
                {
                    field = new CDTextField(index, CDTextField.TextFieldType.Songwriter, dprops.Songwriter);
                    m_disk.AddCDTextField(field);
                }

                m_disk.RemoveCDTextField(CDTextField.TextFieldType.Title, index);
                if (dprops.Title.Length > 0)
                {
                    field = new CDTextField(index, CDTextField.TextFieldType.Title, dprops.Title);
                    m_disk.AddCDTextField(field);
                }
            }
        }

        private void TrackProps_Click(object sender, EventArgs e)
        {
            SongList_DoubleClick(sender, e);
        }
    }
}