//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//

using System;
using System.Collections.Generic;
using System.Text;
using System.Net ;
using System.Net.Sockets ;
using System.Threading;
using System.Diagnostics;
using System.Net.Mail;

namespace Bwg.Database.Freedb
{
    public class FreeDBServer : IDisposable
    {
        #region private variables

        /// <summary>
        /// The site we are connected to
        /// </summary>
        private FreeDBSite m_site;

        /// <summary>
        /// The user information used to initialize the server
        /// </summary>
        private FreeDBUserInfo m_userinfo;

        /// <summary>
        /// This thread does the work in the background
        /// </summary>
        private Thread m_worker;

        /// <summary>
        /// The transport we are using to talk to the server
        /// </summary>
        private FreeDBTransport m_transport;

        /// <summary>
        /// The queue of requests for album information
        /// </summary>
        private IList<FreeDBRequest> m_requests;

        /// <summary>
        /// This is used to lock the request queue when necessary
        /// </summary>
        private object m_lock;

        /// <summary>
        /// This event is used to signal the worker that there is a request
        /// in the queue.
        /// </summary>
        private AutoResetEvent m_event;

        #endregion

        #region public types
        /// <summary>
        /// This enumerated type gives the transport types to try
        /// </summary>
        [FlagsAttribute]
        public enum TransportTypes : byte
        {
            /// <summary>
            /// Try CDDBP
            /// </summary>
            TranportTypeCddbp = 1,

            /// <summary>
            /// Try HTTP
            /// </summary>
            TransportTypeHttp = 2,
        }
        #endregion

        #region public properties
        /// <summary>
        /// The host name of the master freedb database site.  This is just used to get a set
        /// of sites to use
        /// </summary>
        public const string MasterDatabaseHostName = "freedb.freedb.org";

        /// <summary>
        /// This is the port number to use to connect to the database
        /// </summary>
        public const int MasterDatabasePortNumber = 8880;

        /// <summary>
        /// This event is fired when we get a query response back from the freedb server 
        /// </summary>
        public EventHandler<FreeDBResponseArgs> FreeDBResponse;

        /// <summary>
        /// The current site we are connected to
        /// </summary>
        public FreeDBSite Site
        {
            get
            {
                return m_site;
            }
        }
        #endregion

        #region constructors
        public FreeDBServer(FreeDBSite site, FreeDBUserInfo info)
        {
            // Remember the site we are connected to
            m_site = site;

            // Remember the user info
            m_userinfo = info;

            // Create a tranport layer based on the site
            m_transport = CreateTransport(site, info);

            // Create the queue for requests
            m_requests = new List<FreeDBRequest>();

            // Create the lock for controlling access to the request queue
            m_lock = new object();

            // This event is used to signal the worker thread that there are new entries in the
            // request queue
            m_event = new AutoResetEvent(false);

            // Now, create the thread that will handle requests
            m_worker = new Thread(new ThreadStart(WorkerThread));
            m_worker.Start();

        }
        #endregion

        #region static methods
        /// <summary>
        /// This gets set to true if we detect we are behing a firewarell
        /// </summary>
        static public bool BehindFirewall = false;

        static public IList<FreeDBSite> GetSites(FreeDBUserInfo info, TransportTypes t)
        {
            FreeDBTransport transport;

            if (!BehindFirewall)
            {
                if ((t & TransportTypes.TranportTypeCddbp) != 0)
                {
                    using (transport = new FreeDBTransportCddb(info, MasterDatabaseHostName, MasterDatabasePortNumber))
                    {
                        try
                        {
                            int code;
                            object resp;

                            transport.Start();
                            transport.SetProtocol(6);
                            transport.SendCommandGetResponse("sites", out code, out resp);
                            transport.Finish();

                            if (code == 210 && resp is string[])
                                return ParseSitesStrings((string[])resp, false);
                        }
                        catch (Exception)
                        {
                        }
                    }
                }
            }

            BehindFirewall = true;

            if ((t & TransportTypes.TransportTypeHttp) != 0)
            {
                string host = "http://" + MasterDatabaseHostName + ":80/~cddb/cddb.cgi";
                using (transport = new FreeDBTransportHttp(info, host))
                {
                    try
                    {
                        int code;
                        object resp;

                        transport.Start();
                        transport.SetProtocol(6);
                        transport.SendCommandGetResponse("sites", out code, out resp);
                        transport.Finish();

                        if (code == 210 && resp is string[])
                            return ParseSitesStrings((string[])resp, true);
                    }
                    catch (Exception ex)
                    {
                        string s = ex.Message;
                    }
                }
            }

            return null;
        }

        static private IList<FreeDBSite> ParseSitesStrings(string[] resp, bool exclude_cddb)
        {
            IList<FreeDBSite> sites = new List<FreeDBSite>();

            foreach (string line in resp)
            {
                char[] sp = new char[1] { ' ' };
                string[] words = line.Split(sp);
                int port;

                // Now sure what we got here, but it was not what was expected
                if (words.GetLength(0) < 6)
                    continue;

                // If we are excluding CDDB lines, skip this line.  We do this because we
                // could not contact a CDDB server, and therefore are probably behind a firewall
                if (words[1].ToLower() == FreeDBSite.CddbProtocolName && exclude_cddb)
                    continue;

                try
                {
                    // Parse the port #, and skip this line if it does not parse correctly
                    port = Int32.Parse(words[2]);
                }
                catch (Exception)
                {
                    continue;
                }

                // Re-combine everything from the 6th word on
                string loc = "";
                for (int j = 6; j < words.GetLength(0); j++)
                    loc += " " + words[j];

                FreeDBSite site = new FreeDBSite(words[0], port, words[1], words[3], words[4], words[5], loc);
                sites.Add(site);
            }

            return sites;
        }
        #endregion

        #region public methods
        public void Dispose()
        {
            if (m_worker.IsAlive)
                m_worker.Abort();

            if (m_transport != null)
                m_transport.Dispose();
        }

        public void SubmitViaEmail(FreeDBAlbumInfo info)
        {
        }

        public void SubmitViaHttp(FreeDBAlbumInfo info)
        {
        }

        public void QueryRequest(FreeDBAlbumInfo info)
        {
            QueueRequest(new FreeDBQueryRequest(info));
        }

        /// <summary>
        /// This submits a cddb write request to the database.  It turns out that I cannot do this
        /// via cddb protocol so I must use email or http.  For now, there will be no submit support until I
        /// look at this more closely and find a new design for the freedb support.  Mostly likely I will decouple
        /// the query server from the submit process.
        /// </summary>
        /// <param name="info"></param>
        public void SubmitRequest(FreeDBAlbumInfo info)
        {
            QueueRequest(new FreeDBSubmitRequest(info));
        }

        public void CategoryRequest()
        {
            QueueRequest(new FreeDBCategoryRequest()) ;
        }
        #endregion

        #region private functions

        private void QueueRequest(FreeDBRequest req)
        {
            lock (m_lock)
            {
                m_requests.Add(req);
            }
            m_event.Set();
        }

        private FreeDBTransport CreateTransport(FreeDBSite site, FreeDBUserInfo info)
        {
            FreeDBTransport t = null;

            if (site.Protocol.ToLower() == FreeDBSite.CddbProtocolName)
                t = new FreeDBTransportCddb(info, site.HostName, site.Port);
            else if (site.Protocol.ToLower() == FreeDBSite.HttpProtocolName)
            {
                string host = "http://" + site.HostName + ":" + site.Port.ToString() + "/" + site.HttpCgiPath;
                t = new FreeDBTransportHttp(info, host);
            }
            else
            {
                throw new FreeDBInvalidProtocol(site);
            }

            return t;
        }

        private void WorkerThread()
        {

            while (true)
            {
                //
                // Wait for 1 second, or until we are signaled by another thread
                // that has inserted a request into the queue.  This insures we will
                // now spend more than 1 second waiting on a request to pulled from
                // the queue, without having to deal with some of the race conditions
                // that can lead to lockup here.
                //
                m_event.WaitOne(1000, false);

                ProcessAllQueuedRequests() ;
            }
        }

        private void ProcessAllQueuedRequests()
        {
            FreeDBRequest request;

            while (true)
            {
                // Lock the queue long enough to check the request count and if a request
                // is present, pull one requests off of the queue
                lock (m_lock)
                {
                    if (m_requests.Count == 0)
                        break;

                    request = m_requests[0];
                    m_requests.RemoveAt(0);
                }

                // Ok, we have removed the request we are going to process, 
                m_transport.Start();
                m_transport.SetProtocol(6);
                if (request is FreeDBQueryRequest)
                    QueryAlbum((FreeDBQueryRequest)request);
                else if (request is FreeDBSubmitRequest)
                    SubmitAlbum((FreeDBSubmitRequest)request);
                else if (request is FreeDBCategoryRequest)
                    GetCategories((FreeDBCategoryRequest)request) ;

                m_transport.Finish();
            }

            m_transport.Finish();
        }

        private string CreateReadCommand(string key, string category)
        {
            return "cddb read " + category + " " + key;
        }

        private string CreateQueryCommand(FreeDBAlbumInfo album)
        {
            int ntrks = album.TrackCount - 1;
            string cmd = "cddb query " + FreeDbKey(album) + " " + ntrks.ToString();

            for (int i = 0; i < ntrks; i++)
            {
                FreeDBTrackInfo t = album[i];
                int n = t.StartFrame + t.StartSecond * 75 + t.StartMinute * 60 * 75;
                cmd += " " + n.ToString();
            }

            cmd += " " + album.TotalSeconds;
            return cmd;
        }

        private void DoTransportErrorCallback(FreeDBTransport.RequestStatus status, FreeDBRequest request)
        {
            if (FreeDBResponse != null)
            {
                FreeDBResponseArgs.ResponseType t = FreeDBResponseArgs.ResponseType.UnknownError ;

                switch(status)
                {
                case FreeDBTransport.RequestStatus.CannotConnect:
                    t = FreeDBResponseArgs.ResponseType.CouldNotContactServer ;
                    break ;

                case FreeDBTransport.RequestStatus.CannotLogin:
                    t = FreeDBResponseArgs.ResponseType.LoginFailed ;
                    break ;

                case FreeDBTransport.RequestStatus.InvalidReponse:
                    t = FreeDBResponseArgs.ResponseType.InvalidServerResponse ;
                    break ;
                }
            
                FreeDBResponseArgs args = new FreeDBResponseArgs(t, request) ;
                FreeDBResponse(this, args) ;
            }
        }

        private void GetCategories(FreeDBCategoryRequest request)
        {
            int code;
            object resp;
            string cmd;

            cmd = "cddb lscat";
            FreeDBTransport.RequestStatus status = m_transport.SendCommandGetResponse(cmd, out code, out resp);

            if (status != FreeDBTransport.RequestStatus.Success)
                DoTransportErrorCallback(status, request);
            else
            {
                if (code == 210 && resp is string[])
                {
                    request.Categories = (string[])resp;
                }

                if (FreeDBResponse != null)
                {
                    FreeDBResponseArgs args = new FreeDBResponseArgs(FreeDBResponseArgs.ResponseType.Success, request);
                    FreeDBResponse(this, args);
                }
            }
        }

        private void AddTrackFrameOffsets(StringBuilder builder, FreeDBAlbumInfo album)
        {
            builder.Append("# Track frame offsets:\r\n");
            for (int i = 0; i < album.TrackCount - 1; i++)
            {
                FreeDBTrackInfo t = album[i];
                int n = t.StartFrame + t.StartSecond * 75 + t.StartMinute * 60 * 75;
                builder.Append("#         ");
                builder.Append(n.ToString());
                builder.Append("\r\n");
            }
        }

        private void SubmitAlbum(FreeDBSubmitRequest request)
        {
            int code;
            object resp;
            string cmd;

            cmd = CreateSubmitRequest(true, request.Album);
            FreeDBTransport.RequestStatus status = m_transport.SendCommandGetResponse(cmd, out code, out resp);
            if (status != FreeDBTransport.RequestStatus.Success)
                DoTransportErrorCallback(status, request);
            else
            {
                if (FreeDBResponse != null)
                {
                    FreeDBResponseArgs args = new FreeDBResponseArgs(FreeDBResponseArgs.ResponseType.Success, request);
                    FreeDBResponse(this, args);
                }
            }
        }

        private string CreateSubmitRequest(bool hdr, FreeDBAlbumInfo info)
        {
            StringBuilder builder = new StringBuilder(1024) ;

            if (hdr)
            {
                builder.Append("cddb write ");
                builder.Append(info.Category);
                builder.Append(" ");
                builder.Append(FreeDbKey(info));
                builder.Append("\r\n");
            }

            builder.Append("# xmcd\r\n") ;
            builder.Append("#\r\n");

            AddTrackFrameOffsets(builder, info);
            builder.Append("#\r\n");

            builder.Append("# Disc length: ");
            builder.Append(info.TotalSeconds.ToString());
            builder.Append(" seconds\r\n");

            builder.Append("# Submitted via: ");
            builder.Append(m_userinfo.Client);
            builder.Append(" ");
            builder.Append(m_userinfo.ClientVersion);
            builder.Append("\r\n");

            builder.Append("DISCID=");
            builder.Append(FreeDbKey(info));
            builder.Append("\r\n");

            builder.Append("DTITLE=") ;
            if (info.Artist.Length > 0)
            {
                builder.Append(info.Artist);
                builder.Append(" / ");
            }
            builder.Append(info.Title);
            builder.Append("\r\n");

            if (info.Year != 0)
            {
                builder.Append("DYEAR=");
                builder.Append(info.Year.ToString());
                builder.Append("\r\n");
            }

            if (info.Genre.Length > 0)
            {
                builder.Append("DGENRE=");
                builder.Append(info.Genre);
                builder.Append("\r\n");
            }

            for (int i = 0; i < info.TrackCount - 1; i++)
            {
                int track = i + 1;
                FreeDBTrackInfo tinfo = info[i];
                builder.Append("TTITLE");
                builder.Append(track.ToString());
                builder.Append("=");
                if (tinfo.HasArtist)
                {
                    builder.Append(tinfo.Artist);
                    builder.Append(" / ");
                }
                builder.Append(tinfo.Title);
                builder.Append("\r\n");
            }

            builder.Append(".\r\n");

            Debug.WriteLine(builder.ToString());
            return builder.ToString();
        }

        private void QueryAlbum(FreeDBQueryRequest request)
        {
            int code;
            object resp;
            FreeDBResponseArgs.ResponseType rt ;

            // Send the query command to the server
            string cmd = CreateQueryCommand(request.RequestInfo);
            FreeDBTransport.RequestStatus status = m_transport.SendCommandGetResponse(cmd, out code, out resp);

            if (status != FreeDBTransport.RequestStatus.Success)
                DoTransportErrorCallback(status, request);
            else
            {
                if (code == 200 && resp is string)
                {
                    rt = ParseSingleMatch(0, request, (string)resp);
                    request.Status = FreeDBQueryRequest.StatusType.ExactMatch;
                }
                else if (code == 210 && resp is string[])
                {
                    rt = ParseMultipleMatch(request, (string[])resp);
                    request.Status = FreeDBQueryRequest.StatusType.ExactMatch;
                }
                else if (code == 211 && resp is string[])
                {
                    rt = ParseMultipleMatch(request, (string [])resp);
                    request.Status = FreeDBQueryRequest.StatusType.FuzzyMatch;
                }
                else if (code == 202)
                {
                    rt = FreeDBResponseArgs.ResponseType.Success;
                    request.Status = FreeDBQueryRequest.StatusType.NoMatch;
                }
                else
                {
                    rt = FreeDBResponseArgs.ResponseType.InvalidServerResponse;
                }

                if (FreeDBResponse != null)
                {
                    FreeDBResponseArgs args = new FreeDBResponseArgs(rt, request) ;
                    FreeDBResponse(this, args) ;
                }
            }
        }

        private FreeDBResponseArgs.ResponseType ParseMultipleMatch(FreeDBQueryRequest request, string[] lines)
        {
            int i = 0;
            foreach (string l in lines)
            {
                FreeDBResponseArgs.ResponseType st = ParseSingleMatch(i++, request, l);
                if (st != FreeDBResponseArgs.ResponseType.Success)
                    return st;
            }

            return FreeDBResponseArgs.ResponseType.Success;
        }

        private FreeDBResponseArgs.ResponseType ParseSingleMatch(int which, FreeDBQueryRequest request, string line)
        {
            int code;
            object response;
            string category = string.Empty;
            string key = string.Empty;
            int n = 0;

            // Extract the category
            while (n < line.Length && !Char.IsWhiteSpace(line[n]))
                category += line[n++];

            // Skip spaces between category and discid
            while (n < line.Length && Char.IsWhiteSpace(line[n]))
                n++;

            // Extract the disk id
            while (n < line.Length && !Char.IsWhiteSpace(line[n]))
                key += line[n++];


            // We should not be at the end of the line here
            if (n == line.Length)
                return FreeDBResponseArgs.ResponseType.InvalidServerResponse;


            // Create the command to read the disk information
            string cmd = CreateReadCommand(key, category);
            m_transport.SendCommandGetResponse(cmd, out code, out response);

            if (code == 210 && response is string[])
                return ParseDiskInfo(which, category, request, (string[])response);

            return FreeDBResponseArgs.ResponseType.InvalidServerResponse;
        }

        private void SplitArtistTitle(string t, out string artist, out string title)
        {
            int index = t.LastIndexOf(" / ");
            if (index == -1)
            {
                artist = string.Empty;
                title = t;
            }
            else
            {
                artist = t.Substring(0, index);
                title = t.Substring(index + 3);
            }
        }

        private int GetTrackNumber(string str)
        {
            string num = "";
            int i = 0;

            while (i < str.Length && Char.IsDigit(str[i]))
                num += str[i++];

            return Int32.Parse(num);
        }

        private FreeDBResponseArgs.ResponseType ParseDiskInfo(int which, string category, FreeDBQueryRequest request, string [] lines)
        {
            string artist, title ;

            FreeDBAlbumInfo info = request.GetCreateAlbum(which);
            info.Category = category;

            foreach (string l in lines)
            {
                l.Trim();
                if (l.Length == 0 || l[0] == '#')
                    continue;

                int n = l.IndexOf('=');
                if (n == -1)
                    continue;

                string key = l.Substring(0, n);
                string value = l.Substring(n + 1);

                key = key.Trim();
                if (key == "DTITLE")
                {
                    SplitArtistTitle(value, out artist, out title);
                    info = request.GetCreateAlbum(which);
                    info.Title = title;
                    if (artist.Length > 0)
                        info.Artist = artist;
                }
                else if (key == "DYEAR")
                {
                    int year ;
                    info = request.GetCreateAlbum(which);
                    if (Int32.TryParse(value, out year))
                        info.Year = year;
                }
                else if (key == "DGENRE")
                {
                    info = request.GetCreateAlbum(which);
                    info.Genre = value;
                }
                else if (key.Length > 6 && key.Substring(0, 6) == "TTITLE")
                {
                    int tno = GetTrackNumber(key.Substring(6));
                    SplitArtistTitle(value, out artist, out title);

                    if (n < request.TrackCount)
                    {
                        info = request.GetCreateAlbum(which);
                        FreeDBTrackInfo tinfo = info[tno];
                        tinfo.Title = title;
                        if (artist.Length > 0)
                            tinfo.Artist = artist;
                    }
                }
            }

            info.FromFreedb = true;
            return FreeDBResponseArgs.ResponseType.Success;
        }

        private int CddbSum(int n)
        {
            int ret;

            ret = 0;
            while (n > 0)
            {
                ret = ret + (n % 10);
                n = n / 10;
            }

            return (ret);
        }


        /// <summary>
        /// Calculate the disk freedb.org id from the informat stored in a FreeDBAlbumInfo
        /// object.
        /// </summary>
        private string FreeDbKey(FreeDBAlbumInfo d)
        {
            uint n = 0;
            uint t;

            for (int i = 0; i < d.TrackCount - 1; i++)
                n += (uint)CddbSum(60 * d[i].StartMinute + d[i].StartSecond);

            FreeDBTrackInfo leadout = d[d.TrackCount - 1];
            t = (uint)(leadout.StartMinute * 60 + leadout.StartSecond);
            t -= (uint)(d[0].StartMinute * 60 + d[0].StartSecond);

            uint ret = (uint)((n % 0xff) << 24) | (t << 8) | (uint)(d.TrackCount - 1);

            return ret.ToString("x8");
        }
        #endregion
    }
}
