//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//
using System;
using System.Collections.Generic;
using System.Text;
using System.ComponentModel;
using Bwg.Logging ;
using Bwg.Utils ;
using Bwg.Scsi ;
using Bwg.Burner ;

namespace Bwg.DiskImage
{
    /// <summary>
    /// This class is used to read and write a file that contains the contents of 
    /// a complete disk.
    /// </summary>
    public abstract class DiskContents
    {
        #region constants
        public const byte Signature1 = 0x02;
        public const byte Signature2 = 0x05;
        public const byte Signature3 = 0x19;
        public const byte Signature4 = 0x67;
        public const byte MajorVersion = 0x01;
        public const byte MinorVersoin = 0x00;
        #endregion

        #region public types
        public enum ImageType : byte
        {
            Undefined = 0,
            CD = 1,
            DVD = 2
        } ;
        #endregion

        #region public variables
        /// <summary>
        /// This event is signaled when the progress has changed
        /// </summary>
        public ProgressChangedEventHandler ProgressChanged;
        #endregion

        #region private variables
        private Logger m_logger;
        #endregion

        #region protected variables

        /// <summary>
        /// This value gives the type
        /// </summary>
        protected ImageType m_disk_type;

        /// <summary>
        /// If true, we are copying a CD-R or CD-RW that has been left open for more
        /// sessions.
        /// </summary>
        protected bool m_open_disk ;

        /// <summary>
        /// This value holds the number of sessions
        /// </summary>
        internal SessionInfo[] m_sessions;

        #endregion

        #region constructors
        public DiskContents(Logger l)
        {
            m_disk_type = ImageType.Undefined;
            m_sessions = new SessionInfo[0];
            m_logger = l;
        }
        #endregion

        #region public methods

        public DiskOperationError BurnDisk(Drive dest, Drive.BurnType bt)
        {
            DiskOperationError status = Initialize();
            if (status != null)
                return status;

            for (int i = 0; i < m_sessions.GetLength(0); i++)
            {
                bool multi = true;

                if (i == m_sessions.GetLength(0) - 1 && m_open_disk == false)
                    multi = false;

                Bwg.Burner.DiskBurnImage d = new Bwg.Burner.DiskBurnImage(m_logger);
                d.TargetMedia = DiskBurnImage.DiskTargetType.CD_SAO;
                d.AutoGenerateAudioPregap = false;

                int total = m_sessions[i].Tracks.GetLength(0);
                foreach (TrackInfo ti in m_sessions[i].Tracks)
                {
                    TrackBurnImage image = new TrackBurnImage(ti.TrackType, new TrackInfoDataSource(this, ti, total));
                    d.AddTrack(image);
                }

                dest.ProgressChanged += new System.ComponentModel.ProgressChangedEventHandler(ProgressHandler);
                status = dest.BurnDisk(bt, d, multi, null);
                dest.ProgressChanged += new System.ComponentModel.ProgressChangedEventHandler(ProgressHandler);
                if (status != null)
                    return status;
            }

            return null;
        }

        public DiskOperationError WriteFile(OsFileWriter wrtr)
        {
            DiskOperationError status;

            status = Initialize();
            if (status != null)
                return status;

            //
            // Write the disk header
            //
            status = WriteHeader(wrtr);
            if (status != null)
                return status;

            //
            // Write a header for each session
            //
            foreach(SessionInfo info in m_sessions)
            {
                status = info.WriteHeader(wrtr);
                if (status != null)
                    return status;
            }

            //
            // Write the data for a sessions to the disk
            //
            foreach(SessionInfo info in m_sessions)
            {
                info.ProgressChanged += new System.ComponentModel.ProgressChangedEventHandler(ProgressHandler);
                status = info.Write(wrtr, this);
                info.ProgressChanged -= new System.ComponentModel.ProgressChangedEventHandler(ProgressHandler);
                if (status != null)
                    return status;
            }

            return null;
        }

        abstract internal int MaxBufferSize
        {
            get;
        }

        abstract internal DiskOperationError ReadTrackData(int current, int toread, int secsize, IntPtr buf);

        #endregion

        #region protected methods
        /// <summary>
        /// This method is supplied by the derived class and initializes the protected variables
        /// of this class to reflect the disk image, regardless of the image source.
        /// </summary>
        /// <returns></returns>
        protected abstract DiskOperationError Initialize();

        /// <summary>
        /// Write a header for the disk.  This header currently contains the type of image (CD vs DVD)
        /// and the number of sessions in the image.
        /// </summary>
        /// <param name="wrtr">the output file</param>
        /// <returns>null if sucessful, otherwise an object representing the error that occurred</returns>
        protected DiskOperationError WriteHeader(OsFileWriter wrtr)
        {
            int amount;
            byte [] data = new byte[2048] ;

            if (m_sessions.GetLength(0) < 1)
                return new BurnerError("No sessions detected on the disk");

            if (m_sessions.GetLength(0) > 1)
                return new BurnerError("No support for multi-sessions disks (YET)");

            data[0] = Signature1;
            data[1] = Signature2;
            data[2] = Signature3;
            data[3] = Signature4;
            data[4] = MajorVersion;
            data[5] = MinorVersoin;
            data[6] = (byte)m_disk_type ;
            data[7] = (byte)m_sessions.GetLength(0) ;
            data[8] = (byte)(m_open_disk ? 1 : 0);
            if (!wrtr.WriteData(data, out amount) || amount != data.GetLength(0))
                return new DeviceIoctlError("WriteData", (uint)wrtr.LastError);

            return null;
        }

        private void ProgressHandler(object sender, ProgressChangedEventArgs args)
        {
            if (ProgressChanged != null)
                ProgressChanged(sender, args);
        }

        #endregion
    }
}
