//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//
using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using Bwg.Logging;
using Bwg.Utils;
using Bwg.Scsi;
using Bwg.Burner;

namespace Bwg.DiskImage
{
    public sealed class DriveDiskContents : DiskContents
    {
        #region private member variables
        /// <summary>
        /// This is a drive containing a disk that is the source of the disk image
        /// </summary>
        private Drive m_source;

        private int m_read_errors;
        private int m_read_attempts;

        #endregion

        #region constructors
        public DriveDiskContents(Logger l, Drive d) : base(l)
        {
            m_source = d;
            m_read_attempts = 32;
            m_read_errors = 0;
        }
        #endregion

        #region protected methods
        protected override DiskOperationError Initialize()
        {
            // Find the disk type (CD, DVD)
            if (m_source.IsCd)
                m_disk_type = ImageType.CD;
            else if (m_source.IsDvd)
                m_disk_type = ImageType.DVD;
            else
                m_disk_type = ImageType.Undefined;

            // Extract information from the disk
            Disk d = new Disk(m_source);
            if (d.IsEmpty)
            {
                m_disk_type = ImageType.Undefined;
                return null;
            }

            if (d.IsIncomplete)
                m_open_disk = true;
            else
                m_open_disk = false;


            int num = 1;
            foreach(Session s in d.Sessions)
            {
                SessionInfo sinfo = new SessionInfo(num++);

                foreach (Track t in s.Tracks)
                {
                    TrackBurnImage.TrackType type;

                    if (t.IsAudio)
                        type = TrackBurnImage.TrackType.Audio;
                    else if (t.IsDataMode1)
                        type = TrackBurnImage.TrackType.Data_Mode_1;
                    else if (t.IsDataMode2Form1)
                        type = TrackBurnImage.TrackType.Data_Mode_2_Form_1;
                    else if (t.IsDataMode2Form2)
                        type = TrackBurnImage.TrackType.Data_Mode_2_Form_2;
                    else if (t.IsDataMode2Formless)
                        type = TrackBurnImage.TrackType.Data_Mode_2_Formless;
                    else
                    {
                        return new BurnerError("Unknown track type, Session " + sinfo.Number.ToString() +", Track " + t.Info.TrackNumber.ToString());
                    }
                    sinfo.AddTrack(new TrackInfo(type, t.Info.TrackNumber, t.Info.TrackStartAddress, t.Info.TrackSize));
                }

                Array.Resize(ref m_sessions, m_sessions.GetLength(0) + 1);
                m_sessions[m_sessions.GetLength(0) - 1] = sinfo;
            }

            return null;
        }

        internal override int MaxBufferSize
        {
            get
            {
                return m_source.Device.MaximumTransferLength;
            }
        }

        private Device.CommandStatus ReadTrackDataError(int current, int toread, int secsize, IntPtr buf)
        {
            Device.CommandStatus st = Device.CommandStatus.Success;

            while (toread > 0)
            {
                for (int i = 0; i < m_read_attempts; i++)
                {
                    st = m_source.Device.Read(false, false, (uint)current, 1, buf, 2048);
                    if (st == Device.CommandStatus.Success)
                        break;

                    m_read_errors++;
                }

                if (st != Device.CommandStatus.Success)
                    return st;

                toread--;
                buf = new IntPtr(buf.ToInt32() + 2048);
            }

            return Device.CommandStatus.Success;
        }

        internal override DiskOperationError ReadTrackData(int current, int toread, int secsize, IntPtr buf)
        {
            Device.CommandStatus st ;

            if (secsize == 2048)
            {
                st = m_source.Device.Read(false, false, (uint)current, (uint)toread, buf, (ushort)(secsize * toread));
                if (st != Device.CommandStatus.Success)
                {
                    m_read_errors++;
                    st = ReadTrackDataError(current, toread, secsize, buf);
                    if (st != Device.CommandStatus.Success)
                        return Drive.CreateErrorObject(st, m_source.Device);
                }
            }
            else
            {
                byte exp ;
                bool dap = false ;

                if (secsize == 2352)
                {
                    // Audio
                    dap = true ;
                    exp = 1;
                }
                else if (secsize == 2324)
                {
                    // Mode 2, form 2
                    exp = 5;
                }
                else if (secsize == 2336)
                {
                    // Mode 2, formless
                    exp = 3;
                }
                else
                    return new BurnerError("unsupported track data form");

                st = m_source.Device.ReadCD(exp, dap, (uint)current, (uint)toread, buf, secsize * toread);
                if (st != Device.CommandStatus.Success)
                    return Drive.CreateErrorObject(st, m_source.Device);
            }

            return null;
        }

        #endregion
    }
}
