using System;
using System.Collections.Generic;
using System.Text;
using System.ComponentModel;
using System.Diagnostics;

namespace Bwg.DvdMaster
{
    public class VMGM_MAT : DvdDataStructure
    {
        #region private constants
        private const string VIDEO_MGR_ID = "DVDVIDEO-VMG";
        #endregion

        #region private member variables
        private string m_video_manager_id;
        private uint m_last_sector_vmg;
        private uint m_last_sector_vmgi;
        private ushort m_dvd_version;
        private uint m_category_and_mask;
        private ushort m_number_volumes;
        private ushort m_current_volume;
        private byte m_current_side;
        private ushort m_number_video_title_sets;
        private string m_provider_id;
        private ulong m_vmg_pos_code;
        private uint m_end_byte_address_vmgi_mat;
        private uint m_first_play_pgc_addr;
        private uint m_start_sector_vmgm_vobs ;
        private uint m_start_sector_vmg_ptt_srpt ;
        private uint m_start_sector_vmgm_pgci_ut ;
        private uint m_start_sector_vmg_ptl_mait ;
        private uint m_start_sector_vmg_vts_atrt ;
        private uint m_start_sector_vmg_txtdt_mg ;
        private uint m_start_sector_vmgm_c_adt ;
        private uint m_start_sector_vmgm_vobu_admap ;
        private ushort m_video_attributes;
        private byte m_number_of_audio_streams;
        private ulong [] m_audio_stream_attributes;
        private byte m_number_of_subpicture_streams;
        private byte[] m_subpicture_stream_attributes;
        #endregion

        #region constructor
        /// <summary>
        /// Create an empty VMGM_MAT object based on default values
        /// </summary>
        public VMGM_MAT()
        {
            m_video_manager_id = VIDEO_MGR_ID;
        }

        /// <summary>
        /// Create a VMGM_MAT object based on a block of data from the from containing file
        /// </summary>
        /// <param name="data">the data read from the file</param>
        public VMGM_MAT(byte [] data)
        {
            int offset = 0;

            Debug.Assert(data.GetLength(0) >= 376);

            m_video_manager_id = ExtractString(data, ref offset, 12);
            if (m_video_manager_id != VIDEO_MGR_ID)
                throw new BadDataStructureException("Video Manager Identifier", 0, "expected value '" + VIDEO_MGR_ID + "'");

            m_last_sector_vmg = ExtractUInt4(data, ref offset);

            // 12 bytes of padding
            offset += 12;

            m_last_sector_vmgi = ExtractUInt4(data, ref offset);
            m_dvd_version = ExtractUInt2(data, ref offset);
            m_category_and_mask = ExtractUInt4(data, ref offset);
            m_number_volumes = ExtractUInt2(data, ref offset);
            m_current_volume = ExtractUInt2(data, ref offset);
            m_current_side = data[offset++];

            // 19 bytes of padding
            offset += 19;

            m_number_video_title_sets = ExtractUInt2(data, ref offset);
            m_provider_id = ExtractString(data, ref offset, 32);
            m_vmg_pos_code = ExtractUInt8(data, ref offset);

            // 24 bytes of padding
            offset += 24;

            m_end_byte_address_vmgi_mat = ExtractUInt4(data, ref offset);
            m_first_play_pgc_addr = ExtractUInt4(data, ref offset);

            // 56 bytes of padding
            offset += 56;

            m_start_sector_vmgm_vobs = ExtractUInt4(data, ref offset);
            m_start_sector_vmg_ptt_srpt = ExtractUInt4(data, ref offset);
            m_start_sector_vmgm_pgci_ut = ExtractUInt4(data, ref offset);
            m_start_sector_vmg_ptl_mait = ExtractUInt4(data, ref offset);
            m_start_sector_vmg_vts_atrt = ExtractUInt4(data, ref offset);
            m_start_sector_vmg_txtdt_mg = ExtractUInt4(data, ref offset);
            m_start_sector_vmgm_c_adt = ExtractUInt4(data, ref offset);
            m_start_sector_vmgm_vobu_admap = ExtractUInt4(data, ref offset);

            // 32 bytes of padding
            offset += 32;

            m_video_attributes = ExtractUInt2(data, ref offset);

            // 1 byte of padding
            offset++;

            m_number_of_audio_streams = data[offset++];

            m_audio_stream_attributes = new ulong[8];
            for (int i = 0; i < 8; i++)
                m_audio_stream_attributes[i] = ExtractUInt8(data, ref offset);

            // 17 bytes of padding
            offset += 17;

            m_number_of_subpicture_streams = data[offset++];
            m_subpicture_stream_attributes = new byte[32];

            for (int i = 0; i < m_number_of_subpicture_streams; i++)
                m_subpicture_stream_attributes[i] = data[offset++];
        }
        #endregion

        #region public properties
        [Description("The ID string that identifies this data structure as the video manager")]
        public string VideoManagerId
        {
            get { return m_video_manager_id; }
        }

        [Description("The last sector of the video manager (VIDEO_TS.IFO + VIDEO_TS.VOB + VIDEO_TS.BUP)")]
        public uint LastSectorVmg
        {
            get { return m_last_sector_vmg ; }
        }

        [Description("The last sector of the video manager information section (VIDEO_TS.IFO)")]
        public uint LastSectorVmgi
        {
            get { return m_last_sector_vmgi ; }
        }

        [Description("The version number of the DVD specification that this disc follows")]
        public ushort DvdVersion
        {
            get { return m_dvd_version; }
        }

        [Description("The category and region mask for this disc")]
        public uint CategoryAndRegionMask
        {
            get { return m_category_and_mask; }
        }

        [Description("The number of volumes in this DVD set")]
        public ushort NumberOfVolumes
        { 
            get { return m_number_volumes; } 
        }

        [Description("The number of the current volume in this DVD set")]
        public ushort CurrentVolume
        {
            get { return m_current_volume; }
        }

        [Description("The side of the current disc")]
        public byte CurrentSide
        {
            get { return m_current_side; }
        }

        [Description("The number of video title sets stored on this DVD")]
        public ushort NumberOfVideoTitleSets
        {
            get { return m_number_video_title_sets; }
        }

        [Description("The provider of the content on this DVD")]
        public string ProviderID
        {
            get { return m_provider_id; }
        }

        [Description("The VMG Position code ???")]
        public ulong VmgPositionCode
        {
            get { return m_vmg_pos_code; }
        }

        [Description("The end byte address of the VMGI_MAT structure")]
        public uint EndByteAddressOfVmgiMat
        {
            get { return m_end_byte_address_vmgi_mat; }
        }

        [Description("The byte address of the first play program chain (PGC)")]
        public uint FirstPlayPGCByteAddress
        {
            get { return m_first_play_pgc_addr ; }
        }

        [Description("Start sector for the VMGM_VOBS structure")]
        public uint VmgmVobsStartSector
        {
            get { return m_start_sector_vmgm_vobs; }
        }

        [Description("Start sector for the VMG_PTT_SRPT structure")]
        public uint VmgPttSrptStartSector
        {
            get { return m_start_sector_vmg_ptt_srpt; }
        }

        [Description("Start sector for the VMGM_PCGI_UT structure")]
        public uint VmgmPgciUtStartSector
        {
            get { return m_start_sector_vmgm_pgci_ut; }
        }

        [Description("Start sector for the VMG_PTL_MAIT structure")]
        public uint VmgPtlMaitStartSector
        {
            get { return m_start_sector_vmg_ptl_mait; }
        }

        [Description("Start sector for the VMG_VTS_ATRT structure")]
        public uint VmgVtsAtrtStartSector
        {
            get { return m_start_sector_vmg_vts_atrt; }
        }

        [Description("Start sector for the VMG_TXTDT_MG structure")]
        public uint VmgTxtdtMgStartSector
        {
            get { return m_start_sector_vmg_txtdt_mg; }
        }

        [Description("Start sector for the VMGM_C_ADT structure")]
        public uint VmgmCAdtStartSector
        {
            get { return m_start_sector_vmgm_c_adt; }
        }

        [Description("Start sector for the VMGM_VOBU_ADMAP structure")]
        public uint VmgmVobuAdmapStartSector
        {
            get { return m_start_sector_vmgm_vobu_admap; }
        }

        [Description("Video Attributes")]
        public ushort VideoAttributes
        {
            get { return m_video_attributes; }
        }

        [Description("Number Of Audio Streams")]
        public byte NumberOfAudioStreams
        {
            get { return m_number_of_audio_streams; }
        }

        [Description("Audio Stream Attributes For Stream 1")]
        public ulong AudioStream1Attributes
        {
            get { return m_audio_stream_attributes[0]; }
        }

        [Description("Audio Stream Attributes For Stream 2")]
        public ulong AudioStream2Attributes
        {
            get { return m_audio_stream_attributes[1]; }
        }

        [Description("Audio Stream Attributes For Stream 3")]
        public ulong AudioStream3Attributes
        {
            get { return m_audio_stream_attributes[2]; }
        }

        [Description("Audio Stream Attributes For Stream 4")]
        public ulong AudioStream4Attributes
        {
            get { return m_audio_stream_attributes[3]; }
        }

        [Description("Audio Stream Attributes For Stream 5")]
        public ulong AudioStream5Attributes
        {
            get { return m_audio_stream_attributes[4]; }
        }

        [Description("Audio Stream Attributes For Stream 6")]
        public ulong AudioStream6Attributes
        {
            get { return m_audio_stream_attributes[5]; }
        }

        [Description("Audio Stream Attributes For Stream 7")]
        public ulong AudioStream7Attributes
        {
            get { return m_audio_stream_attributes[6]; }
        }

        [Description("Audio Stream Attributes For Stream 8")]
        public ulong AudioStream8Attributes
        {
            get { return m_audio_stream_attributes[7]; }
        }

        [Description("Number Of Sub-Picture Streams")]
        public byte NumberOfSubPictureStreams
        {
            get { return m_number_of_subpicture_streams; }
        }
        #endregion
    }
}
