//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Threading ;
using System.Runtime.InteropServices;
using Bwg.Logging;
using Bwg.Controls;
using Bwg.Scsi;
using Bwg.Burner;

namespace Bwg.EraseDisk
{
    public partial class EraseDiskForm : BwgBurnBaseForm
    {
        #region private member variables
        private Drive m_drive;
        private bool m_quick;
        private DateTime m_start;
        private System.Windows.Forms.Timer m_timer;
        private long m_user_data_area_size;

        private DefectProcessingType m_defect_processing;

        // The results of the erase background thread
        private DiskOperationError m_exit_status;

        #endregion

        #region private types
        private delegate void TimerTickDelegate() ;
        private delegate void DoneDelegate(Device.CommandStatus status, uint errorcode);
        private delegate void ProgressHandlerDelegate(object sender, ProgressChangedEventArgs args);
        private delegate void SetProgressDelegate(int value);

        private enum DefectProcessingType
        {
            KeepExistingDefects,
            RemoveAllDefects,
            RescanForDefects
        } ;
        #endregion

        #region constructor
        /// <summary>
        /// Create the erase disk application form
        /// </summary>
        public EraseDiskForm()
        {
            m_quick = true;

            InitializeComponent();

            LoggerWindow.SetupColumns();
            LoggerWindow.Sink = m_sink;
            DeviceSelector.Logger = m_logger;

            m_logger.LogMessage(new UserMessage(UserMessage.Category.Info, 0, "CD-RW Disk Eraser Version " + GetVersion()));

            DeviceSelector.Logger = m_logger;
            DeviceSelector.InitializeDeviceList(false);

            if (m_logger.HasSink(UserMessage.Category.Debug))
                ShowDebugMessages.Checked = true;
            else
                ShowDebugMessages.Checked = false;

            DebugLevel.SelectedIndex = m_logger.GetLevel(UserMessage.Category.Debug);

            Progress.Value = 0;

            m_timer = new System.Windows.Forms.Timer();

            EraseWorker.DoWork += new DoWorkEventHandler(EraseWorker_DoWork);
            EraseWorker.RunWorkerCompleted += new RunWorkerCompletedEventHandler(EraseWorker_RunWorkerCompleted);
        }
        #endregion

        #region protected functions
        /// <summary>
        /// Called when closing the main form
        /// </summary>
        /// <param name="e">The event arguments</param>
        protected override void OnClosing(CancelEventArgs e)
        {
            if (EraseWorker.IsBusy)
                EraseWorker.CancelAsync();

            while (EraseWorker.IsBusy) ;

            base.OnClosing(e);
        }
        #endregion

        #region private functions

        private void Quick_CheckedChanged(object sender, EventArgs e)
        {
            m_quick = true;
        }

        private void Full_CheckedChanged(object sender, EventArgs e)
        {
            m_quick = false;
        }

        private void EraseWorker_RunWorkerCompleted(object obj, RunWorkerCompletedEventArgs args)
        {
            EnableControls(true);
            m_timer.Stop();
            Progress.Value = Progress.Maximum;
            TopLevelControl.Text = "Erase (" + Progress.Value.ToString() + "% complete)";

            string str = "" ;
            string caption = "Error!" ;
            MessageBoxIcon icon = MessageBoxIcon.Error;

            if (m_exit_status == null)
            {
                    str = "The erase operation completed sucessfully";
                    caption = "Operation Complete" ;
                    icon = MessageBoxIcon.Information;
            }
            else
            {
                str = m_exit_status.Message ;
                caption = "Error!" ;
                icon = MessageBoxIcon.Error ;
            }
            MessageBox.Show(str, caption, MessageBoxButtons.OK, icon);
        }

        private void FormatBDRE()
        {
            FormatParameterList plist = new FormatParameterList();
            plist.IMMED = 1;
            plist.FOV = 1;
            plist.DCRT = 1;
            plist.FormatType = FormatParameterList.UnitFormatType.BDRE_Format_With_Spare_Areas;
            plist.Parameter = 0;
            plist.NumberOfBlocks = (uint)m_user_data_area_size;

            switch (m_defect_processing)
            {
                case DefectProcessingType.KeepExistingDefects:
                    plist.DCRT = 0;
                    plist.FormatSubType = 3;
                    break;

                case DefectProcessingType.RemoveAllDefects:
                    plist.FormatSubType = 1;
                    break;

                case DefectProcessingType.RescanForDefects:
                    plist.DCRT = 0;
                    plist.FormatSubType = 2;
                    break;
            }

            m_exit_status = m_drive.FormatUnit(plist);
        }

        private void EraseWorker_DoWork(object sender, DoWorkEventArgs args)
        {
            if (m_drive.CurrentProfile == Drive.SupportedProfiles.BD_RE)
            {
                FormatBDRE() ;
            }
            else
            {
                if (m_drive.IsEraseSupported)
                {
                    Drive.EraseType et = Drive.EraseType.Minimal;
                    if (!m_quick)
                        et = Drive.EraseType.Full;

                    m_exit_status = m_drive.Erase(et);
                    if (m_exit_status is ScsiDeviceError)
                    {
                        ScsiDeviceError err = (ScsiDeviceError)m_exit_status;
                        if (err.ASC == 0x04 && (err.ASCQ == 0x08 || err.ASCQ == 0x07))
                        {
                            //
                            // LONG WRITE IN PROGRESS
                            //
                            // A bug in a previous program, or stopping a previous program in the debugger has left
                            // a long write operation pending because SYNC_CACHE has not been called.  Call SyncCache
                            // and then retry the erase
                            //
                            m_drive.Device.SynchronizeCache(false);
                            m_exit_status = m_drive.Erase(et);
                        }
                    }
                    m_drive.Device.StartStopUnit(false, Device.PowerControl.NoChange, Device.StartState.EjectDisc);
                }
            }
        }

        private void Erase_Click(object sender, EventArgs e)
        {
            if (m_drive == null)
            {
                if (DeviceSelector.Items.Count == 0)
                {
                    MessageBox.Show("No CD/RW or DVD/RW devices were detecte on your computer.");
                }
                else
                {
                    MessageBox.Show("You must select a device before you can erase a disk.");
                }
            }
            else
            {
                ElapsedTime.Text = "0:00";
                Progress.Minimum = 0;
                Progress.Maximum = 100;

                DiscInformation info;
                Device.CommandStatus st = m_drive.Device.ReadDiscInformation(out info);
                if (st != Device.CommandStatus.Success)
                {
                    string str = "Cannot read the disk information for drive '" + m_drive.DriveModel + "'";
                    MessageBox.Show(str, "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    return;
                }

                if (m_drive.IsEraseSupported || m_drive.CurrentProfile == Drive.SupportedProfiles.BD_RE)
                {
                    EnableControls(false);

                    if (m_drive.CurrentProfile == Drive.SupportedProfiles.BD_RE)
                    {
                        string str = UserDataAreaSize.SelectedItem.ToString();
                        m_user_data_area_size = long.Parse(str);
                    }

                    m_start = DateTime.Now;
                    m_timer.Interval = 1000;
                    m_timer.Tick += new EventHandler(TimerTickHandler);
                    m_timer.Tag = this;
                    m_timer.Start();

                    m_drive.ProgressChanged += new ProgressChangedEventHandler(UpdateProgress);
                    EraseWorker.WorkerSupportsCancellation = true;
                    EraseWorker.RunWorkerAsync();
                }
                else
                {
                    string str = "A disk of type '" + m_drive.CurrentProfile.ToString() + "' is in the drive, this type is not ";
                    str += "supported by the erase utility.";
                    MessageBox.Show(str, "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);

                    m_drive.Device.PreventAllowMediumRemoval(Device.PreventAllow.Allow);
                }
            }
        }

        private void EnableControls(bool val)
        {
            DeviceSelector.Enabled = val;
            Erase.Enabled = val;
            Quick.Enabled = val;
            Full.Enabled = val;
            DebugLevel.Enabled = val;
            ShowDebugMessages.Enabled = val;
        }

        private void UpdateProgress(object sender, ProgressChangedEventArgs args)
        {
            if (InvokeRequired)
            {
                object[] param = new object[2];
                param[0] = sender;
                param[1] = args;
                Invoke(new ProgressHandlerDelegate(UpdateProgress), param);
            }
            else
            {
                Progress.Value = args.ProgressPercentage;
                if (args.UserState is ProgressText)
                {
                    ProgressText pt = (ProgressText)args.UserState;
                    Progress.Text = pt.LongName;
                    TopLevelControl.Text = pt.ShortName + " ( " + args.ProgressPercentage.ToString() + "% )";
                }
                else if (args.UserState is string)
                {
                    Progress.Text = (string)args.UserState;
                    TopLevelControl.Text = (string)args.UserState;
                }
            }
        }

        private void TimerTick()
        {
            TimeSpan elapsed = DateTime.Now - m_start;
            if (elapsed.Hours > 0)
                ElapsedTime.Text = elapsed.Hours.ToString() + ":" + elapsed.Minutes.ToString("D2") + ":" + elapsed.Seconds.ToString("D2");
            else
                ElapsedTime.Text = elapsed.Minutes.ToString() + ":" + elapsed.Seconds.ToString("D2");
        }

        private static void TimerTickHandler(object sender, EventArgs args)
        {
            System.Windows.Forms.Timer t = (System.Windows.Forms.Timer)sender;
            EraseDiskForm form = (EraseDiskForm)t.Tag;
            if (form.InvokeRequired)
            {
                form.Invoke(new TimerTickDelegate(form.TimerTick));
            }
            else
                form.TimerTick();
        }

        private void ShowDebugMessages_CheckedChanged(object sender, EventArgs e)
        {
            if (ShowDebugMessages.Checked)
                m_logger.SetSink(UserMessage.Category.Debug, m_sink);
            else
                m_logger.RemoveSink(UserMessage.Category.Debug);

            SaveLoggerSettings();
        }

        private void DebugLevel_SelectedIndexChanged(object sender, EventArgs e)
        {
            int level = DebugLevel.SelectedIndex;
            m_logger.SetLevel(UserMessage.Category.Debug, level);

            SaveLoggerSettings();
        }

        private void DeviceSelector_SelectedIndexChanged(object sender, EventArgs e)
        {
            DiskOperationError status;

            if (m_drive != null)
                m_drive.Dispose();
            m_drive = null;

            string devname = DeviceSelector.SelectedDeviceName;
            Device dev = new Device(m_logger);
            if (!dev.Open(devname))
            {
                dev.Dispose();
                MessageBox.Show("Cannot open device '" + devname + "'", "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            m_drive = new Drive(dev);
            status = m_drive.Initialize();
            if (status != null)
            {
                m_drive.Dispose();
                m_drive = null;
                MessageBox.Show("Cannot initialize device '" + devname + "' - " + status.Message, "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            BDREGroup.Visible = false ;
            CDRW_DVDRWGroup.Visible = false ;
            Erase.Text = "------" ;
            Erase.Enabled = false;

            if (m_drive.CurrentProfile == Drive.SupportedProfiles.BD_RE)
            {
                BDREGroup.Visible = true ;
                Erase.Text = "Format" ;
                Erase.Enabled = true;

                DiskOperationError err;

                long minimum, maximum, preferred;
                long mediasize;

                err = m_drive.GetBDRESpareSizeValues(out minimum, out maximum, out preferred, out mediasize);
                UserDataAreaSize.Items.Add(minimum.ToString());
                UserDataAreaSize.Items.Add(maximum.ToString());
                UserDataAreaSize.Items.Add(preferred.ToString());
                UserDataAreaSize.SelectedIndex = 1;
                radioButton2.Checked = true;
            }
            else if (m_drive.IsEraseSupported)
            {
                CDRW_DVDRWGroup.Text = m_drive.CurrentProfile.ToString() + " Parameters";
                CDRW_DVDRWGroup.Visible = true;
                Erase.Text = "Erase";
                Erase.Enabled = true;
            }
        }

        private void radioButton1_CheckedChanged(object sender, EventArgs e)
        {
            m_defect_processing = DefectProcessingType.RemoveAllDefects;
        }

        private void radioButton2_CheckedChanged(object sender, EventArgs e)
        {
            m_defect_processing = DefectProcessingType.KeepExistingDefects;
        }

        private void radioButton3_CheckedChanged(object sender, EventArgs e)
        {
            m_defect_processing = DefectProcessingType.RescanForDefects;
        }
        
        #endregion
    }
}