//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//

using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;

namespace Bwg.Isofs
{
    class DirectoryTask : CreatorTask
    {
        private DiskImageElement.NameType m_name_type;
        private IList<DiskImageDirectory> m_dirlist;

        public DirectoryTask(DiskImageElement.NameType type)
        {
            m_name_type = type;
            m_dirlist = new List<DiskImageDirectory>();
        }

        public override bool UpdateLayout(DiskImageCreator creator, DiskInfo info)
        {
            m_dirlist = creator.GetDirectoryList(m_name_type);

            foreach (DiskImageDirectory dir in m_dirlist)
            {
                ExtentAd ex = new ExtentAd(info.CurrentBlock, dir.GetExtentSizeInBlocks(m_name_type) * 2048);
                switch(m_name_type)
                {
                    case DiskImageElement.NameType.ISO9660:
                        dir.SetDirectoryLocation(DiskImageDirectory.DirectoryLocationType.ISO9660, ex);
                        break;
                    case DiskImageElement.NameType.Joliet:
                        dir.SetDirectoryLocation(DiskImageDirectory.DirectoryLocationType.Joliet, ex);
                        break;
                    default:
                        Debug.Assert(false) ;
                        break ;
                }
                info.CurrentBlock += dir.GetExtentSizeInBlocks(m_name_type);
            }

            return true;
        }

        private bool WriteSpecialDir(DiskImageCreator creator, byte name, long loc, long size, IsoWriteStream strm)
        {
            strm.WriteUint8(34);               // Size of . or .. entries
            strm.WriteUint8(0);

            Debug.Assert(loc <= Int32.MaxValue);
            strm.WriteUint32Both((int)loc);

            Debug.Assert(size * 2048 <= Int32.MaxValue);
            strm.WriteUint32Both((int)(size * 2048));
            strm.WriteBinaryDateTime(creator.Settings.RootDirectoryCreateDateTime);
            strm.WriteUint8((byte)(Constants.DirectoryEntryFileFlags.Directory));
            strm.WriteUint8(0);
            strm.WriteUint8(0);
            strm.WriteUint16Both(1);
            strm.WriteUint8(1);                // Length of the name (1  byte)
            strm.WriteUint8(name);             // The one byte name flag

            return true;
        }

        private bool WriteOneElement(DiskImageCreator creator, DiskImageElement elem, IsoWriteStream strm)
        {
            byte len ;
            Encoding enc = GetEncoding(m_name_type);
            byte [] name = enc.GetBytes(elem.GetName(m_name_type)) ;

            len = (byte)(33 + name.GetLength(0));
            if ((len % 2) != 0)
                len++;

            Debug.Assert((len % 2) == 0);

            strm.WriteUint8(len);
            strm.WriteUint8(0);

            long loc;
            long size;
            if (elem is DiskImageDirectory)
            {
                DiskImageDirectory dir = (DiskImageDirectory)elem;
                loc = dir.GetDirectoryLocation(NameTypeToDirLocationType(m_name_type)).Location;
                size = dir.GetExtentSizeInBlocks(m_name_type) * 2048;
            }
            else if (elem is DiskImageFile)
            {
                DiskImageFile file = (DiskImageFile)elem;
                Debug.Assert(file.GetFileExtentCount(DiskImageFile.FileLocationType.DataLocation) == 1);
                loc = file.GetFileExtent(DiskImageFile.FileLocationType.DataLocation, 0).Location;
                size = file.FileSizeInBytes;
            }
            else
            {
                throw new Exception("Unknown IsoFsElement derived type - internal error");
            }

            Debug.Assert(loc <= Int32.MaxValue);
            Debug.Assert(size <= Int32.MaxValue);

            strm.WriteUint32Both((int)loc);
            strm.WriteUint32Both((int)size);

            strm.WriteBinaryDateTime(creator.Settings.RootDirectoryCreateDateTime);
            if (elem is DiskImageDirectory)
                strm.WriteUint8((byte)(Constants.DirectoryEntryFileFlags.Directory));
            else
                strm.WriteUint8(0);
            strm.WriteUint8(0);
            strm.WriteUint8(0);
            strm.WriteUint16Both(1);

            strm.WriteUint8((byte)name.GetLength(0)) ;
            strm.Write(name) ;
            if ((name.GetLength(0) % 2) == 0)
                strm.WriteUint8(0);

            return true;
        }

        private bool WriteOneDirectoryEntry(DiskImageCreator creator, DiskImageDirectory dir, IsoWriteStream strm)
        {
            long start = strm.Position;
            DiskImageDirectory.DirectoryLocationType loctype = NameTypeToDirLocationType(m_name_type);
            Encoding enc = GetEncoding(m_name_type);

            if (!WriteSpecialDir(creator, 0, dir.GetDirectoryLocation(loctype).Location, dir.GetExtentSizeInBlocks(m_name_type), strm))
                return false;

            if (dir.Parent == null)
            {
                if (!WriteSpecialDir(creator, 1, dir.GetDirectoryLocation(loctype).Location, dir.GetExtentSizeInBlocks(m_name_type), strm))
                    return false;
            }
            else
            {
                if (!WriteSpecialDir(creator, 1, dir.Parent.GetDirectoryLocation(loctype).Location, dir.Parent.GetExtentSizeInBlocks(m_name_type), strm))
                    return false;
            }

            foreach (DiskImageElement elem in dir)
            {
                Debug.Assert(strm.Position / 2048 < Int32.MaxValue);
                int curblock = (int)(strm.Position / 2048);
                byte[] name = enc.GetBytes(elem.GetName(m_name_type));
                int dirsize = 33 + name.GetLength(0);
                if ((dirsize % 2) == 0)
                    dirsize++;

                if ((strm.Position + dirsize) / 2048 != curblock)
                {
                    // This directory entry would cross the boundary of a sector, so
                    // pad the current sector to the end before writing the next entry
                    strm.WriteZeros((curblock + 1) * 2048 - strm.Position);
                }
                if (!WriteOneElement(creator, elem, strm))
                    return false;
            }

            // Pad the block with zeros
            uint written = (uint)(strm.Position - start) ;
            strm.WriteZeros(dir.GetExtentSizeInBlocks(m_name_type) * 2048 - written);

            return true;
        }

        public override bool Write(DiskImageCreator creator, DiskInfo info, IsoWriteStream strm)
        {
            foreach (DiskImageDirectory dir in m_dirlist)
                if (!WriteOneDirectoryEntry(creator, dir, strm))
                    return false;

            return true;
        }
    }
}
