//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//

using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Security;
using System.Threading;
using System.Windows.Forms;
using System.ComponentModel;
using System.Diagnostics;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;
using Bwg.Logging;

namespace Bwg.Isofs
{
    //
    // ISO 9660 compliance
    //
    // In general the sum of the file name and the file extension shall not
    // exceed 30 characters
    //
    //     Level 1
    //        - File names must be 8.3
    //        - Directory names must be 8 characters, no extension
    //        - Files must be contained in a single file entry
    //     Level 2
    //        - Files must be contained in a single file entry
    //     Level 3
    //        - No restrictions
    //
    //

    /// <summary>
    /// This class creates a disk image to be written to a disk.
    /// </summary>
    [Serializable()]
    public class DiskImageCreator
    {
        #region private member variables

        /// <summary>
        /// The disk image we are writing to the disk
        /// </summary>
        private DiskImageContents m_image;

        /// <summary>
        /// The starting block for the image
        /// </summary>
        private long m_start_block;

        /// <summary>
        /// The size of the image in blocks
        /// </summary>
        private long m_size_in_blocks;

        /// <summary>
        /// This class is used to check the image to be sure it is valid.
        /// </summary>
        private ImageChecker m_checker;

        /// <summary>
        /// The message logger to send messages to
        /// </summary>
        private Logger m_logger;

        /// <summary>
        /// The count of errors that have occurred in the content definition processes
        /// </summary>
        private int m_error_count;

        #endregion

        #region public memory variables

        /// <summary>
        /// This event is fired when the validation progress has changed.
        /// </summary>
        public ProgressChangedEventHandler ValidationProgressChanged;

        /// <summary>
        /// This event is fired during the creation process to mark progress
        /// </summary>
        public ProgressChangedEventHandler CreateProgressChanged;

        #endregion

        #region public types

        #endregion

        #region constructors
        /// <summary>
        /// Creates a new ISO image creator object
        /// </summary>
        public DiskImageCreator(Logger logger, DiskImageContents image)
        {
            m_logger = logger;
            m_image = image;

            m_size_in_blocks = 0;
            m_checker = new IsoImageChecker(m_logger);
        }

        #endregion

        #region public properties

        /// <summary>
        /// This method returns the logger associated with the image creator
        /// </summary>
        internal Logger MyLogger
        {
            get
            {
                return m_logger;
            }
        }

        /// <summary>
        /// This property returns the number of directories in the image
        /// </summary>
        public int DirectoryCount
        {
            get
            {
                return m_image.DirectoryCount;
            }
        }

        /// <summary>
        /// This property returns the number of files in the image
        /// </summary>
        public int FileCount
        {
            get
            {
                return m_image.FileCount;
            }
        }

        /// <summary>
        /// Return the settings assocaited with this creator
        /// </summary>
        public ImageSettings Settings
        {
            get
            {
                return m_image.Settings;
            }
            set
            {
                m_image.Settings = value;
            }
        }

        /// <summary>
        /// This property returns the starting block of the image on disk
        /// </summary>
        public long StartBlock
        {
            get
            {
                return m_start_block;
            }
        }

        /// <summary>
        /// The maximum directory depth for a file in the disk image
        /// </summary>
        public int MaximumDirectoryDepth
        {
            get
            {
                return m_checker.MaximumDirectoryDepth;
            }

            set
            {
                m_checker.MaximumDirectoryDepth = value;
            }
        }

        /// <summary>
        /// The maximum total path length for a file
        /// </summary>
        public int MaximumTotalPathLength
        {
            get
            {
                return m_checker.MaximumTotalPathLength;
            }
            set
            {
                m_checker.MaximumTotalPathLength = value;
            }
        }

        /// <summary>
        /// This property returns the number of errors detected during each stage of the
        /// image creation process.  During the process of describing the image (via Add() or
        /// AddContentFile() calls) this count reflects the number of errors detected while
        /// describing the image.  During the process of creating the ISO image, this reflects
        /// the number of errors detected during creation.
        /// </summary>
        public int ErrorCount
        {
            get
            {
                return m_error_count;
            }
        }
        /// <summary>
        /// The size of the disk in blocks.
        /// </summary>
        public long SizeInBlocks
        {
            get
            {
                TaskGroup tasklist;

                if (!CreateTaskList(out tasklist))
                    return 0;

                DiskInfo info = new DiskInfo(0, 2048);
                if (!tasklist.UpdateLayout(this, info))
                    return 0;

                m_size_in_blocks = info.CurrentBlock;
                return info.CurrentBlock;
            }
        }

        /// <summary>
        /// The root directory for the disk
        /// </summary>
        public DiskImageDirectory RootDirectory
        {
            get
            {
                return m_image.RootDirectory;
            }
        }
        #endregion

        #region public methods

        /// <summary>
        /// This method returns true if this creator containa  file that is equal to
        /// or greater than the size given
        /// </summary>
        /// <param name="size">the file size limit</param>
        /// <returns>true if a file of the size given is present, false otherwise</returns>
        public bool ContainsFilesSized(long size)
        {
            return m_image.ContainsFilesSized(size);
        }

        /// <summary>
        /// This method removes all files of the size given or larger
        /// </summary>
        /// <param name="size">the size of file in question</param>
        public int RemoveFilesSized(long size)
        {
            return m_image.RemoveFilesSized(size);
        }

        /// <summary>
        /// This method adds an extension to the list of allowed extensions.
        /// </summary>
        /// <param name="ext">an extension to allow on the disk</param>
        public void AddExt(string ext)
        {
            m_image.AddExt(ext);
        }

        /// <summary>
        /// This method validates the image via the Image checker to insure the
        /// image is valid
        /// 
        /// TODO - write the ISO validator.  For now just simulate the existance of one to make sure
        ///        the infrastructure is in place and works.
        /// </summary>
        /// <returns></returns>
        public bool Validate()
        {
            TaskGroup tasklist;

            if (ValidationProgressChanged != null)
            {
                ProgressChangedEventArgs args = new ProgressChangedEventArgs(0, "Validating Disk Image");
                ValidationProgressChanged(this, args);
            }

            if (!CreateTaskList(out tasklist))
                return false;

            DiskInfo info = new DiskInfo(0, 2048);
            if (!tasklist.UpdateLayout(this, info))
                return false;

            return true;
        }

        /// <summary>
        /// Create an image representing the disk image contained in this object.
        /// </summary>
        /// <param name="path">the output file name for the image</param>
        /// <param name="startblock">the starting block of the ISO file system</param>
        /// <returns>true if the image was created, false otherwise</returns>
        public bool CreateImage(string path, long startblock)
        {
            m_error_count = 0;
            FileStream outfile = null;
            string str;
            UserMessage m;

            try
            {
                outfile = new FileStream(path, FileMode.Create, FileAccess.Write);
            }
            catch (SecurityException)
            {
                str = path + ": access denied";
                m = new UserMessage(UserMessage.Category.Error, 1, str);
                m_logger.LogMessage(m);
                m_error_count++;
            }
            catch (ArgumentException)
            {
                str = path + ": invalid path name";
                m = new UserMessage(UserMessage.Category.Error, 1, str);
                m_logger.LogMessage(m);
                m_error_count++;
            }
            catch (UnauthorizedAccessException)
            {
                str = path + ": access denied";
                m = new UserMessage(UserMessage.Category.Error, 1, str);
                m_logger.LogMessage(m);
                m_error_count++;
            }
            catch (PathTooLongException)
            {
                str = path + ": invalid path name, too long";
                m = new UserMessage(UserMessage.Category.Error, 1, str);
                m_logger.LogMessage(m);
                m_error_count++;
            }
            catch (NotSupportedException)
            {
                str = path + ": invalid path name, contains invalid ':' character";
                m = new UserMessage(UserMessage.Category.Error, 1, str);
                m_logger.LogMessage(m);
                m_error_count++;
            }

            if (m_error_count != 0)
                return false;

            bool ret = CreateImage(outfile, startblock);
            outfile.Close();

            return ret;
        }

        /// <summary>
        /// Create an image and write it to a stream
        /// </summary>
        /// <param name="outstrm">the stream to write the image to</param>
        /// <param name="startblock">the starting block for the image</param>
        /// <returns></returns>
        public bool CreateImage(Stream outstrm, long startblock)
        {
            TaskGroup tasklist;
            IsoWriteStream strm = new IsoWriteStream(outstrm, 65536);

            m_start_block = startblock;

            if (!CreateTaskList(out tasklist))
                return false;

            DiskInfo info = new DiskInfo(startblock, 2048);
            if (!tasklist.UpdateLayout(this, info))
                return false;

            info.NextPass();

            if (!tasklist.Write(this, info, strm))
                return false;

            if (CreateProgressChanged != null)
            {
                ProgressChangedEventArgs args = new ProgressChangedEventArgs(100, "Image Creation Complete");
                CreateProgressChanged(this, args);
            }

            return true;
        }

        private void TaskProgressChanged(object sender, ProgressChangedEventArgs args)
        {
        }

        /// <summary>
        /// This method returns a list of directories sorted based on the path table
        /// order.
        /// </summary>
        /// <returns></returns>
        public IList<DiskImageDirectory> GetDirectoryList(DiskImageElement.NameType type)
        {
            IList<DiskImageDirectory> dirs = new List<DiskImageDirectory>();
            dirs.Add(RootDirectory);
            RootDirectory.GetDirectories(ref dirs, true);
            dirs = SortDirectories(dirs, type);
            return dirs;
        }

        #endregion

        #region internal methods
        internal Logger GetLogger()
        {
            return m_logger;
        }

        //internal CreatorTask FindTask(Type t, int which)
        //{
        //    foreach (CreatorTask task in tasklist)
        //    {
        //        if (task.GetType() == t)
        //        {
        //            if (which == 0)
        //                return task;
        //            which--;
        //        }
        //    }
        //    return null;
        //}
        #endregion

        #region private member functions
        /// <summary>
        /// This function is really hairy.  This is especially true in order to support the UDF/ISO 9660 bridge
        /// format.  The ISO 9660 tasks and the UDF tasks are interspersed here in an order taken from the 
        /// ECMA Technical Report TR/71.
        /// </summary>
        /// <param name="tlist">the task group to fill in</param>
        /// <returns>true if sucessful, false otherwise</returns>
        private bool CreateTaskList(out TaskGroup tlist)
        {
            VolumeDescriptorTask primary = null;
            VolumeDescriptorTask joliet = null;
            ElToritoBootVolumeTask bootvolume = null ;
            TaskGroup main_seq = null;
            TaskGroup reserve_seq = null;
            TaskGroup logical_integrity = null;
            TaskGroup main_partition = null ;
            LogicalVolumeIntegrityDescriptorTask integrity_task = null;
            LogicalVolumeDescriptorTask main_logical_volume = null;
            LogicalVolumeDescriptorTask reserve_logical_volume = null;
            PartitionDescriptorTask main_part_desc = null ;
            PartitionDescriptorTask reserve_part_desc = null ;
            int udf_desc_number = 0;

            tlist = new TaskGroup("**MAINLIST**");

            //
            // The first 16 blocks are for "system" use and are not used by the ISO 9660
            // image nor the UDF image.
            //
            tlist.AddTask(new PadToBlockTask(PadToBlockTask.PadMode.PadToSpecificBlock, 16));

            //
            // If we have ISO support, we write the ISO descriptors first
            if (Settings.IsISO9660)
            {
                // The primary volume descriptor
                primary = new VolumeDescriptorTask(Constants.PrimaryVolumeDescriptor, 0, DiskImageElement.NameType.ISO9660);
                tlist.AddTask(primary);

                // The book image descriptor, pointing to El Torito boot block info
                if (Settings.IsBootable)
                {
                    bootvolume = new ElToritoBootVolumeTask();
                    tlist.AddTask(bootvolume);
                }

                if (Settings.IsJoliet)
                {
                    byte[] escapes = new byte[32];

                    joliet = new VolumeDescriptorTask(Constants.SupplementaryVolumeDescriptor, 0, DiskImageElement.NameType.Joliet);

                    //
                    // Attach escape '%/E'
                    //
                    escapes[0] = 0x25;
                    escapes[1] = 0x2f;
                    escapes[2] = 0x45;
                    joliet.EscapeSequences = escapes;

                    tlist.AddTask(joliet);
                }

                // The terminator for the volume descriptors
                tlist.AddTask(new VolumeSetTerminatorTask());
            }

            //
            // Next we write the UDF descriptors
            if (Settings.IsUDF)
            {
                tlist.AddTask(new NamedDescriptor("BEA01"));
                tlist.AddTask(new NamedDescriptor("NSR02"));
                tlist.AddTask(new NamedDescriptor("TEA01"));

                //
                // Now, skip to sector 32 and write the first copy of the descriptors
                //
                tlist.AddTask(new PadToBlockTask(PadToBlockTask.PadMode.PadToSpecificBlock, 32));

                udf_desc_number = 0;
                main_seq = new TaskGroup("MainVolumeDesc", 16);
                main_seq.AddTask(new PrimaryVolumeDescriptorTask(udf_desc_number++));
                main_seq.AddTask(new ImplementationUseVolumeDescriptorTask(udf_desc_number++));
                main_part_desc = new PartitionDescriptorTask(udf_desc_number++) ;
                main_seq.AddTask(main_part_desc);
                main_logical_volume = new LogicalVolumeDescriptorTask(udf_desc_number++);
                main_seq.AddTask(main_logical_volume);
                main_seq.AddTask(new UnallocatedSpaceDescriptorTask(udf_desc_number++));
                main_seq.AddTask(new TerminatorDescriptorTask(udf_desc_number++));
                tlist.AddTask(main_seq);

                //
                // Now, skip to sector 48 and write the second copy of the descriptors
                //
                udf_desc_number = 0;
                reserve_seq = new TaskGroup("ReserveVolumeDesc", 16);
                reserve_seq.AddTask(new PrimaryVolumeDescriptorTask(udf_desc_number++));
                reserve_seq.AddTask(new ImplementationUseVolumeDescriptorTask(udf_desc_number++));
                reserve_part_desc = new PartitionDescriptorTask(udf_desc_number++);
                reserve_seq.AddTask(reserve_part_desc);
                reserve_logical_volume = new LogicalVolumeDescriptorTask(udf_desc_number++);
                reserve_seq.AddTask(reserve_logical_volume);
                reserve_seq.AddTask(new UnallocatedSpaceDescriptorTask(udf_desc_number++));
                reserve_seq.AddTask(new TerminatorDescriptorTask(udf_desc_number++));
                tlist.AddTask(reserve_seq);

                //
                // Now add the logical integrity
                //
                udf_desc_number = 0;
                TaskGroup tg = new TaskGroup("LogicalIntegrity", 8);
                logical_integrity = new TaskGroup("LogicalIntegrityWithoutPadding");
                integrity_task = new LogicalVolumeIntegrityDescriptorTask(udf_desc_number++);
                logical_integrity.AddTask(integrity_task);
                logical_integrity.AddTask(new TerminatorDescriptorTask(udf_desc_number++));
                tg.AddTask(logical_integrity);
                tlist.AddTask(tg);

                main_logical_volume.IntegrityTask = logical_integrity;
                reserve_logical_volume.IntegrityTask = logical_integrity;

                //
                // Now skip to sector 256 and write the anchor
                //
                AnchorVolumeDescriptorPointerTask anchortask = new AnchorVolumeDescriptorPointerTask(0);
                anchortask.MainVolumeDescriptorSequence = main_seq;
                anchortask.ReserveVolumeDescriptorSequence = reserve_seq;
                tlist.AddTask(new PadToBlockTask(PadToBlockTask.PadMode.PadToSpecificBlock, 256));
                tlist.AddTask(anchortask);
            }

            if (Settings.IsISO9660)
            {

                if (Settings.IsBootable)
                {
                    ElToritoBootCatalogTask catalog = new ElToritoBootCatalogTask();
                    ElToritoBootImage image = new ElToritoBootImage();

                    // If we are supporting a bootable image, write both the 
                    // boot catalog and the boot image
                    tlist.AddTask(catalog);
                    tlist.AddTask(image);

                    catalog.BootImageTask = image;
                    bootvolume.CatalogTask = catalog;
                }

                // Add the L path table (little endian)
                primary.LPathTable = new PathTableTask(DiskImageElement.NameType.ISO9660, PathTableTask.PathTableType.LPathTable);

                // Add the M path table (big endian)
                primary.MPathTable = new PathTableTask(DiskImageElement.NameType.ISO9660, PathTableTask.PathTableType.MPathTable);

                tlist.AddTask(primary.LPathTable);
                tlist.AddTask(primary.MPathTable);

                if (Settings.IsJoliet)
                {
                    joliet.LPathTable = new PathTableTask(DiskImageElement.NameType.Joliet, PathTableTask.PathTableType.LPathTable);
                    joliet.MPathTable = new PathTableTask(DiskImageElement.NameType.Joliet, PathTableTask.PathTableType.MPathTable);

                    tlist.AddTask(joliet.LPathTable);
                    tlist.AddTask(joliet.MPathTable);
                }

                primary.Directory = new DirectoryTask(DiskImageElement.NameType.ISO9660);
                tlist.AddTask(primary.Directory);

                if (Settings.IsJoliet)
                {
                    joliet.Directory = new DirectoryTask(DiskImageElement.NameType.Joliet);
                    tlist.AddTask(joliet.Directory);
                }
            }

            main_partition = new TaskGroup("MainDataParition") ;
            tlist.AddTask(main_partition);

            if (Settings.IsUDF)
            {
                FileSetDescriptorTask fdesc;

                main_part_desc.PartitionTask = main_partition;
                reserve_part_desc.PartitionTask = main_partition;

                fdesc = new FileSetDescriptorTask();
                fdesc.Partition = main_partition;
                fdesc.FilesAndDirectories = new UDFDirectoryAndFileTask();
                fdesc.FilesAndDirectories.Partition = main_partition;

                main_partition.AddTask(fdesc);
                main_partition.AddTask(fdesc.FilesAndDirectories);
            }

            // Add all of the files
            main_partition.AddTask(new FileTask());

            //
            // If this is UDF image, write a final copy of the anchor
            //
            if (Settings.IsUDF)
            {
                integrity_task.MainPartitionData = main_partition;

                AnchorVolumeDescriptorPointerTask anchortask = new AnchorVolumeDescriptorPointerTask(0);
                anchortask.MainVolumeDescriptorSequence = main_seq;
                anchortask.ReserveVolumeDescriptorSequence = reserve_seq;
                TaskRepeater anch = new TaskRepeater(anchortask, 259);
                tlist.AddTask(anch);
            }
            else
            {
                tlist.AddTask(new PadToBlockTask(PadToBlockTask.PadMode.PadWithBlockCount, 150));
            }

            return true;
        }

        private int GetParentNumber(IList<DiskImageDirectory> list, DiskImageDirectory dir)
        {
            // This is the root dir, it is #1
            if (dir.Parent == null)
                return 1;

            int i = list.IndexOf(dir);
            if (i == -1)
                return Int32.MaxValue;

            return i;
        }

        private string PadName(string name, int len)
        {
            Debug.Assert(name.Length <= len);

            if (name.Length == len)
                return name;

            string ret = name;
            while (ret.Length < len)
                ret += ' ';

            return ret;
        }

        private int Compare(IList<DiskImageDirectory> list, DiskImageElement.NameType type, DiskImageDirectory d1, DiskImageDirectory d2)
        {
            int i1, i2;

            i1 = d1.Level;
            i2 = d2.Level;

            if (i1 < i2)
                return -1;
            else if (i1 > i2)
                return 1;

            i1 = GetParentNumber(list, d1);
            i2 = GetParentNumber(list, d2);

            if (i1 < i2)
                return -1;
            else if (i1 > i2)
                return 1;

            string n1 = d1.GetName(type);
            string n2 = d2.GetName(type);
            string s1 = PadName(n1, Math.Max(n1.Length, n2.Length));
            string s2 = PadName(n2, Math.Max(n1.Length, n2.Length));

            return s1.CompareTo(s2);
        }

        private void InsertIntoList(ref IList<DiskImageDirectory> list, DiskImageElement.NameType type, DiskImageDirectory dir)
        {
            if (list.Count == 0)
                list.Add(dir);
            else
            {
                bool added = false;

                for (int i = 0; i < list.Count; i++)
                {
                    if (Compare(list, type, dir, list[i]) == -1)
                    {
                        list.Insert(i, dir);
                        added = true;
                        break;
                    }
                }

                if (!added)
                    list.Add(dir);
            }
        }

        private IList<DiskImageDirectory> SortDirectories(IList<DiskImageDirectory> dirlist, DiskImageElement.NameType type)
        {
            IList<DiskImageDirectory> newlist = new List<DiskImageDirectory>();

            foreach (DiskImageDirectory dir in dirlist)
                InsertIntoList(ref newlist, type, dir);

            return newlist;
        }

        #endregion
    }
}
