//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//
using System;
using System.Collections.Generic;
using System.Text;

namespace Bwg.Isofs
{
    /// <summary>
    /// This class contains all of the parameters that controls the creation of an ISO or
    /// UDF image.
    /// </summary>
    [Serializable()]
    public class ImageSettings : ICloneable
    {
        #region public types

        /// <summary>
        /// The possible types of ISO images
        /// </summary>
        public enum ISOImageType
        {
            /// <summary>
            /// No support for an ISO image
            /// </summary>
            None,

            /// <summary>
            /// An ISO 9660 image, level 1
            /// </summary>
            ISO9660_Level1,

            /// <summary>
            /// An ISO 9660 image, level 2
            /// </summary>
            ISO9660_Level2,

            /// <summary>
            /// An ISO 9660 image, level 3
            /// </summary>
            ISO9660_Level3,

            /// <summary>
            /// A basic ISO 9660 format with user defined rules for name lengths and
            /// various name mappings.
            /// </summary>
            ISO9660_Custom
        } ;

        /// <summary>
        /// The possible types of UDF images
        /// </summary>
        public enum UDFImageType
        {
            /// <summary>
            /// No UDF support for this image
            /// </summary>
            None,

            /// <summary>
            /// UDF image 1.02
            /// </summary>
            UDF_1_02,

            /// <summary>
            /// UDF image 1.50, currently not supported
            /// </summary>
            UDF_1_50,

            /// <summary>
            /// USF image 2.00, currently not supported
            /// </summary>
            UDF_2_00,

            /// <summary>
            /// UDF image 2.01, currently not supported
            /// </summary>
            UDF_2_01,

            /// <summary>
            /// UDF image 2.50, currently not supported
            /// </summary>
            UDF_2_50,

            /// <summary>
            /// UDF image 2.60, currently not supported
            /// </summary>
            UDF_2_60,
        } ;

        /// <summary>
        /// The type of emulation for the bootable image
        /// </summary>
        public enum EmulationType : byte
        {
            /// <summary>
            /// No emulation at all
            /// </summary>
            NoEmulation = 0,

            /// <summary>
            /// Emulate a 1.2 Meg floppy disk (5.25 ")
            /// </summary>
            Floppy_1_2 = 1,

            /// <summary>
            /// Emulate a 1.44 Meg floppy disk (3.5 ")
            /// </summary>
            Floppy_1_44 = 2,

            /// <summary>
            /// Emulate a 2.88 Meg floppy disk (3.5 ")
            /// </summary>
            Floppy_2_88 = 3,

            /// <summary>
            /// Emulate a hard disk
            /// </summary>
            HardDisk = 4
        } ;

        /// <summary>
        /// The system type for an Intel/AMD x86 based machine
        /// </summary>
        public const byte SystemType80x86 = 0x00;

        /// <summary>
        /// The system type for a Power PC based system
        /// </summary>
        public const byte SystemTypePowerPC = 0x01;

        /// <summary>
        /// The system type for a MAC
        /// </summary>
        public const byte SystemTypeMac = 0x02;

        #endregion

        #region private member variables

        /// <summary>
        /// If TRUE, this has been changed since saved
        /// </summary>
        [NonSerialized()]
        private bool m_dirty;

        /// <summary>
        /// Settings associated with the ISO 9660 image
        /// </summary>
        private IsoJolietImageSettings m_iso_9660_settings;

        /// <summary>
        /// Settings associated with the Joliet image
        /// </summary>
        private IsoJolietImageSettings m_joliet_settings;

        /// <summary>
        /// Settings associated with the UDF image
        /// </summary>
        private UdfImageSettings m_udf_settings;

        /// <summary>
        /// The ISO type of the image
        /// </summary>
        private ISOImageType m_iso_image_type;

        /// <summary>
        /// The UDF type of the image
        /// </summary>
        private UDFImageType m_udf_image_type;

        /// <summary>
        /// This string contains the name of the file that holds the boot image
        /// </summary>
        private string m_boot_image;


        /// <summary>
        /// The total number of volumes (disks) in the set
        /// </summary>
        private ushort m_volume_set_size;

        /// <summary>
        /// The number of this specific disk in the sequence of volumes.  This is a
        /// one based number.
        /// </summary>
        private ushort m_volume_sequence_number;

        private DateTime m_root_dir_creation_date;

        private ushort m_segment_load_addr;
        private EmulationType m_emulation;
        private byte m_boot_platform_id;

        private bool m_joliet;


        #endregion

        #region constructor
        /// <summary>
        /// Create a new settings object
        /// </summary>
        public ImageSettings()
        {
            m_root_dir_creation_date = DateTime.Now ;

            m_segment_load_addr = 0x7c0;
            m_emulation = EmulationType.Floppy_1_44;
            m_boot_platform_id = SystemType80x86;

            m_boot_image = string.Empty;
            m_volume_set_size = 1;
            m_volume_sequence_number = 1;

            //
            // Default type is an ISO9660 level 3 with Joliet extensions
            //
            m_joliet = true;
            m_iso_image_type = ImageSettings.ISOImageType.ISO9660_Level3;
            m_udf_image_type = ImageSettings.UDFImageType.None;

            m_iso_9660_settings = new IsoJolietImageSettings(m_iso_image_type);
            m_joliet_settings = new IsoJolietImageSettings();
            m_udf_settings = new UdfImageSettings();

            m_dirty = false;
        }
        #endregion

        #region public properties

        /// <summary>
        /// Settings associated with the UDF image
        /// </summary>
        public UdfImageSettings UdfSettings
        {
            get
            {
                return m_udf_settings;
            }
            set
            {
                m_udf_settings = value;
            }
        }

        /// <summary>
        /// Settings associated with the joliet image
        /// </summary>
        public IsoJolietImageSettings JolietSettings
        {
            get
            {
                return m_joliet_settings;
            }
            set
            {
                m_joliet_settings = value;
            }
        }

        /// <summary>
        /// Settings associated with the ISO 9660 image
        /// </summary>
        public IsoJolietImageSettings ISO9660Settings
        {
            get
            {
                return m_iso_9660_settings;
            }
            set
            {
                m_iso_9660_settings = value;
            }
        }

        /// <summary>
        /// The ISO image type
        /// </summary>
        public ISOImageType ISOImage
        {
            get
            {
                return m_iso_image_type;
            }
            set
            {
                m_iso_image_type = value;
            }
        }

        /// <summary>
        /// The UDF image type for the associated image
        /// </summary>
        public UDFImageType UDFImage
        {
            get
            {
                return m_udf_image_type;
            }
            set
            {
                m_udf_image_type = value;
            }
        }

        /// <summary>
        /// Returns TRUE if this is a UDF disc
        /// </summary>
        public bool IsUDF
        {
            get
            {
                return m_udf_image_type != UDFImageType.None;
            }
        }

        /// <summary>
        /// Returns TRUE if this is a ISO disc
        /// </summary>
        public bool IsISO9660
        {
            get
            {
                return m_iso_image_type != ISOImageType.None;
            }
        }

        /// <summary>
        /// This property returns true if the settings are dirty
        /// </summary>
        public bool IsDirty
        {
            get
            {
                return m_dirty || m_iso_9660_settings.Dirty || m_joliet_settings.Dirty || m_udf_settings.Dirty;
            }
            internal set
            {
                m_dirty = value;
                m_iso_9660_settings.Dirty = value;
                m_joliet_settings.Dirty = value;
                m_udf_settings.Dirty = value;
            }
        }

        /// <summary>
        /// This property returns true if a boot image name is present and therefore
        /// the produced image will be bootable.
        /// </summary>
        public bool IsBootable
        {
            get
            {
                return m_boot_image.Length > 0;
            }
        }
       

        /// <summary>
        /// This property contains the name of a file that contains a boot image
        /// for creating a bootable disk
        /// </summary>
        public string BootImageName
        {
            get
            {
                return m_boot_image;
            }
            set
            {
                if (m_boot_image != value)
                {
                    m_boot_image = value;
                    m_dirty = true;
                }
            }
        }
        /// <summary>
        /// This property determines whether or not Joliet support is present on the
        /// disk image created.
        /// </summary>
        public bool IsJoliet
        {
            get
            {
                return m_joliet;
            }
            set
            {
                if (m_joliet != value)
                {
                    m_joliet = value;
                    m_dirty = true;
                }
            }
        }

        /// <summary>
        /// The load address for the boot image
        /// </summary>
        public ushort SegmentLoadAddr
        {
            get
            {
                return m_segment_load_addr;
            }
            set
            {
                if (m_segment_load_addr != value)
                {
                    m_segment_load_addr = value;
                    m_dirty = true;
                }
            }
        }

        /// <summary>
        /// This property returns the emulation type for the boot emulation
        /// </summary>
        public EmulationType Emulation
        {
            get
            {
                return m_emulation;
            }
            set
            {
                if (m_emulation != value)
                {
                    m_dirty = true;
                    m_emulation = value;
                }
            }
        }

        /// <summary>
        /// The platform ID for the disk
        /// </summary>
        public byte BootPlatformID
        {
            get
            {
                return m_boot_platform_id;
            }
            set
            {
                if (m_boot_platform_id != value)
                {
                    m_boot_platform_id = value;
                    m_dirty = true;
                }
            }
        }

        /// <summary>
        /// The size of the volume in disks
        /// </summary>
        public ushort VolumeSpaceSize
        {
            get
            {
                return m_volume_set_size;
            }

            set
            {
                if (m_volume_set_size != value)
                {
                    m_volume_set_size = value;
                    m_dirty = true;
                }
            }
        }

        /// <summary>
        /// The disk number for this disk in the overall volume set.
        /// </summary>
        public ushort VolumeSequenceNumber
        {
            get
            {
                return m_volume_sequence_number;
            }

            set
            {
                if (m_volume_sequence_number != value)
                {
                    m_volume_sequence_number = value;
                    m_dirty = true;
                }
            }
        }

        /// <summary>
        /// The root directory date and time.
        /// </summary>
        public DateTime RootDirectoryCreateDateTime
        {
            get
            {
                return m_root_dir_creation_date;
            }
            set
            {
                if (m_root_dir_creation_date != value)
                {
                    m_root_dir_creation_date = value;
                    m_dirty = true;
                }
            }
        }
        #endregion

        #region public methods
        /// <summary>
        /// This method makes an exact copy of this object
        /// </summary>
        /// <returns></returns>
        public object Clone()
        {
            ImageSettings settings = new ImageSettings();

            settings.BootImageName = BootImageName;
            settings.BootPlatformID = BootPlatformID;
            settings.Emulation = Emulation;
            settings.ISOImage = ISOImage;
            settings.RootDirectoryCreateDateTime = RootDirectoryCreateDateTime;
            settings.SegmentLoadAddr = SegmentLoadAddr;
            settings.UDFImage = UDFImage;
            settings.VolumeSequenceNumber = VolumeSequenceNumber;
            settings.VolumeSpaceSize = VolumeSpaceSize;
            settings.IsJoliet = IsJoliet;
            settings.JolietSettings = (IsoJolietImageSettings)JolietSettings.Clone();
            settings.ISO9660Settings = (IsoJolietImageSettings)ISO9660Settings.Clone();
            settings.UdfSettings = (UdfImageSettings)UdfSettings.Clone();

            return settings;
        }
        #endregion
    }
}
