//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//
using System;
using System.Collections.Generic;
using System.Text;
using System.ComponentModel;
using System.ComponentModel.Design;
using System.Drawing;
using System.Drawing.Design ;

namespace Bwg.Isofs
{
    /// <summary>
    /// This class contains the settings for mapping a filename from the local form of the
    /// name to a for suitable based on these settings
    /// </summary>
    [Serializable()]
    public class IsoJolietImageSettings : ICloneable
    {
        #region private member functions
        /// <summary>
        /// If true, this has been changed since last saved
        /// </summary>
        [NonSerialized()]
        private bool m_dirty;

        /// <summary>
        /// If true, make all case upper case
        /// </summary>
        [NonSerialized()]
        private bool m_iso_image;

        /// <summary>
        /// This variable contains the maximum directory depth allowed for this image
        /// </summary>
        private int m_maximum_directory_depth;

        /// <summary>
        /// The maximum allowed total path length for this image.
        /// </summary>
        private int m_maximum_total_path_length;

        /// <summary>
        /// The maximum length for the name of a directory
        /// </summary>
        private int m_maximum_directory_length;

        /// <summary>
        /// The maximum length for the name of a file
        /// </summary>
        private int m_maximum_filename_length;

        /// <summary>
        /// The maximum length for the extensions on a file
        /// </summary>
        private int m_maximum_file_extension_length;

        /// <summary>
        /// The maximum total length for a file name
        /// </summary>
        private int m_maximum_total_name_length;

        /// <summary>
        /// If TRUE, convert the name to pure upper case
        /// </summary>
        private bool m_convert_to_upper_case;

        /// <summary>
        /// If TRUE, allow non-ISO characters
        /// </summary>
        private bool m_allow_non_iso_chars;

        /// <summary>
        /// The name of the publisher to place in the volume descriptor
        /// </summary>
        private string m_publisher;

        /// <summary>
        /// The name of the preparer to place in the volume descriptor
        /// </summary>
        private string m_preparer;

        /// <summary>
        /// The system identifier, specified in a characters
        /// </summary>
        private string m_system_identifier;

        /// <summary>
        /// The volume identifier, specified in d characters
        /// </summary>
        private string m_volume_identifier;

        /// <summary>
        /// The volume set identifier
        /// </summary>
        private string m_volume_set_identifier;

        /// <summary>
        /// The string identifying the application that created this image
        /// </summary>
        private string m_application_identifier;

        /// <summary>
        /// The name of the top level file that contains copyright information
        /// </summary>
        private string m_copyright_file_identifier;

        /// <summary>
        /// The name of the top level file that contains abstract information
        /// </summary>
        private string m_abstract_file_identifier;

        /// <summary>
        /// The name of the top level file that contains bibliographic information
        /// </summary>
        private string m_bibliographics_file_identifier;

        /// <summary>
        /// The date this volume was created (usually today)
        /// </summary>
        private DateTime m_volume_creation_date;

        /// <summary>
        /// The date this volume was modified (usually today)
        /// </summary>
        private DateTime m_volume_modification_date;

        /// <summary>
        /// The expiration date for this volume
        /// </summary>
        private DateTime m_volume_expiration_date;

        /// <summary>
        /// The effective date of this volume
        /// </summary>
        private DateTime m_volume_effective_date;

        #endregion

        #region constructor
        /// <summary>
        /// Create a settings object for ISO image types with the given ISO image
        /// level
        /// </summary>
        /// <param name="isotype">the level of the ISO image</param>
        public IsoJolietImageSettings(ImageSettings.ISOImageType isotype)
        {
            m_dirty = false;
            switch (isotype)
            {
                case ImageSettings.ISOImageType.ISO9660_Level1:
                    m_maximum_directory_length = 8;
                    m_maximum_filename_length = 8;
                    m_maximum_file_extension_length = 3;
                    m_maximum_total_name_length = 0;
                    m_maximum_directory_depth = 0;
                    m_maximum_total_path_length = 0;
                    m_convert_to_upper_case = true;
                    m_allow_non_iso_chars = false;
                    break;
                case ImageSettings.ISOImageType.ISO9660_Level2:
                    m_maximum_directory_length = 207;
                    m_maximum_filename_length = 0;
                    m_maximum_file_extension_length = 0;
                    m_maximum_total_name_length = 207;
                    m_maximum_directory_depth = 0;
                    m_maximum_total_path_length = 0;
                    m_convert_to_upper_case = true;
                    m_allow_non_iso_chars = false;
                    break;
                case ImageSettings.ISOImageType.ISO9660_Level3:
                    m_maximum_directory_length = 207;
                    m_maximum_filename_length = 0;
                    m_maximum_file_extension_length = 0;
                    m_maximum_total_name_length = 207;
                    m_maximum_directory_depth = 0;
                    m_maximum_total_path_length = 0;
                    m_convert_to_upper_case = true;
                    m_allow_non_iso_chars = false;
                    break;
            }

            m_volume_identifier = "CDROM";
            m_system_identifier = "WIN32";
            m_volume_set_identifier = "";
            m_application_identifier = "BWGBURN";
            m_copyright_file_identifier = "";
            m_abstract_file_identifier = "";
            m_bibliographics_file_identifier = "";
            m_publisher = "";
            m_preparer = "";

            m_volume_creation_date = DateTime.Now;
            m_volume_modification_date = m_volume_creation_date;
            m_volume_effective_date = new DateTime(0);
            m_volume_expiration_date = new DateTime(0);

            m_iso_image = true;
        }

        /// <summary>
        /// Create an image settings object suitable for joliet extenisions to ISO 9660
        /// </summary>
        public IsoJolietImageSettings()
        {
            //
            // Joliet settings
            //
            m_dirty = false;
            m_maximum_directory_length = 64;
            m_maximum_filename_length = 0;
            m_maximum_file_extension_length = 0;
            m_maximum_total_name_length = 64;
            m_maximum_directory_depth = 0;
            m_maximum_total_path_length = 0;
            m_convert_to_upper_case = false;
            m_allow_non_iso_chars = true;

            m_volume_identifier = "CDROM";
            m_system_identifier = "WIN32";
            m_volume_set_identifier = "";
            m_application_identifier = "BwgBurn";
            m_copyright_file_identifier = "";
            m_abstract_file_identifier = "";
            m_bibliographics_file_identifier = "";
            m_publisher = "";
            m_preparer = "";

            m_volume_creation_date = DateTime.Now;
            m_volume_modification_date = m_volume_creation_date;
            m_volume_effective_date = new DateTime(0);
            m_volume_expiration_date = new DateTime(0);

            m_iso_image = false;
        }
        #endregion

        #region public properties

        /// <summary>
        /// This property returns true if this class has been changed, but not saved
        /// </summary>
        [BrowsableAttribute(false)]
        public bool Dirty
        {
            get
            {
                return m_dirty;
            }
            set
            {
                m_dirty = value;
            }
        }

        /// <summary>
        /// This property contains the maximum directory depth for the image being created
        /// </summary>
        [Description("This property sets the maximum depth for the directory hierarchy in the disk image, 0 means no limit"), Category("FileNames")]
        public int MaximumDirectoryDepth
        {
            get
            {
                return m_maximum_directory_depth;
            }
            set
            {
                if (m_maximum_directory_depth != value)
                {
                    m_maximum_directory_depth = value;
                    m_dirty = true;
                }
            }
        }

        /// <summary>
        /// This property contains the maximum total path length allowed for this image.
        /// </summary>
        [Description("This property sets the maximum total path length including the complete directory hierarchy and the filename, 0 means no limit"), Category("FileNames")]
        public int MaximumTotalPathLength
        {
            get
            {
                return m_maximum_total_path_length;
            }
            set
            {
                if (m_maximum_total_path_length != value)
                {
                    m_maximum_total_path_length = value;
                    m_dirty = true;
                }
            }
        }

        /// <summary>
        /// The maximum length of a directory name
        /// </summary>
        [Description("This property sets the maximum length of a directory name, 0 means no limit"), Category("FileNames")]
        public int MaximumDirectoryLength
        {
            get
            {
                return m_maximum_directory_length;
            }
            set
            {
                if (m_maximum_directory_length != value)
                {
                    m_maximum_directory_length = value;
                    m_dirty = true;
                }
            }
        }

        /// <summary>
        /// The maximum length of a file name (not including extension)
        /// </summary>
        [Description("This property sets the maximum length of a file name excluding the extension, 0 means no limit"), Category("FileNames")]
        public int MaximumFileNameLength
        {
            get
            {
                return m_maximum_filename_length;
            }
            set
            {
                if (m_maximum_filename_length != value)
                {
                    m_maximum_filename_length = value;
                    m_dirty = true;
                }
            }
        }

        /// <summary>
        /// Maximum length of a file extension
        /// </summary>
        [Description("This property sets the maximum length of a extension name, 0 means no limit"), Category("FileNames")]
        public int MaximumFileExtensionLength
        {
            get
            {
                return m_maximum_file_extension_length;
            }
            set
            {
                if (m_maximum_file_extension_length != value)
                {
                    m_maximum_file_extension_length = value;
                    m_dirty = true;
                }
            }
        }

        /// <summary>
        /// The maximum total length of a filename including name and extension
        /// </summary>
        [Description("This property sets the maximum total length of a file name including the name, extension and separating '.', 0 means no limit"), Category("FileNames")]
        public int MaximumTotalNameLength
        {
            get
            {
                return m_maximum_total_name_length;
            }
            set
            {
                if (m_maximum_total_name_length != value)
                {
                    m_maximum_total_name_length = value;
                    m_dirty = true;
                }
            }
        }

        /// <summary>
        /// If true, convert all names to upper case
        /// </summary>
        [Description("If true, convert all filenames to all upper case.  If false, do not change the case of the file names."), Category("FileNames")]
        public bool ConvertCaseToUpper
        {
            get
            {
                return m_convert_to_upper_case;
            }
            set
            {
                if (m_convert_to_upper_case != value)
                {
                    m_convert_to_upper_case = value;
                    m_dirty = true;
                }
            }
        }

        /// <summary>
        /// If true, allow non-ISO compliant characters in file and directory names.  If this is
        /// true, all characters but /\:?* are legal in file names.
        /// </summary>
        [Description("If true, allow all ASCII characters in file and directory names except /\\:? and *.  If false, follow the ISO9660 specification exactly."), Category("FileNames")]
        public bool AllowNonISOChars
        {
            get
            {
                return m_allow_non_iso_chars;
            }
            set
            {
                if (m_allow_non_iso_chars != value)
                {
                    m_allow_non_iso_chars = value;
                    m_dirty = true;
                }
            }
        }

        /// <summary>
        /// The volume identifier string
        /// </summary>
        [Description("This string identifies the volume"), Category("Image Properties")]
        public string VolumeIdentifier
        {
            get
            {
                return m_volume_identifier;
            }

            set
            {
                if (m_iso_image && !DiskImageContents.CheckValidDChars(value))
                    throw new Exception("Invalid characters in Volume Identifier, only d-characters allowed (0-9, A-Z, _)");

                if (value.Length > Constants.VolumeIdentifierLength)
                    throw new Exception("Invalid length for Volume Identifier, maximum length is " + Constants.VolumeIdentifierLength.ToString());

                if (m_volume_identifier != value)
                {
                    if (m_iso_image)
                        m_volume_identifier = value.ToUpper();
                    else
                        m_volume_identifier = value;

                    m_dirty = true;
                }
            }
        }

        /// <summary>
        /// The system identifier string
        /// </summary>
        [Description("This string identifies the system that created this image"), Category("Image Properties")]
        public string SystemIdentifier
        {
            get
            {
                return m_system_identifier;
            }

            set
            {
                if (m_iso_image && !DiskImageContents.CheckValidAChars(value))
                    throw new Exception("Invalid characters in System Identifier, only a-characters allowed");

                if (value.Length > Constants.SystemIdentifierLength)
                    throw new Exception("Invalid length for System Identifier, maximum length is " + Constants.SystemIdentifierLength);

                if (m_system_identifier != value)
                {
                    if (m_iso_image)
                        m_system_identifier = value.ToUpper();
                    else
                        m_system_identifier = value;
                    m_dirty = true;
                }
            }
        }

        /// <summary>
        /// The publisher string
        /// </summary>
        [Description("This string identifies the publisher for this volume"), Category("Image Properties")]
        public string Publisher
        {
            get
            {
                return m_publisher;
            }

            set
            {
                if (m_iso_image && !DiskImageContents.CheckValidAChars(value))
                    throw new Exception("Invalid characters in Publisher, only a-characters allowed");

                if (value.Length > Constants.PublisherLength)
                    throw new Exception("Invalid length for Publisher, maximum length is " + Constants.PublisherLength);

                if (m_publisher != value)
                {
                    if (m_iso_image)
                        m_publisher = value.ToUpper();
                    else
                        m_publisher = value;

                    m_dirty = true;
                }
            }
        }

        /// <summary>
        /// The preparer string
        /// </summary>
        [Description("This string identifies the preparer for this volume"), Category("Image Properties")]
        public string Preparer
        {
            get
            {
                return m_preparer;
            }

            set
            {
                if (m_iso_image && !DiskImageContents.CheckValidAChars(value))
                    throw new Exception("Invalid characters in Preparer, only a-characters allowed");

                if (value.Length > Constants.PreparerLength)
                    throw new Exception("Invalid length for Preparer, maximum length is " + Constants.PreparerLength);

                if (m_preparer != value)
                {
                    if (m_iso_image)
                        m_preparer = value.ToUpper();
                    else
                        m_preparer = value;

                    m_dirty = true;
                }
            }
        }

        /// <summary>
        /// The volume set identifier string
        /// </summary>
        [Description("This string identifies the volume set to which this volume belongs"), Category("Image Properties")]
        public string VolumeSetIdentifier
        {
            get
            {
                return m_volume_set_identifier;
            }

            set
            {
                if (m_iso_image && !DiskImageContents.CheckValidDChars(value))
                    throw new Exception("Invalid characters in Volume Set Identifier, only d-characters allowed");

                if (value.Length > Constants.VolumeSetIdentifierLength)
                    throw new Exception("Invalid length for Volume Set Identifier, maximum length is " + Constants.VolumeSetIdentifierLength);

                if (m_volume_set_identifier != value)
                {
                    if (m_iso_image)
                        m_volume_set_identifier = value.ToUpper();
                    else
                        m_volume_set_identifier = value;
                    m_dirty = true;
                }
            }
        }

        /// <summary>
        /// The application identifier string
        /// </summary>
        [Description("This string identifies application that created this volume"), Category("Image Properties")]
        public string ApplicationIdentifier
        {
            get
            {
                return m_application_identifier;
            }
            set
            {
                if (m_iso_image && !DiskImageContents.CheckValidAChars(value))
                    throw new Exception("Invalid characters in Application Identifier, only a-characters allowed");

                if (value.Length > Constants.ApplicationIdentifierLength)
                    throw new Exception("Invalid length for Application Identifier, maximum length is " + Constants.ApplicationIdentifierLength);

                if (m_application_identifier != value)
                {
                    if (m_iso_image)
                        m_application_identifier = value.ToUpper();
                    else
                        m_application_identifier = value;
                    m_dirty = true;
                }
            }
        }

        /// <summary>
        /// The name of the image file that contains the copyright notice
        /// </summary>
        [Description("This string identifies the name of a file that contain copyright information."), Category("Image Properties")]
        public string CopyrightFileIdentifier
        {
            get
            {
                return m_copyright_file_identifier;
            }

            set
            {
                if (m_iso_image && !DiskImageContents.CheckValidDCharsSep1Sep2(value))
                    throw new Exception("Invalid characters in Copyright File Identifier, only d-characters, '.' and ';' allowed");

                if (value.Length > Constants.CopyrightFileIdentifierLength)
                    throw new Exception("Invalid length for Copyright File Identifier, maximum length is " + Constants.CopyrightFileIdentifierLength);

                if (m_copyright_file_identifier != value)
                {
                    if (m_iso_image)
                        m_copyright_file_identifier = value.ToUpper();
                    else
                        m_copyright_file_identifier = value;

                    m_dirty = true;
                }
            }
        }

        /// <summary>
        /// The name of the image file that contains the abstract
        /// </summary>
        [Description("This string identifies the name of a file that contain abstract information."), Category("Image Properties")]
        public string AbstractFileIdentifier
        {
            get
            {
                return m_abstract_file_identifier;
            }

            set
            {
                if (m_iso_image && !DiskImageContents.CheckValidDCharsSep1Sep2(value))
                    throw new Exception("Invalid characters in Abstract File Identifier, only d-characters, '.' and ';' allowed");

                if (value.Length > Constants.AbstractFileIdentifierLength)
                    throw new Exception("Invalid length for Abstract File Identifier, maximum length is " + Constants.AbstractFileIdentifierLength);

                if (m_abstract_file_identifier != value)
                {
                    if (m_iso_image)
                        m_abstract_file_identifier = value.ToUpper();
                    else
                        m_abstract_file_identifier = value;

                    m_dirty = true;
                }
            }
        }

        /// <summary>
        /// The name of the image file that contains the bibliographic information.
        /// </summary>
        [Description("This string identifies the name of a file that contain bibliographic information."), Category("Image Properties")]
        public string BibliographicFileIdentifier
        {
            get
            {
                return m_bibliographics_file_identifier;
            }

            set
            {
                if (m_iso_image && !DiskImageContents.CheckValidDCharsSep1Sep2(value))
                    throw new Exception("Invalid characters in Bibliographic File Identifier, only d-characters, '.' and ';' allowed");

                if (value.Length > Constants.BibliographicFileIdentifierLength)
                    throw new Exception("Invalid length for Bibliographic File Identifier, maximum length is " + Constants.BibliographicFileIdentifierLength);

                if (m_bibliographics_file_identifier != value)
                {
                    if (m_iso_image)
                        m_bibliographics_file_identifier = value.ToUpper();
                    else
                        m_bibliographics_file_identifier = value;

                    m_dirty = true;
                }
            }
        }

        /// <summary>
        /// The creation date and time
        /// </summary>
        [Description("This field gives the creation time and date for the volume."), Category("Dates")]
        [Editor(typeof(DateTimeEditor), typeof(UITypeEditor))]
        public DateTime CreationDateTime
        {
            get
            {
                return m_volume_creation_date;
            }
            set
            {
                if (m_volume_creation_date != value)
                {
                    m_volume_creation_date = value;
                    m_dirty = true;
                }
            }
        }

        /// <summary>
        /// The modification date and time
        /// </summary>
        [Description("This field gives the modification time and date for the volume."), Category("Dates")]
        public DateTime ModificationDateTime
        {
            get
            {
                return m_volume_modification_date;
            }
            set
            {
                if (m_volume_modification_date != value)
                {
                    m_volume_modification_date = value;
                    m_dirty = true;
                }
            }
        }

        /// <summary>
        /// The expiration date and time
        /// </summary>
        [Description("This field gives the expiration time and date for the volume."), Category("Dates")]
        public DateTime ExpirationDateTime
        {
            get
            {
                return m_volume_expiration_date;
            }
            set
            {
                if (m_volume_expiration_date != value)
                {
                    m_volume_expiration_date = value;
                    m_dirty = true;
                }
            }
        }

        /// <summary>
        /// The effective date and time
        /// </summary>
        [Description("This field gives the effective time and date for the volume."), Category("Dates")]
        public DateTime EffectiveDateTime
        {
            get
            {
                return m_volume_effective_date;
            }
            set
            {
                if (m_volume_effective_date != value)
                {
                    m_volume_effective_date = value;
                    m_dirty = true;
                }
            }
        }

        #endregion

        #region public methods
        /// <summary>
        /// This method returns a copy of the given object
        /// </summary>
        /// <returns></returns>
        public object Clone()
        {
            IsoJolietImageSettings settings = new IsoJolietImageSettings();

            settings.AllowNonISOChars = AllowNonISOChars;
            settings.ConvertCaseToUpper = ConvertCaseToUpper;
            settings.MaximumDirectoryDepth = MaximumDirectoryDepth;
            settings.MaximumDirectoryLength = MaximumDirectoryLength;
            settings.MaximumFileExtensionLength = MaximumFileExtensionLength;
            settings.MaximumFileNameLength = MaximumFileNameLength;
            settings.MaximumTotalNameLength = MaximumTotalNameLength;
            settings.MaximumTotalPathLength = MaximumTotalPathLength;

            settings.AbstractFileIdentifier = AbstractFileIdentifier;
            settings.ApplicationIdentifier = ApplicationIdentifier;
            settings.BibliographicFileIdentifier = BibliographicFileIdentifier;

            settings.CopyrightFileIdentifier = CopyrightFileIdentifier;
            settings.CreationDateTime = CreationDateTime;
            settings.EffectiveDateTime = EffectiveDateTime;
            settings.ExpirationDateTime = ExpirationDateTime;
            settings.ModificationDateTime = ModificationDateTime;
            settings.Preparer = Preparer;
            settings.Publisher = Publisher;
            settings.SystemIdentifier = SystemIdentifier;
            settings.VolumeIdentifier = VolumeIdentifier;
            settings.VolumeSetIdentifier = VolumeSetIdentifier;

            settings.m_iso_image = m_iso_image;

            return settings;
        }
        #endregion
    }
}
