//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//

using System;
using System.Collections.Generic;
using System.Text;
using System.IO ;
using System.Diagnostics ;
using System.Runtime.InteropServices;
using Bwg.Utils;

namespace Bwg.Isofs
{
    class IsoWriteStream
    {
        private Stream m_stream ;
        private byte [] m_zeros ;
        private byte[] m_copy_buffer;
        private int m_copy_buffer_used;
        private bool m_copy_file_mode;

        public enum CopyFileStatus
        {
            FileNotFoundSentZeros,
            FileTooShortPaddedZeros,
            FileTooLongTruncated,
            FileJustRight
        } ;

        public long Position
        {
            get
            {
                return m_stream.Position + m_copy_buffer_used;
            }
        }

        public IsoWriteStream(Stream strm, int bufsize)
        {
            m_stream = strm;
            m_zeros = new byte[64] ;

            m_copy_buffer = new byte[bufsize];
            m_copy_buffer_used = 0;
            m_copy_file_mode = false;
        }

        public void StartFileCopy()
        {
            m_copy_file_mode = true;
        }

        public void EndFileCopy()
        {
            FlushCopyBuffer();
        }

        public void WriteUint8(byte b)
        {
            m_stream.WriteByte(b);
        }

        public void WriteUint8(byte b, int cnt)
        {
            for (int i = 0; i < cnt; i++)
                m_stream.WriteByte(b);
        }

        public void WriteUint16Big(ushort b)
        {
            m_stream.WriteByte((byte)((b >> 8) & 0xff));
            m_stream.WriteByte((byte)(b & 0xff));
        }

        public void WriteUint16Small(ushort b)
        {
            m_stream.WriteByte((byte)(b & 0xff));
            m_stream.WriteByte((byte)((b >> 8) & 0xff));
        }

        public void WriteUint16Both(ushort b)
        {
            WriteUint16Small(b);
            WriteUint16Big(b);
        }

        public void WriteUint32Big(int b)
        {
            Debug.Assert(b >= 0) ;
            WriteUint32Big((uint)b) ;
        }

        public void WriteUint32Big(uint b)
        {
            m_stream.WriteByte((byte)((b >> 24) & 0xff));
            m_stream.WriteByte((byte)((b >> 16) & 0xff));
            m_stream.WriteByte((byte)((b >> 8) & 0xff));
            m_stream.WriteByte((byte)(b & 0xff));
        }

        public void WriteUint32Small(int b)
        {
            Debug.Assert(b >= 0);
            WriteUint32Small((uint)b);
        }

        public void WriteUint32Small(uint b)
        {
            m_stream.WriteByte((byte)(b & 0xff));
            m_stream.WriteByte((byte)((b >> 8) & 0xff));
            m_stream.WriteByte((byte)((b >> 16) & 0xff));
            m_stream.WriteByte((byte)((b >> 24) & 0xff));
        }

        public void WriteUint32Both(int b)
        {
            Debug.Assert(b >= 0);
            WriteUint32Both((uint)b);
        }

        public void WriteUint32Both(uint b)
        {
            WriteUint32Small(b);
            WriteUint32Big(b);
        }

        public void WriteString(string s)
        {
            byte [] data = Encoding.ASCII.GetBytes(s) ;
            m_stream.Write(data, 0, data.GetLength(0));
        }

        /// <summary>
        /// Write a string using the supplied encoding.  The string should span
        /// the length given in bytes.
        /// </summary>
        /// <param name="enc">the encoding to use</param>
        /// <param name="s">the string to write</param>
        /// <param name="length">the length of the string</param>
        public void WriteString(Encoding enc, string s, int length)
        {
            WriteString(enc, s, length, ' ');
        }

        /// <summary>
        /// Write a string using the supplied encoding and padding.  The string should be
        /// the length given in bytes
        /// </summary>
        /// <param name="enc"></param>
        /// <param name="s"></param>
        /// <param name="length"></param>
        /// <param name="pad"></param>
        public void WriteString(Encoding enc, string s, int length, char pad)
        {
            byte[] data = enc.GetBytes(s);
            if (data.GetLength(0) >= length)
            {
                m_stream.Write(data, 0, (int)length);
            }
            else
            {
                m_stream.Write(data, 0, data.GetLength(0));
                int towrite = length - data.GetLength(0);
                string padstr = string.Empty ;

                while (true)
                {
                    padstr += pad;
                    data = enc.GetBytes(padstr);
                    if (data.GetLength(0) >= towrite)
                        break;
                }
                m_stream.Write(data, 0, towrite);
            }
        }

        public void Write(byte[] data)
        {
            m_stream.Write(data, 0, data.GetLength(0));
        }

        public void WriteZeros(int cnt)
        {
            while (cnt > 0)
            {
                int towrite = (int)cnt;
                if (towrite > m_zeros.GetLength(0))
                    towrite = m_zeros.GetLength(0);

                m_stream.Write(m_zeros, 0, towrite);
                cnt -= towrite;
            }
        }

        public void WriteZeros(long cnt)
        {
            Debug.Assert(cnt < Int32.MaxValue);
            WriteZeros((int)cnt);
        }

        public void WriteBinaryDateTime(DateTime dt)
        {
            WriteUint8((byte)(dt.Year - 1900));
            WriteUint8((byte)dt.Month);
            WriteUint8((byte)dt.Day);
            WriteUint8((byte)dt.Hour);
            WriteUint8((byte)dt.Minute);
            WriteUint8((byte)dt.Second);

            TimeZone tz = TimeZone.CurrentTimeZone;
            TimeSpan span = tz.GetUtcOffset(dt);
            WriteUint8((byte)(span.TotalMinutes / 15));
        }

        public void WriteAsciiDateTime(DateTime dt)
        {
            if (dt.Ticks != 0)
            {
                string str;

                str = dt.Year.ToString("D4");
                WriteString(Encoding.ASCII, str, 4);
                str = dt.Month.ToString("D2");
                WriteString(Encoding.ASCII, str, 2);
                str = dt.Day.ToString("D2");
                WriteString(Encoding.ASCII, str, 2);
                str = dt.Hour.ToString("D2");
                WriteString(Encoding.ASCII, str, 2);
                str = dt.Minute.ToString("D2");
                WriteString(Encoding.ASCII, str, 2);
                str = dt.Second.ToString("D2");
                WriteString(Encoding.ASCII, str, 2);
                str = ((dt.Millisecond) / 10).ToString("D2");
                WriteString(Encoding.ASCII, str, 2);

                TimeZone tz = TimeZone.CurrentTimeZone;
                TimeSpan span = tz.GetUtcOffset(dt);
                WriteUint8((byte)(span.TotalMinutes / 15));
            }
            else
            {
                // A null date time, means none specfied, write all zeros
                WriteString("0000000000000000");
                WriteUint8(0);
            }
        }

        private void FlushCopyBuffer()
        {
            if (m_copy_buffer_used != 0)
            {
                m_stream.Write(m_copy_buffer, 0, m_copy_buffer_used);
                m_copy_buffer_used = 0;
            }
        }

        public CopyFileStatus CopyFile(string name, long filesize, long copysize)
        {
            CopyFileStatus ret;
            int remaining;
            Debug.Assert(m_copy_file_mode);

            FileStream fs = null;

            try
            {
                fs = new FileStream(name, FileMode.Open, FileAccess.Read);
            }
            catch (Exception)
            {
                WriteZeros(copysize);
                return CopyFileStatus.FileNotFoundSentZeros ;
            }

            long len = fs.Length;

            // Lock the file so it does not change while we do this.
            try
            {
                fs.Lock(0, len);
            }
            catch (Exception)
            {
            }

            if (len < filesize)
                ret = CopyFileStatus.FileTooShortPaddedZeros;
            else if (len > filesize)
                ret = CopyFileStatus.FileTooLongTruncated;
            else
                ret = CopyFileStatus.FileJustRight;

            while (len > 0 && copysize > 0)
            {
                remaining = m_copy_buffer.GetLength(0) - m_copy_buffer_used;
                if (remaining > copysize)
                    remaining = (int)copysize;

                int cnt = fs.Read(m_copy_buffer, m_copy_buffer_used, remaining);
                if (cnt == 0)
                    break;

                m_copy_buffer_used += cnt;
                if (m_copy_buffer_used == m_copy_buffer.GetLength(0))
                    FlushCopyBuffer();

                len -= cnt;
                copysize -= cnt;
            }
            fs.Close();

            //
            // Now, we need to continue to write zero data til copysize is zero
            //
            while (copysize > 0)
            {
                long numzeros = copysize;
                remaining = m_copy_buffer.GetLength(0) - m_copy_buffer_used;

                if (numzeros > remaining)
                    numzeros = remaining;

                for (long i = 0; i < numzeros; i++)
                    m_copy_buffer[m_copy_buffer_used++] = 0;

                copysize -= numzeros;
                if (m_copy_buffer_used == m_copy_buffer.GetLength(0))
                    FlushCopyBuffer();
            }

            return ret;
        }
    }
}
