//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//
using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;

namespace Bwg.Isofs
{
    class TaskGroup : CreatorTask
    {
        #region private member variables

        /// <summary>
        /// A human readable name for this task group
        /// </summary>
        private string m_name;

        /// <summary>
        /// This member holds the total required size of the group.  If this is zero
        /// there is no total size requirement.
        /// </summary>
        private long m_total_size_required;

        /// <summary>
        /// This is a list of tasks that belong to this task group
        /// </summary>
        private IList<CreatorTask> m_tasks ;

        /// <summary>
        /// This task is the pad task that is used to pad the task group to the
        /// required size.
        /// </summary>
        private PadToBlockTask m_padder;
        #endregion

        #region constructors
        /// <summary>
        /// This method creates a task group with no size requirements
        /// </summary>
        /// <param name="name">name of the task group</param>
        public TaskGroup(string name)
        {
            m_total_size_required = 0;
            m_tasks = new List<CreatorTask>();
            m_name = name;
        }

        /// <summary>
        /// This method creats a task group that has a minimum size requirement.
        /// </summary>
        /// <param name="name">name of the task group</param>
        /// <param name="size">total required minimum size of the task group</param>
        public TaskGroup(string name, long size)
        {
            m_total_size_required = size;
            m_tasks = new List<CreatorTask>();
            m_name = name;
        }
        #endregion

        #region public methods
        /// <summary>
        /// This method adds a task to the task group
        /// </summary>
        /// <param name="t"></param>
        public void AddTask(CreatorTask t)
        {
            m_tasks.Add(t);
        }

        /// <summary>
        /// Update the layout of the group of tasks
        /// </summary>
        /// <param name="creator"></param>
        /// <param name="info"></param>
        /// <returns></returns>
        public override bool UpdateLayout(DiskImageCreator creator, DiskInfo info)
        {
            long total = 0 ;

            foreach (CreatorTask t in m_tasks)
            {
                t.BlockAddress = info.CurrentBlock;
                if (!t.UpdateLayout(creator, info))
                    return false;
                t.BlockSize = info.CurrentBlock - t.BlockAddress;
                Debug.WriteLine("TaskGroup '" + m_name + "' : UpdateLayout task '" + t.ToString() + "' - start = " + t.BlockAddress.ToString() + ", length = " + t.BlockSize.ToString());
                total += t.BlockSize;
            }

            if (m_total_size_required > 0 && total < m_total_size_required)
            {
                m_padder = new PadToBlockTask(PadToBlockTask.PadMode.PadWithBlockCount, m_total_size_required - total);
                m_padder.BlockAddress = info.CurrentBlock;
                if (!m_padder.UpdateLayout(creator, info))
                    return false;
                m_padder.BlockSize = info.CurrentBlock - m_padder.BlockAddress;
                // Debug.WriteLine("TaskGroup '" + m_name + "' : UpdateLayout task '" + m_padder.ToString() + "' - start = " + m_padder.BlockAddress.ToString() + ", length = " + m_padder.BlockSize.ToString());
            }
            else
            {
                m_padder = null;
            }

            return true;
        }

        public override bool Write(DiskImageCreator creator, DiskInfo info, IsoWriteStream strm)
        {
            foreach (CreatorTask t in m_tasks)
            {
                Debug.WriteLine("TaskGroup '" + m_name + "' : Write task '" + t.ToString() + "' - start = " + (strm.Position / 2048).ToString());
                if (!t.Write(creator, info, strm))
                    return false;
            }

            if (m_padder != null)
            {
                Debug.WriteLine("TaskGroup '" + m_name + "' : Write task '" + m_padder.ToString() + "' - start = " + (strm.Position / 2048).ToString());
                if (!m_padder.Write(creator, info, strm))
                    return false;
            }

            return true;
        }
        #endregion
    }
}
