//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//

using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using Bwg.Logging;

namespace Bwg.Isofs
{
    class VolumeDescriptorTask : CreatorTask
    {
        #region private variables

        /// <summary>
        /// The name type for this descriptor (either ISO or Joliet)
        /// </summary>
        DiskImageElement.NameType m_name_type;

        /// <summary>
        /// The type of this volume descriptor
        /// </summary>
        byte m_type;

        /// <summary>
        /// The volume flags for the descriptor (0 for primary)
        /// </summary>
        byte m_volume_flags;

        /// <summary>
        /// The L path table
        /// </summary>
        private PathTableTask m_l_path_table;

        /// <summary>
        /// The M path table
        /// </summary>
        private PathTableTask m_m_path_table;

        /// <summary>
        /// This is the root of the directory tree for this descriptor
        /// </summary>
        private DirectoryTask m_directories;

        /// <summary>
        /// This is the escape sequence to write to the block
        /// </summary>
        private byte[] m_escape_sequences;

        #endregion

        #region constructor
        public VolumeDescriptorTask(byte type, byte flags, DiskImageElement.NameType nametype)
        {
            m_name_type = nametype;
            m_volume_flags = flags;
            m_type = type;
        }
        #endregion

        #region public properties

        /// <summary>
        /// This property is the task that generates the directory tree for this
        /// volume descriptor
        /// </summary>
        public DirectoryTask Directory
        {
            get
            {
                return m_directories;
            }
            set
            {
                m_directories = value;
            }
        }

        /// <summary>
        /// The escape sequence for a secondary descriptor
        /// </summary>
        public byte[] EscapeSequences
        {
            get
            {
                return m_escape_sequences;
            }
            set
            {
                m_escape_sequences = value;
                Debug.Assert(m_type == Constants.SupplementaryVolumeDescriptor);
                Debug.Assert(m_escape_sequences.GetLength(0) == 32);
            }
        }

        /// <summary>
        /// The task for the L path table
        /// </summary>
        public PathTableTask LPathTable
        {
            get
            {
                return m_l_path_table;
            }
            set
            {
                m_l_path_table = value;
            }
        }

        /// <summary>
        /// The task for the M path table
        /// </summary>
        public PathTableTask MPathTable
        {
            get
            {
                return m_m_path_table;
            }
            set
            {
                m_m_path_table = value;
            }
        }

        #endregion

        #region public methods
        public override bool UpdateLayout(DiskImageCreator creator, DiskInfo info)
        {
            bool ret = true;

            //
            // The primary volume descriptor must be at block 16 relative to the start
            // of the image
            //
            if (m_type == 1 && info.CurrentBlock != creator.StartBlock +  16)
            {
                Logger logger = creator.GetLogger();
                string str = "internal error - invalid block for 'Primary Volume Descriptor', expected 16, actual ";
                str += info.CurrentBlock.ToString();
                logger.LogMessage(new UserMessage(UserMessage.Category.Error, 0, str));
                ret = false;
            }
            info.CurrentBlock++;

            return ret;
        }

        public override bool Write(DiskImageCreator creator, DiskInfo info, IsoWriteStream strm)
        {
            bool ret = true;
            Encoding enc;

            if (m_name_type == DiskImageElement.NameType.ISO9660)
                enc = Encoding.ASCII;
            else
                enc = Encoding.BigEndianUnicode;

            strm.WriteUint8(m_type);
            strm.WriteString("CD001");                                                  // CD Identifier
            strm.WriteUint8(1);                                                         // Volume Descriptor Version
            strm.WriteUint8(m_volume_flags);                                            // Unused

            if (m_name_type == DiskImageElement.NameType.ISO9660)
            {
                strm.WriteString(enc, MapString(creator.Settings.ISO9660Settings.SystemIdentifier), Constants.SystemIdentifierLength);
                strm.WriteString(enc, MapString(creator.Settings.ISO9660Settings.VolumeIdentifier), Constants.VolumeIdentifierLength);
            }
            else
            {
                strm.WriteString(enc, MapString(creator.Settings.JolietSettings.SystemIdentifier), Constants.SystemIdentifierLength);
                strm.WriteString(enc, MapString(creator.Settings.JolietSettings.VolumeIdentifier), Constants.VolumeIdentifierLength);
            }

            strm.WriteZeros(8);                                                         // Unused
            Debug.Assert(info.ImageSize <= Int32.MaxValue);
            strm.WriteUint32Both((int)info.ImageSize);                                  // Volume Space Size In Blocks

            if (m_escape_sequences == null)
                strm.WriteZeros(32);                                                    // Unused
            else
                strm.Write(m_escape_sequences);

            strm.WriteUint16Both(creator.Settings.VolumeSpaceSize);                     // Volume Set Size = 1
            strm.WriteUint16Both(creator.Settings.VolumeSequenceNumber);                // Volume Sequence Number = 1
            strm.WriteUint16Both(info.BlockSize);                                       // Volume Block Size

            strm.WriteUint32Both(m_l_path_table.PathTableSize);                         // Path table size
            Debug.Assert(m_l_path_table.BlockAddress <= Int32.MaxValue);
            strm.WriteUint32Small((int)m_l_path_table.BlockAddress);                    // L Path table location
            strm.WriteUint32Small(0);                                                   // Optional L Path table location

            Debug.Assert(m_m_path_table.BlockAddress <= Int32.MaxValue);
            strm.WriteUint32Big((int)m_m_path_table.BlockAddress);                      // M Path table location
            strm.WriteUint32Big(0);                                                     // Optional M path table location

            // Write the root directory entry
            strm.WriteUint8(34);                                                        // Root entry size (34 bytes)
            strm.WriteUint8(0);                                                         // No extended attributes

            DiskImageDirectory dir = creator.RootDirectory ;
            long loc = dir.GetDirectoryLocation(NameTypeToDirLocationType(m_name_type)).Location;
            Debug.Assert(loc <= UInt32.MaxValue);
            strm.WriteUint32Both((uint)loc) ;                                           // Extent for the root dir
            strm.WriteUint32Both((uint)(dir.GetExtentSizeInBlocks(m_name_type) * 2048));      
                                                                                        // Size of the extent in bytes

            strm.WriteBinaryDateTime(creator.Settings.RootDirectoryCreateDateTime);      
                                                                                        // Write the date/time
            strm.WriteUint8((byte)Constants.DirectoryEntryFileFlags.Directory);         // Set the file flags
            strm.WriteUint8(0);                                                         // File unit size
            strm.WriteUint8(0);                                                         // Interleave gap size
            strm.WriteUint16Both(1);                                                    // The volume number for this file
            strm.WriteUint8(1);                                                         // Short cut for the name "."
            strm.WriteUint8(0);

            if (m_name_type == DiskImageElement.NameType.ISO9660)
            {
                strm.WriteString(enc, MapString(creator.Settings.ISO9660Settings.VolumeSetIdentifier), Constants.VolumeSetIdentifierLength);
                strm.WriteString(enc, MapString(creator.Settings.ISO9660Settings.Publisher), Constants.PublisherLength);
                strm.WriteString(enc, MapString(creator.Settings.ISO9660Settings.Preparer), Constants.PreparerLength);
                strm.WriteString(enc, MapString(creator.Settings.ISO9660Settings.ApplicationIdentifier), Constants.ApplicationIdentifierLength);
                strm.WriteString(enc, MapString(creator.Settings.ISO9660Settings.CopyrightFileIdentifier), Constants.CopyrightFileIdentifierLength);
                strm.WriteString(enc, MapString(creator.Settings.ISO9660Settings.AbstractFileIdentifier), Constants.AbstractFileIdentifierLength);
                strm.WriteString(enc, MapString(creator.Settings.ISO9660Settings.BibliographicFileIdentifier), Constants.BibliographicFileIdentifierLength);

                strm.WriteAsciiDateTime(creator.Settings.ISO9660Settings.CreationDateTime);                 // Volume creation date/time
                strm.WriteAsciiDateTime(creator.Settings.ISO9660Settings.ModificationDateTime);             // Volume last modified date/time
                strm.WriteAsciiDateTime(creator.Settings.ISO9660Settings.ExpirationDateTime);               // Volume expiration date/time
                strm.WriteAsciiDateTime(creator.Settings.ISO9660Settings.EffectiveDateTime);                // Volume effective date/time
            }
            else
            {
                strm.WriteString(enc, MapString(creator.Settings.JolietSettings.VolumeSetIdentifier), Constants.VolumeSetIdentifierLength);
                strm.WriteString(enc, MapString(creator.Settings.JolietSettings.Publisher), Constants.PublisherLength);
                strm.WriteString(enc, MapString(creator.Settings.JolietSettings.Preparer), Constants.PreparerLength);
                strm.WriteString(enc, MapString(creator.Settings.JolietSettings.ApplicationIdentifier), Constants.ApplicationIdentifierLength);
                strm.WriteString(enc, MapString(creator.Settings.JolietSettings.CopyrightFileIdentifier), Constants.CopyrightFileIdentifierLength);
                strm.WriteString(enc, MapString(creator.Settings.JolietSettings.AbstractFileIdentifier), Constants.AbstractFileIdentifierLength);
                strm.WriteString(enc, MapString(creator.Settings.JolietSettings.BibliographicFileIdentifier), Constants.BibliographicFileIdentifierLength);

                strm.WriteAsciiDateTime(creator.Settings.JolietSettings.CreationDateTime);                 // Volume creation date/time
                strm.WriteAsciiDateTime(creator.Settings.JolietSettings.ModificationDateTime);             // Volume last modified date/time
                strm.WriteAsciiDateTime(creator.Settings.JolietSettings.ExpirationDateTime);               // Volume expiration date/time
                strm.WriteAsciiDateTime(creator.Settings.JolietSettings.EffectiveDateTime);                // Volume effective date/time
            }

            strm.WriteUint8(1);                                                         // File structure version
            strm.WriteUint8(0);                                                         // File structure version
            strm.WriteUint8(0x20, 512);                                                 // Application use area, all spaces
            strm.WriteZeros(653);                                                       // The remainder is padding

            return ret;
        }
        #endregion

        #region private methods
        private string MapString(string s)
        {
            if (m_type == 1)
                return s.ToUpper();

            return s;
        }
        #endregion
    }
}
