//
// BwgBurn - CD-R/CD-RW/DVD-R/DVD-RW burning program for Windows XP
// 
// Copyright (C) 2006 by Jack W. Griffin (butchg@comcast.net)
//
// This program is free software; you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
// for more details.
//
// You should have received a copy of the GNU General Public License along 
// with this program; if not, write to the 
//
// Free Software Foundation, Inc., 
// 59 Temple Place, Suite 330, 
// Boston, MA 02111-1307 USA
//

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Threading;
using System.IO;
using System.Diagnostics;
using Bwg.Controls ;
using Bwg.Burner ;
using Bwg.Logging ;
using Bwg.Scsi ;
using Bwg.Database.Freedb;
using Bwg.Audio ;
using Bwg.Audio.Ripper;
using Bwg.Audio.Acm ;

namespace Bwg.RipAudio
{
    public partial class AudioRipperForm : BwgBurnBaseForm
    {
        #region private member variables

        /// <summary>
        /// The item under the cursor when the right mouse button asked for
        /// the context menu
        /// </summary>
        private ListViewItem m_edit_item ;

        /// <summary>
        /// THe right click context menu for the song menu
        /// </summary>
        private ContextMenu m_song_menu;

        /// <summary>
        /// The edit box for editing song titles or artist
        /// </summary>
        private TextBox m_edit_box;

        /// <summary>
        /// The driver we are reading data from
        /// </summary>
        private Drive m_drive;

        /// <summary>
        /// The object representing the disk in the drive
        /// </summary>
        private Disk m_disk;

        /// <summary>
        /// The result from the freedb.org query on this album
        /// </summary>
        private FreeDBQueryRequest m_result;

        /// <summary>
        /// The album information about the album
        /// </summary>
        private FreeDBAlbumInfo m_album;

        /// <summary>
        /// The thread that is actually ripping the data
        /// </summary>
        private Thread m_rip_thread;

        /// <summary>
        /// The FreeDB server for getting information about the album being ripped
        /// </summary>
        private FreeDBServer m_server;

        /// <summary>
        /// If true, a background thread is running to rip an album
        /// </summary>
        private bool m_ripping;

        /// <summary>
        /// The audio manager for writing data to files
        /// </summary>
        private AudioManager m_mgr;

        /// <summary>
        /// The driver to use to convert data from CD format data to a user selected
        /// output format.
        /// </summary>
        private AcmDriver m_driver;

        /// <summary>
        /// The specific format to use for output
        /// </summary>
        private AcmFormatDetails m_format;

        /// <summary>
        /// The current item being ripped
        /// </summary>
        private int m_current_item;

        /// <summary>
        /// The progress bar for the current item being ripped
        /// </summary>
        private TextProgressBar m_bar;

        /// <summary>
        /// The path to the song to be ripped
        /// </summary>
        private string m_song_path;

        /// <summary>
        /// The timer for timing progress
        /// </summary>
        private System.Windows.Forms.Timer m_timer;

        /// <summary>
        /// The start time for the rip operation
        /// </summary>
        private DateTime m_start;

        /// <summary>
        /// If true, this user wants to close the window
        /// </summary>
        private bool m_want_close;

        /// <summary>
        /// If true, eject the disk when done
        /// </summary>
        private bool m_eject_when_done;

        #endregion

        #region private delegates
        private delegate void UpdateAlbumInfoDelegate(FreeDBResponseArgs args);
        private delegate void SetProgressBarDelegate(int which) ;
        private delegate void RipDoneDelegate();
        private delegate void TimerTickDelegate();
        private delegate void CloseFormDelegate();
        private delegate void SetTrackPathDelegate(string path);
        #endregion

        #region private constants
        const string DiscArtistPropName = "DiscArtist";
        const string DiscTitlePropName = "DiscTitle";
        const string TrackTitlePropName = "TrackTitle";
        const string TrackArtistPropName = "TrackArtist";
        const string MyMusicPropName = "@MyMusic";
        const string MyDocumentPropName = "@MyDocument";
        const string DesktopPropName = "@Desktop";
        const string TrackNumberPropName = "TrackNumber";
        const string PercentPropName = "Percent";
        #endregion

        #region constructor
        public AudioRipperForm()
        {
            InitializeComponent();
            SetupFreeDBServer();

            LoggerWindow.SetupColumns();
            LoggerWindow.Sink = m_sink;

            DeviceSelector.Logger = m_logger;
            DeviceSelector.InitializeDeviceList(false);
            DeviceSelector.SelectedIndexChanged += new EventHandler(DeviceSelector_SelectedIndexChanged);

            m_mgr = new AudioManager(m_logger);

            m_bar = null ;
            m_ripping = false;

            m_album = null;
            m_result = null;

            GetStartupCodecFormat();
            SetCodecControls();

            SongList.MouseDown += new MouseEventHandler(SongList_MouseDown);
         
            m_song_menu = new ContextMenu();
            m_song_menu.MenuItems.Add(new MenuItem("Edit Title", new EventHandler(EditSongTitle)));
            m_song_menu.MenuItems.Add(new MenuItem("Edit Artist", new EventHandler(EditSongArtist)));

            SongList.ColumnWidthChanged += new ColumnWidthChangedEventHandler(SongList_ColumnWidthChanged);
            SongList.ColumnWidthChanging += new ColumnWidthChangingEventHandler(SongList_ColumnWidthChanging);

            m_song_path = "%@MyMusic%\\%DiscArtist%\\%DiscTitle%\\%TrackNumber% %TrackArtist% %TrackTitle%" ;
            m_song_path = GetStartupString(PropertyNameRipAudioMusicLibraryPath, m_song_path);

            m_eject_when_done = GetStartupBool(PropertyNameRipAudioEjectWhenDone, true);
            m_timer = new System.Windows.Forms.Timer();
        }
        #endregion

        #region protected overrides
        protected override void OnClosing(CancelEventArgs e)
        {
            if (m_ripping)
            {
                m_want_close = true;
                e.Cancel = true;

                WaitingOnDrive form = new WaitingOnDrive();
                form.Show(this);
                Application.DoEvents();
            }
            else
            {
                if (m_server != null)
                    m_server.Dispose();

                base.OnClosing(e);
            }
        }

        #endregion

        #region private methods

        private void UpdateSubmitButton()
        {
            //if (m_album == null || m_album.FromFreedb || !m_album.ReadyForSubmit)
            //    SubmitAlbum.Enabled = false;
            //else
            //    SubmitAlbum.Enabled = true;
        }

        private string GetAlbumTitle()
        {
            if (m_album != null)
                return m_album.Title;

            return "";
        }

        private string GetAlbumArtist()
        {
            if (m_album != null)
                return m_album.Artist;

            return "";
        }

        private string GetTrackTitle(int tno)
        {
            if (m_album != null)
                return m_album[tno - 1].Title;

            return "Track " + tno.ToString();
        }

        private string GetTrackArtist(int tno)
        {
            if (m_album != null)
                return m_album[tno - 1].Artist;

            return string.Empty;
        }

        private void SongList_ColumnWidthChanged(object sender, ColumnWidthChangedEventArgs args)
        {
            if (m_bar != null)
            {
                ListViewItem.ListViewSubItem sub = SongList.Items[m_current_item].SubItems[3];
                m_bar.Bounds = sub.Bounds;
            }
        }

        private void SongList_ColumnWidthChanging(object sender, ColumnWidthChangingEventArgs args)
        {
            if (m_bar != null)
            {
                ListViewItem.ListViewSubItem sub = SongList.Items[m_current_item].SubItems[3];
                m_bar.Bounds = sub.Bounds;
            }
        }

        private void SongList_MouseDown(object sender, MouseEventArgs e)
        {
            ListViewHitTestInfo info = SongList.HitTest(e.Location);
            if (info.Item == null || !m_disk.IsAudio)
            {
                SongList.ContextMenu = null;
                m_edit_item = null;
            }
            else
            {
                m_edit_item = info.Item;
                SongList.ContextMenu = m_song_menu;
            }

            base.OnMouseDown(e);
        }


        private void EditSongTitle(object sender, EventArgs args)
        {
            m_edit_box = new TextBox();
            m_edit_box.AcceptsReturn = true;
            m_edit_box.KeyPress += new KeyPressEventHandler(EditTitle_KeyPress);
            m_edit_box.Parent = SongList;
            m_edit_box.Bounds = m_edit_item.GetBounds(ItemBoundsPortion.Label);
            m_edit_box.Text = m_edit_item.SubItems[0].Text;
            m_edit_box.Focus();

            m_edit_box.LostFocus += new EventHandler(EditTitle_LostFocus);
        }

        private void EditTitle_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (e.KeyChar == 0x0d || e.KeyChar == 0x0a)
            {
                e.Handled = true;
                EditTitle_LostFocus(null, null);
            }
            else if (e.KeyChar == 0x1b)
            {
                e.Handled = true;
                m_edit_box.LostFocus -= new EventHandler(EditTitle_LostFocus);
                m_edit_box.Dispose();
            }
        }

        private void EditTitle_LostFocus(object senter, EventArgs args)
        {
            int which = MatchList.SelectedIndex;
            if (which == -1)
                which = 0;

            if (m_album == null)
                m_album = CreateRequest(m_disk);

            m_album.SetTrackTitle(m_edit_item.Index, m_edit_box.Text) ;
            UpdateSubmitButton();
            m_edit_item.SubItems[0].Text = m_edit_box.Text;
            m_edit_box.Dispose();
            m_edit_box = null;
        }

        private void EditSongArtist(object sender, EventArgs args)
        {
            m_edit_box = new TextBox();
            m_edit_box.AcceptsReturn = true;
            m_edit_box.KeyPress += new KeyPressEventHandler(EditArtist_KeyPress);
            m_edit_box.Parent = SongList;
            m_edit_box.Bounds = m_edit_item.SubItems[1].Bounds;
            m_edit_box.Focus();

            m_edit_box.LostFocus += new EventHandler(EditArtist_LostFocus);
        }

        private void EditArtist_LostFocus(object senter, EventArgs args)
        {
            int which = MatchList.SelectedIndex;
            if (which == -1)
                which = 0;

            if (m_album == null)
                m_album = CreateRequest(m_disk);

            m_album.SetTrackArtist(m_edit_item.Index, m_edit_box.Text);
            UpdateSubmitButton();
            m_edit_item.SubItems[1].Text = m_edit_box.Text;
            m_edit_box.Dispose();
            m_edit_box = null;
        }

        private void EditArtist_KeyPress(object sender, KeyPressEventArgs e)
        {
            if (e.KeyChar == 0x0d || e.KeyChar == 0x0a)
            {
                e.Handled = true;
                EditArtist_LostFocus(null, null);
            }
            else if (e.KeyChar == 0x1b)
            {
                e.Handled = true;
                m_edit_box.LostFocus -= new EventHandler(EditArtist_LostFocus);
                m_edit_box.Dispose();
            }
        }

        /// <summary>
        /// Get the default codec to use from the startup information
        /// </summary>
        private void GetStartupCodecFormat()
        {
            string codec = GetStartupString(PropertyNameRipAudioCodec, "");
            string fmt = GetStartupString(PropertyNameRipAudioFormat, "");

            foreach (AcmDriver drv in m_mgr.AcmManager.Drivers)
            {
                if (drv.ToString() == codec)
                {
                    m_driver = drv;
                    break;
                }
            }

            if (m_driver == null)
                return;

            foreach (AcmFormatDetails format in m_driver.FormatDescriptors)
            {
                if (format.ToString() == fmt)
                {
                    m_format = format;
                    break;
                }
            }
        }

        private void SaveStartupCodeFormat()
        {
            if (m_driver != null)
                SetStartupString(PropertyNameRipAudioCodec, m_driver.ToString());
            if (m_format != null)
                SetStartupString(PropertyNameRipAudioFormat, m_format.ToString());
        }

        /// <summary>
        /// Set the text fields that indicate the codec and format that will be used for
        /// the CD rip.
        /// </summary>
        private void SetCodecControls()
        {
            if (m_driver == null)
            {
                CodecControl.Text = "Not Set";
                FormatControl.Text = "Not Set";
            }
            else
            {
                CodecControl.Text = m_driver.ToString();
                if (m_format == null)
                    FormatControl.Text = "Not Set";
                else
                    FormatControl.Text = m_format.ToString();
            }
        }

        /// <summary>
        /// Update the visible album information based on the response received from the
        /// freeDB server.
        /// </summary>
        /// <param name="args">the freedb response arguments</param>
        private void UpdateAlbumInfo(FreeDBResponseArgs args)
        {
            string str;

            if (args.Status == FreeDBResponseArgs.ResponseType.CouldNotContactServer)
            {
                FreeDBStatus.Text = "Could not contact server";
            }
            else if (args.Status == FreeDBResponseArgs.ResponseType.InvalidServerResponse)
            {
                FreeDBStatus.Text = "Invalid server response";
            }
            else if (args.Status == FreeDBResponseArgs.ResponseType.LoginFailed)
            {
                FreeDBStatus.Text = "Login failed";
            }
            else if (args.Status == FreeDBResponseArgs.ResponseType.UnknownError)
            {
                FreeDBStatus.Text = "Unknown Error";
            }
            else if (args.Status == FreeDBResponseArgs.ResponseType.Success && args.Result is FreeDBQueryRequest)
            {
                FreeDBQueryRequest result = (FreeDBQueryRequest)args.Result;

                if (result.Status == FreeDBQueryRequest.StatusType.NoMatch)
                    FreeDBStatus.Text = "No match found";
                else
                {
                    m_result = result;

                    MatchList.Items.Clear();
                    for (int i = 0; i < m_result.MatchCount; i++)
                    {
                        FreeDBAlbumInfo info = m_result[i];
                        MatchList.Items.Add(info);
                    }

                    MatchList.SelectedIndex = 0;

                    str = m_result.MatchCount.ToString() + " ";
                    if (result.Status == FreeDBQueryRequest.StatusType.ExactMatch)
                    {
                        str += "exact ";
                    }
                    else if (result.Status == FreeDBQueryRequest.StatusType.FuzzyMatch)
                    {
                        str += "close ";
                    }

                    if (m_result.MatchCount == 1)
                        str += "match found";
                    else
                        str += "matches found";

                    FreeDBStatus.Text = str;
                }
            }
        }

        /// <summary>
        /// This is the callback for the freedb response.  This method either directly calls
        /// the update routine, or if we are in a different thread (the usual case), we call
        /// through and Invoke call
        /// </summary>
        /// <param name="sender">the object sending the response, should be a FreeDB server</param>
        /// <param name="args">the response from the server</param>
        private void GotFreeDBResponse(object sender, FreeDBResponseArgs args)
        {
            if (InvokeRequired)
            {
                UpdateAlbumInfoDelegate del = new UpdateAlbumInfoDelegate(UpdateAlbumInfo);
                object[] argarr = new object[1];
                argarr[0] = args;
                Invoke(del, argarr);
            }
            else
            {
                if (args.Result is FreeDBQueryRequest)
                    UpdateAlbumInfo(args);
                else if (args.Result is FreeDBSubmitRequest)
                {
                    MessageBox.Show("Submit request completed sucessfully.");
                }
            }
        }

        /// <summary>
        /// This method initialize the FreeDB server when the program is started.  It queries
        /// the startup information from the startup file and then starts the freedb server
        /// thread based on this information
        /// </summary>
        private void SetupFreeDBServer()
        {
            string hostname;
            string protocol;
            int port;
            string cgipath;

            hostname = GetStartupString(PropertyNameFreeDBHostName, "freedb.freedb.org");
            port = GetStartupInteger(PropertyNameFreeDBPort, 8080);
            protocol = GetStartupString(PropertyNameFreeDBProtocol, FreeDBSite.CddbProtocolName);
            cgipath = GetStartupString(PropertyNameFreeDBCgiPath, string.Empty);

            FreeDBUserInfo user = new FreeDBUserInfo("BwgBurn", GetVersion());

            FreeDBSite site;

            if (protocol.ToLower() == FreeDBSite.HttpProtocolName)
                site = new FreeDBSite(hostname, port, cgipath);
            else
                site = new FreeDBSite(hostname, port);

            m_server = new FreeDBServer(site, user);
            m_server.FreeDBResponse += new EventHandler<FreeDBResponseArgs>(GotFreeDBResponse);
        }

        /// <summary>
        /// This method creates a FreeDB album request
        /// </summary>
        /// <param name="dsk"></param>
        /// <returns></returns>
        private FreeDBAlbumInfo CreateRequest(Disk dsk)
        {
            Device.CommandStatus st;
            IList<TocEntry> toc;

            st = dsk.Drive.Device.ReadToc(0, true, out toc);
            if (st != Device.CommandStatus.Success)
                return null;

            uint number = 1;
            FreeDBAlbumInfo req = new FreeDBAlbumInfo();
            req.Title = "?";
            req.Artist = "?";
            foreach (TocEntry entry in toc)
            {
                MinuteSecondFrame msf;

                if (entry.Number != 0xaa)
                {
                    TrackInformation tinfo;

                    st = dsk.Drive.Device.ReadTrackInformation(Device.ReadTrackType.TrackNumber, number, out tinfo);
                    if (st != Device.CommandStatus.Success)
                        return null;

                    msf = new MinuteSecondFrame(tinfo.TrackSize);
                }
                else
                {
                    msf = new MinuteSecondFrame(0, 0, 0);
                }
                string title = "Track " + number.ToString();
                req.AddTrack(title, (byte)entry.StartMSF.Minutes, entry.StartMSF.Seconds, entry.StartMSF.Frames, (byte)msf.Minutes, msf.Seconds, msf.Frames);

                number++;
            }

            return req;
        }

        /// <summary>
        /// This method updates the disk window based on the disk and album information
        /// stored in the class
        /// </summary>
        private void UpdateDiskWindow()
        {
            // Get the total number of matches for this album.  If there is no
            // freedb response, there is a single match
            int which = MatchList.SelectedIndex;
            if (which == -1)
                which = 0;

            // Get the artist for the album
            ArtistControl.Text = GetAlbumArtist();
            TitleControl.Text = GetAlbumTitle();

            //
            // Now, fill in the songs
            //
            for(int i = 1 ; i <= m_disk.Sessions[0].Tracks.GetLength(0) ; i++)
            {
                string title = GetTrackTitle(i);
                ListViewItem item = new ListViewItem(title);

                Track t = m_disk.Sessions[0].Tracks[i - 1] ;
                MinuteSecondFrame msf = new MinuteSecondFrame(t.Info.TrackSize);

                string artist = GetTrackArtist(i);
                item.SubItems.Add(artist);
                item.SubItems.Add(msf.ToString("M:S"));
                item.SubItems.Add("");
                item.Checked = true;
                SongList.Items.Add(item);
            }
        }

        /// <summary>
        /// This method is called when the user changes the device selector.
        /// </summary>
        /// <param name="sender">the control sending the message</param>
        /// <param name="e">the event args, ignored</param>
        private void DeviceSelector_SelectedIndexChanged(object sender, EventArgs e)
        {
            // Close the drive that is currently open and mark it as null
            if (m_drive != null)
                m_drive.Dispose();

            m_drive = null;

            // Mark the disk as null
            m_disk = null;

            // Get the name of the disk from the combo box
            string dname = (string)DeviceSelector.SelectedItem;

            // Open the base device
            Device dev = new Device(m_logger);
            if (!dev.Open(dname[0]))
            {
                UpdateSubmitButton();
                m_album = null;
                MessageBox.Show("Cannot open device '" + dname + "'", "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            // Open/Initialize the driver
            m_drive = new Drive(dev);
            DiskOperationError status = m_drive.Initialize();
            if (status != null)
            {
                m_drive = null;
                UpdateSubmitButton();
                m_album = null;
                MessageBox.Show("Cannot initialize drive '" + dname + "' - " + status.Message, "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return ;
            }

            Drive.FeatureState readfeature = m_drive.GetFeatureState(Feature.FeatureType.CDRead);
            if (readfeature == Drive.FeatureState.Error || readfeature == Drive.FeatureState.NotPresent)
            {
                MessageBox.Show("The drive '" + dname + "' does not support the CD READ feature.  This feature is required to rip audio.", "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            // Clear the list of songs
            SongList.Items.Clear() ;
            m_result = null;
            m_album = null;
            MatchList.Items.Clear();
            MatchList.Text = string.Empty;
            m_disk = new Disk(m_drive);
            if (m_disk.IsAudio)
            {
                // Update the window
                UpdateDiskWindow();

                // Make a request of FreeDB to get the album information
                FreeDBStatus.Text = "Waiting for album information";
                FreeDBAlbumInfo req = CreateRequest(m_disk);
                if (req != null && m_server != null)
                    m_server.QueryRequest(req);
            }
            else
            {
                FreeDBStatus.Text = "Not an audio CD";
                m_disk = null;
            }

            UpdateSubmitButton();
        }

        /// <summary>
        /// Update the title information based on the selection of a new match in the
        /// match list.
        /// </summary>
        /// <param name="sender">the object sending the event</param>
        /// <param name="e">the event args</param>
        private void TitleComboBox_SelectedIndexChanged(object sender, EventArgs e)
        {
            // Set the album to the newly selected one
            m_album = (FreeDBAlbumInfo)MatchList.SelectedItem;

            UpdateSubmitButton();

            // Update the title and artist
            TitleControl.Text = GetAlbumTitle() ;
            ArtistControl.Text = GetAlbumArtist() ;

            // Update the tracks
            for (int i = 1 ; i <= m_disk.Sessions[0].Tracks.GetLength(0) ; i++)
            {
                ListViewItem item = SongList.Items[i - 1];

                // Update the title based on the new selection
                item.SubItems[0].Text = GetTrackTitle(i);

                // Update the artist based on the new selection
                item.SubItems[1].Text = GetTrackArtist(i);
            }
        }

        /// <summary>
        /// This method marks all songs that are checked for being ripped as pending.
        /// </summary>
        private void MarkSongsPending()
        {
            for (int i = 0; i < SongList.Items.Count; i++)
            {
                if (SongList.Items[i].Checked)
                    SongList.Items[i].SubItems[3].Text = "Pending";
                else
                    SongList.Items[i].SubItems[3].Text = "Skipped";
            }
        }

        /// <summary>
        /// This method is called when the user presses the RIP button.
        /// </summary>
        /// <param name="sender">the object sending the event</param>
        /// <param name="e">the event args, ignored</param>
        private void RipButton_Click(object sender, EventArgs e)
        {
            // Make sure we have a disk and it is audio
            if (m_disk == null || !m_disk.IsAudio)
            {
                MessageBox.Show("The disk in this drive is not an audio CD", "Error!", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            // Make sure we have an audio output format
            while (m_driver == null || m_format == null)
            {
                DialogResult ret = MessageBox.Show("No output format is set.  Would you like to set the output format now?", "No Format Set",
                                                   MessageBoxButtons.YesNo, MessageBoxIcon.Question);
                if (ret == DialogResult.No)
                    return;

                Settings_Click(null, null);
            }

            // Disable those tings we don't want active while we are ripping
            RipButton.Enabled = false;
            DeviceSelector.Enabled = false;
            Settings.Enabled = false;
            MatchList.Enabled = false;
            SetupFreeDB.Enabled = false;
            // SubmitAlbum.Enabled = false;
            ArtistControl.Enabled = false;
            TitleControl.Enabled = false;
            LibrarySettings.Enabled = false;

            // Mark all checked songs as pending
            MarkSongsPending();

            // Create the info to send to the rip thread
            int which = MatchList.SelectedIndex;
            if (which == -1)
                which = 0;
            RipInfo info = new RipInfo(which, SongList.Items.Count);
            for (int i = 0; i < SongList.Items.Count; i++)
                info.TrackStatus[i] = SongList.Items[i].Checked;

            m_start = DateTime.Now;
            m_timer.Interval = 1000;
            m_timer.Tick += new EventHandler(TimerTickHandler);
            m_timer.Start();

            // Create the thread for ripping audio and hand the info to that thread
            m_rip_thread = new Thread(new ParameterizedThreadStart(RipProc));
            m_rip_thread.Start(info);
        }

        private void TimerTickHandler(object sender, EventArgs args)
        {
            TimerTickCallback();
        }

        private void TimerTickCallback()
        {
            if (InvokeRequired)
            {
                Invoke(new TimerTickDelegate(TimerTickCallback));
            }
            else
            {
                TimeSpan elapsed = DateTime.Now - m_start;
                ElapsedTime.Text = elapsed.Minutes.ToString() + ":" + elapsed.Seconds.ToString("D2");
            }
        }

        /// <summary>
        /// Convert a general string to something that is acceptable as a file name
        /// </summary>
        /// <param name="str">string to convert</param>
        /// <returns>converted string</returns>
        private string ConvertToFileName(string str)
        {
            string result = string.Empty;
            char [] badchars = Path.GetInvalidPathChars() ;

            foreach (char ch in str)
            {
                bool badchar = false ;

                for(int i = 0 ; i < badchars.GetLength(0) ; i++)
                {
                    if (badchars[i] == ch)
                    {
                        badchar = true ;
                        break ;
                    }
                }

                if (badchar)
                    result += ' ';
                else
                    result += ch;
            }

            result = result.Trim();

            return result;
        }

        /// <summary>
        /// Set the progress bar to the appropriate track based on the item number
        /// </summary>
        /// <param name="item"></param>
        private void SetProgressBar(int item)
        {
            if (InvokeRequired)
            {
                object[] args = new object[1];
                args[0] = item;
                Invoke(new SetProgressBarDelegate(SetProgressBar), args);
            }
            else
            {
                if (m_bar != null)
                {
                    SongList.Items[m_current_item].SubItems[3].Text = "Ripped";
                    m_bar.Dispose();
                    m_bar = null ;
                }

                if (item != -1)
                {
                    ListViewItem.ListViewSubItem sub = SongList.Items[item].SubItems[3];
                    sub.Text = "";
                    m_bar = new TextProgressBar();
                    m_bar.BackColor = Color.Black;
                    m_bar.ColorText = Color.GreenYellow;
                    m_bar.BarOutlineColor = Color.GreenYellow;
                    m_bar.BarInnerColorDark = Color.FromArgb(0, 15, 0);
                    m_bar.BarInnerColorLight = Color.FromArgb(0, 128, 0);
                    m_bar.Text = "Ripping";
                    m_bar.DrawMode = TextProgressBar.DrawingMode.Solid;
                    m_bar.PaintMode = TextProgressBar.PaintingMode.HalfGradient;
                    m_bar.AddPercent = true;
                    m_bar.Minimum = 0;
                    m_bar.Maximum = 100;
                    m_bar.Parent = SongList;
                    m_bar.Bounds = sub.Bounds;
                    m_current_item = item;
                }
            }
        }

        private void SetRipProgress(int value)
        {
            if (InvokeRequired)
            {
                object[] args = new object[1];
                args[0] = value;
                Invoke(new SetProgressBarDelegate(SetRipProgress), args);
            }
            else
            {
                if (m_bar != null)
                {
                    Debug.WriteLine("SET RIP PROGRESS " + value.ToString());
                    m_bar.Value = value;
                    Application.DoEvents();
                }
            }
        }

        private void RipDone()
        {
            if (InvokeRequired)
                Invoke(new RipDoneDelegate(RipDone));
            else
            {
                m_timer.Stop();
                SetProgressBar(-1);

                RipButton.Enabled = true;
                DeviceSelector.Enabled = true;
                Settings.Enabled = true;
                MatchList.Enabled = true;
                SetupFreeDB.Enabled = true;
                // SubmitAlbum.Enabled = true;
                ArtistControl.Enabled = true;
                TitleControl.Enabled = true;
                LibrarySettings.Enabled = true;

                if (m_eject_when_done)
                    m_drive.Device.StartStopUnit(true, Device.PowerControl.NoChange, Device.StartState.EjectDisc);

                MessageBox.Show("The CD was ripped sucessfully");
            }
        }

        private string GetSongFileName(string pathname, IDictionary<string, string> props)
        {
            char [] seps = new char[1] { '\\' } ;
            string[] spath = pathname.Split(seps);
            string result = string.Empty ;

            foreach(string elem in spath)
            {
                bool intoken = false ;
                string elemstr = string.Empty ;
                string token = string.Empty ;

                for(int i = 0 ; i < elem.Length ; i++)
                {
                    if (intoken)
                    {
                        if (elem[i] == '%')
                        {
                            if (token.Length == 0)
                                elemstr += '%' ;
                            else
                            {
                                string tokenresult ;
                                if (!props.TryGetValue(token, out tokenresult))
                                    throw new Exception("Unknown token name '" + token + "' in song path") ;

                                if (token[0] != '@')
                                    tokenresult = ConvertToFileName(tokenresult);

                                elemstr += tokenresult ;
                            }
                            intoken = false;
                        }
                        else
                            token += elem[i] ;
                    }
                    else
                    {
                        if (elem[i] == '%')
                        {
                            token = string.Empty;
                            intoken = true;
                        }
                        else
                            elemstr += elem[i];
                    }
                }

                if (result.Length > 0 && result[result.Length - 1] != '\\')
                    result += '\\' ;
                result += elemstr ;
            }
            return result ;
        }

        private bool IsDrive(string dirpath)
        {
            return dirpath.Length == 3 && Char.IsLetter(dirpath[0]) && dirpath[1] == ':' && dirpath[2] == '\\';
        }

        private void CreatePath(string dirpath)
        {
            if (Directory.Exists(dirpath) && IsDrive(dirpath))
                return;

            string parent = Path.GetDirectoryName(dirpath);
            CreatePath(parent);
            Directory.CreateDirectory(dirpath);
        }

        private void CloseForm()
        {
            if (InvokeRequired)
                Invoke(new CloseFormDelegate(CloseForm));
            else
                Close();
        }

        private void RipProc(object w)
        {
            m_ripping = true;

            if (!(w is RipInfo))
            {
                RipDone();
                return;
            }
            RipInfo info = (RipInfo)w;

            CDRipper ripper = new CDRipper(m_drive.Device, m_logger);
            ripper.Progress += new ProgressChangedEventHandler(RipProgress);

            m_drive.Device.PreventAllowMediumRemoval(Device.PreventAllow.Prevent);
            for (int i = 1; i <= m_disk.Sessions[0].Tracks.GetLength(0); i++)
            {
                if (m_want_close)
                {
                    CloseForm();
                    break;
                }

                if (info.TrackStatus[i - 1])
                {
                    //
                    // Setup the properties for this track
                    //
                    IDictionary<string, object> props = new Dictionary<string, object>();

                    if (m_album != null)
                    {
                        props[AudioWriter.PropNameAlbum] = GetAlbumTitle();
                        props[AudioWriter.PropNameArtist] = GetAlbumArtist();
                        props[AudioWriter.PropNameSongTitle] = GetTrackTitle(i);
                    }
                    props[AcmWriter.PropNameTrackNumber] = i.ToString() + "/" + m_album.TrackCount.ToString();

                    SetProgressBar(i - 1);
                    string songtitle = i.ToString("d2") + " ";
                    IDictionary<string, string> nameprops = new Dictionary<string, string>() ;

                    nameprops[DiscArtistPropName] = GetAlbumArtist();
                    nameprops[DiscTitlePropName] = GetAlbumTitle();
                    nameprops[TrackTitlePropName] = GetTrackTitle(i);
                    nameprops[TrackArtistPropName] = GetTrackArtist(i);
                    nameprops[MyMusicPropName] = Environment.GetFolderPath(Environment.SpecialFolder.MyMusic);
                    nameprops[MyDocumentPropName] = Environment.GetFolderPath(Environment.SpecialFolder.MyDocuments);
                    nameprops[DesktopPropName] = Environment.GetFolderPath(Environment.SpecialFolder.DesktopDirectory);
                    nameprops[TrackNumberPropName] = i.ToString("d2");
                    nameprops[PercentPropName] = "%";

                    songtitle = GetSongFileName(m_song_path, nameprops) ;
                    CreatePath(Path.GetDirectoryName(songtitle));
                    SetTrackPath(songtitle);
                    ripper.RipTrack(i, songtitle, props, m_driver, m_format);
                }
            }
            m_drive.Device.PreventAllowMediumRemoval(Device.PreventAllow.Allow);
            ripper.Progress -= new ProgressChangedEventHandler(RipProgress);

            m_ripping = false;
            RipDone();
        }

        private void SetTrackPath(string path)
        {
            if (InvokeRequired)
            {
                object[] args = new object[1];
                args[0] = path;
                Invoke(new SetTrackPathDelegate(SetTrackPath), args);
            }
            else
            {
                TrackPath.Text = path;
            }
        }

        private void RipProgress(object sender, ProgressChangedEventArgs args)
        {
            SetRipProgress(args.ProgressPercentage);
        }

        private void SetupFreeDB_Click(object sender, EventArgs e)
        {
            FreeDBSetupForm form;

            if (m_server != null && m_server.Site != null)
                form = new FreeDBSetupForm(m_server.Site, "BwgBurn", GetVersion());
            else
                form = new FreeDBSetupForm(null, "BwgBurn", GetVersion());

            form.StartPosition = FormStartPosition.CenterParent;
            form.ShowDialog(this);

            if (m_server != null)
                m_server.Dispose();

            m_server = new FreeDBServer(form.SelectedSite, new FreeDBUserInfo("BwgBurn", GetVersion()));
            m_server.FreeDBResponse += new EventHandler<FreeDBResponseArgs>(GotFreeDBResponse);

            FreeDBStatus.Text = "Waiting for album information";
            Disk dsk = new Disk(m_drive);
            FreeDBAlbumInfo req = CreateRequest(dsk);
            if (req != null && m_server != null)
                m_server.QueryRequest(req);

            SetStartupString(PropertyNameFreeDBHostName, form.SelectedSite.HostName);
            SetStartupInteger(PropertyNameFreeDBPort, form.SelectedSite.Port);
            SetStartupString(PropertyNameFreeDBProtocol, form.SelectedSite.Protocol);
            SetStartupString(PropertyNameFreeDBCgiPath, form.SelectedSite.HttpCgiPath);
        }

        private void Settings_Click(object sender, EventArgs e)
        {
            OutputAudioSettings form = new OutputAudioSettings(m_mgr.AcmManager, m_driver, m_format);
            if (form.ShowDialog() == DialogResult.OK)
            {
                m_driver = form.Driver;
                m_format = form.Format;

                if (m_driver != null)
                    SetStartupString(PropertyNameRipAudioCodec, m_driver.ToString());

                if (m_format != null)
                    SetStartupString(PropertyNameRipAudioFormat, m_format.ToString());

                SetCodecControls();
            }
        }

        private void TitleControl_TextChanged(object sender, EventArgs e)
        {
            int which = MatchList.SelectedIndex ;
            if (which == -1)
                which = 0 ;

            if (m_album == null)
                m_album = CreateRequest(m_disk);

            m_album.Title = TitleControl.Text;
            UpdateSubmitButton();
        }

        private void ArtistControl_TextChanged(object sender, EventArgs e)
        {
            int which = MatchList.SelectedIndex;
            if (which == -1)
                which = 0;

            if (m_album == null)
                m_album = CreateRequest(m_disk);

            m_album.Artist = ArtistControl.Text;
            UpdateSubmitButton();
        }

        private void SubmitAlbum_Click(object sender, EventArgs e)
        {
            SubmitFreedb freedb = new SubmitFreedb(m_server, m_album);
            if (freedb.ShowDialog() == DialogResult.OK)
            {
                m_album.Category = freedb.Category;
                m_album.Year = freedb.Year;
                m_album.Genre = freedb.Genre;

                m_server.SubmitViaEmail(m_album);
            }
        }

        private void LibrarySettings_Click(object sender, EventArgs e)
        {
            MusicLibrarySettingsForm form = new MusicLibrarySettingsForm(m_eject_when_done, m_mgr.AcmManager, m_driver, m_format, m_song_path);
            if (form.ShowDialog() == DialogResult.OK)
            {
                m_driver = form.Driver;
                m_format = form.Format;
                m_song_path = form.SongPath;
                m_eject_when_done = form.EjectWhenDoneValue;

                SetCodecControls();
                SaveStartupCodeFormat();
                SetStartupString(PropertyNameRipAudioMusicLibraryPath, m_song_path);
                SetStartupBool(PropertyNameRipAudioEjectWhenDone, m_eject_when_done);
            }
        }

        #endregion
    }
}